// **********************************************************************
//  PicPicture class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICPICTURE_HXX
#define PICPICTURE_HXX

#include <parray.hxx>
#include <PicList.hxx>
#include <PicRawPicture.hxx>
#include <mtools.hxx>

class PicImageFormat;

/**
  * PicPicture is one of the most important classes in this application.
  * It represents a (logical) image and has 
  * \begin{itemize}
  *   \item ID; Used to uniquely identify it inside the database. Zero means
  *   that the image is not in the database (or the implementation 
  *   does not care).
  *   \item Name; user-given string.
  *   \item Par1; user-given string.
  *   \item Par2; user-given string.
  *   \item Filename; name of the file where the image is stored (if known)
  *    full pathnames are not required ...
  *   \item Image data (opt); as PicRawPicture, which includes width+height
  *   \item Format (opt); pointer to PicImageFormat if a image (file) format 
  *   is defined, used when inserting to database.
  *   \item Original data (opt); in image (file) format pointed to by format
  *   used for storing original data for "lossy" 
  *   PicImageFormat:s so that load+save does
  *   not deteriorate image quality.
  *   \item IdenticalToFilename; boolean: has this image changed since it was 
  *   last loaded/saved?
  *   \item Signatures; array of PicSignature:s.
  *   \item Segmentations; array of PicSegmentation:s.
  *   \item MatchRate; used for passing match rates of query's result images.
  *   \item MatchRanking; used for passing ranking number of query's result 
  *   images.
  * \end{itemize}
  **/

class PicPicture : public PicObject
{
 public:
  PICOBJECTCLASS(PicPicture)

  PICERRORCODES { ok,
    rawNull,
    noPixel
  };

  /**
    * The default constructor.
    **/
  PicPicture ();
  /**
    * The copy constructor. Copying images makes copies of all the members.
    *
    * @param other PicPicture object to be copied.
    **/
  PicPicture (const PicPicture &other);
  /**
    * The destructor.
    **/
  virtual ~PicPicture ();

  /**
    * Updates image from a file trying all available image format 
    * implementations (PicImageFormat:s). First tries to guess
    * the format by filename extension.
    *
    * @param filename Name of the file to be loaded.
    * @return True if successful.
    **/
  PICBOOL loadFromFile(BaseString filename);
  /**
    * Saves image data into file. Decides the image format 
    * (PicImageFormat) to use by filename extension. If none 
    * matches, saves in PPM format.
    *
    * @param filename Name of the file to be saved.
    * @return True if successful.
    **/
  PICBOOL writeToFile(BaseString filename);

  /**
    * Despite the name, used to set image data from data provided
    * by PicGUI.
    *
    * @return True if successful.
    **/
  PICBOOL getBlock(Tk_PhotoImageBlock *);

  /**
    * Sets image data.
    *
    * @param newRaw Pointer to a PicRawPicture object holding the image
    *               raw data.
    **/
  void    setRaw(PicRawPicture *newRaw);

  /// Used for passing match rates of query's result images.
  BaseString        matchRate;
  /// Used for passing ranking number of query's result images.
  PICUSHORT         matchRanking;
  /// List of PicSingatures.
  PicList           signatures;
  /// Array of PicSegmentations.
  PicArray          segmentations;
  /// Picture ID.
  PICULONG          id;
  /// Source database ID
  PICULONG			DBid;
  /// Picture name.
  BaseString        name;
  /// Picture parameter string 1.
  BaseString        par1;
  /// Picture parameter string 2.
  BaseString        par2;
  /// Picture file name.
  BaseString        filename;
  /// Has the image changed since last load/save?
  PICBOOL           identicalToFilename;
  /// Original image data.
  PicBin            origData;
  /// Picture format object.
  PicImageFormat    *format;
  /// Picture raw data.
  PicRawPicture     *raw;

  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;

  /// Returns true when raw == NULL.
  virtual PICBOOL isNull() const;

  /// Sets all attributes to null values.
  virtual void    setNull();

  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// **********************************************************************

inline PICBOOL PicPicture::isDefault() const 
  { return isNull(); }
inline void PicPicture::setDefault()
  { setNull(); }
inline BaseString PicPicture::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicPicture::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }

#endif
