// **********************************************************************
//  PicRawPicture class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICRAWPICTURE_HXX
#define PICRAWPICTURE_HXX

#include <tk.h>
#include <bstring.hxx>
#include <pobject.hxx>

class PicRGBPicture;
class PicGUI;

/**
  * PicRawPicture holds image data (plus width and height of the image),
  * most likely the pixel RGB-values. Usually instances are kept inside 
  * PicPicture:s.
  * Actually PicRawPicture is not supposed to assume anything about the 
  * selected color channel scheme. PicRGBPicture represents an RGB-picture 
  * and you may derive from PicRawPicture to create classes for handling 
  * different color channel schemes (for example YIQ or HSV). However, 
  * since we are using Tcl/Tk and PicGUI directly accesses the member 
  *   Tk_PhotoImageBlock rawImage
  * contained in PicRawPicture, the usage would require (rather painful)
  * conversions and changes to GUI implementation. You'd also need to 
  * change many aspects about the PicSearch core, because it assumes that 
  * raw image data is always interpreted in the same way (i.e. as 
  * RGB-values) and thus saves raw data into database without information 
  * about the color channel scheme. Nevertheless, a clever algorithm 
  * implementation might benefit at run-time from an easy way of 
  * tranforming a PicRGBPicture into some other form. 
  *  Tk_PhotoImageBlock rawImage;
  **/

class PicRawPicture : public PicObject
{
 public:
  PICOBJECTCLASS(PicRawPicture)

  /**
    * The default constructor.
    **/
  PicRawPicture();
  /**
    * The copy constructor. Copying PicRawPictures copies the image data.
    **/
  PicRawPicture(const PicRawPicture &other);
  /**
    * The destructor.
    **/
  virtual ~PicRawPicture();

  /**
    * Returns the width of the image.
    *
    * @return Width of the image.
    **/
  PICUSHORT getWidth() const;
  /**
    * Sets the width of the image. Be sure to set image data accordingly!
    *
    * @param newWidth New width for the image.
    **/
  void      setWidth(PICUSHORT newWidth);
  /**
    * Returns the height of the image.
    *
    * @return Height of the image.
    **/
  PICUSHORT getHeight() const;
  /**
    * Sets the height of the image. Be sure to set image data accordingly!
    *
    * @param newHeight New height for the image.
    **/
  void      setHeight(PICUSHORT newHeight);

  /**
    * Returns pointer to the raw image data. Use function picFree to
    * deallocate memory.
    *
    * @return Pointer to the raw data.
    **/
  unsigned char* getDataPtr() const;
  /**
    * Sets the raw image data pointer for this object. Use function
    * picAllocate to allocate the memory block.
    *
    * @param newPtr Pointer to the new raw data memory block.
    **/
  void           setDataPtr(unsigned char* newPtr);

  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;
  /**
    * Tests if this object is empty.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isNull() const;
  /**
    * Sets this object as empty.
    **/
  virtual void    setNull();
  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

 private:

  /**
    * rawImage holds the actual data pointer, width and height.
    * Defined in tk.h:
    *
    * \begin{verbatim}
         typedef struct {
           unsigned char *pixelPtr;
           int width;
           int height;
           int pitch;
           int pixelSize;
           int offset[3];
         } Tk_PhotoImageBlock;
     \end{verbatim
    **/
  Tk_PhotoImageBlock rawImage;

  friend class PicRGBPicture;
  friend class PicGUI;
};

/*:class ********************************************************************

PicRGBPicture is derived from PicRawPicture to represent RGB-valued
image data.

****************************************************************************/

class PicRGBPicture : public PicRawPicture
{
 public:
  PICOBJECTCLASS(PicRGBPicture)

  /**
    * The default constructor.
    **/
  PicRGBPicture();
  /**
    * Constructor which copies the data from memory buffer.
    *
    * @param width Image width.
    * @param height Image height.
    * @param dataBuf Pointer to image data block.
    **/
  PicRGBPicture(PICULONG width, PICULONG height, unsigned char *dataBuf);
  /**
    * Constructor which copies the data from Tk_PhotoImageBlock.
    **/
  PicRGBPicture(Tk_PhotoImageBlock *block);

};

// **********************************************************************

inline PICBOOL PicRawPicture::isDefault() const 
  { return isNull(); }
inline void PicRawPicture::setDefault()
  { setNull(); }
inline BaseString PicRawPicture::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicRawPicture::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }

#endif

