// **********************************************************************
//  PicSegmentation and PicSegment class definitions
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICSEGMENT_HXX
#define PICSEGMENT_HXX

#include <parray.hxx>

/**
  * PicSegmentation holds a collection of segments computed by some
  * segmentation algorithm.
  **/

class PicSegmentation : public PicObject
{
 public:
  PICOBJECTCLASS(PicSegmentation)

  /**
    * The default constructor.
    **/
  PicSegmentation();
  /**
    * Copy constructor.
    * 
    * @param other Reference to a PicSegmentation object to be copied.
    **/
  PicSegmentation(const PicSegmentation &other);
  /**
    * The destructor.
    **/
  virtual ~PicSegmentation();

  /// Contains the segments (PicSegment objetcs).
  PicArray segments;
  /// ID of the segmentation algorithm that was used to compute the segments.
  PICULONG algId;

  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;
  /**
    * Tests if this object is empty.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isNull() const;
  /**
    * Sets this object as empty.
    **/
  virtual void    setNull();
  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

};

/**
  * PicSegment represents a segment computed by some segmentation 
  * algorithm. The actual format of the data describing the segment
  * is defined by each algorithm, so the users should check the 
  * documentation of a particular segmentation algorithm before using
  * it in computing signatures for segments. (A standard way to 
  * represent segments would be much better...)
  **/

class PicSegment : public PicObject
{
 public:
  PICOBJECTCLASS(PicSegment)

  /**
    * The default constructor.
    **/
  PicSegment();
  /**
    * Copy constructor. When copying PicSegments a new data buffer of size
    * dataSize is allocated and the data is copied.
    *
    * @param other Reference to a PicSegment object to be copied.
    **/
  PicSegment(const PicSegment &other);
  /**
    * The destructor.
    **/
  virtual ~PicSegment();

  /// Should be set to the size of the data buffer.
  PICULONG dataSize;
  /** Should contain the actual data representing this segment. Use function
    * picAllocate to allocate the memory.
    **/
  char     *data;

  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;
  /**
    * Tests if this object is empty.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isNull() const;
  /**
    * Sets this object as empty.
    **/
  virtual void    setNull();
  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// **********************************************************************

inline PICBOOL PicSegmentation::isDefault() const 
  { return isNull(); }
inline void PicSegmentation::setDefault()
  { setNull(); }
inline BaseString PicSegmentation::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicSegmentation::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }

inline PICBOOL PicSegment::isDefault() const 
  { return isNull(); }
inline void PicSegment::setDefault()
  { setNull(); }
inline BaseString PicSegment::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicSegment::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }


#endif


