// **********************************************************************
//  PicSegmentAlgorithm class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICSEGMENTALGORITHM_HXX
#define PICSEGMENTALGORITHM_HXX

#include <PicAlgDefs.hxx>
#include <PicSegment.hxx>
#include <PicPicture.hxx>

/**
  * PicSegmentAlgorithm is the base class for all segmentation algorithms.
  * Segmentation algorithms are used to produce segments out of an image. 
  * The actual representation and meaning of segments (PicSegment:s) depends 
  * on the algorithm, though a general scheme for the segments would be 
  * much better.
  **/

class PicSegmentAlgorithm : public PicObject
{
 public:
  PICABSOBJECTCLASS(PicSegmentAlgorithm)

  /**
    * The default constructor.
    **/
  PicSegmentAlgorithm();

  /// Should be a descriptive name for the algorithm.
  BaseString name;
  /** Should be a unique (within segmentation algorithms) ID identifying
    * this algorithm (define in PicAlgDefs.hxx).
    **/
  PICULONG   id;
  /** Set isAutomatic to boolTrue if you want that this algorithm is used
    * automatically to compute a segmentation for an image before the image
    * is inserted into database (and thus the segments are also inserted
    * into database).
    **/
  PICBOOL    isAutomatic;

  /**
    * This function is meant for accessing the (possible) parameters of this
    * algorithm. Note that you must separately update the GUI to create a
    * suitable window for the parameters.
    *
    * NOTE: The paramaters of a search algorithm should have no 
    *       influence on the signatures that are computed for images
    *       (since those are already in the database and would have
    *       to be recomputed when changing parameters!). 
    *       The paramaters are only meant for giving you at run-time 
    *       a way of controlling how to compare signatures.
    *
    * @param params Reference to a collection object which is to receive the
    *               parameters.
    * @return True if successful.
    **/
  virtual PICBOOL setParams(PicIxCollect &params);
  /**
    * This function is meant for accessing the (possible) parameters of this
    * algorithm. Note that you must separately update the GUI to create a
    * suitable window for the parameters.
    *
    * NOTE: The paramaters of a search algorithm should have no 
    *       influence on the signatures that are computed for images
    *       (since those are already in the database and would have
    *       to be recomputed when changing parameters!). 
    *       The paramaters are only meant for giving you at run-time 
    *       a way of controlling how to compare signatures.
    *
    * @param params Reference to a collection object holding the parameters.
    * @return True if successful.
    **/
  virtual PICBOOL getParams(PicIxCollect &params);

  /**
    * This function should allocate a new PicSegmentation (with 
    * operator new) and set its algId-attributes to the id of this
    * algorithm and insert into its segments-array the segments
    * of the argument pic (i.e. instances of PicSegment:s). The 
    * interpretation of the PicSegment:s depend on the algorithm, 
    * though a general scheme for the segments would be much better.
    *
    * @param pic Reference to a PicPicture object.
    * @return Pointer to the new PicSegmentation object.
    **/
  virtual PicSegmentation* computeSegments(PicPicture &pic) const = 0;
  
  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;
  /**
    * Tests if this object is empty.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isNull() const;
  /**
    * Sets this object as empty.
    **/
  virtual void    setNull();
  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// **********************************************************************

inline PICBOOL PicSegmentAlgorithm::isDefault() const 
  { return isNull(); }
inline void PicSegmentAlgorithm::setDefault()
  { setNull(); }
inline BaseString PicSegmentAlgorithm::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicSegmentAlgorithm::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }


#endif


