// **********************************************************************
//  PicSketchAlgorithm class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICSKETCHALGORITHM_HXX
#define PICSKETCHALGORITHM_HXX

#include <PicSearchAlgorithm.hxx>

/**
  * PicSegmSignature is a class for holding segment signatures.
  * For actual usage, you only need to choose the segmentation 
  * algorithm (just set id to segmAlgId) and implement function
  * algId().
  **/

class PicSegmSignature : public PicSignature
{
 public:
  PICABSOBJECTCLASS(PicSegmSignature)

  PicSegmSignature();
  virtual ~PicSegmSignature();

  virtual char*    getBin(PICULONG &size) const;            // dummy
  virtual PICBOOL  setBin(const char *data, PICULONG size); // dummy
  virtual PICBOOL  isNull() const;
  virtual void     setNull();
  virtual BaseString getImage() const;

//:var
  PICULONG segmAlgId;
  PicArray segmentSigs;
/*
SegmAlgId is the id of the segmentation algorithm used.
SegmentSigs contains the signatures for segments.
*/

};

/**
  * PicSketchAlgorithm is the base class for all search algorithms
  * that use segments as basis for comparison.
  **/

class PicSketchAlgorithm : public PicObject
{
 public:
  PICABSOBJECTCLASS(PicSketchAlgorithm)

  PicSketchAlgorithm();

  /**
    * Calls computeSegmSignatures to ensure that a PicSegmSignature
    * is returned.
    *
    * @param pic Reference to a PicPicture object.
    * @return Pointer to a PicSignature object.
    **/
  virtual PicSignature     *computeSignature(PicPicture &pic);

  /**
    * Implementations of this function should compute signatures for 
    * segments of the given image. Return value should be a newly allocated
    * PicSegmSignature with segmAlgId set to the id of the segmentation
    * algorithm used and with segmentSigs containing the actual signatures
    * for the segments (in the same order as the segments are in the 
    * segmentation).
    *
    * @param pic Reference to a PicPicture object.
    * @return Pointer to a PicSegmSignature object.
    **/
  virtual PicSegmSignature *computeSegmSignatures(PicPicture &pic) =0;
  /**
    * Since function createEmpty for a PicSketchAlgorithm should return
    * an empty instance of PicSegmSignature, this function is needed to 
    * get a dummy instance of the actual PicSignature-based signature
    * class. That is an instance of that segment signature class you 
    * are putting into PicSegmSignature's segmentSigs-array.
    *
    * @return Pointer to a PicSignature object.
    **/
  virtual PicSignature     *createEmptyNonSegmSig() =0;

};

#endif

