// **********************************************************************
//
//      PICClient.h
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICCLIENT_H_
#define PICCLIENT_H_ 

class PICClient;
                
#include <PICDebug.h>
#include <PICNet.h>
#include <PICServerDataBase.h>
#include <parray.hxx>
#include <mmain.hxx> 
#include <signal.h>

#include <mess.h>

#define AUTHORIZED   0
#define UNAUTHORIZED 1
#define SERVERBUSY   2
#define PORTNUMBER 10000
#define SOCKETINITERROR 10
#define SOCKETREADERROR 11
#define INITIALIZATIONERROR 12 


/**
  * PICClient offers network services to PicGUI.
  */

class PICClient
{

// =============================================================
//  PUBLIC
// =============================================================

private:

  static void SignalHandler(int); 

	//
	// New in v3! Returns the database id corresponding to the given
	// name, or -1 if no database by that name was found.
	//
	static long dbNameToId(const char *dbName);

public:	

  //SP: added this hack to get around Translate() in server.
  //static bool SendMessageEscape();
  
  /// Information about servers databases and users.  
  static PICServerData svrData;

  /// the PICNet object
  static PICNet net;  

  /// xdr stuff.
  static xdrHandler *xdrH;

  /// Connection variable (initialized in HandleConnection)
  static bool connected;

  /// Current database id (insert).
  static long curDb;

  /// Latest query database id (delete, ReadImageFromDb).
  static long latDb;

  // For PICCache information
  static cacheArrayDescr *descArray;

  /** Sends the querymessage with query picture.
    *
    * @param algName Algorithm name.
    * @param pic Query picture.
    *
    * @return True, if the message was sent successfully, otherwise false.  
    */
  static PICBOOL StartQuery(BaseString algName, PicPicture &pic);

  /** Pack general options to the query packet.
    *
    * @param packet Query packet.
    * @param generalOptions Array of general options.
    *
    * @return True on ok, otherwise false.
    */
  static PICBOOL AddGeneralOptions(query_msg &msg, 
                                   PicArray &generalOptions);

  /** A client uses this message to abort a previously started query.
    *
    * @return True, if the message was sent successfully, otherwise false. 
    */
  static PICBOOL StopQuery(void);

  /** To get more result thumbnails, the client sends this message 
    * to the server and the server sends the next n result thumbnails
    * from the numOfFirst. This message contains no more information.
    *
    * @param numOfFirst Ranking number of the first thumbnail to get. 
    * @param genOpts General options of the client.
    *
    * @return True, if the message was sent successfully, otherwise false.
    */
  static PICBOOL GetMoreThumbNails(int numOfFirst, PicArray &genOpts);

  /** This message is used to retrieve an image from the server database.
    *
    * @param imageId ID of the picture in the database to retrieve. 
    * @param indexAt tells the location where to store the data
	*
	* @return Pointer to the location, NULL if failed.    
    */
  static PicPicture *ReadImageFromDb(PICULONG imageId, PICULONG indexAt);

  /** This message is used to insert a picture to the database on the server. 
    *
    * @param pic Picture.
    * @param dbId ID of the database which the picture will be inserted to.
    * 
    * @return True, if the message was sent successfully, otherwise false.  
    */
  static PICBOOL InsertImageToDb(PicPicture &pic);

  /** A client uses this message to delete a picture from a database 
    * on a server. 
    * 
    * @param dbId ID of the database where to delete the picture from. 
    * @param imageId ID of the picture to delete in the database.  
    *
    * @return True, if the message was sent successfully, otherwise false.    
    */
  static PICBOOL DeleteImageFromDb(PICULONG imageId);

  /** Makes a connection to the database server.
    *
    * @param serverAddress Servers http address.
    * @param userName Users name.
    * @param password Users password.
    * 
    * @return Servers reply value:  AUTHORIZED = 0, UNAUTHORIZED = 1,
    *         SERVERBUSY = 2, SOCKETREADERROR = 11 (or undefined error)
    */
  static int Connect(BaseString serverAddress, 
			 BaseString userName,
			 BaseString password);

  /** Returns the names of databases the user has access to.
    *
    * @return the names of databases the user has access to.
    */
  static BaseStringArray GetDbNames(void);

  /** Returns all databases.
    *
    */
  static BaseStringArray GetAllDbNames(void);

  /** Returns list of the supported algoritms of the
    * selected database.
    *
    * @param dbName Database name.
    *
    * @return False if operation failed, otherwise true.
    */
  static BaseStringArray GetDbAlgorithms(BaseString dbName);

  /** Asks the list of all users.
    * Calls setDbUsers in PICGui with all users as a parameter.
    *
    * @return False if operation failed, otherwise true.
    */
  static BaseStringArray GetDbUsers();

  /** Asks the server and certain database ringths of a specific user.
    *
    * @param dbName Name of the database.
    * @param userName Users name.
    *
    * @return False if operation failed, otherwise true.
    */
//  static BaseStringArray GetDbUserRights(char* dbName, char* userName);

	//
	// New in v3! Returns a user's server authorization.
	//
	static BaseStringArray getUserServerAuthorization(char* userName);

	//
	// New in v3! Returns a user's database authorization.
	//
	static BaseString getUserDbAuthorization(char* userName, char* dbName);

  /** A client uses this message to break the network connection 
    * between itself and a server (Client also purges its 
    * datastructures).
    *
    */
  static PICBOOL Disconnect(void);

  /** This message is sent to the client when server's timeout 
    * alarm goes off. When a client receives this message, it can 
    * expect the network connection to the server to break 
    * immediately and it is not allowed to send any further requests 
    * to the server.
    */
  static void TimeOut();

// =============================================================
//
//   RECEIVE METHOD
//
// =============================================================
    
   /** Reads the type of the Servers message and calls
     * a method which will handle the message.
     *
     * @return False if there is nothing to read from the socket
     *         or user is not connected, otherwise true.
     */
  static bool ReadServerMessage();  

// =============================================================
//
//   Servers Messages and Replies
//
// =============================================================

  /** Reads the servers init message and initializes DataBaseArray
    * and AlgorithmArray.
    *
    * @param msg Received message.
    */
  static void ReceiveInit(init_msg &init);

  /** Closes the connection (closes the server socket).
    */
  static void CloseConnection(); 

  /** This message is used by a server as a progress report reply 
    * to Add/Remove Algorithms. One progress report message is sent 
    * after each algorithm has been added or removed from the 
    * database as specified in the request.
    *
    * @param msg Received message.
    */
  static void AlgorithmProgressReport(algProgress_msg &msg);

  /** This message is sent by a server during a query to report the 
    * progress of the query to a client. The message is sent after 
    * every 10 images scanned by the server. A client can expect some 
    * Query Progress Report messages after sending a StartQuery message
    * to the server.
    *
    * @param msg Received message.
    */
  static void QueryProgressReport(progReport_msg &msg);
   
  /** This message is sent to a client when a query has been 
    * completed by the server and some result images have been found 
    * to the query. This message is a reply either to Message 
    * StartQuery() or GetMoreThumbnails().
    *
    * @param msg Received message.
    */
  static void QueryResults(queryResults_msg &msg);
  
  /** This message is sent to a client as a result to a query when 
    * no matches with the specified parameters were found from a database.
    *
    */
  static void NoQueryMatches();

  /** This is a general failure message used by the server to 
    * inform a client that it could not complete the request sent 
    * by the client. 
    *
    */
  static void FailedToCompleteRequest();
  
  /** The server sents this message to a client as a reply to 
    * GetMoreThumbnails, if no more matching images 
    * were found from the database during the previous query. 
    *
    */
  static void NoMoreQueryResults();

  /** This is used by the server as a general success message 
    * when an operation has been successfully completed and the 
    * server does not need to send the client any further information.
    *
    */
  static void OperationComplete();
  
  /** This message is sent by the server it has received 
    * GetPicture() from a client. This message contains the 
    * full sized image data for the requested image. 
    *
    * @param msg Received message.
    * @param exPic Reference to picture data.
    *
    * @return True, if the message was read successfully, otherwise false.  
    */
  static bool PictureData(picData_msg &msg, PicPicture &exPic);

  /** This message contains the usernames of all users registered 
    * on the server and their authorization levels. This message 
    * is a reply to GetUsers().
    *
    * @param msg Received message.
    */
  static void UserData(userData_msg &msg);

  /** This message is used by the server as a reply to AddUser()
    * to inform the client about the new user's authorization levels.    
    *
    * @param msg Received message.
    */
  static void NewUserRights(newUser_msg &msg);
  
  /** A client process may get this message from a server if it 
    * does not have superuser server authorization level to the 
    * server and the user tried to change his/her password but 
    * misspelled the old password which is required from normal 
    * users to change their password.
    */
  static void PasswordIncorrect();
  
  /** This message is sent to the client by the server when a 
    * new database is either created (CreateDatabase()) or added 
    * (AddDatabase()).
    *
    * @param msg Received message.
    */
  static void NewDatabase(newDB_msg &msg);
  
  /** This message contains all options for the server. This 
    * message is a reply to Message GetServerOptions().
    *
    * @param msg Received message.
    */
  static void ServerOptions(servOpts_msg &msg);

  /** This message is sent by the server it has received 
    * InsertPicture() from a client.
    *
    * @param msg Received message.
    */
  static void NewPictureId(newPic_msg &msg);

// =============================================================
//
//   Clients Requests
//
// =============================================================

  /** A client uses this message to get more result thumbnails from a 
    * server after a query has been completed and the server sends the 
    * next n result thumbnails. This message contains no  more information. 
    *
    * @param Ranking number of the first thumbnail to get.
    * @param Number of thumbnails to get. 
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool GetMoreThumbnails(int first, int numThumbs);

  /** This message is used to insert a picture to the database on 
    * the server. A batch insert is also done with this message - a 
    * client will send this message to the server for each picture to be
    * inserted to the database.
    *
    * @param filename Picture filename.
    * @param size Size (in bytes) of picture data.
    * @param data Picture data. 
    * @param param1 Picture parameter string 1.
    * @param param2 Picture parameter string 2.    
    * @param dbId Database id of the database .
    * @param picId Picture id.
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool InsertPicture(const char *filename, long size, 
                            const char *data, const char *param1, 
                            const char *param2, long dbId, long picId); 
 
  /** This message is used to retrieve an full sized image from the 
    * server database.
    *
    * @param databaseId ID of the database where to get the image from. 
    * @param pictureId ID of the picture in the database to retrieve. 
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool GetPicture(long databaseId, long pictureId);

  /** A client uses this message to delete a picture from a database 
    * on a server. The picture file itself is not deleted from the 
    * server (provided that it is separate from the database file)
    * but its entry from the database will be removed.
    *
    * @param databaseId ID of the database where to delete the picture from. 
    * @param pictureId ID of the picture to delete in the database. 
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool DeletePicture(long databaseId, long pictureId); 

  static bool AddAlgorithm(char *libPath, char *paramWinPath, int algClass);
  static bool RemoveAlgorithm(char *algName);
  static bool ModifyAlgorithm(char *libPath, char *paramWinPath,
  							  char *algName, int algClass, bool recalcSigs);

  /** This message is used by the client when on a superuser server 
    * authorization level to add or remove algorithms used in a database.
    *
    * @param databaseId ID of the database which to add or remove algorithms.
    * @param algArray Algorithms of this database
	*
	* @return True if the packet was sent successfully. Otherwise false.
    */
  static bool setDatabaseAlgorithms(char *dbName, PicArray &algArray);

  /** A client uses this message when invoking the user management 
    * section of the client. This request returns (if successful) a 
    * list of all users on the server and their authorization levels. 
    */
  static bool GetUsers();
  
  /** Used by the client to register changes done in the user management 
    * to the server. The message contains information about all users 
    * whose information has been somewhat altered.
    *
    * @param dbName Database.
    * @param userName Username.
    * @param authSrvr Server authorization level. 
    * @param authDb Database authorization level. 
    * @param cPassword can the user change password.
    *
    */
//  static bool ChangeUserRights(char* dbName, char *userName,
//			       int authSrvr, int authDb, int cPassword );

	//
	// New in v3! Set a user's database access rights.
	//
	static PICBOOL changeUserDbAuthorization(char *userName, char *dbName,
		int dbAuth);
	//
	// New in v3! Set a user's server access rights.
	//
	static PICBOOL changeUserServerAuthorization(char *userName,
		int serverAuth, int canChangePw);


  /** Cancels the changes made in the user management dialog
    */
  static void CancelRights();

  /** Sends the changes in the user management dialog to the server.
    * 
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool AcceptRights();

  /** A client uses this message to add an username and a password 
    * to the server. The new user will be given the anonymous user 
    * account authorization levels on the server and all databases 
    * (the levels will be returned by the server).
    *
    * @param userName Username. 
    * @param userName Password.            
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool AddUser(char *userName, char *passWord);
  
  /** When a server receives this message from a client, it removes 
    * the specified user from the userlist of the server thus denying 
    * any further connection attempts with the specified username.
    *
    * @param userName Username.     
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool RemoveUser(char *userName);

  /** This message is used by a client to change a user's password 
    * to a new one. This can either be used by a superuser level 
    * client which can change any user's password or an normal user 
    * to change his/her own password.
    *
    * @param userName Username. 
    * @param newPass New password. 
    * @param oldPass Old password (not checked if client has 
    *        superuser server authorization level).
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool ChangePassword(char *userName, char *newPass, char *oldPass);

  /** The client can create a new database to the server with this message. 
    *
    * @param path Directory name to create for the database. 
    * @param databaseName Database name to use on the server. 
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool CreateDatabase(char *path, char *databaseName);

  /** This message informs the server to physically delete a 
    * database from the server and remove it from the database list. 
    * If the client wishes to only remove the database from the 
    * server's database list but not physically delete the database, it uses
    * RemoveDatabase().
    *
    * @param dbName Name of the database to be deleted. 
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool DeleteDatabase(BaseString dbName);    
  
  /** With this message the client can obtain the server's options.
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool GetServerOptions();
  
  /** The client can change the server options with this message.
    *
    * @param maxClients Maximum number of clients.
    * @param timeout Timeout value.
    * @param httpAddress URL of the servers homepage.
    * @param commercial Servers advertisement text.
    *
    * @return True if the packet was sent successfully. Otherwise false.
    */
  static bool SetServerOptions(char *maxClients, char *timeout, 
                               char *httpAddress, char *commercial);

  /** The PICClient returns the server's http address to PicGUI.
    */
  static BaseString GetAddress();

  /** The PICClient returns the server's advertisement to PicGUI.
    */
  static BaseString GetServerAdvertisement();
  
  /** Writes the description of the algoithm to the info screen.
    * 
    * @param alg the name of the algorithm.
    *
    * @return true if the algorithm was found. Otherways false.
    */
  static BaseString GetAlgorithmInfo( char* alg );

  /** Purges client's datastructures.
    */
  static bool PurgeDataStructures();

  /** Returns init data to Tcl/Tk.
    */
  static BaseStringArray GetInitData();

   /** The constructor.
     * 
     */
   PICClient();
   
  /** The destructor.
    */
  ~PICClient();
  
  static bool Shutdown();

  /** This will handle a login to the selected server.
    *
    * @param host Hosts address.  
    * @param port Port number.
    * @param userName Users Name.
    * @param passWord Users passWord.  
    *
    * @return Server's reply: AUTHORIZED = 0, UNAUTHORIZED = 1,
    *   SERVERBUSY = 2, SOCKETREADERROR = 11 (or undefined error)  
    */
  static int HandleConnection(char* host, int port, char* userName, 
                              char* passWord);

  static int xdrTranslate(void);

	static int algNameConvToTk(int algId, char *oldName, char *newName);

};

#endif // PICCLIENT_H_   
