// **********************************************************************
//
//      PICUser.cpp
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <PICClient.h>
#include <PICServerDataBase.h>
#include <mmain.hxx>
#include <message.h>

PicArray changed;
PicArray usr;

// =============================================================
// a debugging procedure
void printUserArray( PicArray* a ) {
  PicIterator ii;

  ITERATE( *a, ii ) {
    PICUserInfo &info = AS(PICUserInfo, a->at(ii));
    DEBUG_MSG("[CLNT] " << info.name.chars() <<" "<< info.authLevel );
    if( info.canChangePassword == 1 ) DEBUG_MSG(" 1");
    if( info.canChangePassword == 0 ) DEBUG_MSG(" 0");
    DEBUG_MSG(endl);

    PicIterator iii;
    ITERATE( info.dbRights, iii ) {
      PICDbRight &dbRight =
	AS(PICDbRight, info.dbRights.at(iii));
      DEBUG_MSG("[CLNT] dbAuth:" << dbRight.id << " " 
		<< dbRight.authLevel << endl);
    }
  }
}

// copy constructor of the PICUserInfo class is not invoked when PicArray is
// copied, so the copying has to be done manually. The operator= works fine
// for the drRight class.
void copyUserInfo( PICUserInfo* a, PICUserInfo* b )
{
  b->name = a->name;
  b->authLevel = a->authLevel;
  b->canChangePassword = a->canChangePassword;
  b->dbRights = a->dbRights;
}

// =============================================================

//
// New in v3! An auxiliary method to convert database names into
// corresponding database id's.
//
long PICClient::dbNameToId(const char *dbName) {
  long dbId = -1;
  
  //
  // Search for the id of the database with dbName.
  //
  PicIterator iter;
  ITERATE(svrData.allDatabases, iter) {
    PICServerDbData &dbData =
      AS(PICServerDbData, svrData.allDatabases.at(iter));
    
    if (!strcmp(dbData.name.chars(), dbName)) {
      dbId = dbData.id;
    }
  }
  return dbId;
}

// =============================================================

BaseStringArray PICClient::GetDbUsers() {
  DEBUG_MSG("[CLNT] in GetDbUsers\n");

  usr.deleteAll();
  changed.deleteAll();

  // copy the contents of the svrData.users into usr.user
  // set every usr.changed into zero.
  // add every username into the results array, and return
  // the data for the GUI

	//
	// !V3 NOTE!
	// This method returns _nothing_ to the gui, if there
	// are no databases and thus no users are visible to the GUI.
	// This is a quirk of v2, and its effects can be avoided by
	// creating at least one database before invoking user management
	// in the GUI.
	//

  // count
  BaseStringArray results;

  DEBUG_MSG("[CLNT] users " << svrData.users.size() << endl);

  // read all data from the svrData.users
  PicIterator iter;
  ITERATE( svrData.users, iter )
  {
    PICUserInfo newInfo;
    usr.add( newInfo );
  }

  ITERATE( svrData.users, iter )
  {
    PICUserInfo &info = AS(PICUserInfo, svrData.users.at(iter));
    PICUserInfo &newInfo = AS(PICUserInfo, usr.at(iter));
    copyUserInfo( &info, &newInfo );
    results.add( newInfo.name );
  }

  // only for debug...
  printUserArray( &svrData.users );
  printUserArray( &usr );

  return results;
}

// =============================================================

//
// New in v3! This method combined with getUserDbAuthorization()
// replaces the functionality of GetDbUserRights().
//
BaseStringArray PICClient::getUserServerAuthorization(char* userName) {
  DEBUG_MSG("[CLNT] In getUserServerAuthorization.\n");
  
  BaseStringArray userAuthorization;

  //
  // Search for the user's authorization information.
  //
  char serverAuth[2] = {0,0};
  char canChangePassword[2] = {0,0};

  PicIterator iter;
  ITERATE(usr, iter) {
    PICUserInfo &u = AS(PICUserInfo, usr.at(iter));
    
    if( !strcmp(u.name.chars(), userName) ) {
      
      //
      // Server authorization level.
      //
      *serverAuth = (char)('0'+u.authLevel);
      userAuthorization.add(serverAuth);

      //
      // Can the user change his/her password?
      //
      *canChangePassword = u.canChangePassword == 1 ? '1' : '0';
      userAuthorization.add(canChangePassword);
    }
  }
  return userAuthorization;  
}


//
// New in v3! This method combined with getUserServerAuthorization()
// replaces the functionality of GetDbUserRights().
//
BaseString PICClient::getUserDbAuthorization(char* userName, char* dbName) {
  DEBUG_MSG("[CLNT] In getUserDbAuthorization.\n");

  long dbId = dbNameToId(dbName);
  BaseString userDbAuthorization;  

  //
  // Bail out if there is no database by the given name.
  //
  if (dbId < 0) {
    DEBUG_MSG("[CLNT] Db not found.\n");
    return userDbAuthorization;
  }

  //
  // Search for the user's database authorization.
  //
  bool iterationOk = false;
  char dbAuth[2] = {'0',0};

  PicIterator iter2;
  ITERATE(usr, iter2) {
    PICUserInfo &u = AS(PICUserInfo, usr.at(iter2));
    
    if( !strcmp(u.name.chars(), userName)) {
      DEBUG_MSG("[CLNT] looking for db: " << dbName << " " << dbId << endl);
      
      PicIterator iter3;
      ITERATE(u.dbRights, iter3) {
	PICDbRight &dbRight = AS(PICDbRight, u.dbRights.at(iter3));
	
	if (dbRight.id == dbId) {
	  *dbAuth = (char)('0'+dbRight.authLevel);
	  iterationOk = true;
	}
      }
    }
  }
  
  //
  // If no authorization was found, the default value ('0') gets.
  //
  if (!iterationOk) {
    DEBUG_MSG ("[CLNT] No rights defined for the db." << endl);
  }
  userDbAuthorization = dbAuth;
  
  return userDbAuthorization;  
}

// =============================================================
//
//   Servers Replies
//
// =============================================================

void PICClient::UserData(userData_msg &msg)
{
  DEBUG_MSG("[CLNT] In UserData Method..."<<endl);

  // Number of users.
  unsigned int userCount = msg.users.users_len;
  DEBUG_MSG("[CLNT] " << userCount << " users in this server" << endl);

  svrData.users.deleteAll();
  for(unsigned int i = 0; i < userCount; i++)
  {
    PICUserInfo usrInfo;    
    svrData.users.add( usrInfo );
  }

  PicIterator iterator;
  unsigned int userc=0;
  ITERATE( svrData.users, iterator )
  {
    PICUserInfo &info = AS(PICUserInfo, svrData.users.at(iterator));
    const UserRights &user=msg.users.users_val[userc++];

    // Username.
    info.name = strdup(user.name);
    DEBUG_MSG("[CLNT] " << info.name.chars() << endl );

    // Server authorization level. 
    info.authLevel = user.authLevel;

    // Can user change password or not?
    info.canChangePassword = user.changePW;

    // Number of databases where user has rights.
    unsigned int dbCount = user.DBs.DBs_len;
    for(unsigned int i = 0; i < dbCount; i++ ) {
      PICDbRight dbRight;
      info.dbRights.add(dbRight);
    }

    PicIterator iterator2;
    unsigned int DBc=0;
    ITERATE( info.dbRights, iterator2 )
    {
      const DBRights &DB=user.DBs.DBs_val[DBc++];

      PICDbRight &right = AS(PICDbRight, info.dbRights.at(iterator2));
      right.id = DB.ID;
      right.authLevel = DB.auth;
      DEBUG_MSG("[CLNT] client db auth to db " << right.id << " is " << 
		right.authLevel << endl );
    }
  }
  printUserArray( &svrData.users );
}

// =============================================================

void PICClient::NewUserRights(newUser_msg &msg)
{
  DEBUG_MSG("[CLNT] In NewUserRights Method..." << endl);

  PICUserInfo newUser;
  const UserRights &user=msg.user;

  // Username.
  newUser.name = strdup(user.name);

  // adding a PICUserInfo to PicArray seemed to copy only the name and the 
  // authLevel fields, so the copying is done manually. First a blank 
  // PICUserInfo is added with only the name filled in. Then the object is
  // searched (according to the name) from the array, and the other fields
  // are filled in correctly.
  // The same thing is done with the usr array (the buffer array that has
  // the knowledge of the wanted changes.
  // FIXME: now _that_ is a royal mess.

  svrData.users.add( newUser );
  usr.add( newUser );

  PicIterator ii;

  ITERATE( svrData.users, ii )
  {
    PICUserInfo &info = AS( PICUserInfo, svrData.users.at( ii ) );
    if( info.name == newUser.name ) {
      info.authLevel = user.authLevel;
      info.canChangePassword = user.changePW;
      unsigned int dbCount = user.DBs.DBs_len;
      if (dbCount > 0) {
	for (unsigned int j = 0; j < dbCount; j++) {
	  PICDbRight dbRight;
	  const DBRights &DB=user.DBs.DBs_val[j];

	  dbRight.id = DB.ID;
	  dbRight.authLevel = DB.auth;
	  info.dbRights.add(dbRight);
	}
      }
      PicIterator iii;
      ITERATE( usr, iii )
      {
	PICUserInfo &info2 = AS( PICUserInfo, usr.at( iii ) );
	if( info2.name == info.name ) {
	  copyUserInfo( &info, &info2 );
	}
      }
    }
  }

  // print debug
  printUserArray( &svrData.users );
  printUserArray( &usr );
}  

// =============================================================

void PICClient::PasswordIncorrect()
{
  INFO_PRINT("PASSWORD INCORRECT. Password not changed.\n");
} 


// =============================================================
// 
//   Clients Requests
//
// =============================================================

bool PICClient::GetUsers()
{
  getUsers_msg msg;

  DEBUG_MSG("[CLNT] Writing GetUsers request.\n");
  xdrH->send(&msg);
  return true;
}

// =============================================================

bool PICClient::AddUser(char *userName, char *passWord)
{
  addUser_msg msg;

  msg.name=userName;
  msg.passW=passWord;
  xdrH->send(&msg);
  DEBUG_MSG("[CLNT] AddUser message sent.\n");
  return true;
  // adding to the memory structures happends when the user information is 
  // returned from the server in NewUserRights.
} 

// =============================================================

bool PICClient::RemoveUser(char *userName)
{
  deleteUser_msg msg;

  msg.name=userName;
  xdrH->send(&msg);
  DEBUG_MSG("[CLNT] Remove message sent.\n");
  PicIterator i;
  ITERATE( svrData.users, i )
  {
    PICUserInfo &u = AS( PICUserInfo, svrData.users.at( i ) );

    if( !strcmp(u.name.chars(), userName) ) {
      u.name = "";
      DEBUG_MSG("[CLNT] Removing user from svrData\n");
    }
  }
  ITERATE( usr, i )
  {
    PICUserInfo &u = AS( PICUserInfo, usr.at( i ) );

    if( !strcmp(u.name.chars(), userName) ) {
      u.name = "";
      DEBUG_MSG("[CLNT] Removing user from usr\n");
    }
  }

  ITERATE( changed, i )
  {
    PicString &u = AS( PicString, changed.at( i ) );

    if( !strcmp(u.chars(), userName) ) {
      u = "";
      DEBUG_MSG("[CLNT] Removing user from changed\n");
    }
  }

  return true;
}

// =============================================================

bool PICClient::ChangePassword(char *userName, char *newPass, char *oldPass) {
  changePW_msg msg;

  msg.name=userName;
  msg.newPW=newPass;
  msg.oldPW=oldPass; 
  //FIXME: there might be a problem here with a lacking null-terminator.
  xdrH->send(&msg);
  DEBUG_MSG("[CLNT] ChangePassword message sent.\n");
  return true;
}

// =============================================================
/*
bool PICClient::ChangeUserRights(char* dbName, char *userName,
			       int authSrvr, int authDb, int cPassword )
{
  DEBUG_MSG("[CLNT] In ChangeUserRights\n");

  long dbId = -1;
  bool iterationOk = false;

  // Search id of the database with dbName.
  PicIterator iter;
  ITERATE(svrData.allDatabases, iter)
  {
    PICServerDbData &dbData =
      AS(PICServerDbData, svrData.allDatabases.at(iter));

    if (!strcmp(dbData.name.chars(), dbName))
    {
      // Id of the selected database.
      dbId = dbData.id;
    }
  }

  // If the database not found...
  if (dbId < 0) {
    DEBUG_MSG("[CLNT] DBname not found\n");
    return false;
  }

  // Search users rights.
  
  PicIterator ii;

  ITERATE( usr, ii )
  {
    PICUserInfo &u = AS( PICUserInfo, usr.at( ii ) );
    if( !strcmp(u.name.chars(), userName) )
    {
      PicIterator iii;
      bool iterOk = false;
      ITERATE( changed, iii )
      {
	PicString &s = AS( PicString, changed.at( iii ) );
	if( s == u.name ) {
	  iterOk = true;
	}
      }
      if( !iterOk ) changed.add( u.name );

      // Server authorization level.
      u.authLevel = authSrvr;

      PicIterator iter1;
  
      ITERATE(u.dbRights, iter1)
      {
        PICDbRight &dbRight = AS(PICDbRight, u.dbRights.at(iter1));
        if (dbRight.id == dbId)
        {
          // Database authorization level.
	  dbRight.authLevel = authDb;
          iterationOk = true;
        }
      }
      if (!iterationOk)
      {
	DEBUG_MSG ("[CLNT] no previously defined rights." << endl);
	PICDbRight newDbRight;
	newDbRight.id = dbId;
	newDbRight.authLevel = authDb;

	u.dbRights.add( newDbRight );
	//	return false;
      }

      // Server authorization level.
      u.canChangePassword = (char) cPassword;
    }
  }
  return true;
}
*/

//
// New in v3! This method combined with changeUserServerAuthorization()
// replaces the functionality of ChangeUserRights().
//
PICBOOL PICClient::changeUserDbAuthorization(char *userName, char *dbName,
					     int dbAuth) {
  long dbId = dbNameToId(dbName);
  bool iterationOk = false, iteration2Ok = false;

  DEBUG_MSG("[CLNT] In changeUserDbAuthorization().\n");

  //
  // If the database was not found...
  //
  if (dbId < 0) {
    DEBUG_MSG("[CLNT] Database id for '" << dbName << "' not found.\n");
    return boolFalse;
  }

  //
  // Search for the user's rights.
  //
  PicIterator iter;
  ITERATE(usr, iter) {
    PICUserInfo &u = AS( PICUserInfo, usr.at(iter) );
    if( !strcmp(u.name.chars(), userName) ) {
      PicIterator iter2;
      ITERATE(changed, iter2) {
	PicString &s = AS( PicString, changed.at(iter2) );
	if( s == u.name ) {
	  iteration2Ok = true;
	}
      }
      if (!iteration2Ok) {
	changed.add( u.name );
      }
      
      PicIterator iter3;
      ITERATE(u.dbRights, iter3) {
	PICDbRight &dbRight = AS(PICDbRight, u.dbRights.at(iter3));
	if (dbRight.id == dbId) {
	  dbRight.authLevel = dbAuth;
	  iterationOk = true;
	}
      }
      if (!iterationOk) {
	DEBUG_MSG ("[CLNT] no previously defined rights." << endl);
	PICDbRight newDbRight;
	newDbRight.id = dbId;
	newDbRight.authLevel = dbAuth;
	
	u.dbRights.add( newDbRight );
      }
    }
  }
  return boolTrue;
}

//
// New in v3! This method combined with changeUserDbAuthorization()
// replaces the functionality of ChangeUserRights().
//
PICBOOL PICClient::changeUserServerAuthorization(char *userName,int serverAuth,
						 int canChangePw) {
  bool iterationOk = false, iteration2Ok = false;
  
  DEBUG_MSG("[CLNT] In changeUserServerAuthorization().\n");
  
  //
  // Search for the user's rights.
  //
  PicIterator iter;
  ITERATE(usr, iter) {
    PICUserInfo &u = AS( PICUserInfo, usr.at(iter) );
    if( !strcmp(u.name.chars(), userName) ) {
      iterationOk = boolTrue;
      PicIterator iter2;
      ITERATE(changed, iter2) {
	PicString &s = AS( PicString, changed.at( iter2 ) );
	if( s == u.name ) {
	  cout << "user found in 'changed'." << endl;
	}
      }
      if (!iteration2Ok) {
	changed.add( u.name );
      }
      
      //
      // Server authorization level.
      //
      u.authLevel = serverAuth;
      u.canChangePassword = (char)canChangePw;
    }
  }
  return iterationOk;
}

// =============================================================
// clear the usr datastructure from the previous changes.
void PICClient::CancelRights() {
  DEBUG_MSG("[CLNT] In cancel rights (" << usr.size() << " users)\n");
  PicIterator i;

  ITERATE( usr,i )
  {
    PICUserInfo &u = AS( PICUserInfo, usr.at( i ) );
    u.dbRights.deleteAll();
  }
  usr.deleteAll();
}

// =============================================================

bool PICClient::AcceptRights() {
  changeRights_msg msg;

  //FIXME: I think this'll leak mem.
  msg.users.users_len=changed.size();
  msg.users.users_val=new UserRights[msg.users.users_len];

  // flag that tells wether the message was constructed correctly or not.
  bool send = true;

  PicIterator i;
  unsigned int userc=0;
  ITERATE( changed, i )
  {
    PicString &n = AS( PicString, changed.at( i ) );

    // if the item was "" decrement the count.
    if(!strcmp(n.chars(),"")) {
      msg.users.users_len--;
    } else {
      PicIterator ii;
      // go trough all changed items in the usr array
      ITERATE( usr, ii ) {
	PICUserInfo &u = AS( PICUserInfo, usr.at( ii ) );
   
	if( u.name == n ) {

	  // find the same user from the svrData
	  PicIterator iiii;
	  bool iterateOk = false;
	  ITERATE( svrData.users, iiii ) {
	    PICUserInfo &svrInfo = AS(PICUserInfo, svrData.users.at( iiii ));
	    
	    if( svrInfo.name == u.name ) {
	      // User's name.
	      UserRights &usr=msg.users.users_val[userc++];
	      
	      usr.name=(char *)u.name.chars();
	      svrInfo.name = u.name;
	  
	      // User's authorization level.
	      usr.authLevel=(ServerAuth)u.authLevel;
	      svrInfo.authLevel=u.authLevel;
	      
	      // Can user change password or not?
	      usr.changePW=u.canChangePassword;
	      svrInfo.canChangePassword = u.canChangePassword;
	      
	      // Number of databases where user has rights 
	      usr.DBs.DBs_len=u.dbRights.size();
	      usr.DBs.DBs_val=new DBRights[usr.DBs.DBs_len];

	      unsigned int DBc=0;
	      PicIterator iii;
	      ITERATE(u.dbRights, iii)
	      {
		PICDbRight &dbRight = AS(PICDbRight, u.dbRights.at(iii));
		DBRights &DB=usr.DBs.DBs_val[DBc++];

		DB.ID=dbRight.id;
		DB.auth=(DatabaseAuth)dbRight.authLevel;
	      }
	      svrInfo.dbRights = u.dbRights;
	      iterateOk = true;
	    }
	  }
	  if( !iterateOk ) send = false;
	}
      }
    } 
  }

  if(send && msg.users.users_len > 0) {
    xdrH->send(&msg);
    DEBUG_MSG("[CLNT] ChangeUserRights message sent.\n");     
  } else {
    DEBUG_MSG("[CLNT] ChangeUserRights message NOT sent!.\n");     
    DEBUG_MSG("[CLNT] Number of users " << msg.users.users_len << endl);
  }
  return true;
}












