// **********************************************************************
//  PicMain class definition
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef MMAIN_HXX
#define MMAIN_HXX

#include <palloc.hxx>

#include <PicSketchAlg.hxx>
#include <PicSegmentAlgorithm.hxx>
#include <PicImageFormat.hxx>
#include <searchtype.hxx>

#include <PicGUI.hxx>

#include <mshared.hxx>
#include <merrors.hxx>
#include <mthreads.hxx>
#include <mtools.hxx>

/*:class ********************************************************************

PicMain is a placeholder class full of static member functions, 
most of which are called by PicGUI. One PicMain object is 
constructed in the main-function to initialize the system.

****************************************************************************/

class PicMain
{
 public:

  PICCLASS()
  PICERRORCODES {  ok,
    initFailed,
    unKnownImageFormat,
    unKnownSearchType,
    unKnownSearchAlg,
    unKnownSegmAlg,
    unKnownIPAlg,
    unKnownExampleImage,
    threadSuspendFailure,
    threadAbortFailure,
    invalidParams,
    imageInsertFailed,
    imageReadFailed,
    deleteImageFailed,
    exImageCreateFailed,
    exImageCloseFailed,
    exImageClearFailed
  };

//:fun
  PicMain();
  ~PicMain();
/*
Constructor initializes the system by creating 
    search types 
and available 
    PicImageFormat:s, 
    PicSearchAlgorithm:s and 
    PicSegmentAlgorithm:s 
*/

//:fun
  PICBOOL startGUI(int argc, char *argv[]);
/*
Starts the graphical interface by calling gui.start(...)
*/

//:fun
  static PICBOOL guiHasStarted(void);
/*
This function performs various start-up routines after the GUI 
is up. First the database is reserved and after a successful 
connection the segmentation and search algorithms are checked 
against database. If some algorithm is not found in the database,
the database update thread is launched.
PicGUI is returned the names of available search types. Before
returning, the application parameters are read from the database. 

This function is called by PicGUI as soon as it has successfully 
started up.
*/

//:fun
  static PICBOOL changeSearchType(BaseString searchTypeName,
    BaseStringArray &algNames);
/*
This function informs PicGUI about the search algorithms 
belonging to the given search type by setting the names of the
search algorithms to the BaseStringArray given as the 2nd parameter.

Called by PicGUI when the user has selected a search type.
Returns boolFalse on error, otherwise returns boolTrue.

arg: searchTypeName is the name of the selectedsearch type.
*/

//:fun
  static PICBOOL newExampleImage(PICULONG listIndex);
/*
This function adds a new (null) PicPicture to examplePics,
identified hereafter by listIndex.

Called by PicGUI when user tries to create a new example image. 
Returns boolFalse on error, otherwise returns boolTrue.
*/

//:fun
  static PICBOOL exImageClosed(PICULONG listIndex);
/*
This function destructs the example image identified by 
listIndex.

Called by PicGUI when user closes the active example image.
Returns boolFalse on error, otherwise returns boolTrue.
*/

//:fun
  static PICBOOL exImageCleared(PICULONG listIndex);
/*
This function sets the given example image to null, thus freeing
its PicRawPicture component and setting all parameters to 
default.

Called by PicGUI when user clears the active example image.
Returns boolFalse on error, otherwise boolTrue.
*/

//:fun 
  static PICBOOL setImageAttributes(PicIxCollect &params,
                                    PICULONG listIndex);
/*
This function sets the parameters of an example image.
Parameters are interpreted as follows:
  1 = name 
  2 = par1 
  3 = par2
  4 = width
  5 = height

Called by PicGUI when user has set image parameters. On success
returns boolTrue, otherwise returns boolFalse. The actual values
accepted can be found in argument variable params.
*/

//:fun 
  static PICBOOL loadImage(BaseString filename,
                           PicPicture &resultImage,
                           PICULONG   listIndex,
                           PICBOOL    updating);
/*
This function reads image data from a file to the given 
PicPicture. The filename attribute is also updated.

Called by PicGUI when user has selected a file to be loaded.
Returns boolFalse on error and sets result image into null state,
otherwise returns boolTrue.

arg: filename is the name of the file to be read, suffix is used
     as the primary means to identify the image format
arg: resultImage contains the image data after loading
arg: listIndex is the index of the example image in question
     NOTE: listIndex==0 means that an image was grabbed from 
     an editor window for image processing. In this case the
     example image is not touched.
arg: updating means when true, that the image is actually being
     exported from an image editor and this is not an ordinary
     load
*/

//:fun
  static PICBOOL saveImage(BaseString filename,
                           PICULONG listIndex,
                           PICBOOL overwrite);
/*
This function writes (synchronously) the given image into a file.
If the file already exists, a confirmation is asked from the user 
(with PicGUI::simpleDialog). The type of the output file is 
determined by filename's extension, see PicPicture::writeToFile.

Called by PicGUI when user has given a filename to which the 
active example image should be written. Returns boolFalse on 
error, otherwise returns boolTrue and set image's image format
component and filename.

arg: filename is the name of the file to be written
arg: listIndex is the index of the example image in question
arg: buttonNbr is used to deliver the order number of the
     button the user pressed in the confirmation dialog window
*/

//:fun 
  static BaseString getSignatureImage(BaseString algName,
                                      PICULONG listIndex);
/*
This function gets a string representation for a signature 
produced by a search algorithm. If no signature is available, 
returns a descriptive text like "Not available".

Called by PicGUI when user has selected an algorithm in the 
"Signatures" -dialog window. 

arg: algName is the name of the search algorithm
arg: listIndex is the index of the example image in question
*/

	static PICDbAlgorithm *getSearchAlg(BaseString algName);

//:fun
//  static PICBOOL getSegmAlgParams(BaseString algName,
//                                  PicIxCollect &params);
//  static PICBOOL setSegmAlgParams(BaseString algName,
//                                  PicIxCollect &params);
/*
These functions set and get parameters of a segmentation 
algorithm.

Called by PicGUI when user opened / pushed "OK"-button in the 
(algorithm-specific) "Segmentation algorithm parameters"-dialog. 
Returns boolFalse on error, otherwise returns boolTrue. Results 
can be found in argument "params".

arg: algName is the name of the segmentation algorithm.
arg: params contains the parameters of the algorithm 
     (as PicString-objects), they are simply passed to the 
     algorithm implementation, which determines how to 
     interpret them.

NOTE: Even when setting parameters, they may be changed by the 
      function (since some validity checking may be done)
*/

//:fun
//  static PICBOOL segmentImage(BaseString algName,
//                              PicPicture &image,
//                              PICULONG listIndex);
/*
This function segments (synchronously) the given image with a 
segmentation algorithm. 

Called by PicGUI when user pushed "Execute"-button in the
segmentation dialog window. Returns boolFalse on error, 
otherwise returns boolTrue. Results may be queried by function 
getSegmentImage(...)

arg: algName is the name of the segmentation algorithm.
arg: image is the image to be segmented
arg: listIndex is the index of the example image in question
*/

//:fun 
//  static BaseString getSegment(BaseString algName,
//                               PICULONG listIndex,
//                               PICUSHORT &segNum);
/*
This function gets a string representation for a segment
produced by a segmentation algorithm. If no segment is available,
returns a descriptive text like "Not available".

Called by PicGUI when user has selected an algorithm in the 
"Segmentation" -dialog window or scrolled segments. 

arg: algName is the name of the segmentation algorithm.
arg: listIndex is the index of the example image in question
arg: segNum is order number of the segment (1 is the first).
     It may be changed if user tries to move past the end, so the 
     caller should check the value afterwards.
*/

//:fun 
  static PICBOOL browseQueryResults(PICUSHORT numOfFirst);
/*
This function may be used to scroll results of a finished query.

Called by PicGUI when user scrolls query's result images. 
Returns boolFalse if no finished query is available, otherwise 
returns boolTrue and uses 
  PicGUI::sendQueryResults(const PicIxCollect &images).

arg: numOfFirst is the order number (in query's result set) of 
     the first result image that should be visible after the
     scroll. For example numOfFirst==1 means that the best 
     match and the PicParams::dbImagesOnScreen images following
     it are to be displayed.
*/

//:fun 
  static PICBOOL suspendQuery();
/*
This function suspends the running query.

Called by PicGUI when user tries to suspend a query.
Returns boolTrue if a running query was really suspended,
otherwise returns boolFalse.

NOTE: the execution of the query can be resumed by calling 
      resumeQuery.
*/

//:fun 
  static PICBOOL resumeQuery();
/*
This function resumes a suspended query.

Called by PicGUI when user tries to resume a query.
Returns boolFalse on error, otherwise returns boolTrue.
i*/

//:fun 
/*
  static PICBOOL insertImageToDb(PicPicture &image,
                                 PICULONG listIndex,
                                 PICUSHORT buttonNbr);
*/
  static PICBOOL insertImageToDb(PicPicture &image, PICULONG listIndex,
    PICBOOL updating);
/*
This function insert the given image to database. If the given 
image has a non-zero database-id, a confirmation for insertion is
done ("Update / Save as new / Cancel"). After a successful 
operation the parameter image has a valid id.

Called by PicGUI when user has issued the command to insert the 
active example image into database. Returns boolFalse if the user
cancels the action or operation fails, otherwise returns boolTrue 
and uses PicGUI::sendQueryResults(const PicIxCollect &pics) to show 
(the new copy of) the example image also in a database image frame.

arg: image is the image to be inserted into database
arg: listIndex is the index of the example image in question
arg: buttonNbr is used to deliver the order number of the
     button the user pressed in the confirmation dialog window
*/

//:fun 
  static PICBOOL readImageFromDb(PICULONG imageId,
                                 PICULONG listIndex,
                                 PicPicture &resultImage);
/*
This function copies a database image into given image. All the
attributes, signatures and segmentations are also read.

Called by PicGUI when user has issued the command to copy 
database image into active example image. Returns boolFalse on
error, otherwise returns boolTrue.

arg: imageId is the image id of the desired image.
arg: listIndex is the index of the example image in question
arg: resultImage contains the image data and attributes
     after loading, on failure "resultImage.isNull()" is true.
*/

//:fun 
  static PICBOOL deleteImageFromDb(PICULONG imageId,
                                   PICUSHORT buttonNbr);
/*
This function deletes a database image from database a.

Called by PicGUI when user has issued the command to delete a
database image. Returns boolFalse on error, otherwise returns 
boolTrue (the caller should then clear the database frame).

arg: imageId is the image id of the desired image.
arg: buttonNbr is obsolete (confirmation dialog window handled
     within Tcl/Tk)
*/

//:fun 
  static PICBOOL batchInsert(BaseString filename, PICBOOL buttonNbr);

/*
This function starts the batch insert thread which inserts a set
of images into database. If the thread is already running, asks 
confirmation from the user. 

Called by PicGUI when user has given a filename for batch insert. 

arg: filename is interpreted as follow:
         1. if has suffix "LST" (case-insensitive) is a 
            file containing a list of filenames to insert
         2. if has a wildcard "*" insert all matching files
         3. otherwise insert just the given file
     filename's suffix is used as the primary means to identify
     the image format
arg: buttonNbr is used to deliver the order number of the
     button the user pressed in the confirmation dialog window

NOTE: this function does not wait until the thread has finished.
*/

//:fun 
  static PICBOOL getParams(PicIxCollect &params);
  static PICBOOL setParams(PicIxCollect &params);
/*
These functions set and get various parameters of the system.

Called by PicGUI when user has changed system parameters or
opened the "Options"-dialog (+ once on start-up). Returns 
boolFalse on error, otherwise returns boolTrue.

arg: params is a collection of parameters as PicString:s, 
     following indexes are checked:
        1 -- show results of aborted query, "0"/"1"
        2 -- ask confirmation when closing editor window 
	     with a changed image, "0"/"1"
        3 -- save database images physically to database, 
	     not only full pathnames, "0"/"1"
        4 -- show PicGUI debug mesgs, "0"/"1"
        5 -- show PicCore debug mesgs, "0"/"1"
        6 -- show PicAlg debug mesgs, "0"/"1"
        7 -- show PicDb debug mesgs, "0"/"1"
        8 -- default width of example images, PICUSHORT
        9 -- default height of example images, PICUSHORT

NOTE: Even when setting parameters, they may be changed if they 
are not in allowed range (e.g. image size too big), so the caller 
should update his values accordingly.
*/

//:fun
  static PICUSHORT getNumOfDbFrames(void);
/*
With this function PicGUI can ask PicCore for the number of
database frames PicGUI is able to show on the main screen.
*/

//:fun
  static void setNumOfDbFrames(PICUSHORT frames);
/*
With this function PicGUI informs PicCore about the number of the
database frames PicGUI is able to show on the main screen. After 
a query PicCore sends this many images as result images.
*/

//:fun 
  static PICBOOL exitProgram(void);
/*
This function determines whether it is ok to shut down the 
application. If a batch insert or database update thread is 
running, asks confirmation from the user. (If the user confirms
the exit) stops all running threads.

Called by PicGUI when user has requested the application to exit.
Returns boolFalse if it is not ok to exit GUI, otherwise returns 
boolTrue.

arg: buttonNbr is used to deliver the order number of the
     button the user pressed in the confirmation dialog window
*/

// ***************************************************************

//:var
  static PicMainSharedData globData;
/*
PicMain::globData contains the publicly available shared data.

see: PicMainSharedData
*/

//:fun
  static PICBOOL calcAllSegments(PicPicture &pic);
/*
Computes segmentations with all known (isAutomatic) segmentation
algorithms.
*/
//:fun
  static PICBOOL calcAllSignatures(PicPicture &pic);
/*
Computes signatures with all known search algorithms.
*/

//:fun
  static PICBOOL checkSegmAlgInDb(PicSegmentAlgorithm &alg);
  static PICBOOL checkSearchAlgInDb(PicSearchAlgorithm &alg);
/*
Check whether a given algorithm is found in the database.
*/

//:fun
  static void debugMesgGUI(BaseString mesg);
  static void debugMesgCore(BaseString mesg);
  static void debugMesgAlg(BaseString mesg);
  static void debugMesgDb(BaseString mesg);
  static void debugMesgCache(BaseString mesg);

  static void printErrors(void);
/*
Used to print to info screen.
*/

//:fun
  static PicGUI& getGUI();
/*
Can be used to retrieve a reference to the PicGUI.
*/

//:fun
  static void guiLeavingC();
  static void guiEnteringC();
/*
Used for controlling access to Tcl/Tk-functions (and thereby
Tcl/Tk's data variables).
*/

  static bool threadRunning();
  static void serveChild(PicThread &thread);

  static PicPicture *getExamplePic(PICULONG listIndex);
 
 private:

//:fun

  static PICBOOL abortThread(PicEnhancedThread &thread);
  static PICBOOL suspendThread(PicEnhancedThread &thread);

//:var
  static PicArray      examplePics;
  static PicArray      exampleListIndexes;
/*
Array of example images, exampleListIndexes holds the index
values used by PicGUI (those listIndex parameters in the API).
*/

  static PicGUI        gui;

  static PicMassInsertThread batchThread;
};

// **************************************************************************

inline PicGUI& PicMain::getGUI() { return gui; }

#endif
