// **********************************************************************
//  Shared data members
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef MSHARED_HXX
#define MSHARED_HXX

#include <parray.hxx>

class PicSearchType;
class PicSearchAlgorithm;
class PicSegmentAlgorithm;
class PicImageFormat;

/*:class ********************************************************************

PicParams contains the general parameters of the system.

****************************************************************************/

enum PICPARAMS_LIST {
  PICPARAMS_ASK_CHANGE_CONF = 1,
  PICPARAMS_GUI_DEBUG,
  PICPARAMS_CORE_DEBUG,
  PICPARAMS_ALG_DEBUG,
  PICPARAMS_DB_DEBUG,
  PICPARAMS_CACHE_DEBUG,
  PICPARAMS_DEF_WIDTH,
  PICPARAMS_DEF_HEIGHT,
  PICPARAMS_WWWBROWSER,
  PICPARAMS_NUM_OF_RESULT_IMAGES,
  PICPARAMS_HIT_PERC,
  PICPARAMS_STAMP_WIDTH,
  PICPARAMS_STAMP_HEIGHT,
  PICPARAMS_QUERY_SIZE,
  PICPARAMS_DB_IMAGE_FRAMES_ON_SCREEN,
  PICPARAMS_SHOW_ABORTED_QUERY,
  PICPARAMS_CACHE_SIZE,
  PICPARAMS_WINDOW_SIZE,
  PICPARAMS_WINDOW_POSITION_X,
  PICPARAMS_WINDOW_POSITION_Y,
  PICPARAMS_PLATFORM,
  PICPARAMS_PERFORMANCE_INDEX,
  PICPARAMS_NETWORK_PERFORMANCE_INDEX,
  PICPARAMS_SCALE_QUERY_IMAGE,
  PICPARAMS_ASK_ALGORITHM_PARAMETERS,
  PICPARAMS_SHOW_SERVER_ADVERT,
  PICPARAMS_IMAGE_EDITOR,
  PICPARAMS_LOGON_USER,
  PICPARAMS_THREAD_TIMEOUT,
  PICPARAMS_SERVER_LIST
};

/* BEGIN: New defines & constants introduced by PICSearch v3 */

typedef struct {
  PICULONG nonNull;
  PICULONG elements;
  PICULONG *descArray;
} cacheArrayDescr;

enum CONFIG_FILE_SECTIONS {
  GUI_PARAMS,
  RUNTIME_PARAMS,
  USED_SERVERS_PARAMS,
  USER_PARAMS,
  CACHE_PARAMS,
  QUERY_PARAMS,
  DEBUG_PARAMS,
  MISC_PARAMS
};

enum CONFIG_ERRORS {
  CONFIG_ERROR_NEW_CONFIG
};

// could be made to a command line option?
#define DEFAULT_CONFIG "picsearch.cfg"
// how many last used servers to have PICSearch remember
#define SERVER_LIST_SIZE 10

#define PER_CHAIN 8
#define PER_CHAIN_SHIFT 3 // 2^PER_CHAIN_SHIFT = PER_CHAIN
#define HASH_CONST 0.6180339887

#define MIN_CACHE_SIZE 64
#define MAX_CACHE_SIZE 10000

/* END: New defines & constants introduced by PICSearch v3 */

class PicParams
{
 public:
  /* BEGIN: New public methods introduced by PICSearch v3 */

  PicParams(); // constructor
  PicParams(char *); // constructor
  PicParams(const PicParams&); // initialization
  ~PicParams(); // destructor

  void closeConfig(); // because configFp in data section
  void restoreSettings(const PicParams&); // if not saved
  void showSettings(); // shows settings, debug only

  PicParams &operator=(const PicParams&); //  a = b

  PICBOOL defaultSettings(); // no config file, create defaults
  PICBOOL readSettings(); // reads settings from config file
  PICBOOL writeSettings(); // writes settings to config file
  PICBOOL isOptSec(char *); // is the line a section in the config
  PICBOOL isOpt(char *); // has the line a option
  PICBOOL getOpt(char *, char *); // get option
  PICBOOL getOptVal(char *, char *); // get value for option
  PICBOOL getOptValString(char *, char *); // get value for option (spaces)
  PICBOOL addServer(char *); // add new server to LRU server list
  PICBOOL addUser(BaseString); // add latest logonuser

  PICSHORT getOptSec(char *); // get option section
  PICSHORT detectPlatform(); // self explanatory

  /* END: New public methods introduced by PICSearch v3 */

//:fun
  PICBOOL fromColl(PicIxCollect &coll);
  PICBOOL toColl(PicIxCollect &coll);

  /* BEGIN: New variables (parameters) introduced by PICSearch v3 */

  PICUINT cacheSize; // amount of pictures stored in cache
  PICUSHORT windowSize; // 0 = 640x480, 1 = 800x600, 2 = 1024x768
  PICUSHORT posX; // x-cordinate of left top corner
  PICUSHORT posY; // y-cordinate of left top corner
  PICSHORT globalErrorFlag; // For checking error status
  PICSHORT platform; // -1 = unknown, 0 = windows, 1 = unix
  PICSHORT perfIndex; // 10 = base, in 1 increments +/-
  PICSHORT networkPerformance; // 10 = base, in 1 increments +/-
  PICBOOL picCacheDebug; // whether to show PICCache debug messages
  PICBOOL scaleQueryImage; // whether to scale query images
  PICBOOL askAlgParameters; // wheter to ask algorithm parameters
  PICBOOL showAdvert; // whether to show server ad when connected
  BaseString imageEditor; // the image editor
  BaseString logonUser; // logonuser

  BaseString serverList[SERVER_LIST_SIZE];

  /* END: New variables (parameters) introduced by PICSearch v3 */

/*
These functions set/get the following parameters 
from/to a collection of PicStrings (in this order):
  showAbortedQuery, askChangeConfirm,
  picGUIDebug, picCoreDebug, picAlgDebug, picDbDebug,
  defWidth and defHeight.
Validity of defWidth and defHeight are checked.
*/

//:var
  PICBOOL showAbortedQuery;  // not used any more
/*
If a query gets aborted (by a newly launched query) do we 
show the results anyway? 
*/
//:var
  PICBOOL askChangeConfirm;
/*
If an image has been edited but the changes not exported from
the image editor window, do we ask a confirmation?
(Implemented by PicGUI).
*/
//:var
  PICBOOL picGUIDebug;
  PICBOOL picCoreDebug;
  PICBOOL picAlgDebug;
  PICBOOL picDbDebug;
/*
Flags controlling the debug modes. Subsystems print many debugging
messages when their flag is on.
*/

  BaseString wwwBrowser;
  PICULONG   resultImages;
  PICUSHORT  hitPercentage;

//:var
  PICUSHORT getDefWidth();
  PICUSHORT setDefWidth(PICUSHORT newWidth);
  PICUSHORT getDefHeight();
  PICUSHORT setDefHeight(PICUSHORT newHeight);
/*
Get and set default dimensions of example images.
Set methods have range checking and return the actual value in effect.
*/

//:var
  PICUSHORT threadAbortTimeout;
  PICUSHORT querySetSize;
  PICUSHORT dbImageFramesOnScreen;
  PICUSHORT stampWidth;
  PICUSHORT stampHeight;
  PicImageFormat *formatForScaledImages;
/*
These parameters are mainly for internal use. They are not set
or accessed by the user or written into database.
 - threadAbortTimeout indicates the amount of seconds to wait for 
   a thread to respond in PicMain::abortThread and suspendThread
 - querySetSize is the size of the result set of queries, used
   by PicQueryArray
 - dbImageFramesOnScreen is the number of frames for database images
   on the screen. This may be changed by PicGUI calling
   PicMain::setNumOfDbFrames
 - stampWidth is the maximum width for thumbnail representations of
   database images
 - stampHeight is the maximum height for thumbnail representations of
   database images
 - formatForScaledImages is a pointer to the PicImageFormat to be
   used for thumbnail representations of database images 
   (if NULLPTR, then raw RGB values are used)
   Used by PicMainDbUser::insertBasicData and updateBasicData.
*/

 private:

  PICUSHORT defWidth;
  PICUSHORT defHeight;

  /* BEGIN: New variables (parameters) introduced by PICSearch v3 */

  FILE *configFp;

  /* BEGIN: New variables (parameters) introduced by PICSearch v3 */
};

/*:class ********************************************************************

CacheObj is the definition for one cache object

****************************************************************************/

class CacheObj {
  // Here is the definition of the CacheObj object
public:
  CacheObj(); // constructor
  ~CacheObj(); // destructor
  
  CacheObj *next; // next in chain, NULL if last
  CacheObj *prev; // previous in chain, NULL if first

  long databaseId; // for hashing
  long pictureId; // ---""---
  time_t inserted; // time of creation
  time_t lastUsed; // last time of usage

  long hitCount; // usage counter
  long inThisChain; // how many cache objects in the chain
  long collisions; // collisions

  PicPicture *thisPic; // the picture itself
};

/*:class ********************************************************************

PICCache is the client side cache introduced by PICSearch v3

****************************************************************************/

class PICCache {
  // Here is the definition of the PICCache object
  CacheObj **array; // array for the chains
public:
  PICCache(); // constructor
  PICCache(long); // constructor
  ~PICCache(); // destructor

  void purgeCache();

  int globalErrorFlag; // For checking error status

  PICUINT cacheSize; // configurable in client options
  PICULONG arraySize; // cacheSize >> PER_CHAIN_SHIFT
  PICULONG cacheHits; // succesfull lookups = hits
  PICULONG cacheMisses; // un-succesfull lookups = misses
  PICULONG getHashKey(PICULONG, PICULONG); // method to return hash value
  PICULONG chainSize(PICULONG); // for returning from private area
  PICULONG elementsInChain(PICULONG); // may differ from chainSize()
  PICULONG chainCollisions(PICULONG); // for returning from private area

  bool destroyCache(); // clean up

  PicPicture *getThumb(long, long);
  bool isThumbInCache(long, long);
  bool putThumb(long, long, const PicPicture &);

  cacheArrayDescr *getArrayDescr();
};

/*:class ********************************************************************

PicMainSharedData contains various globally accessible data objects.

****************************************************************************/

class PicMainSharedData
{
 public:

  // constructor needed for new PicParams -SK
  PicMainSharedData() {
    unsigned long size_shift;

    picPars = new PicParams(DEFAULT_CONFIG);
    picParsStored = new PicParams(*picPars);
    if(picPars->cacheSize != 0) {
      if(picPars->cacheSize < MIN_CACHE_SIZE)
	picPars->cacheSize = MIN_CACHE_SIZE;
      if(picPars->cacheSize > MAX_CACHE_SIZE)
	picPars->cacheSize = MAX_CACHE_SIZE;
      // This is for rounding up to the next 2^, non-optimal
      while(1) {
	size_shift = picPars->cacheSize >> PER_CHAIN_SHIFT;
	if((size_shift * PER_CHAIN) != picPars->cacheSize)
	  picPars->cacheSize++;
	else
	  break;
      }
      cache = new PICCache(picPars->cacheSize);
      cacheIsDisabled = false;
    }
    else {
      // The cacheSize = 0 is a special case, disable cache
      cacheIsDisabled = true;
    }
  };
  // destructor needed for new PicParams -SK
  ~PicMainSharedData() {
    if(cacheIsDisabled == false) {
      cache->purgeCache();
      delete cache;
    }
    delete picParsStored;
    delete picPars;
  };
//:fun
  PicImageFormat*        getImageFormat(PICULONG id);
  PicSearchType*         getSearchType(BaseString name);
//  PicSegmentAlgorithm*   getSegmentAlg(BaseString name);
//  PicSegmentAlgorithm*   getSegmentAlg(PICULONG id);
/*
These functions can be used to retrieve a pointer to algorithms,
search types and image formats. Retrieval can be done by id,
name or both.
*/

//:var
  PicArray imageFormats;
  PicArray searchTypes;
  PicArray searchAlgs;
  PicArray imageProcessAlgs;
  PicArray segmentationAlgs;
/*
These collections hold "staticly" the prevalent algorithms and
image format capabilities of the system. The objects inside them
are created at program start-up.

imageFormats        -- collection of PicImageFormat:s
searchTypes         -- collection of PicSearchType:s
searchAlgs          -- collection of PicSearchAlgorithm:s
segmentationAlgs    -- collection of PicSegmentAlgorithm:s
*/

//:var
  BaseString currentSearchType;
/*
Name of the current search type.
*/

  // Method for runtime re-initialization, has to be here
  bool reInitializeCache(long);

//:var
  /* New for v3, backup also */
  PicParams *picPars;
  PicParams *picParsStored;

  /* This is the HOT v3 stuff, client side cache */
  bool cacheIsDisabled;
  PICCache *cache;
/*
General parameters for the system.
*/
};

#endif


