// **********************************************************************
//  This file contains miscellaneous routines of PicMain
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <mmain.hxx>
#include <pinteger.hxx>
#include <ptext.hxx>
#include <pstrstm.hxx>

#include <PICClient.h>

// **********************************************************************

PICBOOL PicMain::newExampleImage(PICULONG listIndex)
{
  PicErrorContext ecxt;
  PICBOOL ret = boolFalse;

  DEBUG_CORE("Creating new query image with index " 
             & asString(listIndex) & "\n");
  if(exampleListIndexes.add(PicInteger(listIndex))){
    PicPicture pic;
    pic.raw = new PicRawPicture();
    if(examplePics.add(pic)){
      ret = boolTrue;
    }else{
      exampleListIndexes.deleteAt(exampleListIndexes.size());
    }
  }
  if(!ret){
    if(DEBUG_CORE_ON){
      EHOLD(PicMain, exImageCreateFailed);
    }else{
      ERAISE(PicMain, exImageCreateFailed);
    }
  }
  FLUSH_ERRORS();
  DEBUG_CORE("New query image created: " & BaseString(ret ? "" : "not ") & "ok.\n");
  return ret;
}

// **********************************************************************

PICBOOL PicMain::exImageClosed(PICULONG listIndex)
{
  PicErrorContext ecxt;
  PicIterator     iter;
  PICBOOL         ret = boolFalse;

  DEBUG_CORE("Closing query image with index " 
             & asString(listIndex) & "\n");

  if(examplePics.size() < 2)
    return exImageCleared(listIndex);

  ITERATE(exampleListIndexes, iter)
  {
    PicInteger &curr = AS(PicInteger, exampleListIndexes.at(iter));
    if((PICULONG)curr.asLong() == listIndex){
      PICINDEX ix = exampleListIndexes.indexAt(iter);
      examplePics.deleteAt(ix);
      exampleListIndexes.deleteAt(ix);
      if(!PicError::isDetected())
        ret = boolTrue;
      break;
    }
  }
  if(!ret){
    if(DEBUG_CORE_ON){
      EHOLD(PicMain, exImageCloseFailed);
    }else{
      ERAISE(PicMain, exImageCloseFailed);
    }
  }
  FLUSH_ERRORS();
  DEBUG_CORE("Closed query image: " & BaseString(ret ? "" : "not ") & "ok.\n");
  return ret;
}

// **********************************************************************

PICBOOL PicMain::exImageCleared(PICULONG listIndex)
{
  PicErrorContext ecxt;
  PicIterator iter;
  PICBOOL ret = boolFalse;

  DEBUG_CORE("Clearing query image with index " 
             & asString(listIndex) & "\n");

  PicPicture *pic = getExamplePic(listIndex);
  if(pic){
    pic->setNull();
    pic->raw = new PicRawPicture();
    ret = boolTrue;
  }
  if(!ret){
    if(DEBUG_CORE_ON){
      EHOLD(PicMain, exImageClearFailed);
    }else{
      ERAISE(PicMain, exImageClearFailed);
    }
  }
  FLUSH_ERRORS();
  DEBUG_CORE("Cleared query image " & BaseString(ret ? "" : "not ") & "ok.\n");
  return ret;
}

// **********************************************************************

PICBOOL PicMain::getParams(PicIxCollect &params)
{
  DEBUG_CORE("Getting system parameters\n");
  PICBOOL ret = globData.picPars->toColl(params);
  FLUSH_ERRORS();
  return ret;
}

PICBOOL PicMain::setParams(PicIxCollect &params)
{
  DEBUG_CORE("Setting system parameters\n");

  PICULONG oldCacheSize = globData.picPars->cacheSize;
  PICBOOL ret = globData.picPars->fromColl(params);
  
  if(!ret) { 
    EHOLD(PicMain, invalidParams);
  }
  else {
    if(globData.picPars->cacheSize != oldCacheSize) {
      globData.reInitializeCache(globData.picPars->cacheSize);
      // Change the PicIxCollect, PicGUI will notice it, hopefully
      params.putAt(PICPARAMS_CACHE_SIZE, PicString(asString(globData.picPars->cacheSize)));
    }
  }
  FLUSH_ERRORS();
  return ret;
}

PICUSHORT PicMain::getNumOfDbFrames(void) {
  return globData.picPars->dbImageFramesOnScreen;
}

void PicMain::setNumOfDbFrames(PICUSHORT frames)
{
  DEBUG_CORE("Database frames on screen now = " & asString(frames) & "\n");
  globData.picPars->dbImageFramesOnScreen = frames;
}

// **********************************************************************

extern PICBOOL picSearchIsExiting;  // defined in maininit.cxx

PICBOOL PicMain::exitProgram(void)
{
  PICBOOL batchOn = batchThread.state() == PicThread::running;

  DEBUG_CORE("About to exit system.\n");

  if(batchOn){
    INFO_PRINT("Aborting " & batchThread.name & " thread.\n");
    abortThread(batchThread);
    FLUSH_ERRORS();
  }
  batchOn = batchThread.state() == PicThread::running;

  FLUSH_ERRORS();
  PICBOOL ret = (!batchOn);
  if(DEBUG_CORE_ON){
    if(!ret){ 
      INFO_PRINT("Not "); 
    }
    INFO_PRINT("OK to exit program.\n"); 
  }
  batchThread.wait();
  FLUSH_ERRORS();
  if(ret){
    picSearchIsExiting = boolTrue;
  }
  // Write client settings into picsearch.cfg
  globData.picPars->writeSettings();
  return ret;
}

// **********************************************************************

PICBOOL PicMain::abortThread(PicEnhancedThread &thread)
{
  PICUSHORT attempts = globData.picPars->threadAbortTimeout * 2;

  if(thread.state() != PicThread::running)
    return boolTrue;

  while(thread.state() == PicThread::running){
    DEBUG_CORE("Signaling abort to thread \"" & thread.name & "\".\n");
    thread.command = PicEnhancedThread::abort;
    PicTools::sendThreadCommand(thread);
    PicThread::sleep(500); // sleep 0.5 sec
    
    if(attempts > 0)
      attempts--;
    else
      break;
  }
  if(thread.state() == PicThread::running){
    ERAISEFMT(PicMain, threadAbortFailure) 
      << globData.picPars->threadAbortTimeout; 
    return boolFalse;
  }
  DEBUG_CORE("Thread \"" & thread.name & "\" aborted successful.\n");
  return boolTrue;
}

// **********************************************************************

PICBOOL PicMain::suspendThread(PicEnhancedThread &thread)
{
  PICUSHORT attempts = globData.picPars->threadAbortTimeout * 2;

  if(thread.state() != PicThread::running){
    return boolFalse;
  }else{
    if(thread.command == PicEnhancedThread::go)
      thread.command = PicEnhancedThread::suspend;
  }

  PicTools::sendThreadCommand(thread);
  
  DEBUG_CORE("Thread \"" & thread.name & "\" suspended successful.\n");
  return boolTrue;
}

// **********************************************************************

PicPicture* PicMain::getExamplePic(PICULONG listIndex)
{
  PicIterator iter;
  ITERATE(exampleListIndexes, iter)
  {
    PicInteger &curr = AS(PicInteger, exampleListIndexes.at(iter));
    if((PICULONG)curr.asLong() == listIndex){
      PICINDEX ix = exampleListIndexes.indexAt(iter);
      if(examplePics.contains(ix)){
        return &AS(PicPicture, examplePics.at(ix));
      }else{
        break;
      }
    }
  }
  ERAISEFMT(PicMain, unKnownExampleImage) << listIndex;

  return NULLPTR;
}

// **********************************************************************

void PicMain::debugMesgGUI(BaseString mesg)
{
  mesg.prepend("GUI debug: ");
  INFO_PRINT(mesg);
}

void PicMain::debugMesgCore(BaseString mesg)
{
  mesg.prepend("Core debug: ");
  INFO_PRINT(mesg);
}

void PicMain::debugMesgAlg(BaseString mesg)
{
  mesg.prepend("Alg debug: ");
  INFO_PRINT(mesg);
}

void PicMain::debugMesgDb(BaseString mesg)
{
  mesg.prepend("Db debug: ");
  INFO_PRINT(mesg);
}

void PicMain::debugMesgCache(BaseString mesg)
{
  mesg.prepend("Cache debug: ");
  INFO_PRINT(mesg);
}

void PicMain::printErrors(void)
{
  if(PicError::isDetected()){
    PICUSHORT errors = 0;
    const PicError *curr = &PicError::current();
    while(curr){
      if(errors > 0){ 
        INFO_PRINT(": "); 
      }
      INFO_PRINT(curr->text());
      errors++;
      curr = curr->nextItem();
    }
    INFO_PRINT("\n");
  }
}

// **********************************************************************

PicImageFormat* PicMainSharedData::getImageFormat(PICULONG id)
{
  PicIterator iter;
  ITERATE(imageFormats, iter)
  {
    PicImageFormat &curr = AS(PicImageFormat, imageFormats.at(iter));

    if(id == curr.id)
      return &curr;
  }
  ERAISE(PicMain, unKnownImageFormat);
  return NULLPTR;
}

// **********************************************************************

PICDbAlgorithm *PicMain::getSearchAlg(BaseString name)
{
	PicIterator iter;

	ITERATE(PICClient::svrData.algorithms, iter)
	{
		PICDbAlgorithm &dbAlg =
			AS(PICDbAlgorithm, PICClient::svrData.algorithms.at(iter));

		if(dbAlg.name==name)
			return(&dbAlg);
	}

	return(NULL);
}

// **********************************************************************

PicSearchType* PicMainSharedData::getSearchType(BaseString name)
{
  PicIterator iter;
  ITERATE(searchTypes, iter)
  {
    PicSearchType &curr = AS(PicSearchType, searchTypes.at(iter));
    DEBUG_CORE("Iterating searchTypes, found \"" & curr.name & "\"\n");
    if(name == curr.name)
      return &curr;
  }
  ERAISEFMT(PicMain, unKnownSearchType) << name;
  return NULLPTR;
}

// **********************************************************************

bool PicMainSharedData::reInitializeCache(long size) {
  PICULONG size_shift;
  
  // First destroy old PICCache
  cache->purgeCache();
  delete cache;
  if(size != 0) {
    if(size < MIN_CACHE_SIZE)
      size = MIN_CACHE_SIZE;
    if(size > MAX_CACHE_SIZE)
      size = MAX_CACHE_SIZE;
    picPars->cacheSize = size;
    // This is for rounding up to the next 2^, non-optimal
    while(1) {
      size_shift = picPars->cacheSize >> PER_CHAIN_SHIFT;
      if((size_shift * PER_CHAIN) != picPars->cacheSize)
	picPars->cacheSize++;
      else
	break;
    }
    cache = new PICCache(picPars->cacheSize);
    cacheIsDisabled = false;
  }
  else {
    // The size = 0 is a special case, disable PICCache
    cacheIsDisabled = true;
  }
  return true;
}

static void stdoutPrintErrors(void)
{
  if(PicError::isDetected()){
    PICUSHORT errors = 0;
    const PicError *curr = &PicError::current();
    while(curr){
      if(errors > 0){ 
        printf(": "); 
      }
      printf("%s", curr->text().chars());
      errors++;
      curr = curr->nextItem();
    }
    printf("\n");
  }
}

