// **********************************************************************
//  Implementation of the auxiliarity threads for PicCore
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <mmain.hxx>
#include <PicSignature.hxx>

#include <perrmgr1.hxx>
#include <pfilestm.hxx>
#include <pstrstm.hxx>
#include <ptext.hxx>
#include <pixmap.hxx>

#include <PICClient.h>

// **********************************************************************

MAKE_PICVIRTCLASS(PicEnhancedThread, PicThread)

PicEnhancedThread::PicEnhancedThread(PicPriority priority)
  : PicThread(priority),
    command(go),
    name("Noname")
{

}

// **********************************************************************

MAKE_PICVIRTCLASS(PicMassInsertThread, PicEnhancedThread)

ERROR_MESSAGES(PicMassInsertThread)
PARAMSTR(noFileName, "Batch insertion: No file name specified")
PARAMSTR(fileNotExist, "Batch insertion: File @* not found")
END_PARAM_STRINGS

PicMassInsertThread::PicMassInsertThread(PicPriority priority)
  : PicEnhancedThread(priority),
    insertedPics(0), failedPics(0), notFound(0),
    inputStream(NULLPTR)
{
  name = "Batch insertion";
}

// **********************************************************************

PICBOOL PicMassInsertThread::doInits()
{
  if(listFile.isNull()){
    ERAISE(PicMassInsertThread, noFileName);
    FLUSH_ERRORS();
    return boolFalse;
  }
  if(inputStream){
    delete inputStream;
    inputStream = NULLPTR;
  }
  INFO_PRINT("Batch insertion started.\n");
  if(listFile.suffix().equal("LST")){
    INFO_PRINT("Batch insert: reading from list file.\n");
    inputStream = new PicFileStream(listFile, PicFile::forReading);
  }else{
    INFO_PRINT("Batch insert: using file name mask.\n");
    tempBuf = listFile.fullName().copy();
    inputStream = new PicInStringStream(tempBuf);
  }
  insertedPics = failedPics = notFound = 0;
  return (inputStream != NULLPTR);
}

// **********************************************************************

PICBOOL PicMassInsertThread::extractParams(BaseString line, 
                                           BaseString &fName, 
                                           PicPicture &newPic)
{
  BaseString  name, par1, par2;
  PICPOSITION fileNamePos = 1, namePos=0, par1Pos=0, par2Pos=0;

  fName = line; 
  if((namePos = line.positionOf('#', fileNamePos))){
    fName.deleteAfter(namePos-1);
    if((par1Pos = line.positionOf('#', namePos+1))){
      if((par2Pos = line.positionOf('#', par1Pos+1))){
        par2 = line.after(par2Pos);
        par1 = line.between(par1Pos+1, par2Pos-1);
      }else{
        par1 = line.after(par1Pos);
      }
      name = line.between(namePos+1, par1Pos-1);
    }else{
      name = line.after(namePos);
    }
  }
  fName.trim(); name.trim(); par1.trim(); par2.trim();
  newPic.name = name; 
  newPic.par1 = par1; 
  newPic.par2 = par2;
  return boolTrue;  
}

// **********************************************************************

PICBOOL PicMassInsertThread::printStats()
{
  PicStringText doneTxt;
  doneTxt("Batch insertion: @* image") << insertedPics ENDFMT;
  if(insertedPics != 1){ doneTxt("s"); }
  doneTxt(" inserted successfully to database.\n");
  if(failedPics > 0){
    doneTxt("Batch insertion: Failed to insert @* image") << failedPics ENDFMT;
    if(failedPics != 1){ doneTxt("s"); }
    doneTxt(".\n");
  }
  if(notFound > 0){
    doneTxt("Batch insertion: @* image file") << notFound ENDFMT;
    if(notFound != 1){ doneTxt("s"); }
    doneTxt(" not found.\n");
    doneTxt.flush();
  }
  doneTxt("") ENDFMT;
  INFO_PRINT(doneTxt);
  return boolTrue;
}
// **********************************************************************

PICBOOL PicMassInsertThread::insertPic(PicPicture &newPic)
{
  PICBOOL doingOk = boolTrue;

  doingOk = newPic.loadFromFile(newPic.filename);
  if(doingOk){
      doingOk = PICClient::InsertImageToDb(newPic);
  }
  if(doingOk){
    insertedPics++;
    if((insertedPics % 5)==0){
      INFO_PRINT("Batch insertion: " & 
                 asString(insertedPics) & " images done...\n");
    }
  }else{
    failedPics++;
  }
  return doingOk;
}

// **********************************************************************

PICTHREADRESULT PicMassInsertThread::threadFunction(PICTHREADARG arg)
{
  DEBUG_CORE(name & " thread running!\n");

  if(!doInits())
    return 0;

  PicText     input(*inputStream);
  BaseString  fName, line;

  while(!input.endOfText()){
    input.rdln(line);
    line.trim();
    if(line.isSpace())
      continue;

    PicPicture  newPic;
    extractParams(line, fName, newPic);

    PicFileIterator iter(fName);
    PicFile pictureFile;
    PICBOOL needsIteration = (line.positionOf('*') != outOfRange);

    while(1){
      if(!needsIteration){
        pictureFile.setFullName(fName);
        DEBUG_CORE("Batch insert trying file \"" & line & "\"\n");
      }else{
        if(iter.next()){
          pictureFile = iter.file;
        }else{
          break;
        }
      }

      if(!pictureFile.exists()){
        ERAISEFMT(PicMassInsertThread, fileNotExist)
            << pictureFile.fullName();
        FLUSH_ERRORS();
        notFound++;
      }else{
        newPic.filename = pictureFile.fullName();
        newPic.id = 0;
        insertPic(newPic);
        if(PicError::isDetected()){
          INFO_PRINT("Batch insertion for file \"" &
                     pictureFile.fullName() & "\":\n");
          FLUSH_ERRORS();
        }
      }
    }
  }
endThread:

  printStats();
  if(inputStream){
    delete inputStream;
    inputStream = NULLPTR;
  }
  FLUSH_ERRORS();

  return 0;
}

// **********************************************************************
// **********************************************************************

