// --------------------------------------------------------------------------
//  BASE DATE
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef BDATE_HXX
#define BDATE_HXX

#include <picbase.hxx>
#include <picbase.erh>
#include <picerror.erh>
#include <bstring.hxx>

class BaseStringArray;
class PicFmtText;

typedef PICSHORT  PICYEAR;        // Valid years for PicDate (1600..2500).
typedef char      PICMONTH;       // Months of a year (1..12).
typedef char      PICDAY;         // Days of a month (1..31).

typedef PICSHORT  PICWEEKDAY;     // Weekday numbers (1..7)
typedef PICSHORT  PICWEEKNUMBER;  // Week numbers in a year (1..52).

typedef PICLONG   PICDAYCOUNT;    // Date interval in days.
typedef PICLONG   PICWEEKCOUNT;   // Date interval in weeks.
typedef PICLONG   PICMONTHCOUNT;  // Date interval in months.
typedef PICLONG   PICYEARCOUNT;   // Date interval in years.

const PICYEAR     minYear          = 1600; // Minimum value of PICYEAR.
const PICYEAR     maxYear          = 2500; // Maximum value of PICYEAR.
const PICMONTH    minMonth         = 1;    // Minimum value of PICMONTH.
const PICMONTH    maxMonth         = 12;   // Maximum value of PICMONTH.
const PICDAY      minDay           = 1;    // Minimum value of PICDAY.
const PICDAY      maxDay           = 31;   // Maximum value of PICDAY.
const PICWEEKDAY  minWeekday       = 1;    // Minimum value of PICWEEKDAY.
const PICWEEKDAY  maxWeekday       = 7;    // Maximum value of PICWEEKDAY.
const PICDAYCOUNT daysInNormalYear = 365L; // Length of normal year in days.

/*:class ====================================================================
BaseDate implements the basic date handling operations. Date values
include a day, month and year component. Null date is used to represent
unknown or missing value. At any moment after creation BaseDate instance
is either null or represents a valid date.

Date format can be selected with the PicDateIniter class.
*/

class BaseDate
{
    public:
// --------------------------------------------------------------------------
//      DATE FORMATTING TYPES
//:enum ---------------------------------------------------------------------

    enum DResolution
    {
        yearsOnly        = FORMAT_YEARS_ONLY,
        yearsToMonths    = FORMAT_YEARS_TO_MONTHS,
        yearsToDays      = FORMAT_YEARS_TO_DAYS
    };

    enum DStyle
    {
        yearFirst        = FORMAT_YEAR_FIRST,
        monthFirst       = FORMAT_MONTH_FIRST,
        dayFirst         = FORMAT_DAY_FIRST,
        monthByName      = FORMAT_MONTH_BY_NAME,
        monthByShortName = FORMAT_MONTH_BY_SHORT_NAME
    };

/*
Date format is defined as a combination of these two parameters.
Resolution means the number of components included in the textual
representation:

    yearsOnly          1990
    yearsToMonths      1990.12
    yearsToDays        1990.12.24

Style means the way the components are presented:

    yearFirst          1990.12.24
    monthFirst         12.24.1990
    dayFirst           24.12.1990
    monthByName        December 24, 1990
    monthByShortName   Dec 24, 1990

*/

//:enum ---------------------------------------------------------------------

    enum Relative
    {
        today,
        thisWeek,
        thisMonth,
        thisYear
    };

/*
PicRelativeDate specifies the step of the date interval: day, week,
month or year.
*/

// ---------------------------------------------------------------------

    enum YearMWStyle
    {
        year4            = FORMAT_YEAR4,
        year2            = FORMAT_YEAR2,
        year4db          = FORMAT_YEAR4DB,
        year2db          = FORMAT_YEAR2DB
    };

// --------------------------------------------------------------------------
//      DEFAULT DATE FORMAT
//:fun ----------------------------------------------------------------------

    static DStyle      defaultStyle       ();
    static DResolution defaultResolution  ();
    static char        defaultDelimiter   ();
    static char        defaultLeader      ();
    static YearMWStyle defaultYearMWStyle ();

/*
These methods return the components of the currently selected default
date format.
*/

// --------------------------------------------------------------------------
//      ATTRIBUTES FROM GIVEN VALUE
//:fun ----------------------------------------------------------------------

    static BaseString monthName        (PICMONTH);
    static BaseString shortMonthName   (PICMONTH);
    static BaseString weekdayName      (PICWEEKDAY);
    static BaseString shortWeekdayName (PICWEEKDAY);

/*
These methods return the currently selected names of weekdays and months.

:err illegalComponent: Month or weekday number is out of range. Return value
                       is a null string.
*/

//:fun ----------------------------------------------------------------------

    static PICDAYCOUNT daysInYear (PICYEAR);

/*
This method returns the number of days in the given year.

:err illegalComponent: Year is out of range. Return value is zero.
*/

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    BaseDate ();
    BaseDate (Relative, PICLONG unitInterval = 0);
    BaseDate (PICYEAR, PICMONTH, PICDAY);
    BaseDate (const BaseString&, DStyle, char delimiter);
    BaseDate (const BaseString&, DStyle);
    BaseDate (const BaseString&);
    BaseDate (const BaseDate&);

/*
BaseDates can be created in the following ways:

    1. BaseDate a;
    2. BaseDate b (BaseDate::today, +2);
    3. BaseDate c (1991, 12, 24);
    4. BaseDate d ("24.12.1991");
    5. BaseDate e ("1991-12-24", BaseDate::yearFirst, '-');
    6. BaseDate f (aBaseDate);

1. Creates a null BaseDate.
2. Creates a BaseDate with the value 'day after tomorrow'
   (today + 2 days). This constructor uses the current system date.
3. Creates a BaseDate with the value 24.12.1991 given as components.
4. Creates a BaseDate with the value 24.12.1991 in default format.
5. Creates a BaseDate with the value 24.12.1991 in a specified format.
6. Copies another BaseDate.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid date.
*/

// ----------------------------------------------------------------------

    BaseDate& operator = (const BaseDate&);

// --------------------------------------------------------------------------
//      ERROR HANDLING
// --------------------------------------------------------------------------

    PICCLASS()

    PICERRORCODES { ok,
        nullInstance     = ERR_NULL_INSTANCE,
        formatError      = ERR_DATE_FORMAT,
        deformatError    = ERR_DATE_DEFORMAT,
        illegalComponent = ERR_DATE_COMPONENT,
        dateIllegal      = ERR_DATE_ILLEGAL,
        tooEarly         = ERR_DATE_TOO_EARLY,
        tooLate          = ERR_DATE_TOO_LATE,
    };

// --------------------------------------------------------------------------
//      NULL HANDLING
// --------------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

// --------------------------------------------------------------------------
//      INTERVALS
//:fun ----------------------------------------------------------------------

    PICDAYCOUNT   daysBetween   (const BaseDate&) const;
    PICWEEKCOUNT  weeksBetween  (const BaseDate&) const;
    PICMONTHCOUNT monthsBetween (const BaseDate&) const;
    PICYEARCOUNT  yearsBetween  (const BaseDate&) const;

/*
These methods return the count of full units between the receiver and the
argument. If the argument is greater than the receiver the resulting count
is negative. Week is defined to be 7 days.

:err nullInstance: Receiver or argument is null.
*/

//:fun ----------------------------------------------------------------------

    void addDays   (PICDAYCOUNT);
    void addWeeks  (PICWEEKCOUNT);
    void addMonths (PICMONTHCOUNT);
    void addYears  (PICYEARCOUNT);

/*
These methods add the given interval to the receiver. Note that when
months or years are added to the receiver, day part might also change.
E.g. adding one year to 29.02.1988 results in 28.02.1989 and adding
one month to 31.10.1990 becomes 30.11.1990. Due to this, note also:

        BaseDate a (1991, 10, 31);
        BaseDate b (a);

        // a == b

        a.addMonth(1);  // Now a = 1991.11.30
        a.addMonth(-1); // Now a = 1991.10.30

        // a != b

The same problem arises with addYear and February 29th.

:err nullInstance: Receiver or argument is null. Receiver is left
                   unchanged.
:err dateIllegal:  Overflow or underflow occurred. Receiver is left
                   unchanged.
*/

// --------------------------------------------------------------------------
//      LOGIC COMPARISONS
//:fun ----------------------------------------------------------------------

    PICBOOL operator == (const BaseDate&) const;
    PICBOOL operator != (const BaseDate&) const;
    PICBOOL operator >  (const BaseDate&) const;
    PICBOOL operator >= (const BaseDate&) const;
    PICBOOL operator <  (const BaseDate&) const;
    PICBOOL operator <= (const BaseDate&) const;

/*
BaseDates are compared along the time axis. Later dates are considered
greater. Two null dates are equal and null date is smaller than any other
valid date.
*/

// --------------------------------------------------------------------------
//      FORMATTING & DEFORMATTING
//:fun ----------------------------------------------------------------------

    BaseString getImage () const;
    BaseString getImage (
        DResolution, DStyle,
        char delimiter,
        char leader) const;

/*
These methods return the textual representation of the receiver. Default
format is used unless a specific format is given. For null date a null
string is returned.

Delimiter separater the components of the date value. Common delimiters
are point (24.12.1991) and minus (24-12-1991).

Leader defines the character which is used to fill date image when a
component is less than 10. If it is '\0' filling is not done.

    leader = '0'    -> (1d 2m 1991y) -> "01-02-1991"
    leader = '\0'   -> (1d 2m 1991y) -> "1-2-1991"
*/

//:fun ----------------------------------------------------------------------

    void setImage (const BaseString&);
    void setImage (const BaseString&, DStyle, char delimiter);

/*
These methods set the value of the receiver according to the given textual
date image. A date style and delimiter are used to help parsing. If no
format is given, formatting is based on default values.

See BaseDateTime::deformat for more details.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid date.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const BaseDate&);

/*
Output operator to write a BaseDate to PicFmtText.
*/

// --------------------------------------------------------------------------
//      SET DATE
//:fun ----------------------------------------------------------------------

    void setDate (PICYEAR, PICMONTH, PICDAY);
    void setDate (PICYEAR, const BaseString& monthName, PICDAY);

/*
These methods set the value of the receiver according to the given date
components. Month name may be a long or a short version.

:err dateIllegal: A date component is out of range or the date as a whole
                  is unacceptable. The receiver is left unchanged.
*/

//:fun ----------------------------------------------------------------------

    void setRelativeDate (Relative, PICLONG unitInterval);

/*
This method sets the value of the receiver based on the current system
date. Intervals of days, weeks, months or years can be added to or
subtracted from the current date.

:err dateIllegal: Overflow or underflow occurred. Receiver is left
                  unchanged.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL setWeek (PICYEAR, PICWEEKNUMBER);

/*
This method sets the date to the Monday of the given week of the
given year. Note: the year may change if the week is 1, 52 or 53.
True is returned if the operation is successful. Otherwise the
date is not changed.
*/

// --------------------------------------------------------------------------
//      DATE COMPONENTS
//:fun ----------------------------------------------------------------------

    PICYEAR  year  () const;
    PICMONTH month () const;
    PICDAY   day   () const;

/*
These methods return the components of the date value. For null
date zero is returned.
*/

// --------------------------------------------------------------------------
//      DATE ATTRIBUTES
//:fun ----------------------------------------------------------------------

    PICWEEKDAY    weekday        () const;
    PICWEEKNUMBER weekNumber     () const;
    PICDAYCOUNT   dayNumber      () const;
    PICBOOL       isLeapYear     () const;
    PICDAY        lastDayOfMonth () const;

/*
For null date zero, null string or boolFalse is returned.

:ret weekday:    Order number of the weekday (Monday == 1).
:ret dayNumber:  Order number of the day starting from January the 1st.
:ret weekNumber: Week number starting from January the 1st.
:ret isLeapYear: boolTrue if the receiver's year is a leap year.
:ret lastDayOfMonth: The last day of the current month.
*/

    private:
// ==========================================================================
//      INSTANCE VARIABLES

    PICYEAR  theYear;
    PICMONTH theMonth;
    PICDAY   theDay;

    friend class BaseDateTime;
};

typedef BaseDate::DResolution PicDateResolution;
typedef BaseDate::DStyle      PicDateStyle;
typedef BaseDate::Relative    PicRelativeDate;
typedef BaseDate::YearMWStyle PicYearMWStyle;

// --------------------------------------------------------------------------
//      DATE CLASS INITIALIZATION
/*:class --------------------------------------------------------------------

BaseDate/PicDate uses default parameters until an PicDateIniter is created
(either as an automatic or static object). Given values are used until
another PicDateIniter is created. When PicDateIniter is destroyed the
previous values are restored.

If NULLPTR is given as BaseStringArray*, previous values are retained for
that array. Note that PicDateIniter will not delete the arrays.
*/


class PicDateIniter
{
    public:

    PicDateIniter (
         PicDateStyle,
         PicDateResolution,
         char delimiter,
         char leader,
         const BaseStringArray* longMonthNames,
         const BaseStringArray* shortMonthNames,
         const BaseStringArray* weekdayNames,
         const BaseStringArray* shortWeekdayNames,
         PicYearMWStyle = BaseDate::year4);

   ~PicDateIniter ();

    private:

    PicDateStyle      dStyle;
    PicDateResolution dResol;
    char              dDelim;
    char              dLeader;
    char             *longNames[13];
    char             *shortNames[13];
    char             *days[8];
    char             *shortDays[8];
    PicYearMWStyle    ymwStyle;

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PICBOOL BaseDate::isNull () const
{
    // Day == 0 is an indication of a null date:
    return theDay ? boolFalse : boolTrue;
}

inline void BaseDate::setNull ()
       { theYear = theMonth = theDay = 0; }

inline PICBOOL BaseDate::operator != (const BaseDate &right) const
       { return !(*this == right); }
inline PICBOOL BaseDate::operator >= (const BaseDate &right) const
       { return (*this == right || *this > right); }
inline PICBOOL BaseDate::operator < (const BaseDate &right) const
       { return !(*this >= right); }
inline PICBOOL BaseDate::operator <= (const BaseDate &right) const
       { return !(*this > right); }

inline PICWEEKCOUNT BaseDate::weeksBetween (const BaseDate &right) const
       { return PICWEEKCOUNT(daysBetween(right) / maxWeekday); }

inline void BaseDate::addWeeks (PICWEEKCOUNT weeks)
       { addDays(weeks * maxWeekday); }

inline PICYEAR BaseDate::year () const
       { return /* isNull() ? 0 : */ theYear;  /* Always zero when null */ }
inline PICMONTH BaseDate::month () const
       { return /* isNull() ? 0 : */ theMonth; /* Always zero when null */ }
inline PICDAY BaseDate::day () const
       { return /* isNull() ? 0 : */ theDay;   /* Always zero when null */ }

#endif
