// --------------------------------------------------------------------------
//  DATE-TIME
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef BASEDATETIME_HXX
#define BASEDATETIME_HXX

#include <bdate.hxx>
#include <btime.hxx>

class PicFmtText;
class PicTextFmt;
class BaseStringArray;
class PicObject;

/*:class ********************************************************************

BaseDateTime is a simple combination of the classes BaseDate and BaseTime.
It allows free access to its two components.

For more details, see BaseDate and BaseTime.

****************************************************************************/

class BaseDateTime
{

    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//       - copy constructor & assignment by default
//:fun ----------------------------------------------------------------------

    BaseDateTime ();
    BaseDateTime (const BaseDate&);
    BaseDateTime (const BaseTime&);
    BaseDateTime (const BaseDate&, const BaseTime&);
    BaseDateTime (
        PicRelativeDate,
        PICLONG      dateInterval = 0,
        PICHOURCOUNT timeInterval = 0);
    BaseDateTime (
        PICYEAR, PICMONTH,  PICDAY,
        PICHOUR, PICMINUTE, PICSECOND);

    PICBOOL isNull  () const;
    void    setNull ();

/*
In a null BaseDateTime both date and time components are null.
If you create BaseDateTime with only BaseDate, the time component will
be null. If you create BaseDateTime with only BaseTime, the date
component will be null.

Copy constructor and assignment are compiler-generated.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL operator == (const BaseDateTime&) const;
    PICBOOL operator != (const BaseDateTime&) const;
    PICBOOL operator <  (const BaseDateTime&) const;

/*
Comparison operators use both the date and the time part.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const BaseDateTime&);

    PICOFFSET writeFmt (
        PicTextFmt&, PICOFFSET,
        const char* typeName) const;

/*
Output operator to write a BaseDateTime to PicFmtText. The method writeFmt
implements formatted output for all date and time classes. A date/time
format may contain the following components:

    D         Date in default format
    T         Time in default format

    yy        Year (00 .. 99, i.e. 1950 .. 2049)
    yyyy      Year (1600 .. 2500)

    m         Month (1 .. 12)
    mm        Month (01 .. 12)
    +m        Month (use default leader)
    mmm       Month (Jan .. Dec)
    MMM       Month (JAN .. DEC)
    month     Month (January .. December, i.e. default case)
    MONTH     Month (JANUARY .. DECEMBER)

    d         Day (1 .. 31)
    dd        Day (01 .. 31)
    +d        Day (use default leader)

    w         Weekday (1 .. 7)
    www       Weekday (Mon .. Sun)
    WWW       Weekday (MON .. SUN)
    weekday   Weekday (Monday .. Sunday, i.e. default case)
    WEEKDAY   Weekday (MONDAY .. SUNDAY)

    #         Week number (1 .. 53)
    ##        Week number (01 .. 53)

    h         Hour (0 .. 23)
    hh        Hour (00 .. 23)
    +h        Hour (use default leader)
    a         Hour (0 .. 11)
    aa        Hour (00 .. 11)
    +a        Hour (use default leader)

    ap        am/pm
    AP        AM/PM

    nn        Minute (00 .. 59)

    ss        Second (00 .. 59)

The following separators are allowed between components:

    space . , : ; - / \

Default date separator is expressed as | and default time separator as ~.
Other characters can be inserted as literals between single quotes.

Examples:   [D]                   11.6.1993
            [*]                   11.6.1993 16:17:49
            [dd|mm|yy]            11.06.93
            [yyyy-mmm-d]          1993-Jun-11
            [weekday, d.m.yyyy]   Friday, 11.6.1993
            [dd.mm, aa~nn~ss AP]  11.06, 04:17:49 PM
            [a. 'hour']           4. hour
            ['week' # 'of' -yy]   week 23 of -93

*/

//:fun ----------------------------------------------------------------------

    BaseString getImage () const;
    void       setImage (const BaseString&);

    PICOFFSET deformat (
        const char* image,
        PICOFFSET      beginAt,
        PicDateStyle   dStyle,
        PICSHORT       fDateTime);

/*
Image formatting and deformatting for DateTime, i.e. both date and time
components are expected. If components are missing, the interpretation
is carried out in certain preference order:

Components  Interpretation preference order

1           hour -> day -> year
2           hour.min -> day.month -> month.year -> day hour
3           hour.min.sec -> day.mon.year -> day.mon hour -> day hour.min
4           day.mon hour.min -> day.mon.year hour -> day hour.min.sec
5           day.mon hour.min.sec -> day.mon.year hour.min
6           day.mon.year hour.min.sec

Order of date components is assumed to follow the default date style.
However, unambiguous components like a month name may change the
interpretation. Also the default date and time delimiters are taken
as a clue if they appear in the image.

Some interpretations:

    - Years 00..99 fall between 1950..2049
    - * means current date and/or time

BaseDateTime::deformat is the general deformatting procedure of date and
time values. The procedure attempts to read as many date/time components
as possible, up to 6 (or 7 with AM/PM), or 3 (4) with date or time only.
All non-alphanumeric characters are considered separators; only one non-
space separator is allowed between components. Moreover, only whitespace
is allowed between date and time parts of the image.

:arg image:     The date/time image string. Do not give empty string.
:arg beginAt:   Index (zero-based) from where to start interpreting the image.
:arg dStyle:    Date style to aid in deformatting. BaseDate::defaultStyle()
                is usually used here.
:arg fDateTime: Specifies which components are expected: 1 - date only,
                2 - time only, 3 - both date and time (in this order).

:ret 0:         The image was not interpretable as a date/time value.
:ret 0xffff:    The image was interpreted but the resulting date/time value
                was not meaningful (e.g. 31th of February or 15 AM).
:ret any other: The offset of the next nonspace character after the
                interpreted image. If the whole string was interpreted,
                this offset points to the terminating zero byte.

*/

//:fun ----------------------------------------------------------------------

    void addHours   (PICHOURCOUNT);
    void addMinutes (PICMINUTECOUNT);
    void addSeconds (PICSECONDCOUNT);

/*
These methods add an interval to a date-time. The operation may affect
also the date part.

:err nullInstance: Date or time is null.
*/

// ----------------------------------------------------------------------

    BaseDate date;
    BaseTime time;

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline BaseDateTime::BaseDateTime () : date(), time() {}

inline BaseDateTime::BaseDateTime (const BaseDate& dt) :
    date(dt), time() {}

inline BaseDateTime::BaseDateTime (const BaseTime& tm) :
    date(), time(tm) {}

inline BaseDateTime::BaseDateTime (const BaseDate& dt, const BaseTime& tm) :
    date(dt), time(tm) {}

inline BaseDateTime::BaseDateTime (
    PicRelativeDate erd,
    PICLONG         dateInterval,
    PICHOURCOUNT    timeInterval)
    :
    date(erd, dateInterval),
    time(timeInterval) {}

inline BaseDateTime::BaseDateTime (
    PICYEAR year, PICMONTH month,   PICDAY day,
    PICHOUR hour, PICMINUTE minute, PICSECOND second)
    :
    date(year, month, day),
    time(hour, minute, second) {}

inline PICBOOL BaseDateTime::isNull () const
       { return date.isNull() && time.isNull(); }

inline void BaseDateTime::setNull ()
       { date.setNull(); time.setNull(); }

inline PICBOOL BaseDateTime::operator == (const BaseDateTime& bdt) const
       { return date == bdt.date && time == bdt.time; }
inline PICBOOL BaseDateTime::operator != (const BaseDateTime& bdt) const
       { return date != bdt.date || time != bdt.time; }

#endif
