// --------------------------------------------------------------------------
//  REAL NUMBER base class
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef BREAL_HXX
#define BREAL_HXX

#include <bstring.hxx>
#include <picerror.erh>
class PicFmtText;
class PicTextFmt;
class PicObject;

const PICDIGIT maxDigit  = 0x3fff;    // Upper limit of real number precision (16383).
const PICDIGIT minDigit  = -maxDigit; // Lower limit of real number precision (-16383).
const PICDIGIT floatReal = minDigit - 1; // Precision of a floating-point number.
const PICDIGIT decimals0 =  0; // Real number precision of zero decimals.
const PICDIGIT decimals1 = -1; // Real number precision of one decimal.
const PICDIGIT decimals2 = -2; // Real number precision of two decimals.
const PICDIGIT decimals3 = -3; // Real number precision of three decimals.
const PICDIGIT decimals4 = -4; // Real number precision of four decimals.
const PICDIGIT decimals5 = -5; // Real number precision of five decimals.

/*:class ********************************************************************

BaseReals are either fixed point or floating point numbers. Floating point
BaseReals behave exactly like the basic C++ type double.

Fixed point BaseReals represent discrete number space consisting of numbers
of form N * 10^D, where both D and N are integers. Current implementation
guarantees that the absolute value of N can be anything below 10^16, and D
in range -16383..16383.

The precision is determined by constructor (PICDIGIT parameter) and can be
later changed only by an explicit routine. The precision is expressed in ten
base e.g. precision -2 means 2 decimals.

BaseReals are usually not used as such, but only in implementing higher-
level arithmetic classes.

****************************************************************************/

class BaseReal
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    BaseReal ();
    BaseReal (PICDIGIT);
    BaseReal (PICFLOAT, PICDIGIT = floatReal);
    BaseReal (const BaseString&);
    BaseReal (const BaseString&, PICDIGIT);
    BaseReal (const BaseReal&);

/*
BaseReals can be constructed in the following ways:

    1. BaseReal a;
    2. BaseReal b (decimals2);
    3. BaseReal c (34.1672);
    4. BaseReal d (34.1672, decimals4);
    5. BaseReal e ("34.1672");
    6. BaseReal f ("34.1672", -6);
    7. BaseReal g (f);

1. Creates a null floating point number.
2. Creates a null fixed-point number with precision of 2 decimals.
3. Creates a floating point number that is approximately 34.1672.
4. Creates a fixed-point number that is exactly 34.1672 (precision -4).
5. Creates a fixed-point number that is exactly 34.1672 (precision -4 is
   determined by examining the string).
6. Creates a fixed-point number that is exactly 34.167200 (precision -6).
7. Copy of another BaseReal.

Note1: When using the string constructor (5 & 6) you must spell the number
with the currently selected decimal separator. 

Note2: creating a BaseReal out of a literal number constant (like 34.1672)
may lead to rounding errors in some rare cases. 

:err deformatError: A string doesn't represent a real number.
*/

/*:fun ----------------------------------------------------------------------

:fun BaseReal (const BaseReal&);
:fun BaseReal& operator = (const BaseReal&);

These methods are defined by the C++ compiler. They copy both the number
and its precision. If you need to copy the number only (and change its
precision), use the assignValue-methods.

*/

//:fun ----------------------------------------------------------------------

    BaseReal& operator = (PICFLOAT);

    BaseReal& assignValue (const BaseReal&);
    BaseReal& assignValue (PICLONG);

/*
These methods copy a real number and adjust it to the receiver's precision.
*/

// --------------------------------------------------------------------------
//      ERROR HANDLING
// --------------------------------------------------------------------------

    PICCLASS()

    PICERRORCODES { ok,
        nullInstance  = ERR_NULL_INSTANCE,
        formatError   = ERR_DECIMAL_FORMAT,
        deformatError = ERR_DECIMAL_DEFORMAT,
        underflow     = ERR_NUMBER_UNDERFLOW,
        overflow      = ERR_NUMBER_OVERFLOW,
        divideByZero  = ERR_NUMBER_DIVIDE_BY_ZERO
    };

// --------------------------------------------------------------------------
//      TYPE CONVERSIONS
//:fun ----------------------------------------------------------------------

    PICFLOAT asFloat () const;

/*
:ret PICFLOAT: Receiver's value as PICFLOAT (double). For fixed point
               numbers the nearest possible PICFLOAT value is returned.

:err nullInstance: Null BaseReal has no PICFLOAT counterpart.
*/

// --------------------------------------------------------------------------
//      NULL HANDLING
//:fun ----------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

/*
BaseReal has special null value which means that the value is unknown or
missing. In general operations on null BaseReals are errors. However,
attributes can be asked also from nulls.
*/

// --------------------------------------------------------------------------
//      PRECISION
//:fun ----------------------------------------------------------------------

    PICDIGIT getPrecision () const;
    void     setPrecision (PICDIGIT);

    PICBOOL isFloat () const;
    PICBOOL isFixed () const;

/*
All BaseReals have precision, also the null-valued ones. For floating point
BaseReals precision is floatReal.

BaseReal::setPrecision changes the precision and rounds the value.

:ret isFloat: boolTrue if the receiver is a floating point number.
:ret isFixed: boolTrue if the receiver is a fixed-point number.
*/

// --------------------------------------------------------------------------
//      COMPARISONS
//:fun ----------------------------------------------------------------------

    friend PICBOOL operator == (const BaseReal&, const BaseReal&);
    friend PICBOOL operator != (const BaseReal&, const BaseReal&);
    friend PICBOOL operator >  (const BaseReal&, const BaseReal&);
    friend PICBOOL operator >= (const BaseReal&, const BaseReal&);
    friend PICBOOL operator <  (const BaseReal&, const BaseReal&);
    friend PICBOOL operator <= (const BaseReal&, const BaseReal&);

/*
These are the usual comparisons of real numbers. If either of arguments
is a floating point number both arguments are converted floating point
numbers before comparing. Two null reals are equal and null is smaller
than any other real.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL isZero     () const;
    PICBOOL isPositive () const;
    PICBOOL isNegative () const;

/*
:ret isZero:     boolTrue if the receiver is zero.
:ret isPositive: boolTrue if the receiver is greater than zero.
:ret isNegative: boolTrue if the receiver is less than zero.
*/

// --------------------------------------------------------------------------
//      ARITHMETIC OPERATIONS
//:fun ----------------------------------------------------------------------

    friend BaseReal operator + (const BaseReal&, const BaseReal&);
    friend BaseReal operator - (const BaseReal&, const BaseReal&);
    friend BaseReal operator * (const BaseReal&, const BaseReal&);
    friend BaseReal operator / (const BaseReal&, const BaseReal&);

/*
These operators return a real number having the greater precision (smaller
digit-value) of the operands. Result is rounded according to that precision.

:err nullInstance: Either of the operands is null. The return value is
                   undefined and shouldn't be used.
:err divideByZero: Second argument of operator / is zero.
:err overflow:     Result of the operation falls outside the range of
                   BaseReal.
*/

// --------------------------------------------------------------------------
//      ASSIGNMENT + ARITHMETIC OPERATIONS
//:fun ----------------------------------------------------------------------

    BaseReal& operator += (const BaseReal&);
    BaseReal& operator -= (const BaseReal&);
    BaseReal& operator *= (const BaseReal&);
    BaseReal& operator /= (const BaseReal&);

    BaseReal& operator += (PICLONG);
    BaseReal& operator -= (PICLONG);
    BaseReal& operator *= (PICLONG);
    BaseReal& operator /= (PICLONG);

/*
These methods change the value of the receiving instance, but its precision
is not altered. The result is rounded according to the receiver.

:err nullInstance: Either the receiver or the argument is null.
:err divideByZero: Argument of operator /= is zero.
:err overflow:     Result of the operation falls outside the range of
                   BaseReal.
*/

// --------------------------------------------------------------------------
//      ABSOLUTE VALUE
//:fun ----------------------------------------------------------------------

    void makeAbs ();

/*
This method makes an absolute value of the receiver.

:err nullInstance: The receiver is null.
*/

// --------------------------------------------------------------------------
//      NEGATE
//:fun ----------------------------------------------------------------------

    void negate ();

/*
This method negates of the receiver.

:err nullInstance: The receiver is null.
*/

// --------------------------------------------------------------------------
//      ROUND & TRUNCATE
//:fun ----------------------------------------------------------------------

    void round    (PICDIGIT);
    void truncate (PICDIGIT);

/*
These methods round or truncate the receiver at the digit counted from
decimal point so that 0 rounds to integer, -1 to one decimal, 1 to tenths.
If the argument is floatReal, a floating point BaseReal is created.

Effect is the following:

    digit    a        a.round     a.truncate
    --------------------------------------
    -2    12345.67    12345.67    12345.67
    -1    12345.67    12345.70    12345.60
     0    12345.67    12346.00    12345.00
     1    12345.67    12350.00    12340.00
     2    12345.67    12300.00    12300.00
     f    12345.67    12345.6700000000009805497...

:err nullInstance: The receiver is null.
:err overflow:     Result of the operation falls outside the range of
                   BaseReal.
*/

// --------------------------------------------------------------------------
//      ROOF & FLOOR etc.
//:fun ----------------------------------------------------------------------

    PICLONG  roof        () const;
    PICLONG  floor       () const;
    PICLONG  integerPart () const;
    BaseReal fraction    () const;

/*
:ret roof:        Integer which is always greater or equal compared to
                  the receiver.
:ret floor:       Integer which is always less or equal compared to the
                  receiver.
:ret integerPart: Integer which is always towards zero.
:ret fraction:    BaseReal equal to x - x.integerPart(). Precision is
                  that of the receiver.

:err overflow:     If long integer range was exceeded.
:err nullInstance: Possible error for roof, floor and integerPart.
*/

// --------------------------------------------------------------------------
//      FORMATTING and DEFORMATTING
//:fun ----------------------------------------------------------------------

    BaseString getImage () const;
    BaseString getImage (PICDIGIT, char decPoint = '\0') const;

/*
These methods return the textual representation of the receiver. The number
of decimals and decimal separator can be given as arguments. A null string
is returned if the receiver is null.
*/

//:fun ----------------------------------------------------------------------

    void setImage (const BaseString& image);

/*
This method sets the value of the receiver based on the given numeric
string.

:err deformatError: The given string doesn't represent a real number.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const BaseReal&);

    PICOFFSET writeFmt (
        PicTextFmt&, PICOFFSET,
        const char* typeName) const;

/*
Output operator to write a BaseReal to PicFmtText. The method writeFmt
implements formatted output for BaseReal.
*/

    private:
// ==========================================================================
//      INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS
//
//      imp        Unscaled value as double.
//      flags      Contains precision and nullity flag.

    PICFLOAT  imp;
    PICUSHORT flags;

};

/*:class ********************************************************************

BaseReal uses default parameters until an PicRealIniter is created (either
as an automatic or static object). Given values are used until another
PicRealIniter is created. When PicRealIniter is destroyed the previous
values are restored.

****************************************************************************/

class PicRealIniter
{
    public:

    PicRealIniter (char decPoint);
   ~PicRealIniter ();

    private:

    char prevDecPoint;
};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PICBOOL operator != (const BaseReal& value1, const BaseReal& value2)
       { return !(value1 == value2); }
inline PICBOOL operator < (const BaseReal& value1, const BaseReal& value2)
       { return value2 > value1; }
inline PICBOOL operator <= (const BaseReal& value1, const BaseReal& value2)
       { return value2 >= value1; }

#endif
