// --------------------------------------------------------------------------
//  MEMORY ALLOCATION
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PALLOC_HXX
#define PALLOC_HXX

#include <picbase.hxx>

// --------------------------------------------------------------------------
//      MEMORY ALLOCATION ROUTINES
//:fun ----------------------------------------------------------------------

PICUINT picCalculateSize (PICUINT bytes);

/*
Memory buffers are allocated in chunks of 16/32/64/128/256/512/1024/2048/
4096/+2048 .. maxNewSize bytes. The reasoning for this allocation method
is that it will reduce heap fragmentation. The function picCalculateSize
implements the calculation. You should check the allocation size against
maxNewSize before using this routine!

:arg const PICUSHORT: Requested size.
:ret PICUSHORT: Calculated chunck size.
*/

//:fun ----------------------------------------------------------------------

void* picAllocate (PICUINT bytes, PICBOOL clear = boolTrue);

/*
Allocation returns pointer to buffer able to store the requested amount
of bytes (use picCalculateSize to calculate allocation by chunks).

:arg bytes: Number of bytes to allocate.
:arg clear: If boolTrue allocate zeros the buffer.
:ret void*: Pointer to allocated buffer.
:err MEMORY_ALLOCATION_ERROR: Allocation failed. NULLPTR is returned.
*/

//:fun ----------------------------------------------------------------------

void* picResize (
    void*   oldBuffer,
    PICUINT oldBytes,
    PICUINT newBytes,
    PICBOOL clear = boolTrue);

/*
Resize allocates buffer with the given size and copies bytes from the
old buffer to newly created buffer. If new buffer is smaller than the
old buffer only sizeof(new buffer) is copied.

:arg oldBuffer: Pointer to old buffer.
:arg oldBytes:  Size of the old buffer. Must be zero if the old buffer
                doesn't exist.
:arg newBytes:  New buffer size.
:arg clear:     If boolTrue resize zeros the part of new buffer that
                is not copied from the old buffer.
:ret void*: Pointer to allocated buffer.
:err MEMORY_ALLOCATION_ERROR: Allocation failed. NULLPTR is returned.
*/

//:fun ----------------------------------------------------------------------

void picFree (void* buffer);

/*
The picFree-function should be used to free the buffer allocated
with picAllocate. Do not use it to free buffers allocated with
new or malloc!

:arg void*: Pointer to buffer to be deleted.
*/

#endif 
