// --------------------------------------------------------------------------
//  ARRAY OBJECTS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PARRAY_HXX
#define PARRAY_HXX

#include <pcollect.hxx>
#include <pobjarr.hxx>

/*:class *** ARRAY OBJECT ***************************************************

PicArrayObject is an abstract base class for two collections of PicObjects:
PicArray and PicSortedArray.

****************************************************************************/

class PicArrayObject : public PicSeqIxCollect
{

    public:
// --------------------------------------------------------------------------
//:fun

    virtual PicObject&      operator = (const PicObject&);
            PicArrayObject& operator = (const PicArrayObject&);

/*
Assignment copies indexed collections.
*/

// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicArrayObject)

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL operator == (const PicObject&) const;

    PICBOOL isSubsetOf (const PicArrayObject&) const;

/*
These methods perform a deep comparison of two arrays. Arrays are equal if
they contain equal components in the same order. A subset array contains
equal components in any order.
*/

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

/*
Null array is considered empty. It does not allocate dynamic memory.
*/

// --------------------------------------------------------------------------

    virtual PicObject& access  (const PicIterator&);
    virtual PicObject* replace (PicIterator&, PicObject*) =0;

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents);

    virtual PICINDEX indexAt   (const PicIterator&)     const;
    virtual PICBOOL  findIndex (PICINDEX, PicIterator&) const;
    virtual PICBOOL  insert    (PICINDEX, PicIterator&);
    virtual PICBOOL  put       (PICINDEX, PicIterator&);

    virtual void deleteAll ();
    virtual void removeAll ();

    virtual PICBOOL next     (PicIterator&) const;
    virtual PICBOOL previous (PicIterator&) const;

// --------------------------------------------------------------------------
//:fun

    void purgeOnOrder ();

/*
Purges all empty locations. The order is preserved, but component indices
may change.
*/

// --------------------------------------------------------------------------
//:fun

    virtual PICSIZE size    () const;
            PICBOOL isEmpty () const;

    PICINDEX firstIndex () const;
    PICINDEX lastIndex  () const;
    PICINDEX minIndex   () const;
    PICINDEX maxIndex   () const;

    PICBOOL  inRange    (PICINDEX) const;
    PICINDEX firstFree  ()         const;

    PICINDEX smallest   () const;
    PICINDEX greatest   () const;

/*
:ret firstIndex: The index of the first component or outOfIndex if the array
                 is empty.
:ret lastIndex:  The index of the last component or outOfIndex - 1 if the
                 array is empty.
:ret minIndex:   The first index available (always 1).
:ret maxIndex:   The last index available without resizing.
:ret inRange:    True if the given index is within minIndex and lastIndex.
:ret firstFree:  The first unused location in the array.
:ret smallest:   The index of the smallest component or outOfIndex if
                 the array is empty.
:ret greatest:   The index of the greatest component or outOfIndex if
                 the array is empty.
*/

    protected:
// ==========================================================================
//:fun

    PicArrayObject ();
    PicArrayObject (const PicArrayObject&);
    PicArrayObject (const PicObjectPtrArray&);

    const PicObjectPtrArray& objectArray () const;
          PicObjectPtrArray& objectArray ();

/*
PicArrayObject has the following protected constructors:

    1. PicArrayObject array;
    2. PicArrayObject array (anArray);
    3. PicArrayObject array (anObjectPtrArray);

1. Creates a null array that doesn't allocate memory for components.
2. Creates an array with components copied from the argument.
3. Creates an array out of an object pointer array. The internal array
   can be accessed with the method objectArray.
*/

    PICBOOL equals (const PicArrayObject&) const;

    private:
// ==========================================================================
//      INSTANCE VARIABLES

    PicObjectPtrArray array;

};


/*:class *** ARRAY **********************************************************

PicArray is a sequentially indexed collection of pointers to PicObjects.
The first available index is always 1. The size of the array may change
dynamically as components are added. The insertion order of components is
preserved, although the indices may change.

****************************************************************************/

class PicArray : public PicArrayObject
{

    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicArray ();
    PicArray (const PicArray&);
    PicArray (const PicCollection& fromCollection);

/*
PicArrays can be created as follows:

    1. PicArray array;
    2. PicArray array (anArray);
    3. PicArray array (collection);

1. Creates a null array that doesn't allocate memory for components.
3. Creates an array with components copied from the argument.
3. Components are copied from a collection.
*/

//:fun ----------------------------------------------------------------------

   ~PicArray ();

/*
Deletes the array and all its components.
*/

// --------------------------------------------------------------------------

    PicObject& operator = (const PicObject&);
    PicArray&  operator = (const PicArray&);

    PICOBJECTCLASS(PicArray)

// --------------------------------------------------------------------------

    PICBOOL operator == (const PicObject&) const;
    PICBOOL operator == (const PicArray&)  const;

    virtual PicObject* replace (PicIterator&, PicObject*);
    virtual PICBOOL    purge   (PicIterator&, PICBOOL inOrder = boolTrue);

//:fun ----------------------------------------------------------------------

    PICBOOL putPtrAt (PICINDEX, PicObject*);

/*
Puts the given object at the given index. If there is a component at
the index, it is deleted. Returns true if the operation is successful.
*/

//:fun ----------------------------------------------------------------------

    void purge ();

/*
Purges all empty locations without preserving the order of components.
*/

// --------------------------------------------------------------------------
//:fun

    void sort (PICOBJSORTER* = PicSeqIxCollect::ascending);

/*
Array is sorted according to a compare function of the following form:

    PICSHORT compare (const PicObject& o1, const PicObject& o2)
    {
        // return 1 if o1 sorts after o2
        //        0 if o1 == o2
        //       -1 if o1 sorts before o2
    }

:err BaseCollection::unsortable: The array contains objects which are not
                                 comparable.
*/

// --------------------------------------------------------------------------

    PicArray& operator *= (const PicCollection&);
    PicArray& operator -= (const PicCollection&);
    PicArray& operator += (const PicIxCollect&);

    protected:

    PicArray (const PicObjectPtrArray&);
};


/*:class *** SORTED ARRAY ***************************************************

PicSortedArray is a sequentially indexed collection of pointers to
PicObjects. The first available index is always 1. The size of the array
may change dynamically as components are added. The components are kept
in order of deep comparison, so all components must be comparable. In
addition, a sorted array doesn't accept duplicate components. Because of
sorting the component indices may change at any time.

****************************************************************************/

class PicSortedArray : public PicArrayObject
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicSortedArray ();
    PicSortedArray (const PicSortedArray&);
    PicSortedArray (const PicCollection& fromCollection);

/*
PicSortedArrays can be created as follows:

    1. PicSortedArray array;
    2. PicSortedArray array (anArray);
    3. PicSortedArray array (collection);

1. Creates a null array that doesn't allocate memory for components.
3. Creates an array with components copied from the argument.
3. Components are copied from a collection.

The default sorting order is PicSeqIxCollect::ascending.
*/

//:fun ----------------------------------------------------------------------

   ~PicSortedArray ();

/*
Deletes the array and all its components.
*/

// --------------------------------------------------------------------------

    PicObject&      operator = (const PicObject&);
    PicSortedArray& operator = (const PicSortedArray&);

    PICOBJECTCLASS(PicSortedArray)

// --------------------------------------------------------------------------

    PICBOOL operator == (const PicObject&) const;
    PICBOOL operator == (const PicSortedArray&) const;

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL    insert  (PICINDEX, PicIterator&);
    virtual PICBOOL    put     (PICINDEX, PicIterator&);
    virtual PicObject* replace (PicIterator&, PicObject*);
    virtual PICBOOL    purge   (PicIterator&, PICBOOL inOrder = boolTrue);

    virtual PICBOOL findObject (const PicObject&, PicIterator& here) const;

/*
Replace sorts the new component to its proper location. Insert always
inserts at the end, because replace sorts the array anyway. Put doesn't
leave empty locations, because they would be sorted at the end. Searching
is implemented as a binary search of ordered components.
*/

// --------------------------------------------------------------------------
//:fun

    PICBOOL addPtr (PicObject*);

/*
This method adds the given object to the array and sorts it to the proper
location. It returns boolTrue if the operation was successful.
*/

// --------------------------------------------------------------------------
//:fun

    void sort (PICOBJSORTER* = PicSeqIxCollect::ascending);

    virtual PICBOOL isSorted () const;

/*
This method sets the sorting criterion for the array. The array is sorted
according to the new order and it retains the order from that point on.

:ret isSorted: boolTrue.
*/

// --------------------------------------------------------------------------

    PicSortedArray& operator *= (const PicCollection&);
    PicSortedArray& operator -= (const PicCollection&);
    PicSortedArray& operator += (const PicIxCollect&);

    protected:

    PicSortedArray (const PicObjectPtrArray&);

    private:

    PICOBJSORTER* order;
};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicArrayObject::PicArrayObject () : PicSeqIxCollect(), array() {}
inline PicArrayObject::PicArrayObject (const PicArrayObject& arr)
 : PicSeqIxCollect(), array(arr.array) {}
inline PicArrayObject::PicArrayObject (const PicObjectPtrArray& objarr)
 : PicSeqIxCollect(), array(objarr) {}

inline PicArrayObject& PicArrayObject::operator = (const PicArrayObject& arr)
       { array = arr.array; return *this; }
inline PICBOOL PicArrayObject::equals (const PicArrayObject& arr) const
       { return (array == arr.array); }
inline PICBOOL PicArrayObject::isSubsetOf (const PicArrayObject& arr) const
       { return arr.array.contains(this->array); }

inline void PicArrayObject::purgeOnOrder ()     { array.purge(boolTrue); }
inline PICBOOL PicArrayObject::isEmpty () const { return array.isEmpty(); }
inline PICBOOL PicArrayObject::allocate (PICSIZE size)
       { return array.put(size); }

inline PICINDEX PicArrayObject::firstIndex () const
       { PICPOSITION pos = array.firstPos(); return (PICINDEX)(pos ? pos : outOfIndex); }
inline PICINDEX PicArrayObject::lastIndex () const
       { PICPOSITION pos = array.lastPos(); return (PICINDEX)(pos ? pos : outOfIndex - 1); }
inline PICINDEX PicArrayObject::minIndex () const
       { return 1; }
inline PICINDEX PicArrayObject::maxIndex () const
       { return (PICINDEX)array.maxPos(); }
inline PICBOOL PicArrayObject::inRange (PICINDEX index) const
       { return array.inRange(index); }
inline PICINDEX PicArrayObject::firstFree () const
       { PICPOSITION pos = array.firstFree(); return (PICINDEX)(pos ? pos : outOfIndex); }

inline PICINDEX PicArrayObject::smallest () const
       { PICPOSITION pos = array.smallest(); return (PICINDEX)(pos ? pos : outOfIndex); }
inline PICINDEX PicArrayObject::greatest () const
       { PICPOSITION pos = array.greatest(); return (PICINDEX)(pos ? pos : outOfIndex); }

inline const PicObjectPtrArray& PicArrayObject::objectArray () const
       { return array; }
inline PicObjectPtrArray& PicArrayObject::objectArray ()
       { return array; }

// --------------------------------------------------------------------------

inline PicObject& PicArray::operator = (const PicObject& obj)
       { return PicArrayObject::operator = (obj); }
inline PicArray& PicArray::operator = (const PicArray& arr)
       { PicArrayObject::operator = (arr); return *this; }

inline PICBOOL PicArray::operator == (const PicObject& obj) const
       { return PicArrayObject::operator == (obj); }
inline PICBOOL PicArray::operator == (const PicArray& arr) const
       { return PicArrayObject::equals(arr); }

inline void PicArray::purge () { objectArray().purge(boolFalse); }

inline PicArray& PicArray::operator *= (const PicCollection& cl)
       { return (PicArray&)PicCollection::operator *= (cl); }
inline PicArray& PicArray::operator -= (const PicCollection& cl)
       { return (PicArray&)PicCollection::operator -= (cl); }
inline PicArray& PicArray::operator += (const PicIxCollect& ixcl)
       { return (PicArray&)PicSeqIxCollect::operator += (ixcl); }

inline PICBOOL PicArray::putPtrAt (PICINDEX index, PicObject* objPtr)
       { return objectArray().putAt(index, objPtr); }

// --------------------------------------------------------------------------

inline PicObject& PicSortedArray::operator = (const PicObject& obj)
       { return PicArrayObject::operator = (obj); }
inline PicSortedArray& PicSortedArray::operator = (const PicSortedArray& arr)
       { PicSeqIxCollect::operator = (arr); return *this; }

inline PICBOOL PicSortedArray::operator == (const PicObject& obj) const
       { return PicArrayObject::operator == (obj); }
inline PICBOOL PicSortedArray::operator == (const PicSortedArray& arr) const
       { return PicArrayObject::equals(arr); }

inline PicSortedArray& PicSortedArray::operator *= (const PicCollection& cl)
       { return (PicSortedArray&)PicCollection::operator *= (cl); }
inline PicSortedArray& PicSortedArray::operator -= (const PicCollection& cl)
       { return (PicSortedArray&)PicCollection::operator -= (cl); }
inline PicSortedArray& PicSortedArray::operator += (const PicIxCollect& ixcl)
       { return (PicSortedArray&)PicSeqIxCollect::operator += (ixcl); }

inline PICBOOL PicSortedArray::isSorted () const { return boolTrue; }

#endif
