// --------------------------------------------------------------------------
//  INDEX-OBJECT ASSOCIATION
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PASSOC_HXX
#define PASSOC_HXX

#include <pobject.hxx>
class BaseString;

/*:class ====================================================================
PicAssoc is an association of general object and numeric index. It is
mainly used with map collections.
*/

class PicAssoc
{

    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun

    PicAssoc ();
    PicAssoc (const PicAssoc&);
    PicAssoc (PICINDEX, PicObject*);

   ~PicAssoc ();

    PicAssoc& operator = (const PicAssoc&);

/*
PicAssoc can be created as

    1. PicAssoc association;
    2. PicAssoc association(index,anyObject);
    3. PicAssoc association(anotherAssociation);

1. Create null association
2. Associate index with given object (object is not copied, it becomes
   part of the association).
3. Copy given association (both index and the object).

Destructor deletes the object.
Assignment copies the association (both index and the object).
*/

// --------------------------------------------------------------------------

    PICCLASS()

// --------------------------------------------------------------------------
//      NULL ASSOCIATION
//:fun ----------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

/*
Null association does not allocate memory. Performing setNull deletes
the object in the association.
*/

// --------------------------------------------------------------------------
//      DEFAULT ASSOCIATION
//:fun ----------------------------------------------------------------------

    PICBOOL isDefault  () const;
    void    setDefault ();

/*
Association is in default state, when the object is. A null association
is in default state.
*/
// --------------------------------------------------------------------------
//      COMPARISON
//:fun ----------------------------------------------------------------------

    PICBOOL operator == (const PicAssoc&) const;
    PICBOOL operator <  (const PicAssoc&) const;

/*
Equality comparison compares both indexes and objects. Less than-operator
compares indexes only. Null associations are considered equal.
*/
// --------------------------------------------------------------------------
//      ACCESS
//:fun ----------------------------------------------------------------------

    PicObject* object () const;
    PICINDEX   index  () const;

/*
Unconstrained access to the contents of the association.
*/
//:fun ----------------------------------------------------------------------

    void setObject (PicObject*);
    void setIndex  (PICINDEX);


/*
Change index or object of the association. The old object (if any)
is deleted.
*/
//:fun ----------------------------------------------------------------------

    PicObject* removeObject ();  // Association is null afterwards
    void       swapObject   (PicObject*&);


/*
removeObject removes the object from the association and returns it.
The association is null afterwards.
swapObject swaps the associated object with the given one.
*/

    protected:
// ==========================================================================
//     INSTANCE VARIABLES

    PICINDEX   ind;
    PicObject* obj;

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicAssoc::PicAssoc () : ind(0), obj(NULLPTR) {}

inline PicAssoc::PicAssoc (PICINDEX ind, PicObject* obj) :
    ind(ind), obj(obj) {}

inline PICBOOL PicAssoc::isNull () const { return !obj; }

inline PICBOOL PicAssoc::isDefault () const
       { return (isNull() || obj->isDefault()); }
inline void PicAssoc::setDefault ()
       { if (obj) obj->setDefault(); }

inline PicObject* PicAssoc::object () const { return obj; }

inline PICINDEX PicAssoc::index () const { return ind; }

inline void PicAssoc::setIndex (PICINDEX index) { ind = index; }

inline void PicAssoc::setObject (PicObject* objPtr)
       { if (obj) delete obj; obj = objPtr; }

inline PicObject* PicAssoc::removeObject ()
{
    PicObject* object = obj;
    obj = NULLPTR;
    ind = 0;
    return object;
}

inline void PicAssoc::swapObject (PicObject*& newObj)
{
    PicObject* oldObj = obj;
    obj = newObj;
    newObj = oldObj;
}

inline PICBOOL PicAssoc::operator <  (const PicAssoc& ass) const
       { return ind < ass.ind; }

#endif
