// --------------------------------------------------------------------------
//  COLLECTION OBJECT base classes
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PICCOLLECTION_HXX
#define PICCOLLECTION_HXX

#include <pobject.hxx>
#include <piterate.hxx>
#include <bstring.hxx>
class PicIxCollect;
class PicNmCollect;

typedef PICSHORT PICIXSORTER(PICINDEX, PICINDEX);
typedef PICSHORT PICNMSORTER(const BaseString&, const BaseString&);
typedef PICSHORT PICOBJSORTER(const PicObject&, const PicObject&);

/*:class *** COLLECTION *****************************************************

PicCollection is an abstract base class for all the collection PicObjects,
i.e. objects containing other objects as components (like an array).

PicCollection defines methods that are applicable to all kinds of
collections; derived classes add methods for indexed or named collections.

****************************************************************************/

class PicCollection : public PicObject
{

    public:
// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicCollection)

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&) =0;

/*
Assignment operator must be defined for every derived class. Usually an
assignment of collection means copying the components of the argument
to the receiver (after the receiver has been emptied). Collections must
be similar enough, i.e. both indexed or both named.
*/

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL operator == (const PicObject&) const =0;
    virtual PICBOOL operator <  (const PicObject&) const;

/*
Equality comparison must be defined for every derived class. Usually two
collections are considered equal, when they contain equal components in
the same order. Objects of different collection classes are never equal.

Less than -operator is generally not defined for collections.
*/

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

    static PicCollection& null();

/*
A null collection is typically one that doesn't contain components and
doesn't allocate memory for them. It is normally treated like an empty
collection.

PicCollection::null() is a global null instance of PicNullCollection that
can be used e.g. as a return value when no valid collection can be
returned. It is always in null state.
*/

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();

/*
For collections the default state is a component-wise default state.
The default implementation of these methods uses iterator to query
the defaultness of each component or to set each component to its
default state.
*/

// --------------------------------------------------------------------------
//:fun

          PicObject& at (const PicIterator&);
    const PicObject& at (const PicIterator&) const;

          PicIxCollect& ixc (const PicIterator&);
    const PicIxCollect& ixc (const PicIterator&) const;
          PicNmCollect& nmc (const PicIterator&);
    const PicNmCollect& nmc (const PicIterator&) const;

    PICBOOL putAt (PicIterator&, const PicObject&);

          PicObject& first ();
    const PicObject& first () const;
          PicObject& last  ();
    const PicObject& last  () const;

/*
:list Method
:item at:    Returns the component at the given location. The iterator
             must point to an existing component. If the collection is
             constant, the returned reference is also constant.
:item ixc:   Accesses a component that is an indexed collection.
:item nmc:   Accesses a component that is a named collection. If the
             component is not a named (or indexed) collection, a fatal
             error is raised.
:item putAt: Stores a copy of the object at the given location. If there
             is a component in the location, that object is deleted.
:item first: Returns the first component in the collection or
             PicObject::null() if collection is empty.
:item last:  Returns the last component in the collection or
             PicObject::null() if collection is empty.
:end
*/

// --------------------------------------------------------------------------
//:fun  SEARCHING FOR COMPONENTS

    virtual PICBOOL findObject (const PicObject&, PicIterator&) const;

    PICBOOL find      (const PicObject&, PicIterator&) const;
    PICBOOL contains  (const PicObject&) const;

    PICBOOL findFirst (PicIterator&) const;
    PICBOOL findLast  (PicIterator&) const;

    PICBOOL findMin   (PicIterator&) const;
    PICBOOL findMax   (PicIterator&) const;

    PICBOOL findPosition (PICPOSITION n, PicIterator&) const;

/*
These methods search for a component and set an iterator to mark the
found object. If searching fails, false is returned. Typically these
methods are used as follows:

    PicIterator here;
    if (collection.find(object, here))
    {
        const PicObject& component = collection.at(here);
        ...
    }

:list Method
:item findObject:   Searches the object by using equality comparison.
:item find:         Same as findObject. This name is intended for
                    normal use and it is overloaded in derived classes;
                    findObject is virtual method to be redefined if
                    needed.
:item contains:     Same as find but without the iterator.
:item findFirst:    Finds the first component of the collection.
:item findLast:     Finds the last component of the collection.
:item findMin:      Finds the smallest component of the collection by using
                    PicObject::operator <.
:item findMax:      Finds the greatest component of the collection by using
                    PicObject::operator <.
:item findPosition: Finds the n:th component from the collection.
:end
*/

// --------------------------------------------------------------------------
//:fun  DELETING COMPONENTS

    PICBOOL deleteAt     (PicIterator&,     PicPurgeMode = noPurge);
    void    deleteObject (const PicObject&, PicPurgeMode = noPurge);

    virtual void deleteAll ();
    virtual void removeAll ();

/*
:list Method
:item deleteAt: Deletes the component at the given location and moves the
                iterator to the location of the previous object. If there
                is no previous object, returns false.
:item deleteObject: Searches a component by using equality comparison and
                deletes the object if it is found and if it is not the
                same object as the argument.
:item deleteAll: Deletes all objects in the collection.
:item removeAll: Removes all objects from the collection but doesn't delete
                 them (beware of memory leaks when using this method).
:end
*/

// --------------------------------------------------------------------------
//:fun  ASKING ATTRIBUTES

    virtual PICSIZE size    () const;
            PICBOOL isEmpty () const;

    virtual PICBOOL isSorted () const;

/*
Size is the number of components in the collection.

A sorted collection keeps its components in a specific order at all times.
By default, isSorted returns boolFalse.
*/

// --------------------------------------------------------------------------
//:fun  STANDARD ITERATOR

    virtual void    startIteration (PicIterator&) const;
    virtual PICBOOL next           (PicIterator&) const;
    virtual PICBOOL previous       (PicIterator&) const;

/*
These methods implement the standard iterator. They are usually used via
the macros ITERATE and ITERATE_BACKWARDS.

:list Method
:item startIteration: Initializes the iterator to a class-specific starting
                      value. After starting next moves to the first
                      component and previous to the last one. The
                      default implementation uses zero as starting value.
:item next:           Moves the iterator to point to the next component.
                      If there is no next component, returns false. The
                      default implementation only returns false.
:item previous:       Moves the iterator to point to the previous component.
                      If there is no previous component, returns false. The
                      default implementation only returns false.
:end
*/

// --------------------------------------------------------------------------
//:fun  SET OPERATORS

    PicCollection& operator *= (const PicCollection&);
    PicCollection& operator -= (const PicCollection&);

/*
Operators are provided to combine collections:

    a *= b    Intersection  (remove components not found in b).
    a -= b    Difference    (remove components found in b).

Operators are implemented by using equality comparison of components
(so they don't care about component indices or names). Both operators
return a reference to the receiver.
*/

// --------------------------------------------------------------------------
//:fun  VIRTUAL INTERFACE

    virtual PicObject& access (const PicIterator&) =0;

/*
All accessing is ultimately done via this method. It returns a reference
to the component at the given location. The iterator should always point
to an existing component when calling this method.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject* replace (PicIterator&, PicObject*) =0;

/*
This method replaces the component at the given location and returns the
old component. After a call to this method the argument belongs to the
collection and the returned object belongs to the caller.

If there was no old component at the location, NULLPTR is returned. If the
argument is NULLPTR the component is effectively removed. A collection that
stores whole objects must copy the argument, so it returns the argument
back to the caller.

The iterator may be modified if the collection sorts the new component to
another location. If a component is removed, the location may be purged if
the collection cannot have empty locations; in such case the iterator is
set to point to a location after the last component.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;

/*
This method discards the given location from the collection. If it contains
a component, that object is deleted. The iterator is moved to point to the
previous component, if any.

:arg inOrder: When this is true (the default), the collection is kept in
              order when purging. By allowing order to be disturbed the
              operation may be faster.

:ret PICBOOL: True if the iterator points to a valid location after purging.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL allocate (PICSIZE forComponents) =0;

/*
This method asks the collection to allocate memory for at least the
given number of components. This can be used to optimize filling if
you know the size beforehand.
*/

// --------------------------------------------------------------------------

    virtual BaseString getImage () const;
    virtual BaseString getImage (const PicObject& format) const;

    virtual void setImage (const BaseString& image);
    virtual void setImage (
        const PicObject& format,
        const BaseString& image);

    private:

    PICBOOL findMinMax (PicIterator&, PICBOOL) const;

};


/*:class *** INDEXED COLLECTION ********************************************

PicIxCollect is an abstract base class for collections whose components
are identified by an integer index.

****************************************************************************/

class PicIxCollect : public PicCollection
{
    public:
// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicIxCollect)

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&) =0;

/*
Assignment operator must be defined by derived classes.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator == (const PicObject&) const =0;

/*
Equality comparison must be defined by derived classes.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

    static PicIxCollect& null();

/*
PicIxCollect::null() is a global null instance of PicNullIxCollect that
can be used e.g. as a return value when no valid indexed collection can
be returned. It is always in null state.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& access  (const PicIterator&)       =0;
    virtual PicObject* replace (PicIterator&, PicObject*) =0;

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents) =0;

/*
These basic operations of collections must be implemented in derived
classes.
*/

//:fun ----------------------------------------------------------------------

          PicObject& at (const PicIterator&);
    const PicObject& at (const PicIterator&) const;

          PicIxCollect& ixc (const PicIterator&);
    const PicIxCollect& ixc (const PicIterator&) const;
          PicNmCollect& nmc (const PicIterator&);
    const PicNmCollect& nmc (const PicIterator&) const;

          PicObject& at (PICINDEX index);
    const PicObject& at (PICINDEX index) const;

          PicIxCollect& ixc (PICINDEX index);
    const PicIxCollect& ixc (PICINDEX index) const;
          PicNmCollect& nmc (PICINDEX index);
    const PicNmCollect& nmc (PICINDEX index) const;

/*
These methods return the component at the given location or index. An
iterator must point to an existing component. If an index doesn't refer
to an existing component, PicObject::null() is returned and an error is
raised. If the component is not an object of the desired class, a fatal
error is raised.
*/

//:fun ----------------------------------------------------------------------

    virtual PICINDEX indexAt    (const PicIterator&) const =0;
    virtual PICBOOL  findIndex  (PICINDEX, PicIterator&) const;

    PICBOOL find (PICINDEX, PicIterator&) const;
    PICBOOL find (const PicObject&, PicIterator&) const;

    PICBOOL contains (PICINDEX)         const;
    PICBOOL contains (const PicObject&) const;

    PICINDEX indexOf (const PicObject&) const;

    PICINDEX firstIndex () const;
    PICINDEX lastIndex  () const;

/*
The method indexAt returns the index of the given component. The iterator
must point to an existing component.

The method findIndex searches the component having the given index and
returns true if successful. The default implementation uses iterator and
indexAt; derived classes may provide more efficient implementations.

These two methods are intended for redefinition in derived classes. For
normal use the rest of the methods offer more convenient variants.

:list Method
:item find:     Searching by index or by value.
:item contains: Same as find but without the iterator argument.
:item indexOf:  Searches the given object by value and returns its index
                if successful, or outOfIndex otherwise.
:item firstIndex: Index of the first component or outOfIndex if the
                  collection is empty.
:item lastIndex:  Index of the last component or outOfIndex if the
                  collection is empty.
:end
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL putAtIndex (PICINDEX, const PicObject&, PicIterator&);

    PICBOOL putAt (PicIterator&, const PicObject&);
    PICBOOL putAt (PICINDEX, const PicObject&, PicIterator&);
    PICBOOL putAt (PICINDEX, const PicObject&);

/*
The method putAtIndex puts a copy of the given object to the collection
at the given index. If the index is already in use, the old component is
deleted. The location of the index is set to the iterator. Return value is
true if the operation was successful. The default implementation uses
findIndex to locate the place where to put the object, so it cannot
create new indices. The derived classes PicKeyIxCollect and PicSeqIxCollect
add the ability to create indices.

The overloaded variants of putAt are intended for normal use, whereas
putAtIndex is intended to be redefined in derived classes.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL    deleteAt (PicIterator&,   PicPurgeMode = noPurge);
    void       deleteAt (PICINDEX index, PicPurgeMode = noPurge);
    PicObject* remove   (PICINDEX index, PicPurgeMode = noPurge);

/*
:list Method
:item deleteAt: Deletes the component at the given location or index. If
                the index is not found, nothing is done.
:item remove:   Removes the component at the given index. If the index is
                not found, a NULLPTR is returned.
:end
*/

//:fun ----------------------------------------------------------------------

    static PICSHORT ascending  (PICINDEX, PICINDEX);
    static PICSHORT descending (PICINDEX, PICINDEX);

/*
These two functions are provided for sorting purposes. They sort an
indexed collection in ascending or descending order according to the
component indices.
*/

// --------------------------------------------------------------------------
//:fun  SET OPERATOR

    PicIxCollect& operator &= (const PicIxCollect&);

/*
Operator is provided to combine indexed collections:

    a &= b   Update a from b (replace already existing components).

Operator is implemented by comparing component indices, so it doesn't
care about the actual component objects. The operator returns a reference
to the receiver.
*/

};


/*:class *** INDEX-KEYED COLLECTION *****************************************

PicKeyIxCollect is an abstract base class for collections whose components
are identified by an integer index that is a unique key. In other words,
the collection contains index-value pairs.

****************************************************************************/

class PicKeyIxCollect : public PicIxCollect
{
    public:
// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicKeyIxCollect)

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);

/*
The assignment operator copies an indexed collection by using the
method putAt.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator == (const PicObject&) const =0;

/*
Equality comparison must be defined by derived classes.
*/

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

//:fun ----------------------------------------------------------------------

    virtual PICINDEX   indexAt (const PicIterator&) const =0;
    virtual PicObject& access  (const PicIterator&)       =0;
    virtual PicObject* replace (PicIterator&, PicObject*) =0;

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents) =0;

/*
These basic operations of collections must be implemented in derived
classes.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL insert     (PICINDEX, PicIterator&) =0;
    virtual PICBOOL putAtIndex (PICINDEX, const PicObject&, PicIterator&);

/*
The method insert creates a new location with the given index. It
returns true after successful operation (either the index existed or it
was successfully inserted). New location is initially empty, so you should
fill it as soon as possible by calling replace.

The method putAtIndex combines insert and replace in a more convenient
form.
*/

// --------------------------------------------------------------------------
//:fun  SET OPERATORS

    PicKeyIxCollect& operator |= (const PicIxCollect&);
    PicKeyIxCollect& operator += (const PicIxCollect&);

/*
Operators are provided to combine indexed collections:

    a |= b      Supplement a from b (add non-existing components).
    a += b      Add b to a (&= and |= together).

Operators are implemented by comparing component indices, so they don't
care about the actual component objects. Each operator returns a reference
to the receiver.
*/

};


/*:class *** SEQUENTIAL INDEXED COLLECTION **********************************

PicSeqIxCollect is an abstract base class for collections whose components
are identified by a sequential index, i.e. the components are not tied to
any particular index but the indices may change at any time when components
are added or removed.

****************************************************************************/

class PicSeqIxCollect : public PicIxCollect
{
    public:
// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicSeqIxCollect)

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);

/*
The assignment operator copies an indexed collection by using the
method putAt.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator == (const PicObject&) const =0;

/*
Equality comparison must be defined by derived classes.
*/

// ----------------------------------------------------------------------

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

//:fun ----------------------------------------------------------------------

    virtual PICINDEX   indexAt (const PicIterator&) const =0;
    virtual PicObject& access  (const PicIterator&)       =0;
    virtual PicObject* replace (PicIterator&, PicObject*) =0;

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents);

/*
These basic operations of collections must be implemented in derived
classes.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL insert     (PICINDEX, PicIterator&) =0;
    virtual PICBOOL put        (PICINDEX, PicIterator&) =0;
    virtual PICBOOL putAtIndex (PICINDEX, const PicObject&, PicIterator&);

    PICBOOL add   (const PicObject&);
    PICBOOL add   (const PicObject&, PicIterator&);

    PICBOOL addAt (PicIterator&, const PicObject&);
    PICBOOL addAt (PICINDEX,     const PicObject&, PicIterator&);
    PICBOOL addAt (PICINDEX,     const PicObject&);

/*
The methods insert and put create a new location at the given index.
They return true after successful operation (either the location existed
or it was successfully inserted). The iterator is set to point to the
location. New location is initially empty, so you should fill it as soon
as possible by calling replace.

:list Method
:item insert: Inserts an empty location at the given index moving components
              towards the end of the collection. Inserting to the index
              outOfIndex means inserting at the end.
:item put:    Ensures that a location is available at the given index.
              The location may or may not contain a component before the
              operation.
:end

The method putAtIndex combines put and replace in a more convenient
form. Likewise add and addAt combine insert and replace.
*/

// --------------------------------------------------------------------------
//:fun  SET OPERATORS

    PicSeqIxCollect& operator += (const PicCollection&);

/*
This is the union-operator: it adds all components of the argument to
the receiver. If the receiver doesn't accept duplicates, they are
deleted in the process. The operator returns a reference to the receiver.
*/

//:fun ----------------------------------------------------------------------

    static PICSHORT ascending  (const PicObject&, const PicObject&);
    static PICSHORT descending (const PicObject&, const PicObject&);

/*
These two functions are provided for sorting purposes. They sort a
sequential collection in ascending or descending order according to
the operator < of the component objects.
*/

};


/*:class *** NAMED COLLECTION ***********************************************

PicNmCollect is an abstract base class for collections whose components
are identified by a textual key, i.e. the components are name-value pairs.

****************************************************************************/

class PicNmCollect : public PicCollection
{
    public:
// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicNmCollect)

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);

/*
The assignment operator copies a named collection by using the
method putAt.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator == (const PicObject&) const =0;

/*
Equality comparison must be defined by derived classes.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

    static PicNmCollect& null();

/*
PicNmCollect::null() is a global null instance of PicNullNmCollect that
can be used e.g. as a return value when no valid named collection can be
returned. It is always in null state.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& access  (const PicIterator&)       =0;
    virtual PicObject* replace (PicIterator&, PicObject*) =0;

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents) =0;

/*
These basic operations of collections must be implemented in derived
classes.
*/

//:fun ----------------------------------------------------------------------

          PicObject& at (const PicIterator&);
    const PicObject& at (const PicIterator&) const;

          PicIxCollect& ixc (const PicIterator&);
    const PicIxCollect& ixc (const PicIterator&) const;
          PicNmCollect& nmc (const PicIterator&);
    const PicNmCollect& nmc (const PicIterator&) const;

          PicObject& at (const BaseString& name);
    const PicObject& at (const BaseString& name) const;
          PicObject& at (const char*    name);
    const PicObject& at (const char*    name) const;

          PicIxCollect& ixc (const char* name);
    const PicIxCollect& ixc (const char* name) const;
          PicNmCollect& nmc (const char* name);
    const PicNmCollect& nmc (const char* name) const;

/*
These methods return the component at the given location or name. An
iterator must point to an existing component. If a name doesn't refer
to an existing component, PicObject::null() is returned and an error is
raised. If the component is not an object of the desired class, a fatal
error is raised.
*/

//:fun ----------------------------------------------------------------------

    virtual const char* nameAt (const PicIterator&) const =0;

    virtual PICBOOL findName (const char* name, PicIterator&) const;

    PICBOOL find (const BaseString& name, PicIterator&) const;
    PICBOOL find (const char*    name, PicIterator&) const;
    PICBOOL find (const PicObject&,       PicIterator&) const;

    PICBOOL contains (const BaseString& name) const;
    PICBOOL contains (const char*    name) const;
    PICBOOL contains (const PicObject&)       const;

    BaseString nameOf (const PicObject&) const;

    BaseString firstName () const;
    BaseString lastName  () const;

/*
The method nameAt returns the name of the given component. The iterator
must point to an existing component.

The method findName searches the component having the given name and
returns true if successful. The default implementation uses iterator and
nameAt; derived classes may provide more efficient implementations.

These two methods are intended for redefinition in derived classes. For
normal use the rest of the methods offer more convenient variants.

:list Method
:item find:      Searching by name or by value.
:item contains:  Same as find but without the iterator argument.
:item nameOf:    Searches the given object by value and returns its name
                 if successful, or a null string otherwise.
:item firstName: Name of the first component or null string if the
                 collection is empty.
:item lastName:  Name of the last component or null string if the
                 collection is empty.
:end
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL insert (const char* name, PicIterator&) =0;

    PICBOOL putAt (PicIterator&,           const PicObject&);
    PICBOOL putAt (const BaseString& name, const PicObject&);
    PICBOOL putAt (const BaseString& name, const PicObject&, PicIterator&);
    PICBOOL putAt (const char*    name, const PicObject&);
    PICBOOL putAt (const char*    name, const PicObject&, PicIterator&);

/*
The method insert creates a new location with the given name. It
returns true after successful operation (either the name existed or it
was successfully inserted). New location is initially empty, so you should
fill it as soon as possible by calling replace.

The method putAt(name) puts a copy of the given object to the collection
at the given name. If the name is already in use, the old component is
deleted. The location of the name is set to the iterator. Return value is
true if the operation was successful. The default implementation uses
insert and replace.

The overloaded variants of putAt are intended for normal use, whereas
insert is intended to be redefined in derived classes.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL    deleteAt (PicIterator&,           PicPurgeMode = noPurge);
    void       deleteAt (const BaseString& name, PicPurgeMode = noPurge);
    void       deleteAt (const char*    name, PicPurgeMode = noPurge);
    PicObject* remove   (const BaseString& name, PicPurgeMode = noPurge);
    PicObject* remove   (const char*    name, PicPurgeMode = noPurge);

/*
:list Method
:item deleteAt: Deletes the component at the given location or name. If
                the name is not found, nothing is done.
:item remove:   Removes the component at the given name. If the name is
                not found, a NULLPTR is returned.
:end
*/

// --------------------------------------------------------------------------
//:fun  SET OPERATORS

    PicNmCollect& operator &= (const PicNmCollect&);
    PicNmCollect& operator |= (const PicNmCollect&);
    PicNmCollect& operator += (const PicNmCollect&);

/*
Operators are provided to combine collections:

    a &= b      Update a from b (replace already existing components).
    a |= b      Supplement a from b (add non-existing components).
    a += b      Add b to a (&= and |= together).

Operators are implemented in terms of component names, i.e. they don't
care about the actual component objects. Each operator returns a reference
to the receiver.
*/

//:fun ----------------------------------------------------------------------

    static PICSHORT ascending  (const BaseString&, const BaseString&);
    static PICSHORT descending (const BaseString&, const BaseString&);

/*
These two functions are provided for sorting purposes. They sort a
named collection in ascending or descending order according to the
component names.
*/

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PICBOOL PicCollection::find (const PicObject& obj, PicIterator& here) const
       { return findObject(obj, here); }

#define SELF ((PicCollection*)this)

inline PicObject& PicCollection::at (const PicIterator& i)
       { return access(i); }
inline const PicObject& PicCollection::at (const PicIterator& i) const
       { return SELF->access(i); }

inline PicIxCollect& PicCollection::ixc (const PicIterator& i)
       { return AS(PicIxCollect, access(i)); }
inline const PicIxCollect& PicCollection::ixc (const PicIterator& i) const
       { return AS(PicIxCollect, SELF->access(i)); }
inline PicNmCollect& PicCollection::nmc (const PicIterator& i)
       { return AS(PicNmCollect, access(i)); }
inline const PicNmCollect& PicCollection::nmc (const PicIterator& i) const
       { return AS(PicNmCollect, SELF->access(i)); }

inline const PicObject& PicCollection::first () const
       { return SELF->first(); }
inline const PicObject& PicCollection::last () const
       { return SELF->last(); }

#undef SELF

inline PICBOOL PicCollection::contains (const PicObject& obj) const
       { PicIterator here; return find(obj, here); }

inline PICBOOL PicCollection::isEmpty () const
       { return size() == 0; }

inline PICBOOL PicCollection::isSorted () const { return boolFalse; }

inline PICBOOL PicCollection::findFirst (PicIterator& here) const
       { startIteration(here); return next(here); }
inline PICBOOL PicCollection::findLast (PicIterator& here) const
       { startIteration(here); return previous(here); }

inline PICBOOL PicCollection::findMin (PicIterator& here) const
       { return findMinMax(here, boolTrue); }
inline PICBOOL PicCollection::findMax (PicIterator& here) const
       { return findMinMax(here, boolFalse); }

// --------------------------------------------------------------------------

inline PicObject& PicIxCollect::at (const PicIterator& i)
       { return PicCollection::at(i); }
inline const PicObject& PicIxCollect::at (const PicIterator& i) const
       { return PicCollection::at(i); }

#define SELF ((PicIxCollect*)this)

inline const PicObject& PicIxCollect::at (PICINDEX index) const
       { return SELF->at(index); }

inline PicIxCollect& PicIxCollect::ixc (const PicIterator& i)
       { return AS(PicIxCollect, access(i)); }
inline const PicIxCollect& PicIxCollect::ixc (const PicIterator& i) const
       { return AS(PicIxCollect, SELF->access(i)); }
inline PicNmCollect& PicIxCollect::nmc (const PicIterator& i)
       { return AS(PicNmCollect, access(i)); }
inline const PicNmCollect& PicIxCollect::nmc (const PicIterator& i) const
       { return AS(PicNmCollect, SELF->access(i)); }

inline PicIxCollect& PicIxCollect::ixc (PICINDEX index)
       { return AS(PicIxCollect, at(index)); }
inline const PicIxCollect& PicIxCollect::ixc (PICINDEX index) const
       { return AS(PicIxCollect, SELF->at(index)); }
inline PicNmCollect& PicIxCollect::nmc (PICINDEX index)
       { return AS(PicNmCollect, at(index)); }
inline const PicNmCollect& PicIxCollect::nmc (PICINDEX index) const
       { return AS(PicNmCollect, SELF->at(index)); }

#undef SELF

inline PICBOOL PicIxCollect::find (PICINDEX index, PicIterator& here) const
       { return findIndex(index, here); }
inline PICBOOL PicIxCollect::find (const PicObject& obj, PicIterator& here) const
       { return PicCollection::find(obj, here); }

inline PICBOOL PicIxCollect::contains (const PicObject& obj) const
       { return PicCollection::contains(obj); }
inline PICBOOL PicIxCollect::contains (PICINDEX index) const
       { PicIterator here; return find(index, here); }

inline PICBOOL PicIxCollect::putAt (PicIterator& here, const PicObject& obj)
       { return PicCollection::putAt(here, obj); }
inline PICBOOL PicIxCollect::putAt (PICINDEX index, const PicObject& obj, PicIterator& here)
       { return putAtIndex(index, obj, here); }
inline PICBOOL PicIxCollect::putAt (PICINDEX index, const PicObject& obj)
       { PicIterator here; return putAtIndex(index, obj, here); }

inline PICBOOL PicIxCollect::deleteAt (PicIterator& here, PicPurgeMode purge)
       { return PicCollection::deleteAt(here, purge); }

// --------------------------------------------------------------------------

inline PICBOOL PicSeqIxCollect::add (const PicObject& obj)
       { PicIterator here; return add(obj, here); }

inline PICBOOL PicSeqIxCollect::addAt (PicIterator& here, const PicObject& obj)
       { return addAt(indexAt(here), obj, here); }
inline PICBOOL PicSeqIxCollect::addAt (PICINDEX index, const PicObject& obj)
       { PicIterator here; return addAt(index, obj, here); }

inline PICBOOL PicSeqIxCollect::allocate (PICSIZE size)
       { PicIterator i; return put(size,i); }

// --------------------------------------------------------------------------

inline PicObject& PicNmCollect::at (const PicIterator& i)
       { return PicCollection::at(i); }
inline const PicObject& PicNmCollect::at (const PicIterator& i) const
       { return PicCollection::at(i); }
inline PicObject& PicNmCollect::at (const BaseString& name)
       { return at(name.chars()); }

#define SELF ((PicNmCollect*)this)

inline const PicObject& PicNmCollect::at (const char* name) const
       { return SELF->at(name); }
inline const PicObject& PicNmCollect::at (const BaseString& name) const
       { return SELF->at(name.chars()); }

inline PicIxCollect& PicNmCollect::ixc (const PicIterator& i)
       { return AS(PicIxCollect, access(i)); }
inline const PicIxCollect& PicNmCollect::ixc (const PicIterator& i) const
       { return AS(PicIxCollect, SELF->access(i)); }
inline PicNmCollect& PicNmCollect::nmc (const PicIterator& i)
       { return AS(PicNmCollect, access(i)); }
inline const PicNmCollect& PicNmCollect::nmc (const PicIterator& i) const
       { return AS(PicNmCollect, SELF->access(i)); }

inline PicIxCollect& PicNmCollect::ixc (const char* name)
       { return AS(PicIxCollect, at(name)); }
inline const PicIxCollect& PicNmCollect::ixc (const char* name) const
       { return AS(PicIxCollect, SELF->at(name)); }
inline PicNmCollect& PicNmCollect::nmc (const char* name)
       { return AS(PicNmCollect, at(name)); }
inline const PicNmCollect& PicNmCollect::nmc (const char* name) const
       { return AS(PicNmCollect, SELF->at(name)); }

#undef SELF

inline PICBOOL PicNmCollect::find (const char* name, PicIterator& here) const
       { return findName(name, here); }
inline PICBOOL PicNmCollect::find (const BaseString& name, PicIterator& here) const
       { return findName(name.chars(), here); }
inline PICBOOL PicNmCollect::find (const PicObject& obj, PicIterator& here) const
       { return PicCollection::find(obj, here); }

inline PICBOOL PicNmCollect::contains (const PicObject& obj) const
       { return PicCollection::contains(obj); }
inline PICBOOL PicNmCollect::contains (const char* name) const
       { PicIterator here; return findName(name, here); }
inline PICBOOL PicNmCollect::contains (const BaseString& name) const
       { PicIterator here; return findName(name.chars(), here); }

inline PICBOOL PicNmCollect::putAt (PicIterator& here, const PicObject& obj)
       { return PicCollection::putAt(here, obj); }
inline PICBOOL PicNmCollect::putAt (const char* name, const PicObject& obj)
       { PicIterator here; return putAt(name, obj, here); }
inline PICBOOL PicNmCollect::putAt (const BaseString& name, const PicObject& obj, PicIterator& here)
       { return putAt(name.chars(), obj, here); }
inline PICBOOL PicNmCollect::putAt (const BaseString& name, const PicObject& obj)
       { PicIterator here; return putAt(name.chars(), obj, here); }

inline PICBOOL PicNmCollect::deleteAt (PicIterator& here, PicPurgeMode purge)
       { return PicCollection::deleteAt(here, purge); }
inline void PicNmCollect::deleteAt (const BaseString& name, PicPurgeMode purge)
       { deleteAt(name.chars(), purge); }

inline PicObject* PicNmCollect::remove (const BaseString& name, PicPurgeMode purge)
       { return remove(name.chars(), purge); }

#endif
