// --------------------------------------------------------------------------
//  DATE OBJECT
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PDATE_HXX
#define PDATE_HXX

#include <pobject.hxx>
#include <bdate.hxx>

/*:class ********************************************************************

PicDate implements the basic date handling operations. Date values include
a day, month and year component. Null date is used to represent unknown or
missing value. At any moment after creation PicDate instance is either null
or represents a valid date.

Date format can be selected with the PicDateIniter class.

PicDate belongs to the kernel level 2 library. It has a counterpart BaseDate
in the level 1. Most date handling operations are defined for BaseDate and
inherited by PicDate.

****************************************************************************/

class PicDate : public PicPrimitive, public BaseDate
{
    public:

    PICOBJECTCLASS(PicDate)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicDate ();
    PicDate (PicRelativeDate, PICLONG unitInterval = 0);
    PicDate (PICYEAR, PICMONTH, PICDAY);
    PicDate (
        const BaseString&,
        PicDateStyle,
        char delimiter);
    PicDate (
        const BaseString&,
        PicDateStyle);
    PicDate (const BaseString&);
    PicDate (const BaseDate&);
    PicDate (const PicDate&);

/*
PicDates can be created in the following ways:

    1. PicDate a;
    2. PicDate b (PicDate::today, +2);
    3. PicDate c (1991, 12, 24);
    4. PicDate d ("24.12.1991");
    5. PicDate e ("1991-12-24", PicDate::yearFirst, picMinus);
    6. PicDate f (aBaseDate);
    7. PicDate g (anPicDate);

1. Creates a null PicDate.
2. Creates an PicDate with the value 'day after tomorrow' (today + 2 days).
   This constructor uses the current system date.
3. Creates an PicDate with the value 24.12.1991 given as components.
4. Creates an PicDate with the value 24.12.1991 in default format.
5. Creates an PicDate with the value 24.12.1991 in a specified format.
6. Copies a BaseDate.
7. Copies another PicDate.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid date.

*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);
            PicDate&   operator = (const PicDate&);
            PicDate&   operator = (const BaseDate&);

/*
These methods copy the date value.

:arg PicObject&: Reference to an PicDate.
:ret PicDate&, PicObject&: Reference to the receiver.
:err typeMismatch: PicObject&-argument is not an PicDate.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&) const;
            PICBOOL operator != (const PicObject&) const;
    virtual PICBOOL operator <  (const PicObject&) const;

            PICBOOL operator == (const PicDate&)   const;
            PICBOOL operator != (const PicDate&)   const;
            PICBOOL operator <  (const PicDate&)   const;

            PICBOOL operator == (const BaseDate&)  const;
            PICBOOL operator != (const BaseDate&)  const;
            PICBOOL operator <  (const BaseDate&)  const;

/*
PicDates are compared along the time axis. Later dates are considered
greater.
Two null dates are equal and null date is smaller than any other
non-null date.
*/

// ----------------------------------------------------------------------

    PICBOOL is (const PicDate&)   const;
    PICBOOL is (const BaseDate&)  const;
    PICBOOL is (const PicObject&) const;

// --------------------------------------------------------------------------

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

// --------------------------------------------------------------------------
//:fun

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();

/*
Default PicDate is null.
*/

// --------------------------------------------------------------------------
//:fun  TEXTUAL IMAGE

    virtual BaseString getImage () const;
    virtual BaseString getImage (const PicObject& format) const;

    BaseString getImage (
        PicDateResolution,
        PicDateStyle,
        char delimiter,
        char leader) const;

/*
These methods return the textual representation of the receiver. Default
format is used unless a specific format is given. For null date a null
string is returned.

Formatting object may contain the following components:

    {
        STYLE      = <FORMAT_YEARS_FIRST or FORMAT_MONTH_FIRST or
                      FORMAT_DAY_FIRST or FORMAT_MONTH_BY_NAME or
                      FORMAT_MONTH_BY_SHORT_NAME>;
        RESOLUTION = <FORMAT_YEARS_ONLY or FORMAT_YEARS_TO_MONTH or
                      FORMAT_YEARS_TO_DAYS>;
        DELIMITER  = <char>;
        LEADER     = <char>;
        SIZE       = <PICSIZE>; // Field width
        JUSTIFY    = <FORMAT_LEFT or FORMAT_CENTER or FORMAT_RIGHT>;
        FILLER     = <char>;
    }

Constants are defined in picbase.erh.

:arg delimiter: Delimiter of the date components, e.g. '.' (12.24.1991).
:arg leader:    Leader for one-character numbers, e.g. '0' (02.07.1991).
*/

//:fun ----------------------------------------------------------------------

    virtual void setImage (const BaseString& image);
    virtual void setImage (
        const PicObject&  format,
        const BaseString& image);

    void setImage (
        const BaseString&,
        PicDateStyle,
        char delimiter);

/*
These methods set the value of the receiver according to the given textual
date image. A date format is needed for parsing. If no format is given, the
default is used. A formatting object may contain the following components:

    {
        STYLE      = <FORMAT_YEARS_FIRST or FORMAT_MONTH_FIRST or
                      FORMAT_DAY_FIRST or FORMAT_MONTH_BY_NAME or
                      FORMAT_MONTH_BY_SHORT_NAME>;
        RESOLUTION = <FORMAT_YEARS_ONLY or FORMAT_YEARS_TO_MONTH or
                      FORMAT_YEARS_TO_DAYS>;
        DELIMITER  = <char>;
        LEADER     = <char>;
    }

Constants are defined in picbase.erh.

The date image must contain only the components defined by the resolution
and nothing else. When resolution is yearsOnly or yearsToMonths missing
components are set to 1, e.g:

    setImage("1990", BaseDate::yearsOnly)  ->  1.1.1990

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid date.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const PicDate&);

    virtual PICOFFSET writeFmt (PicTextFmt&, PICOFFSET) const;

/*
Implementation of formatted output.
*/

// --------------------------------------------------------------------------
//:fun  INCREMENT/DECREMENT

    virtual PICBOOL increment (PICSHORT);

/*
Implementation of increment/decrement operation.
*/

// --------------------------------------------------------------------------
//:fun  UNDEFINED OPERATIONS

    virtual PICSHORT       asInt     () const;
    virtual char           asChar    () const;
    virtual PICBOOL        asBoolean () const;
    virtual const char*    asChars   () const;

/*
These raise PicObject::undefinedOperation when called.
*/

    private:
// ==========================================================================
//:fun  PRIVATE METHODS

    virtual PICBOOL setTypedValue (const PicObject& constraints);

/*
Checks the current value of the receiver for PicDate-specific type
definition constraints. The constraint object must be an indexed
dictionary; it may contain the following constraints:

    {
        NULL_IS_CURRENT = <days>;
        // Null value is converted to the current date + <days>.
        NOT_BEFORE = <days>;
        // Dates before today + <days> are illegal.
        NOT_AFTER  = <days>;
        // Dates after today + <days> are illegal.
    }

The constants are defined in picbase.erh.
*/

};


// --------------------------------------------------------------------------
//  INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicDate& PicDate::operator = (const PicDate& date)
       { BaseDate::operator = (date); return *this; }
inline PicDate& PicDate::operator = (const BaseDate& date)
       { BaseDate::operator = (date); return *this; }

inline PICBOOL PicDate::isNull  () const { return BaseDate::isNull(); }
inline void    PicDate::setNull ()       { BaseDate::setNull(); }

inline PICBOOL PicDate::isDefault  () const { return BaseDate::isNull(); }
inline void    PicDate::setDefault ()       { BaseDate::setNull(); }

inline BaseString PicDate::getImage () const
       { return BaseDate::getImage(); }

inline BaseString PicDate::getImage (
    PicDateResolution resolution,
    PicDateStyle      style,
    char           delimiter,
    char           leader) const
    { return BaseDate::getImage(resolution, style, delimiter, leader); }

inline void PicDate::setImage (const BaseString& image)
       { BaseDate::setImage(image); }

inline void PicDate::setImage (
    const BaseString& image,
    PicDateStyle      style,
    char           delimiter)
    { BaseDate::setImage(image, style, delimiter); }

inline PICBOOL PicDate::operator != (const PicObject& obj) const
       { return !(PicDate::operator == (obj)); }
inline PICBOOL PicDate::operator == (const PicDate& date) const
       { return BaseDate::operator == (date); }
inline PICBOOL PicDate::operator != (const PicDate& date) const
       { return BaseDate::operator != (date); }
inline PICBOOL PicDate::operator < (const PicDate& date) const
       { return BaseDate::operator < (date); }
inline PICBOOL PicDate::operator == (const BaseDate& date) const
       { return BaseDate::operator == (date); }
inline PICBOOL PicDate::operator != (const BaseDate& date) const
       { return BaseDate::operator != (date); }
inline PICBOOL PicDate::operator < (const BaseDate& date) const
       { return BaseDate::operator < (date); }

inline PICBOOL PicDate::is (const PicDate& date) const
       { return this == &date; }
inline PICBOOL PicDate::is (const BaseDate& date) const
       { return ((BaseDate*)this) == &date; }
inline PICBOOL PicDate::is (const PicObject& obj) const
       { return this == &obj; }

inline PicFmtText& operator << (PicFmtText& txt, const PicDate& obj)
       { return operator << (txt, (const PicObject&)obj); }

inline PICBOOL PicDate::increment (PICSHORT inc)
       { addDays(inc); return boolTrue; }

inline PICBOOL PicDate::asBoolean () const
       { undefined("asBoolean"); return boolFalse; }
inline PICSHORT PicDate::asInt () const
       { undefined("asInt"); return 0; }
inline char PicDate::asChar () const
       { undefined("asChar"); return '\0'; }
inline const char* PicDate::asChars () const
       { undefined("asChars"); return NULLPTR; }

#endif
