// --------------------------------------------------------------------------
//  DATE-TIME OBJECT
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PDATETIM_HXX
#define PDATETIM_HXX

#include <pobject.hxx>
#include <bdatetim.hxx>

/*:class ********************************************************************

PicDateTime is an PicPrimitive containing a BaseDateTime. Thus, it
represents a full time value with both date and time component.

:abs PicDateTime is a combination of BaseDate and BaseTime.

****************************************************************************/

class PicDateTime : public PicPrimitive
{

    public:
// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicDateTime ();
    PicDateTime (const BaseDate&);
    PicDateTime (const BaseTime&);
    PicDateTime (const BaseDate&, const BaseTime&);
    PicDateTime (
        PicRelativeDate,
        PICLONG      dateInterval = 0,
        PICHOURCOUNT timeInterval = 0);
    PicDateTime (
        PICYEAR, PICMONTH,  PICDAY,
        PICHOUR, PICMINUTE, PICSECOND);
    PicDateTime (const BaseDateTime&);
    PicDateTime (const PicDateTime&);

/*
PicDateTimes can be created in the following ways:

    1. PicDateTime a;
    2. PicDateTime b (baseDate);
    3. PicDateTime c (baseTime);
    4. PicDateTime d (baseDate, baseTime);
    5. PicDateTime e (PicDate::today, +2, 0);
    6. PicDateTime f (1991, 12, 24, 12, 0 ,0);
    7. PicDateTime g (aBaseDateTime);
    8. PicDateTime h (anPicDateTime);

1. Creates a null PicDateTime.
2. Creates PicDateTime out of a BaseDate (or PicDate). The time component
   will be null.
3. Creates PicDateTime out of a BaseTime (or PicTime). The date component
   will be null.
4. Creates PicDateTime out of a BaseDate (or PicDate) and a BaseTime (or
   PicTime).
5. Creates an PicDateTime with the value 'day after tomorrow' (today + 2
   days). This constructor uses the current system date and time.
6. Creates an PicDateTime with the value 24.12.1991 12:00:00.
7. Copies a BaseDateTime.
8. Copies another PicDateTime.
*/

    PICOBJECTCLASS(PicDateTime)

//:fun ----------------------------------------------------------------------

    virtual PicObject&   operator = (const PicObject&);
            PicDateTime& operator = (const PicDateTime&);
            PicDateTime& operator = (const BaseDateTime&);

/*
These methods copy the date and time values.

:arg PicObject&: Reference to an PicDateTime.
:ret PicDateTime&, PicObject&: Reference to the receiver.
:err typeMismatch: PicObject&-argument is not an PicDateTime.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&)    const;
            PICBOOL operator != (const PicObject&)    const;
    virtual PICBOOL operator <  (const PicObject&)    const;

            PICBOOL operator == (const PicDateTime&)  const;
            PICBOOL operator != (const PicDateTime&)  const;
            PICBOOL operator <  (const PicDateTime&)  const;

            PICBOOL operator == (const BaseDateTime&) const;
            PICBOOL operator != (const BaseDateTime&) const;
            PICBOOL operator <  (const BaseDateTime&) const;

/*
PicDateTimes are compared along the time axis. Later times are considered
greater.
Two null date-times are equal and null date-time is smaller than any other
non-null date-time.
*/

// ----------------------------------------------------------------------

    PICBOOL is (const PicDateTime&)  const;
    PICBOOL is (const BaseDateTime&) const;
    PICBOOL is (const PicObject&)    const;

// --------------------------------------------------------------------------

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

// --------------------------------------------------------------------------
//:fun  DEFAULT

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();

/*
Default PicDateTime is null.
*/

// --------------------------------------------------------------------------
//:fun  TEXTUAL IMAGE

    virtual BaseString getImage () const;
    virtual BaseString getImage (const PicObject& format) const;

/*
These methods return the textual representation of the receiver. Default
format is used unless a specific format is given. For null date-time a null
string is returned.
*/

//:fun ----------------------------------------------------------------------

    virtual void setImage (const BaseString& image);
    virtual void setImage (
        const PicObject&  format,
        const BaseString& image);

/*
These methods set the value of the receiver according to the given textual
date-time image. A date format aids the parsing. If no format is given, the
default is used. A formatting object may contain the same components as for
PicDate and PicTime.
*/

//:fun ----------------------------------------------------------------------

    void addHours   (PICHOURCOUNT);
    void addMinutes (PICMINUTECOUNT);
    void addSeconds (PICSECONDCOUNT);

/*
These methods add an interval to a date-time. The operation may affect
also the date part. Be sure to use these instead of BaseTime's versions.

:err nullInstance: Date or time is null.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const PicDateTime&);

    virtual PICOFFSET writeFmt (PicTextFmt&, PICOFFSET) const;

/*
Implementation of formatted output.
*/

// --------------------------------------------------------------------------
//:fun  INCREMENT/DECREMENT

    virtual PICBOOL increment (PICSHORT);

/*
Implementation of increment/decrement operation.
*/

// --------------------------------------------------------------------------
//:fun  UNDEFINED OPERATIONS

    virtual PICSHORT       asInt     () const;
    virtual char        asChar    () const;
    virtual PICBOOL        asBoolean () const;
    virtual const char* asChars   () const;

/*
These raise PicObject::undefinedOperation when called.
*/

//:var ----------------------------------------------------------------------

    BaseDateTime dt;

/*
The embedded BaseDateTime is directly accessible. You can apply operations
directly to the components, but only those that are not implemented
explicitly for PicDateTime.
*/

    private:
// ==========================================================================
//:fun  PRIVATE METHODS

    virtual PICBOOL setTypedValue (const PicObject& constraints);

/*
Checks the current value of the receiver for PicDateTime-specific type
definition constraints. The constraint object must be an indexed
dictionary; it may contain the following constraints:

    {
        NULL_IS_CURRENT = <days>;
        // Null value is converted to the current date&time + <days>.
        NOT_BEFORE = <days>;
        // Dates before today + <days> are illegal.
        NOT_AFTER  = <days>;
        // Dates after today + <days> are illegal.
    }

The constants are defined in picbase.erh.
*/

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PICBOOL PicDateTime::isNull  () const { return dt.isNull(); }
inline void    PicDateTime::setNull ()       { dt.setNull(); }

inline PICBOOL PicDateTime::isDefault  () const { return dt.isNull(); }
inline void    PicDateTime::setDefault ()       { dt.setNull(); }

inline PicDateTime& PicDateTime::operator = (const PicDateTime& edt)
       { dt = edt.dt; return *this; }
inline PicDateTime& PicDateTime::operator = (const BaseDateTime& bdt)
       { dt = bdt; return *this; }

inline PICBOOL PicDateTime::operator != (const PicObject& obj) const
       { return !(PicDateTime::operator == (obj)); }

inline PICBOOL PicDateTime::operator == (const PicDateTime& edt) const
       { return dt == edt.dt; }
inline PICBOOL PicDateTime::operator == (const BaseDateTime& bdt) const
       { return dt == bdt; }

inline PICBOOL PicDateTime::operator < (const PicDateTime& edt) const
       { return dt < edt.dt; }
inline PICBOOL PicDateTime::operator < (const BaseDateTime& bdt) const
       { return dt < bdt; }

inline PICBOOL PicDateTime::operator != (const PicDateTime& edt) const
       { return dt != edt.dt; }
inline PICBOOL PicDateTime::operator != (const BaseDateTime& bdt) const
       { return dt != bdt; }

inline BaseString PicDateTime::getImage () const
       { return dt.getImage(); }
inline void PicDateTime::setImage (const BaseString& image)
       { dt.setImage(image); }

inline PICBOOL PicDateTime::is (const PicDateTime& edt) const
       { return this == &edt; }
inline PICBOOL PicDateTime::is (const BaseDateTime& bdt) const
       { return &dt == &bdt; }
inline PICBOOL PicDateTime::is (const PicObject& obj) const
       { return this == &obj; }

inline void PicDateTime::addHours   (PICHOURCOUNT c)   { dt.addHours(c); }
inline void PicDateTime::addMinutes (PICMINUTECOUNT c) { dt.addMinutes(c); }
inline void PicDateTime::addSeconds (PICSECONDCOUNT c) { dt.addSeconds(c); }

inline PicFmtText& operator << (PicFmtText& txt, const PicDateTime& obj)
       { return operator << (txt, (const PicObject&)obj); }

inline PICBOOL PicDateTime::increment (PICSHORT inc)
       { dt.date.addDays(inc); return boolTrue; }

inline PICBOOL PicDateTime::asBoolean () const
       { undefined("asBoolean"); return boolFalse; }
inline PICSHORT PicDateTime::asInt () const
       { undefined("asInt"); return 0; }
inline char PicDateTime::asChar () const
       { undefined("asChar"); return '\0'; }
inline const char* PicDateTime::asChars () const
       { undefined("asChars"); return NULLPTR; }

#endif
