// --------------------------------------------------------------------------
//  ERROR MANAGER - ABSTRACT BASE CLASS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PERRMGR_HXX
#define PERRMGR_HXX

#include <picbase.hxx>
#include <picerror.hxx>
#include <picerror.erh>
class BaseString;
class PicError;
class PicObject;

// --------------------------------------------------------------------------
//      ERROR MANAGER CLASS ID

#define PICERRORMGRCLASS() \
    PICCLASS()

#define MAKE_PICERRORMGRCLASS(className) \
    MAKE_PICCLASS(className)

/*:class *** ERROR MANAGER PARAMETERS ***************************************

PicErrMgrParam is a structure containing parameters for the error manager:

:list Parameter
:item elogCircular:         Set boolTrue to use the circular log.
:item elogFilter:           The current error log filter. See
                            PicErrorLogFilter for details.
:item elogSuspend:          When nonzero the error log is suspended
                            temporarily. If you want to suspend error logging
                            increment this value and decrement to release
                            suspension.
:item elogFileName:         Pointer to the error log file name.
:item elogRecLength:        Length of circular log record in bytes.
:item elogFileSize:         Size of circular log file in records.
:item elogStamps:           Pointers to the error log stamp strings. The
                            array is indexed by PicErrorLogStamp.
:item emgrRaiseTemplate:    Pointer to the raise template string.
:item emgrHoldTemplate:     Pointer to the hold template string.
:item emgrNoDescription:    Default error description when no error message
                            text is found.
:end

The default values are as follows:

:list Parameter
:item elogCircular:         False
:item elogFilter:           raised | held | discarded.
:item elogSuspend:          0.
:item elogFileName:         ERROR.LOG
:item elogRecLength:        220 bytes.
:item elogFileSize:         100 records.
:item elogStamps[0]:        CLASS@* error STATUS@* in FILE@*(LINE@*)
:item elogStamps[1]:        Error cleared:    TEXT@*
:item elogStamps[2]:        Error discarded:  TEXT@*
:item elogStamps[3]:        Error held:       TEXT@*
:item elogStamps[4]:        Error raised:     TEXT@*
:item elogStamps[5]:        -------- Error manager reset --------
:item elogStamps[6]:        Fatal error:      TEXT@*
:item elogStamps[7]:        Previous error was internal to error manager
                            (FILE@* on line LINE@*).
:item emgrRaiseTemplate:    FILE@*(LINE@*) : CLASS@* : TEXT@*
:item emgrHoldTemplate:     -- FILE@*(LINE@*) : CLASS@* : TEXT@*
:item emgrNoDescription:    -- No description available --
:end

****************************************************************************/

struct PicErrMgrParam
{
    PICBOOL        elogCircular;
    PICUSHORT      elogFilter;
    PICUSHORT      elogSuspend;
    const char*    elogFileName;
    PICUSHORT      elogRecLength;
    PICUSHORT      elogFileSize;
    const char*    elogStamps[8];
    const char*    emgrRaiseTemplate;
    const char*    emgrHoldTemplate;
    const char*    emgrNoDescription;
};

/*:class *** NOTIFIER *******************************************************

PicNotifier is an auxiliary class for the error manager. It defines a
number of hook methods to handle notifications from low-level library
classes that should be displayed for the end-user. PicErrorManager has
a default notifier suitable for character-based applications.

****************************************************************************/

class PicNotifier
{
    public:
// --------------------------------------------------------------------------

    virtual ~PicNotifier ();

    PICCLASS()

// --------------------------------------------------------------------------
//:fun  FATAL & MEMORY ERRORS

    virtual void fatalExit (
        PICUSHORT   lineNumber,
        const char* fileName) =0;

    virtual void memoryExit (
        const char* message,
        PICUSHORT   lineNumber,
        const char* fileName) =0;

/*
The method PicNotifier::fatalExit is called from PicError::exitOnFatalError.
By redefining it you can alter the fatal exit behavior e.g. in graphical
environment. Error information is in the current context. Arguments refer
to the point of exitOnFatalError-call. After returning the program is
exited immediately. You cannot prevent exiting.

Similarly the method PicNotifier::memoryExit is called in out-of-memory
situations. It is a separate function because it must not try to allocate
memory. Also for that reason the error information is given as an ASCIIZ
string.
*/

// --------------------------------------------------------------------------
//:fun  NOTIFICATIONS

    virtual void status   (const BaseString& msg) =0;
    virtual void progress (const BaseString& msg, PICUSHORT readyness) =0;

/*
These methods display notification messages to the user. The first variant
is a simple one-line message, suitable for the status line of a window.
The second variant is more complex: it displays a progress indicator for
which you have to provide sensible readiness percentages (from 0 to 100).
When the readiness reaches 100, the notification window is removed. Be
sure to end the operation with progress(..., 100)!

Note that only one progress message may be active at a time.
*/
    protected:  
        PicNotifier () {}
};


/*:class *** ERROR MANAGER **************************************************

An abstract base class for Error Manager. Most methods are pure virtual;
only a few are implemented here. 

****************************************************************************/

class PicErrorManager : public PicErrorContext
{
    public:
// --------------------------------------------------------------------------

    PICERRORMGRCLASS()

    PICERRORCODES { ok,
        duplicateManager = ERR_DUPLICATE_ERRORMGR,
        cannotOpenLog    = ERR_CANNOT_OPEN_ERRLOG,
        logWriteFailed   = ERR_ERRLOG_WRITE_FAILED,
        error
    };

// --------------------------------------------------------------------------
//:fun  GETTING ERROR MANAGER

    static PicErrorManager* get ();

/*
This method returns a pointer to the error manager of the current thread.
*/

// --------------------------------------------------------------------------
//:fun  NOTIFIER

    void         notifier (PicNotifier&);
    PicNotifier& notifier ();

/*
These methods set and access the notification handler of the error manager.
The error manager doesn't delete the notifier.
*/

// --------------------------------------------------------------------------
//:var  ERROR MANAGER PARAMETERS

    static PicErrMgrParam& param ();

/*
The error manager parameters are global, so if you have multiple threads
and multiple error managers for them, they all share the same parameters.
You can change the parameters before creating the error manager 
(and most of them also at other times).
*/

    protected:
// ==========================================================================
//      CONSTRUCTOR & DESTRUCTOR
//
//      Error managers are automatic objects, global pointers to them
//      are maintained by protected constructor and destructor. 
//      The get-method can be used to access the error manager.

             PicErrorManager ();
    virtual ~PicErrorManager ();

// --------------------------------------------------------------------------
//      CLEARING ERRORS
//
//      clearCurrent      Clears the errors of the current context.
//      reset             Resets the error manager.

    virtual void clearCurrent () =0;
    virtual void reset        () =0;

// --------------------------------------------------------------------------
//      GETTING ERROR INFORMATION
//
//      currentError   Returns pointer to the current error, or a NULLPTR
//                     if the current context is clear.
//      isRaised       Returns whether a non-fatal error has been raised
//                     in the current context.
//      outOfMemory    Returns boolTrue, if a memory allocation
//                     error has been detected.
//      isDetected  == isRaised || outOfMemory

    PicError* currentError () const;
    PICBOOL   isRaised     () const;
    PICBOOL   outOfMemory  () const;
    PICBOOL   isDetected   () const;

// --------------------------------------------------------------------------
//      RAISING ERRORS
//
//      raise           Raises a new error in the current context. Raising
//                      an error will set the error status of the context
//                      and discard all information concerning the previous
//                      error.
//      hold            Holds the current error situation while pushing
//                      new error information concerning the error.
//      cumulate        Appends textual error information in the description.
//
//      raiseMemoryError  Raises a memory allocation error. Does not try to
//                        allocate more memory.
//      checkMemoryError  Checks whether a memory allocation error has
//                        occurred. Logs an additional explanation. Does
//                        not try to allocate more memory.

    virtual void raise (const PicErrorText& message, PICBOOL hold) =0;

    virtual void cumulate (const BaseString& cumulationText) =0;

    virtual void raiseMemoryError (const char* explanation) =0;

    virtual PICBOOL checkMemoryError (
        PICBOOL      wantToContinue,
        const char*  cumulationText,
        PICUSHORT    lineNumber,
        const char*  fileName) =0;

// --------------------------------------------------------------------------
//      CONTEXT HANDLING
//
//      createContext  Links a new context in the context stack.
//      deleteContext  Removes the current context from the context stack.
//                     The previous context becomes current.
//      removeError    Removes the error from the current context and
//                     returns it. In effect, the context is cleared.
//      reverseError   Reverses the order of held information in the
//                     current context.

    void createContext (PicErrorContext* newContext);
    void deleteContext ();

    virtual PicError* removeError () =0;

    void reverseError ();

// --------------------------------------------------------------------------
//      READING ERROR INFO FROM OBJECTS/RESOURCES
//
//      get*ErrorText      Fetches error message text from resources and
//                         performs keyword replacement, if necessary.
//                         Fetches delayed messages.

    virtual BaseString getObjErrorText (
        const PicObject& messages,
        const PicClass&  errorClass,
        PICSTATUS        errorStatus) const =0;

// --------------------------------------------------------------------------
//      INSTANCE VARIABLES
//
//      memoryError     -- This is boolTrue, if memory error is detected.
//      currentContext  -- The current error context (points to the
//                         end of the context list, i.e. stack top).
//      pNotifier       -- Handler for notifications. Never a NULLPTR.

    PICBOOL           memoryError;
    PicErrorContext*  currentContext;
    PicNotifier*      pNotifier;

    private:
// ==========================================================================
//      PRIVATE METHODS
//
//      transfer           Transfers the error item in the given context
//                         to the previous context in the stack. Possible
//                         error item in the previous context is discarded.

    virtual void transfer (PicErrorContext* current) =0;

//:fun ----------------------------------------------------------------------

    virtual PICBOOL logErrorHook (BaseString& logText) const;

/*
The method logErrorHook is called every time something is written to the
error log. By redefining it the error log can be redirected to desired media. 
The default implementation only returns boolFalse.

:arg logText:   The text to be written to error log. It can be modified
                before being written.

:ret boolFalse: Text is written to the ordinary error log.
:ret boolTrue:  Text is not written to the ordinary log.
*/

    friend class PicErrorLog;
    friend class PicError;
    friend class PicErrorContext;
    friend class PicErrorText;
};


// --------------------------------------------------------------------------
//  INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicNotifier::~PicNotifier () {}

inline void PicErrorManager::notifier (PicNotifier& ntfr)
       { pNotifier = &ntfr; }
inline PicNotifier& PicErrorManager::notifier ()
       { return *pNotifier; }

inline PICBOOL PicErrorManager::isRaised () const
       { return currentContext && currentContext->error; }
inline PICBOOL PicErrorManager::outOfMemory () const
       { return memoryError; }
inline PICBOOL PicErrorManager::isDetected () const
       { return isRaised() || outOfMemory(); }

inline PicError* PicErrorManager::currentError () const
       { return currentContext ? currentContext->error : NULLPTR; }

inline void PicErrorManager::createContext (PicErrorContext* newContext)
{
    if (newContext)
    {
        newContext->previous = currentContext;
        currentContext = newContext;
    }
}

inline void PicErrorManager::deleteContext ()
{
    if (currentContext)
    {
        if (currentContext->error)
            transfer (currentContext);
        currentContext = currentContext->previous;
    }
}

inline void PicErrorManager::reverseError ()
{
    if (currentContext->error)
        currentContext->error = currentContext->error->reverse();
}

inline PICBOOL PicErrorManager::logErrorHook (BaseString&) const
       { return boolFalse; }

#endif
