// --------------------------------------------------------------------------
//  FILE STREAM
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PFILESTM_HXX
#define PFILESTM_HXX

#include <pstream.hxx>
#include <pfile.hxx>
class BaseString;

typedef int PICHFILE;

/*:class ********************************************************************

PicFileStream operates on contents of a named file. File name uniquely
identifies the file, so it cannot contain wild-card characters.

Before any operation can be done the file must be opened for proper mode.
There are three modes available: reading, writing and appending (writing
at the end of the file). Once opened read and write methods can be called.
When the stream is not needed any more it should be closed, either explictly 
by close-method or implictly by destructor.

****************************************************************************/

class PicFileStream : public PicBufferStream
{
    public:

    PICSTREAMCLASS(PicFileStream)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicFileStream (const PicFile&);
    PicFileStream (
        const PicFile&,
        PicOpenMode,
        PICUINT attributes = PicFile::normalFile);

    virtual ~PicFileStream ();

/*
Constructor creates file stream with given name. The stream is opened
only if open mode is given. Later stream can be opened explicitly with
open, append or create method.

Destructor closes the file if it's still open.

:arg const PicFile&: File name.
:arg PicOpenMode: Defines operations on the file.
:arg PICUINT: Attributes for new file. Attributes are used only if
              the file is created.
:err PicFile::fileNotFound: File was opened for reading but it doesn't
                            exist.
:err PicFile::accessError: Requested access mode is not allowed for the
                           file.
:err PicFile::sharingError: Other process has prevented requested access to
                            the file.
*/

//:fun ----------------------------------------------------------------------

    static PicFileStream& in  ();
    static PicFileStream& out ();
    static PicFileStream& err ();

/*
These static methods access predefined streams associated with standard
input, output and error devices. These streams may be redirected.

:ret in:  Standard input.
:ret out: Standard output
:ret err: Standard error output.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE write (const void* from, PICSIZE length);

/*
This method writes length bytes from the given buffer to the file. The
return value is the number of characters written to the file; it may
differ from the buffer length, if newlines were translated.

:err PicFile::fileNotOpen: The file is not open for writing.
:err PicStream::writeError: The file is not writeable or system error.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE read (void* to, PICSIZE length);

/*
This method reads length bytes from the file and copies to buffer pointed
by to. The return value is the number of characters written to the buffer
(it may differ from the read amount if newline sequences are translated).

:err PicFile::fileNotOpen: The file is not open for reading.
:err PicStream::readError: The file is not readable or system error.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const;

/*
Status reflects whether the file has been opened for reading or writing.
File stream is always positionable.

*/
    virtual BaseString name () const;

// --------------------------------------------------------------------------
//:fun  FILE

    PicFile file       () const;
    void    changeFile (const PicFile&);

/*
These methods change/access the file associated with the stream. When
file is changed the old file stream is closed. New file stream must be
explicitly opened before any I/O operation can be done.

:arg const PicFile&: New file associated with file stream.
:ret PicFile:        The file associated with file stream.
*/

//:fun ----------------------------------------------------------------------

    void open ();

/*
This method explicitly opens file stream for reading. The file must
exist.

:err PicFile::fileNotFound: Requested file doesn't exist. File is
                            not opened.
:err PicFile::fileAlreadyOpen: The stream has already been opened.
:err PicFile::sharingError: Other process has prevented reading for
                            the file.
*/

//:fun ----------------------------------------------------------------------

    void append (
        PICUINT attributes = PicFile::normalFile);

/*
This method opens file stream for writing at the end of file. It creates
the file if it doesn't exist.

:arg const PICUSHORT: Attributes for new file. Attributes are used only if
                      file is created.
:err PicFile::fileAlreadyOpen: The stream has already been opened.
:err PicFile::accessError:     The file is read-only.
:err PicFile::sharingError:    Other process has prevented reading for
                               the file.
*/
//:fun ----------------------------------------------------------------------

    void create (
        PICUINT attributes = PicFile::normalFile);

/*
This method opens file stream for writing. File is created if it doesn't
exist and truncated if it exists. Initial file size is 0.

:arg const PICUSHORT: Attributes for new file. Attributes are used only if
                      file is created.
:err PicFile::fileAlreadyOpen: The stream has already been opened.
:err PicFile::accessError: File is read-only.
:err PicFile::sharingError: Other process has prevented reading for
                            the file.
*/

//:fun ----------------------------------------------------------------------

    void flush ();

/*
This method explicitly writes all pending changes to disk leaving the
stream open.
*/

//:fun ----------------------------------------------------------------------

    void close ();

/*
This method explicitly closes the stream. All pending changes are flushed
to disk.

:err PicFile::fileNotOpen: Stream has not been opened.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL isOpen () const;

/*
This method checks if the stream is successfully opened.

:ret boolTrue: Stream is open.
:ret boolFalse: Stream is not open.
*/

//:fun ----------------------------------------------------------------------

    PicOpenMode openMode () const;

/*
:ret PicOpenMode: Open mode for which the file is opened.
:err PicFile::fileNotOpen: The file has not been opened.
*/

//:fun ----------------------------------------------------------------------

    virtual void move (PICLONG, PicStreamPosition = current);

/*
Normally streams are accessed sequentially. With file stream also random
access can be used. Move sets the file pointer of the stream. Next read
or write operation takes place at the new position. End of file condition
is unspecified after change and will be set after next read operation.

:arg PicStreamPosition: Defines starting point for moving.
:arg PICLONG: Defines distance for move. If it's positive pointer
              moves forward through the file, otherwise backward.
:err PicFile::fileNotOpen: File has not been opened.
*/

    protected:
// ==========================================================================
// Protected constructor is used to create standard input/output
//   streams. 

    PicFileStream (PICHFILE, PICHFILE /*dummy*/);

    void fOpen (PicOpenMode, PICUINT attr);

    private:
// ==========================================================================
//      INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS

    PicFile     theFile;
    PICHFILE    fHandle;            // file handle.

    PICBOOL     endOfFile;          // True if end of file has been reached.
    PicOpenMode saveOpenFlags;      // Mode of the last file open.

    DENY_COPYING_FOR(PicFileStream)

};

const PICHFILE picHNullFile = (PICHFILE)-1; // file has not been opened.

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicFile PicFileStream::file () const { return theFile; }

inline void PicFileStream::open ()
    { fOpen(PicFile::forReading,PicFile::normalFile);
      saveOpenFlags = PicFile::forReading; }

inline void PicFileStream::append (PICUINT attributes)
    { fOpen(PicFile::forAppending,attributes);
      saveOpenFlags = PicFile::forAppending; }

inline void PicFileStream::create (PICUINT attributes)
    { fOpen(PicFile::forWriting,attributes);
      saveOpenFlags = PicFile::forWriting; }

inline PICBOOL PicFileStream::isOpen () const
    { return fHandle != picHNullFile; }

#endif
