// --------------------------------------------------------------------------
//  BASIC PIC DECLARATIONS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PICBASE_HXX
#define PICBASE_HXX

extern "C" {
#include <limits.h>
}

#include <stddef.h>
#include <new.h>
//  inline void* operator new (size_t, void* p) { return p; }

#include <picerror.erh>

/*:decl *********************************************************************

The header file PICBASE.HXX contains common declarations for all the PIC
classes and applications. It is included almost always. It automatically
includes the standard C++ header limits.h and stddef.h. 
Note that the types PICINDEX, PICPOSITION, PICSIZE, PICINT
and PICUINT follow the size of int. On the other hand, PICSHORT and
PICLONG are guaranteed to retain their size.

****************************************************************************/

// --------------------------------------------------------------------------
//:popup  BASIC TYPES

typedef int                PICINT;     // A 16-bit or 32-bit signed integer.
typedef unsigned int       PICUINT;    // A 16-bit or 32-bit unsigned integer.
typedef short int          PICSHORT;   // A 16-bit signed integer.
typedef unsigned short int PICUSHORT;  // A 16-bit unsigned integer.
typedef long int           PICLONG;    // A 32-bit signed integer.
typedef unsigned long int  PICULONG;   // A 32-bit unsigned integer.
typedef double             PICFLOAT;   // A double-precision floating-point number.
typedef PICSHORT           PICDIGIT;   // Precision of a fixed-point real number (PicReal).
typedef unsigned char      PICFLAG;    // An 8-bit value, often a set of binary flags.
typedef unsigned char      PICBYTE;    // An 8-bit value.
typedef PICSHORT           PICSTATUS;  // An error status code.
typedef PICUINT            PICOFFSET;  // A 16-bit or 32-bit offset.

const PICSHORT  minShortInt  = SHRT_MIN;  // Minimum value of PICSHORT.
const PICSHORT  maxShortInt  = SHRT_MAX;  // Maximum value of PICSHORT.
const PICUSHORT maxUShortInt = USHRT_MAX; // Maximum value of PICUSHORT.
const PICLONG   minLongInt   = LONG_MIN;  // Minimum value of PICLONG.
const PICLONG   maxLongInt   = LONG_MAX;  // Maximum value of PICLONG.
const PICULONG  maxULongInt  = ULONG_MAX; // Maximum value of PICULONG.

typedef void* PICHANDLE; // A 32-bit general handle.

// --------------------------------------------------------------------------
//:popup  NULL POINTER

#define NULLPTR 0 // Null pointer refers to no valid object.

// --------------------------------------------------------------------------

// --------------------------------------------------------------------------
//:popup  THIS

#define THIS (*this) // Reference to this object.

// --------------------------------------------------------------------------
//:popup  BINARY LOGIC

typedef PICSHORT PICBOOL; // Type for binary logic with values boolTrue and boolFalse.

const PICBOOL boolFalse = 0; // False value of binary logic (PICBOOL).
const PICBOOL boolTrue  = 1; // True value of binary logic (PICBOOL).

// --------------------------------------------------------------------------
//:popup  INDEX, POSITION, SIZE

typedef int          PICINDEX;    // Type for indexes in the range -PICINT..PICINT.
typedef unsigned int PICPOSITION; // Type for positions (ordinal numbers) in the range 1..PICUINT.
typedef unsigned int PICSIZE;     // Type for sizes in the range 0..PICUINT.

//:var
const PICINDEX maxIndex = INT_MAX;       // Maximum value of PICINDEX.
const PICINDEX minIndex = -maxIndex;     // Minimum value of PICINDEX.
const PICINDEX outOfIndex = 0;

const PICSIZE  minSize  = 0;             // Minimum value of PICSIZE.
const PICSIZE  maxSize  = UINT_MAX-1;    // Maximum value of PICSIZE.

const PICPOSITION firstPos   = 1;           // Minimum value of PICPOSITION (first position = 1).
const PICPOSITION lastPos    = UINT_MAX-1;  // Maximum value of PICPOSITION.
const PICPOSITION outOfRange = 0;           // Invalid PICPOSITION value.

/* 
minIndex is Minimum value of PICINDEX. 
maxIndex is Maximum value of PICINDEX. 
minIndex is Minimum value of PICSIZE. 
maxIndex is Maximum value of PICSIZE. 
firstPos is Minimum value of PICPOSITION. 
lastPos  is Maximum value of PICPOSITION. 
*/

// --------------------------------------------------------------------------
//:enum   PURGE MODE

enum PicPurgeMode
{
    noPurge,
    withPurge,
    withPurgeOnOrder
};

/*
Purging means that empty locations are removed from a collection.

:list Purge mode
:item noPurge:          Components are not purged. Empty locations may be
                        left in collection.
:item withPurge:        Components are purged; their ordering may change in
                        the process.
:item withPurgeOnOrder: Components are purged; their ordering is preserved.
                        Operation is typically slower than
                        non-order-preserving purging.
:end
*/

// --------------------------------------------------------------------------
//:popup  MEMORY ALLOCATION

const PICSIZE maxNewSize = 64000; // The largest size that operator new can ever allocate.

const PICSIZE maxNewWordSize = maxNewSize / sizeof (short int); // The largest number of 16-bit values that can be allocated.
const PICSIZE maxNewPtrSize  = maxNewSize / sizeof (void*);     // The largest number of far pointers that can be allocated.
const PICSIZE defaultSize    = 16;                              // A default allocation size.

// --------------------------------------------------------------------------
/*:des    DENY COPYING

    DENY_COPYING_FOR(keyword className)
    DENY_CREATION_FOR(keyword className)

By using these macros in the private part of a class declaration you can
deny copying/creation of objects of that class. Otherwise you should 
explicitly define copy & default constructors and assignment operator (=)
as public member functions or at least comment their presence if they are
implemented automatically by the C++ compiler.

For example:

    class MyClass
    {
        public:
        ...
        private:
        DENY_COPYING_FOR(MyClass)
    };

*/

#define DENY_COPYING_FOR(className) \
    className (const className&);   \
    className& operator = (const className&);

#define DENY_CREATION_FOR(className) \
    className ();                    \
    DENY_COPYING_FOR(className)

/*:class ********************************************************************

PicClass provides a simple form of run-time type information. In addition
to the identity of the class, it tells the name of the class and the size
of the class instances.

PicClass objects are static. A class identifier is created once for every
class (perhaps not including trivial 'helper' classes). Also other class
identifiers may exist for special purposes. The physical address of an
PicClass object is the actual identification, so the equality comparison is
defined to compare pointers (or references).

You create an PicClass identifier by using two macros:

    // In the header file:

    class MyClass
    {
        public:
        PICCLASS()
        ...
    };

    // In the implementation file:

    MAKE_PICCLASS(MyClass)

These macros declare and implement a static member function classId for
the given class. They ensure the proper initialization of the PicClass
object. The className argument is given without quotes.

A more complex form of class identification is provided for classes in
virtual hierarchies: PicVirtClass and its descendants.

*/

class PicClass
{
    public:

    //:fun

    const char*    className    () const;
    PICUSHORT      instanceSize () const;

    friend PICBOOL operator == (const PicClass&, const PicClass&);
    friend PICBOOL operator != (const PicClass&, const PicClass&);

    static const PicClass& classId   ();
    static const PicClass& nullClass ();

    PICBOOL isNull () const;

/*
A simple PicClass identifier records the name of the class and the size of
the class instances in bytes. In addition, PicClass identifiers can be
compared for equality and inequality. Two class identifiers are equal if
they are the same object.

PicClass has an identifier of its own: PicClass::classId(). A null class
identifier PicClass::nullClass() is guaranteed to refer to no class.
The method isNull returns whether the receiver is the null class id.
*/

    const char* pClassName;   // These are public to enable static
    PICUSHORT   usInstSize;   // initialization (no need to call
                              // constructors during start-up).

};

inline const char* PicClass::className    () const { return pClassName; }
inline PICUSHORT      PicClass::instanceSize () const { return usInstSize; }

inline PICBOOL operator == (const PicClass& c1, const PicClass& c2)
       { return &c1 == &c2; }
inline PICBOOL operator != (const PicClass& c1, const PicClass& c2)
       { return &c1 != &c2; }

inline PICBOOL PicClass::isNull () const { return this == &nullClass(); }

#define PICCLASS() \
  static const PicClass& classId();

#define MAKE_PICCLASS(className) \
  const PicClass& className::classId () \
  { static PicClass id = { #className, sizeof(className) }; return id; }

#define MAKE_PICCLASS2(className, quotedClassName) \
  const PicClass& className::classId () \
  { static PicClass id = { quotedClassName, sizeof(className) }; return id; }

// --------------------------------------------------------------------------
/*:des   PARAMETER STRINGS

Strings can be parameterized as static data:

    PARAM_STRINGS(SomeClass::strings)
    PARAMSTR(STR_STRING1, "string1")
    PARAMSTR(STR_STRING2, "string2")
    ...
    PARAMSTR(STR_STRINGN, "stringN")
    END_PARAM_STRINGS

The above macros create code for a function SomeClass::strings that returns
the given string from the array of number-string pairs. You declare that
function as follows:

    class SomeClass
    {
        ...
        static PICPARAMSTRINGS(string)
        ...
    };

You use the strings as follows:

    aString = SomeClass::string(STR_STRING1);

The most common application for parameterized strings is error messages.
*/

struct PicNumStringPair
{
    PICUSHORT      num;
    const char* str;
};

typedef const char* PIC_PARAMSTR_FUNC (PICUSHORT);

const char* picParamString (const PicNumStringPair*, PICUSHORT);

#define PICPARAMSTRINGS(name) \
    const char* name (PICUSHORT);

#define PARAM_STRINGS(name) \
    const char* name (PICUSHORT _id) \
    { static const PicNumStringPair _strs[] = {

#define PARAMSTR(num, str) { num, str },

#define END_PARAM_STRINGS  \
    { 0, NULLPTR } }; return ::picParamString(_strs, _id); }

struct PicParamStringIniter
{
    PicParamStringIniter (PIC_PARAMSTR_FUNC*& rpfs, PIC_PARAMSTR_FUNC* pfs)
    {
        rpfs = pfs;
    }
};

#define PARAM_STRINGS_INIT(name, initp) \
    static PicParamStringIniter init_##name (initp, &name); \
    PARAM_STRINGS(name)

// --------------------------------------------------------------------------
/*:des    ERROR CODES

    PICERRORCODES { ok, err1 = 1, err2 = 2, ... };

This macro is used in a class declaration. It declares a static method
ClassName::errors that defines error messages and begins an enumeration
containing the status codes. The error messages must be defined in an
implementation file as follows:

    ERROR_MESSAGES(ClassName)
    PARAMSTR(err1, "Error 1")
    PARAMSTR(err2, "Error 2")
    ...
    END_PARAM_STRINGS

*/

#define PICERRORCODES static const char* errors (PICUSHORT); enum
#define ERROR_MESSAGES(className) PARAM_STRINGS(className::errors)

// DATE STRINGS --------------------------------------------------------

extern const char** picLongMonths    ();
extern const char** picShortMonths   ();
extern const char** picWeekdays      ();
extern const char** picShortWeekdays ();

enum PicSortOrder { ascending, descending };

#endif
