// --------------------------------------------------------------------------
//  PUBLIC INTERFACE OF ERROR MANAGER
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PICERROR_HXX
#define PICERROR_HXX

#include <picerror.erh>
#include <picbase.erh>
#include <perrctxt.hxx>
#include <ptext.hxx>
class PicObject;
class PicErrorManager;

typedef const PicClass& PICFCLASSID ();

// --------------------------------------------------------------------------
//      MACROS
/*:des ----------------------------------------------------------------------

    void ERAISE    (keyword className, PICSTATUS status);
    void EHOLD     (keyword className, PICSTATUS status);
    void ERAISEFMT (keyword className, PICSTATUS status) << data...;
    void EHOLDFMT  (keyword className, PICSTATUS status) << data...;

These macros are the most common way to raise errors:

    ERAISE(MyClass, status1);
    ERAISEFMT(MyClass, status2) << someKeyword;
    EHOLDFMT(MyClass, status3) >>"MORE" << more >>"KEYWORDS" << keywords;

They assume the following:

    * your class has a static method MyClass::errors that returns
      PICPARAMSTRINGS containing error message texts indexed by
      the status code.
    * your class has a static method MyClass::classId that returns the
      class identifier.
    * the status code is declared in an enumeration inside the class,
      e.g. MyClass::status1 refers to a status code.

If the status code doesn't identify a message text in MyClass::errors,
the format is left empty. Then you can give it as hard-coded:

    ERAISEFMT(MyClass, status)("Error @* in @*.") << errorCode << objName;

And if a format was found in MyClass::errors, that format is flushed and
the hard-coded one is used from that on.

*/

/*:des ----------------------------------------------------------------------

    void ERAISE_LINE_FILE (
        PICPARAMSTRINGS() pfErrors,
        const PicClass&() pfClassId,
        PICSTATUS         status,
        PICUSHORT         lineNumber,
        const char*    fileName);

    void EHOLD_LINE_FILE (
        PICPARAMSTRINGS() pfErrors,
        const PicClass&() pfClassId,
        PICSTATUS         status,
        PICUSHORT         lineNumber,
        const char*    fileName);

    void ERAISEFMT_LINE_FILE (
        PICPARAMSTRINGS() pfErrors,
        const PicClass&() pfClassId,
        PICSTATUS         status,
        PICUSHORT         lineNumber,
        const char*    fileName) << data...;

    void EHOLDFMT_LINE_FILE (
        PICPARAMSTRINGS() pfErrors,
        const PicClass&() pfClassId,
        PICSTATUS         status,
        PICUSHORT         lineNumber,
        const char*    fileName) << data...;

These macros raise errors. They are useful when you want to memorize
the file name and line number information in an earlier place in the
source code or when the error messages, class id and status code are
not all found in the same class.

:arg pfErrors:      Pointer to function that takes one argument (status
                    code) and returns the corresponding error message as
                    const char*.
:arg pfClassId:     Pointer to function that takes no arguments and returns
                    a reference to a class identifier.
:arg status:        The error status code.
:arg lineNumber:    Line number where the error occurred.
:arg fileName:      Source file where the error occurred.
:arg data:          Replacement values for placeholders in the error message.
                    See PicFmtText for more details.

*/

#define ERAISEFMT(className, status) \
    if (0) ; else PicRaiseFmtError(className::errors, (PICFCLASSID*)className::classId, \
        className::status, __LINE__, __FILE__).text()

#define EHOLDFMT(className, status) \
    if (0) ; else PicHoldFmtError(className::errors, (PICFCLASSID*)className::classId, \
        className::status, __LINE__, __FILE__).text()

#define ERAISEFMT_LINE_FILE(pfErrors, pfClassId, status, line, file) \
    if (0) ; else PicRaiseFmtError(pfErrors, (PICFCLASSID*)pfClassId, status, line, file).text()

#define EHOLDFMT_LINE_FILE(pfErrors, pfClassId, status, line, file) \
    if (0) ; else PicHoldFmtError(pfErrors, (PICFCLASSID*)pfClassId, status, line, file).text()

#define ERAISE(className, status) \
    PicRaiseFmtError::raise(className::errors, (PICFCLASSID*)className::classId, \
        className::status, __LINE__, __FILE__, 0)

#define EHOLD(className, status) \
    PicRaiseFmtError::raise(className::errors, (PICFCLASSID*)className::classId, \
        className::status, __LINE__, __FILE__, 1)

#define ERAISE_LINE_FILE(pfErrors, pfClassId, status, line, file) \
    PicRaiseFmtError::raise(pfErrors, (PICFCLASSID*)pfClassId, status, line, file, 0)

#define EHOLD_LINE_FILE(pfErrors, pfClassId, status, line, file) \
    PicRaiseFmtError::raise(pfErrors, (PICFCLASSID*)pfClassId, status, line, file, 1)

/*:des ----------------------------------------------------------------------

    void    CHECK_MEMORY_ERROR (const char* description);
    PICBOOL MEMORY_ERROR       (const char* description);
    void    OUT_OF_MEMORY_EXIT (const char* description);

These macros are intended for handling out-of-memory errors. 
CHECK_MEMORY_ERROR exits if out of memory error is detected.
MEMORY_ERROR returns boolTrue if out of memory error is detected,
but does not exit.
OUT_OF_MEMORY_EXIT exits always.

:arg description: Descriptive texts are given as ASCIIZ strings to avoid
                  allocating memory for BaseStrings.

:ret PICBOOL: boolTrue if an out-of-memory error has occurred.

*/

#define CHECK_MEMORY_ERROR(description) \
  PicError::checkMemoryError(boolFalse, description, __LINE__, __FILE__)

#define MEMORY_ERROR(description) \
  PicError::checkMemoryError(boolTrue, description, __LINE__, __FILE__)

#define OUT_OF_MEMORY_EXIT(message) \
  PicError::exitOnMemoryError(message, __LINE__, __FILE__)

/*:des ----------------------------------------------------------------------

    void ERAISE_FATAL     (keyword className, PICSTATUS status);
    void ERAISEFMT_FATAL  (keyword className, PICSTATUS status) << data;
    void FATAL_EXIT       ();

    void DEBUG_REQUIRE    (condition);
    void DEBUG_NOTNULL    (object);
    void DEBUG_NOTNULLPTR (pointer);

These macros cause a fatal exit. If error information is given, the error
is raised first.

:arg condition: Any boolean expression that is valid in the context where
                the macro is invoked. If it evaluates to false, a fatal exit
                is performed. The condition becomes part of the error
                message text.
:arg object:    The condition !object.isNull() must hold.
:arg pointer:   The condition pointer != NULLPTR must hold.

*/

#define ERAISE_FATAL(className, status) \
    PicRaiseFmtError::raise(className::errors, (PICFCLASSID*)className::classId, \
        className::status, __LINE__, __FILE__, 2)

#define ERAISEFMT_FATAL(className, status) \
    if (0) ; else PicRaiseFatalError(className::errors, (PICFCLASSID*)className::classId, \
        className::status, __LINE__, __FILE__).text()

#define FATAL_EXIT() \
    PicError::exitOnFatalError(__LINE__, __FILE__)

#define DEBUG_REQUIRE(condition) \
    if (!(condition)) ERAISEFMT_FATAL(PicError, assertCondition) << #condition

#define DEBUG_NOTNULL(object) \
    if ((object).isNull()) ERAISEFMT_FATAL(PicError, assertNotNull) << #object

#define DEBUG_NOTNULLPTR(pointer) \
    if (!(pointer)) ERAISEFMT_FATAL(PicError, assertNotNullPtr) << #pointer

/*:class *** ERROR TEXT *****************************************************

PicErrorText is a helper class that encapsulates the basic error
message information. Normally you don't use the class directly, but
through the various error raising macros. See ERAISE etc. for more
details.

****************************************************************************/

class PicErrorText
{
    public:

//:fun ----------------------------------------------------------------------

    PicErrorText (
        const PicClass& errorClass,
              PICSTATUS errorStatus,
              PICUSHORT errorLine,
        const char*  errorFile);

   ~PicErrorText ();

/*
The error class, status code and source file information are given when
PicErrorText object is constructed. The error message text is given later
with the method PicErrorText::msg.
*/

//:fun ----------------------------------------------------------------------

    PicErrorText& msg (const char*       errorMsg);
    PicErrorText& msg (const BaseString& errorMsg);

          BaseString& msg ();
    const BaseString& msg () const;

    const PicClass& id     () const;
    PICSTATUS       status () const;
    PICUSHORT       line   () const;
    const char*  file   () const;

/*
These methods set and access the components of the error message.
*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES & PRIVATE MEMBERS

    BaseString      errorText;
    const PicClass* errorClass;
    PICSTATUS       errorStatus;
    PICUSHORT       errorLine;
    const char*     errorFile;

    DENY_COPYING_FOR(PicErrorText)
};


/*:class *** FORMATTED ERROR TEXT *******************************************

PicFmtErrorText is a helper class derived from PicErrorText.
An embedded PicFmtText allows the use of formatted output. This
class is usually not used directly, but through the macros ERAISE,
EHOLD etc.

****************************************************************************/

class PicFmtErrorText : public PicErrorText
{
    public:

   ~PicFmtErrorText ();

    const PicFmtText& text () const;
          PicFmtText& text ();

    protected:
// ==========================================================================
//:fun  CONSTRUCTOR

    PicFmtErrorText (
        PIC_PARAMSTR_FUNC* messages,
        const PicClass&    errorClass(),
        PICSTATUS          errorStatus,
        PICUSHORT          errorLine,
        const char*        errorFile);

/*
The constructor takes the message format string as a string in
ERROR_MESSAGES at the index errorStatus.

Note that the error class id and PIC_PARAMSTR_FUNC are given as
function pointers, e.g:

    PicFmtErrorText(&MyClass::errors, &MyClass::classId, st, line, file);

*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES

    PicStringStream strStream;
    PicFmtText      fmtText;

    DENY_COPYING_FOR(PicFmtErrorText)
};

class PicRaiseFmtError : public PicFmtErrorText
{
    public:

    PicRaiseFmtError (
        PIC_PARAMSTR_FUNC* messages,
        const PicClass&    errClass(),
        PICSTATUS          errStatus,
        PICUSHORT          errLine,
        const char*     errFile) :
        PicFmtErrorText(messages, errClass, errStatus, errLine, errFile) {}

   ~PicRaiseFmtError ();

    static void raise (
        PIC_PARAMSTR_FUNC* messages,
        const PicClass&    errClass(),
        PICSTATUS          errStatus,
        PICUSHORT          errLine,
        const char*     errFile,
        PICUSHORT          errKind); // 0=raise 1=hold 2=fatal

};

class PicHoldFmtError : public PicFmtErrorText
{
    public:

    PicHoldFmtError (
        PIC_PARAMSTR_FUNC* messages,
        const PicClass&    errClass(),
        PICSTATUS          errStatus,
        PICUSHORT          errLine,
        const char*     errFile) :
        PicFmtErrorText(messages, errClass, errStatus, errLine, errFile) {}

   ~PicHoldFmtError ();

};

class PicRaiseFatalError : public PicFmtErrorText
{
    public:

    PicRaiseFatalError (
        PIC_PARAMSTR_FUNC* messages,
        const PicClass&    errClass(),
        PICSTATUS          errStatus,
        PICUSHORT          errLine,
        const char*     errFile) :
        PicFmtErrorText(messages, errClass, errStatus, errLine, errFile) {}

   ~PicRaiseFatalError ();

};


/*:class *** ERROR MESSAGE **************************************************

Class PicError contains the public interface of error manager as static
functions. Many of them are used through macros that hide the line number
and file name information.

PicError objects represent single error messages containing the held/
cumulated information. The error information is accessed through the non-
static methods of PicError class.

****************************************************************************/

class PicError
{

    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicError (const PicErrorText& message, PicErrorManager* = NULLPTR);
    PicError (const PicError&);

   ~PicError ();

/*
With the public constructors you can create PicError objects yourself.

Error information is given as an argument. All the placeholders in the
error message must have been replaced at this time.

Copy constructor copies the whole chain of error information and destructor
deletes it.
*/

    PICCLASS()

    PICERRORCODES { ok,
        functionNotImplemented = ERR_FUNCTION_NOT_IMPLEMENTED,
        outOfMemory            = ERR_OUT_OF_MEMORY,
        memoryAllocation       = ERR_MEMORY_ALLOCATION,
        nullInstance           = ERR_NULL_INSTANCE,
        assertCondition        = ERR_ASSERT_CONDITION,
        assertNotNull          = ERR_ASSERT_NOT_NULL,
        assertNotNullPtr       = ERR_ASSERT_NOT_NULLPTR,
        error,
        fatalError
    };

// --------------------------------------------------------------------------
//      COMPARISON
//:fun ----------------------------------------------------------------------

    PICBOOL operator == (const PicError&) const;
    PICBOOL operator != (const PicError&) const;
    PICBOOL is          (const PicError&) const;

/*
PicError objects may be compared against sameness. Thus operator == and
the is-method perform the same operation.
*/

// --------------------------------------------------------------------------
//      CLEARING ERRORS
//:fun ----------------------------------------------------------------------

    static void clearCurrent ();

/*
This method clears all the errors in the current context.
*/

//:fun ----------------------------------------------------------------------

    static void resetErrorManager ();

/*
This method resets the error manager of the current thread: it clears all
the errors in the context stack and frees all the dynamically allocated
memory. It is recommended to be called when finishing a thread in
multithread environment.
*/

// --------------------------------------------------------------------------
//      GETTING ERROR INFORMATION
//:fun ----------------------------------------------------------------------

    BaseString description (
        const char* raiseTemplate = NULLPTR,
        const char* holdTemplate  = NULLPTR) const;

    void description (PicFmtText&, const char* descrTemplate) const;

/*
The first method produces a textual error description that is combined
from the individual error information components. Description is formatted
according to a template containing certain keywords:

    TEXT@f  CLASS@f  STATUS@f  LINE@f  FILE@f  NUMBER@f

Here 'f' is any format accepted by a string or integer, depending on
the data item.

The keywords are replaced with the corresponding error information items.
If no template is given the default is used. If error information has been
cumulated with EHOLD, it is appended to the description. If the current
context is clear, an empty string is returned.

The second method produces a description of one PicError object using the
given template. The description is written to the given PicFmtText.

:arg raiseTemplate: The template for the first row of the description.
:arg holdTemplate:  The template for the subsequent rows, i.e. the held
                    information. If this is absent, raiseTemplate is
                    used for all rows.
*/

//:fun ----------------------------------------------------------------------

    const BaseString& text   () const;
    const PicClass&   id     () const;
    PICSTATUS         status () const;
    PICUSHORT         line   () const;
    const char*       file   () const;
    PICUSHORT         number () const;

/*
These methods return information about the current error. If error
information has been cumulated with EHOLD, return values concern the latest
error. The whole chain of held information can be accessed with the method
PicError::nextItem.

:ret text:   Textual error message given with ERAISE or EHOLD.
:ret id:     Reference to the error class identifier.
:ret status: The error status code.
:ret line:   Line number where the error was raised.
:ret file:   Source file where the error was raised.
:ret number: Serial number of the error object.
*/

//:fun ----------------------------------------------------------------------

    static PICBOOL isDetected ();
    static PICBOOL isDetected (const PicClass& errorClass);

/*
These methods ask whether an error has been detected in the current context.

:arg errorClass: Ask whether an error of the given class has been detected.
*/

//:fun ----------------------------------------------------------------------

    static const PicError& current ();

/*
This method returns the PicError of the current context, or nullError if
the context is clear.
*/

//:fun ----------------------------------------------------------------------

    static const PicError& nullError ();

/*
Null error is used to signify that no errors have been detected.
*/

// --------------------------------------------------------------------------
//      RAISING ERRORS
//:fun ----------------------------------------------------------------------

    static void raise (const PicErrorText&, PICBOOL hold = boolFalse);

/*
This method raises or holds an error in the current context. The previous
error, if any, is discarded. This method is normally not used directly,
but through the macros ERAISE etc.
*/

//:fun ----------------------------------------------------------------------

    static void cumulate (const BaseString& cumulationText);

/*
This method adds textual error information in the error message text.
Every cumulation is appended to a separate line, so if the text must be
continuous, you must catenate it first and then cumulate with a single
call.

:arg cumulationText: Any text to be appended to the error message.
*/

//:fun ----------------------------------------------------------------------

    static void raiseMemoryError (const char* explanation);

/*
This method is used in the operator new to raise an out-of-memory error.
It is rarely needed in application programs.
*/

//:fun ----------------------------------------------------------------------

    static PICBOOL checkMemoryError (
        PICBOOL        wantToContinue,
        const char* cumulationText,
        PICUSHORT      lineNumber,
        const char* fileName);

/*
This method is normally used through the macros CHECK_MEMORY_ERROR and
MEMORY_ERROR. It checks whether an out-of-memory error has been raised.
*/

//:fun ----------------------------------------------------------------------

    static void exitOnFatalError (PICUSHORT line, const char* file);
    static void exitOnFatalError (const PicErrorText&);

/*
These methods are normally used through the macros ERAISE_FATAL and
FATAL_EXIT. They cause the program to terminate in a controlled manner.
*/

//:fun ----------------------------------------------------------------------

    static void exitOnMemoryError (
        const char*  message,
        PICUSHORT       lineNumber,
        const char*  fileName);

/*
This method terminates the program safely in an out-of-memory situation.
It is normally used through the macro OUT_OF_MEMORY_EXIT.
*/

// --------------------------------------------------------------------------
//      WRITE TO ERROR LOG
//:fun ----------------------------------------------------------------------

    static void log (BaseString& logText);

/*
This method writes text directly to the error log. Information about
process, thread and time are prepended before writing to the log.

:arg logText: The text to be written to error log. Note that the string is
              modified in the function.
*/

// --------------------------------------------------------------------------
//      MANIPULATING ERROR OBJECT CHAIN
//:fun ----------------------------------------------------------------------

    static void reverseCurrent ();

/*
This method reverses the order of held information of the current error.
Normally error information is cumulated so that the latest error is at the
top. In some situations it is more useful to have the error information in
the same order the errors were raised.
*/

//:fun ----------------------------------------------------------------------

    static PicError* remove ();

/*
This method clears the current error and returns the error information.
If the current context is clear, a NULLPTR is returned.
*/

//:fun ----------------------------------------------------------------------

    void appendError (PicError* nextError);

/*
This method appends the given error object to the end of the receiver's
error message chain. From now on that item belongs to the chain and is
deleted by it. Error holding uses this method.
*/

//:fun ----------------------------------------------------------------------

    const PicError* nextItem () const;

/*
This method returns the next item in the chain if held error messages.
By applying the same method to the returned PicError object you get the
next item and so on. End of the chain is indicated by returning a NULLPTR.
*/

    protected:
// =========================================================================
//      PROTECTED METHODS
//:fun ----------------------------------------------------------------------

    PicError* reverse ();

/*
This method reverses the chain of error message items and returns a pointer
to the new first item.
*/

//:fun ----------------------------------------------------------------------

    void appendText (const BaseString& cumulationText);

/*
This method appends text to the end of the receiver. PicError::cumulate
uses this method.

:arg cumulationText: Text to be appended to the error message. A newline
                     is prepended if such doesn't exist in the end of
                     the message.
*/

    private:
// =========================================================================
//      INSTANCE VARIABLES, PRIVATE METHODS & FRIENDS
//
//      Private methods are used internally by the error manager.

    // No assignment operator for PicError
    PicError& operator = (const PicError&);

          BaseString       errorText;
    const PicClass&        errorClass;
          PICSTATUS        errorStatus;
          PICUSHORT        errorLine;
    const char*            errorFile;
          PICUSHORT        errorNum;
          PicErrorManager* manager;
          PicError*        next;
          PicError*        last;

    friend class PicL1ErrorManager;
    friend class PicErrorManager;

};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline const BaseString& PicErrorText::msg () const { return errorText; }
inline       BaseString& PicErrorText::msg ()       { return errorText; }

inline const PicClass& PicErrorText::id     () const { return *errorClass; }
inline PICSTATUS       PicErrorText::status () const { return errorStatus; }
inline PICUSHORT       PicErrorText::line   () const { return errorLine; }
inline const char*  PicErrorText::file   () const { return errorFile; }

inline PicErrorText& PicErrorText::msg (const char* errorMsg)
       { errorText = errorMsg; return *this; }
inline PicErrorText& PicErrorText::msg (const BaseString& errorMsg)
       { errorText = errorMsg; return *this; }

inline const PicFmtText& PicFmtErrorText::text () const { return fmtText; }
inline       PicFmtText& PicFmtErrorText::text ()       { return fmtText; }

// --------------------------------------------------------------------------

inline PICBOOL PicError::operator == (const PicError& right) const
       { return this == &right; }
inline PICBOOL PicError::operator != (const PicError& right) const
       { return this != &right; }
inline PICBOOL PicError::is (const PicError& right) const
       { return this == &right; }

inline const BaseString& PicError::text   () const { return errorText; }
inline const PicClass&   PicError::id     () const { return errorClass; }
inline PICSTATUS         PicError::status () const { return errorStatus; }
inline PICUSHORT         PicError::line   () const { return errorLine; }
inline const char*    PicError::file   () const { return errorFile; }
inline PICUSHORT         PicError::number () const { return errorNum; }

inline void PicError::exitOnFatalError (const PicErrorText& message)
{
    raise(message, boolTrue);
    exitOnFatalError(message.line(), message.file());
}

inline const PicError* PicError::nextItem () const { return next; }

#endif
