// --------------------------------------------------------------------------
//  INTEGER OBJECT
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PINTEGER_HXX
#define PINTEGER_HXX

#include <pobject.hxx>
#include <bstring.hxx>
#include <picerror.erh>
class BaseReal;

/*:class ********************************************************************

PicInteger class replaces basic C++ integer. It has same methods and
associative rules than basic int. The objective of PicInteger is to
abstract the underlying representation and provide safety and additional
methods needed in applications.

Currently PicInteger is implemented as 32-bit long integer. It's range
is -2147483647 .. 2147483647.

****************************************************************************/

class PicInteger : public PicPrimitive
{
    public:

    PICOBJECTCLASS(PicInteger)

    PICERRORCODES { ok,
        nullInstance  = ERR_NULL_INSTANCE,
        formatError   = ERR_INTEGER_FORMAT,
        deformatError = ERR_INTEGER_DEFORMAT,
        underflow     = ERR_NUMBER_UNDERFLOW,
        overflow      = ERR_NUMBER_OVERFLOW,
        divideByZero  = ERR_NUMBER_DIVIDE_BY_ZERO,
        uncomparable  = ERR_OBJECT_UNCOMPARABLE
    };

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, COPY & ASSIGNMENT

    PicInteger ();
    PicInteger (PICLONG);
    PicInteger (const BaseString&);
    PicInteger (const PicInteger&);

/*
PicIntegers can be constructed in the following ways:

    1. PicInteger a;
    2. PicInteger b (34);
    3. PicInteger c ("34");
    4. PicInteger e (anPicInteger);

1. Creates a null PicInteger.
2. Creates an PicInteger having the value 34.
3. Creates an PicInteger having the value 34.
4. Copy of another PicInteger.

:err deformatError: A string doesn't represent an integer.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject&  operator = (const PicObject&);
            PicInteger& operator = (const PicInteger&);
            PicInteger& operator = (PICLONG);

/*
These methods copy the integer value and nullity.

:arg PicObject&: Reference to an PicInteger.
:ret PicInteger&, PicObject&: Reference to the receiver.
:err typeMismatch: PicObject&-argument is not an PicInteger.
*/

// --------------------------------------------------------------------------
//:fun  NULL

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

/*
PicInteger has special null value which means that the value is unknown
or missing. In general operations on null PicIntegers are errors. However,
attributes can be asked also from nulls.
*/

// --------------------------------------------------------------------------
//:fun  DEFAULT

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();
    
/*
Default PicInteger is 0.
*/

// --------------------------------------------------------------------------
//:fun  CONVERSIONS

    virtual PICSHORT asInt  () const;
    virtual char     asChar () const;

    PICLONG asLong () const;

/*
:ret asInt:  An integer within minShortInt..maxShortInt as PICSHORT.
:ret asChar: An integer within 0..UCHAR_MAX as char.
:ret asLong: Value of the receiver as PICLONG.

:err nullInstance: The receiver is null.
:err overflow:     The receiver is invalid as PICSHORT or char.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&) const;
            PICBOOL operator != (const PicObject&) const;
    virtual PICBOOL operator <  (const PicObject&) const;

    friend  PICBOOL operator == (const PicInteger&, const PicInteger&);
    friend  PICBOOL operator != (const PicInteger&, const PicInteger&);
    friend  PICBOOL operator >= (const PicInteger&, const PicInteger&);
    friend  PICBOOL operator >  (const PicInteger&, const PicInteger&);
    friend  PICBOOL operator <  (const PicInteger&, const PicInteger&);
    friend  PICBOOL operator <= (const PicInteger&, const PicInteger&);

/*
These are the usual comparisons of integers. Two null integers are
equal and null integer is smaller than any other non-null integer.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL isZero     () const;
    PICBOOL isPositive () const;
    PICBOOL isNegative () const;

/*
:ret isZero:     boolTrue if the receiver is zero.
:ret isPositive: boolTrue if the receiver is greater than zero.
:ret isNegative: boolTrue if the receiver is less than zero.
*/

// --------------------------------------------------------------------------
//:fun  ARITHMETIC OPERATIONS

    friend PicInteger operator + (const PicInteger&, const PicInteger&);
    friend PicInteger operator - (const PicInteger&, const PicInteger&);
    friend PicInteger operator * (const PicInteger&, const PicInteger&);
    friend PicInteger operator / (const PicInteger&, const PicInteger&);

    friend PicInteger operator % (const PicInteger&, const PicInteger&);

    friend PicInteger operator & (const PicInteger&, const PicInteger&);
    friend PicInteger operator | (const PicInteger&, const PicInteger&);
    friend PicInteger operator ^ (const PicInteger&, const PicInteger&);

/*
These are the usual integer arithmetic operators:

    +   sum
    -   difference
    *   product
    /   integer division
    %   remainder of integer division (modulo)

    &   bitwise and
    |   bitwise or
    ^   bitwise exclusive or

:err nullInstance: Either of the operands is null. The return value is
                   undefined and shouldn't be used.
:err divideByZero: Second argument of operator / or % is zero.
:err overflow:     Result of the operation falls outside the range of
                   PicInteger.
*/

// --------------------------------------------------------------------------
//:fun  ASSIGNMENT + ARITHMETIC OPERATIONS

    PicInteger& operator += (const PicInteger&);
    PicInteger& operator -= (const PicInteger&);
    PicInteger& operator *= (const PicInteger&);
    PicInteger& operator /= (const PicInteger&);

    PicInteger& operator %= (const PicInteger&);

    PicInteger& operator &= (const PicInteger&);
    PicInteger& operator |= (const PicInteger&);
    PicInteger& operator ^= (const PicInteger&);

/*
These methods change the value of the receiving instance.

:err nullInstance: Either the receiver or the argument is null.
:err divideByZero: Argument of operator /= or %= is zero.
:err overflow:     Result of the operation falls outside the range of
                   PicInteger.
*/

// --------------------------------------------------------------------------
//:fun  INCREMENT & DECREMENT

    PicInteger& operator ++ ();
    PicInteger& operator -- ();

/*
Increment and decrement have no separate pre/postfix operations.

:err nullInstance: The receiver is null.
:err overflow:     Result of the operation falls outside the range of
                   PicInteger.
*/

// --------------------------------------------------------------------------
//:fun  ROUND & TRUNCATE

    PicInteger& round     (PICDIGIT);
    PicInteger& truncate  (PICDIGIT);

    PicInteger  rounded   (PICDIGIT) const;
    PicInteger  truncated (PICDIGIT) const;

/*
The receiver is rounded or truncated at the given digit. The effect is
as follows:

    round(16, 1)    -> 20,   round(12, 1)    -> 10
    truncate(16, 1) -> 10,   truncate(12, 1) -> 10

:ret PicInteger&: Reference to the receiver after it has been rounded or
                  truncated.
:ret PicInteger:  A copy of the receiver as rounded or truncated.

:err nullInstance: The receiver is null.
:err overflow:     Result of the operation falls outside the range of
                   PicInteger.
*/

// --------------------------------------------------------------------------
//:fun  ABSOLUTE VALUE

    PicInteger& makeAbs ();
    PicInteger  abs     () const;

/*
PicInteger::makeAbs makes an absolute value of the receiver.
PicInteger::abs returns another PicInteger that is the absolute
value of the receiver.

:err nullInstance: The receiver is null.
*/

// --------------------------------------------------------------------------
//:fun  NEGATE

    PicInteger& negate     ();
    PicInteger  operator - () const;

/*
PicInteger::negate negates of the receiver. The unary operator - returns
another PicInteger that is negation of the receiver.

:err nullInstance: The receiver is null.
*/

// --------------------------------------------------------------------------
//:fun  ODD and PICN

    PICBOOL odd  () const;
    PICBOOL even () const;

/*
Null PicInteger is neither odd nor even.

:ret odd:  boolTrue if receiver % 2 != 0.
:ret even: boolTrue if receiver % 2 == 0.
*/

// --------------------------------------------------------------------------
//:fun  TEXTUAL IMAGE

    virtual BaseString getImage ()                        const;
    virtual BaseString getImage (const PicObject& format) const;
            BaseString getImage (PICUSHORT radix)         const;

/*
These methods return the textual representation of the receiver. Radix
can be given as an argument of getImage, but it cannot be parameterized.
A null string is returned if the receiver is null.

A formatting object may contain the following components:

    {
        RADIX   = <PICUSHORT>;
        SIZE    = <PICSIZE>; // Field width
        JUSTIFY = <FORMAT_LEFT or FORMAT_CENTER or FORMAT_RIGHT>;
        FILLER  = <char>;
    }

Constants are defined in picbase.erh.
*/

//:fun ----------------------------------------------------------------------

    virtual void setImage (const BaseString& image);
    virtual void setImage (
        const PicObject& format,
        const BaseString& image);

/*
This method sets the value of the receiver based on the given numeric
string. Formatting object is not used. For empty, null or all blank
string the receiver is set null.

:err deformatError: The given string doesn't represent an integer.
*/

//:fun ----------------------------------------------------------------------

    virtual PICOFFSET writeFmt (PicTextFmt&, PICOFFSET) const;

/*
Implementation of formatted output.

*/

// --------------------------------------------------------------------------
//:fun  INCREMENT/DECREMENT

    virtual PICBOOL increment (PICSHORT);

/*
Implementation of increment/decrement operation.
*/

// --------------------------------------------------------------------------
//:fun  UNDEFINED OPERATIONS

    virtual PICBOOL     asBoolean () const;
    virtual const char* asChars   () const;

/*
These raise PicObject::undefinedOperation when called.
*/

    protected:
// ==========================================================================
//  PROTECTED METHODS

    void multiplyTo (BaseReal&) const; // These methods affect
    void divideFrom (BaseReal&) const; // the argument BaseReal


    private:
// ==========================================================================
//      INSTANCE VARIABLES
//
//      * imp is the value of PicInteger instance.
//      * isNullInteger is boolTrue if the instance is null.

    PICLONG imp;
    PICBOOL isNullInteger;

    friend class PicMoney;
    friend class PicMeasure;
    friend class PicPercent;

};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicInteger& PicInteger::operator = (const PicInteger& val)
       { imp = val.imp; isNullInteger = val.isNullInteger; return *this; }

inline PicInteger& PicInteger::operator = (PICLONG val)
       { imp = val; isNullInteger = boolFalse; return *this; }

inline PICBOOL PicInteger::operator != (const PicObject& obj) const
       { return !((*this) == obj); }
inline PICBOOL operator != (const PicInteger& val1, const PicInteger& val2)
       { return !(val1 == val2); }
inline PICBOOL operator >= (const PicInteger& val1, const PicInteger& val2)
       { return val2 <= val1; }
inline PICBOOL operator > (const PicInteger& val1, const PicInteger& val2)
       { return val2 < val1; }

inline PICBOOL PicInteger::isZero () const
       { return isNullInteger ? boolFalse : (imp == 0); }
inline PICBOOL PicInteger::isPositive () const
       { return isNullInteger ? boolFalse : (imp > 0); }
inline PICBOOL PicInteger::isNegative () const
       { return isNullInteger ? boolFalse : (imp < 0); }

inline PICBOOL PicInteger::isNull     () const { return isNullInteger; }
inline PICBOOL PicInteger::isDefault  () const { return isZero(); }
inline void    PicInteger::setDefault ()       { (*this) = 0L; }

inline PicInteger operator + (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) += val2; }
inline PicInteger operator - (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) -= val2; }
inline PicInteger operator * (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) *= val2; }
inline PicInteger operator / (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) /= val2; }
inline PicInteger operator % (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) %= val2; }
inline PicInteger operator & (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) &= val2; }
inline PicInteger operator | (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) |= val2; }
inline PicInteger operator ^ (const PicInteger& val1, const PicInteger& val2)
       { return PicInteger(val1) ^= val2; }

inline PicInteger PicInteger::rounded (PICDIGIT digit) const
       { return PicInteger(*this).round(digit); }
inline PicInteger PicInteger::truncated (PICDIGIT digit) const
       { return PicInteger(*this).truncate(digit); }

inline PicInteger PicInteger::abs () const
       { return PicInteger(*this).makeAbs(); }
inline PicInteger PicInteger::operator - () const
       { return PicInteger(*this).negate(); }

inline BaseString PicInteger::getImage () const
       { return getImage(10); }
inline void PicInteger::setImage (const PicObject&, const BaseString& str)
       { setImage(str); }

inline PICBOOL PicInteger::increment (PICSHORT incr)
       { PicInteger i (incr); *this += i; return boolTrue; }

inline PICBOOL PicInteger::asBoolean () const
       { undefined("asBoolean"); return boolFalse; }
inline const char* PicInteger::asChars () const
       { undefined("asChars"); return NULLPTR; }

#endif
