// --------------------------------------------------------------------------
//  LOGIC OBJECT
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PLOGIC_HXX
#define PLOGIC_HXX

#include <pobject.hxx>
#include <picerror.erh>

/*:class ********************************************************************

PicLogic implements three valued logic with values true, unknown and
false. The meaning of unknown is 'not known' or 'not meaningful'.
PicLogic has also an undefined ('null') value. Default textual images
of logic values are 'unknown', 'false' and 'true'. They can be
parameterized.

If three-state values are not needed, PICBOOL (binary logic) is usually
used instead.

****************************************************************************/

class PicLogic : public PicPrimitive
{
    public:

    PICOBJECTCLASS(PicLogic)

    PICERRORCODES { ok,
        nullInstance  = ERR_NULL_INSTANCE,
        deformatError = ERR_LOGIC_DEFORMAT,
        valueError    = ERR_LOGIC_VALUE
    };

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, COPY & ASSIGNMENT

    PicLogic ();
    PicLogic (PICBOOL);
    PicLogic (const BaseString&);
    PicLogic (const PicLogic&);

/*
PicLogics can be constructed in the following ways:

    1. PicLogic a;
    2. PicLogic b (boolTrue);
    3. PicLogic c ("true");
    4. PicLogic e (anPicLogic);

1. Creates a null PicLogic.
2. Creates an PicLogic having the value true.
3. Creates an PicLogic having the value true.
4. Copy of another PicLogic.

:err deformatError: A string doesn't represent a logic value.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);
            PicLogic&  operator = (const PicLogic&);
            PicLogic&  operator = (PICBOOL);

/*
These methods copy the logic value.

:arg PicObject&: Reference to an PicLogic.
:ret PicLogic&, PicObject&: Reference to the receiver.
:err typeMismatch: PicObject&-argument is not an PicLogic.
*/

// --------------------------------------------------------------------------
//:fun  LOGIC CONSTANTS

    static const PicLogic& False   ();
    static const PicLogic& Unknown ();
    static const PicLogic& True    ();
    static const PicLogic& Null    ();

/*
These are the valid values of PicLogic.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&) const;
            PICBOOL operator != (const PicObject&) const;

            PICBOOL operator == (const PicLogic&)  const;
            PICBOOL operator != (const PicLogic&)  const;

/*
Logic values can be compared against equality and unequality.
Two nulls are considered equal.
*/

// --------------------------------------------------------------------------
//:fun  NULL

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

/*
PicLogic has special null value which means that the value is unknown
or missing (don't confuse with the value unknown). In general operations
on null PicLogics are errors. However, attributes can be asked also from
nulls.
*/

// --------------------------------------------------------------------------
//:fun  DEFAULT

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();

/*
Default PicLogic is unknown.
*/

// --------------------------------------------------------------------------
//:fun  CONVERSIONS

    virtual PICSHORT    asInt     () const;
    virtual PICBOOL     asBoolean () const;
    virtual const char* asChars   () const;

/*
:ret asInt:       The internal representation of PicLogic.
:ret asBoolean:   For true returns boolTrue; for all others boolFalse.
:ret asChars:     Image as char*. If the receiver is null, NULLPTR is
                  returned.
*/

// --------------------------------------------------------------------------
//:fun  LOGIC OPERATIONS

    const PicLogic& operator !  ()                const;
    const PicLogic& operator || (const PicLogic&) const;
    const PicLogic& operator && (const PicLogic&) const;

/*
These are the three-valued boolean operations:

    NOT |      OR  | F | T | U   AND | F | T | U
    --------   ---------------   ---------------
     T  | F     F  | F | T | U    F  | F | F | F
     U  | U     T  | T | T | T    T  | F | T | U
     F  | T     U  | U | T | U    U  | F | U | U

:err nullInstance: The receiver or the argument is null.
*/

// --------------------------------------------------------------------------
//:fun  SURELY, SURELY NOT, UNCERTAIN

    friend PICBOOL surely    (const PicLogic&);
    friend PICBOOL surelyNot (const PicLogic&);
    friend PICBOOL uncertain (const PicLogic&);

/*
These are the conversions to binary logic. They are used mainly in
if-expressions.Note especially that surely(x) is not the same as
!surelyNot(x).

:ret surely:    boolTrue if the receiver is true.
:ret uncertain: boolTrue if the receiver is unknown.
:ret surelyNot: boolTrue if the receiver is false.
*/

// --------------------------------------------------------------------------
//:fun  TEXTUAL IMAGE

    virtual BaseString getImage () const;
    virtual BaseString getImage (const PicObject& format) const;

/*
These methods return the textual representation of the receiver. A null
string is returned if the receiver is null.

A formatting object may contain the following components:

    {
        SIZE    = <PICSIZE>; // Field width
        JUSTIFY = <FORMAT_LEFT or FORMAT_CENTER or FORMAT_RIGHT>;
        FILLER  = <char>;
    }

Constants are defined in picbase.erh.
*/

//:fun ----------------------------------------------------------------------

    virtual void setImage (const BaseString& image);
    virtual void setImage (
        const PicObject& format,
        const BaseString& image);

/*
This method sets the value of the receiver based on the given logic
string. Formatting object is not used. For empty, null or all blank
string the receiver is set null. Comparison is case-insensitive, so
"UNKNOWN" and "unKnoWn" are the same but " unknown " is not.

:err deformatError: The given string doesn't represent a logic value.
*/

// --------------------------------------------------------------------------
//:fun  INCREMENT/DECREMENT

    virtual PICBOOL increment (PICSHORT);

/*
Implementation of increment/decrement operation.
*/

// --------------------------------------------------------------------------
//:fun  UNDEFINED OPERATIONS

    virtual char asChar     () const;
    virtual PICBOOL operator < (const PicObject&) const;

/*
These raise PicObject::undefinedOperation when called.
*/

    private:
// ==========================================================================
//      INSTANCE VARIABLES

    PICSHORT val;

};

/*:class ********************************************************************

PicLogic uses default parameters until an PicLogicIniter is created
(either as an automatic or static object). Given values are used until
another PicLogicIniter is created. When PicLogicIniter is destroyed the
previous values are restored. Note that if you give the keywords with
the first constructor, you must control their allocation and deallocation.

****************************************************************************/

class PicLogicIniter
{
    public:

    PicLogicIniter (
        const char* trueKeyword,
        const char* falseKeyword,
        const char* unknownKeyword);
   ~PicLogicIniter ();

    private:

    const char* prevTrueKeyword;
    const char* prevFalseKeyword;
    const char* prevUnknownKeyword;

};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicLogic& PicLogic::operator = (const PicLogic& logic)
       { val = logic.val; return *this; }

inline PICBOOL PicLogic::operator == (const PicLogic& logic) const
       { return val == logic.val; }
inline PICBOOL PicLogic::operator != (const PicObject& obj) const
       { return !(*this == obj); }
inline PICBOOL PicLogic::operator != (const PicLogic& logic) const
       { return !(*this == logic); }

inline PICBOOL PicLogic::isDefault () const
       { return uncertain(*this); }

inline PICSHORT PicLogic::asInt () const
       { return val; }
inline PICBOOL PicLogic::asBoolean () const
       { return surely(*this); }
inline char PicLogic::asChar () const
       { undefined ("asChar"); return '\0'; }

inline void PicLogic::setImage (const PicObject&, const BaseString& image)
       { setImage(image); }

#endif
