// --------------------------------------------------------------------------
//  PicObject
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef POBJECT_HXX
#define POBJECT_HXX

#include <pvirtobj.hxx>
#include <picerror.erh>
class PicIterator;
class PicObject;
class PicFmtText;
class PicTextFmt;

typedef PicObject* PICFCREATENEW ();

/*:class ********************************************************************

PicObjectClass is a class identifier for classes derived from PicObject.
It provides the features of PicVirtObject plus the ability to create
PicObjects dynamically when a class id is known.

PicObjectClass is used through the following macros:

    // In the class header:
    public:
    PICOBJECTCLASS(className)

    // In the class implementation:
    MAKE_PICOBJECTCLASS(className, baseClass)

In addition to the usual PicVirtClass support these macros declare and
implement the following member functions:

    For the class identifier class (function pointers):
    PicObject* createNew ();

    For the derived PicObject class:
    virtual PicObject* copy () const;

The first function creates an PicObject dynamically. The second one
copies an PicObject.

Corresponding macros PICABSOBJECTCLASS and MAKE_PICABSOBJECTCLASS create
a class id for an abstract class.
*/

class PicObjectClass : public PicVirtClass
{
    public:

//:fun ----------------------------------------------------------------------

    const PicObjectClass* derived (const char* className) const;

/*
This method is redefined to return PicObjectClass* instead of general
PicVirtClass*.
*/

//:var ----------------------------------------------------------------------

    PICFCREATENEW* createNew;

/*
This function is used to create PicObjects when a class identifier is
known. It creates an PicObject in dynamic memory. The user
must delete it after use. 

    PicObject* createNew ();

The macros PICOBJECTCLASS and MAKE_PICOBJECTCLASS implement this method
properly and initialize the function pointer, so you need not worry about
that.
*/

    PICCLASS()

    PicObjectClass (
        PicVirtClassInfo* info,
        PICFCREATENEW*    pfCreateNew)
        :
        PicVirtClass (info),
        createNew    (pfCreateNew)
    {}

};

// --------------------------------------------------------------------------
//  Macros to initialize PicObjectClass

#define PICANYOBJECTCLASS(className)      \
    PICANYCLASS(className, PicObjectClass, PicObject)

#define PICOBJECTCLASS(className)         \
    virtual PicObject* copy () const;     \
    PICANYOBJECTCLASS(className)

#define PICABSOBJECTCLASS(className)         \
    virtual PicObject* copy () const =0;     \
    PICANYOBJECTCLASS(className)

#define MAKE_PICOBJECTCLASSID(className, base)                           \
    const PicObjectClass& className::classId() {                         \
      static PicVirtClassInfo info = { #className,                       \
            sizeof(className), (PICFVIRTCLASSID*)base::classId };        \
      static PicObjectClass id (&info, createNew##className);            \
      return id; }                                                       \
    MAKE_PICCLASSID(className)

#define MAKE_PICOBJECTCLASS(className, base)                         \
    PicObject* className::copy () const { return new className(*this); } \
    static PicObject* createNew##className () { return new className(); } \
    MAKE_PICOBJECTCLASSID(className, base)

#define MAKE_PICABSOBJECTCLASS(className, base)                   \
    static PicObject* createNew##className () { return NULLPTR; } \
    MAKE_PICOBJECTCLASSID(className, base)


/*:class ********************************************************************

PicObject is a base class for dynamically manipulateable objects. It
contains many methods of various derived PicObjects as pure virtual
functions. PicObjects cannot be created, but pointers and references
may be used.

****************************************************************************/

class PicObject : public PicVirtObject
{
    public:

    PICABSOBJECTCLASS(PicObject)

    PICERRORCODES { ok,
        staticDeleted = ERR_OBJECT_STATIC_DELETED,
        uncomparable  = ERR_OBJECT_UNCOMPARABLE
    };

    virtual ~PicObject ();

/*:fun ----------------------------------------------------------------------

    virtual PicObject* copy () const =0;

The copy-method is a so called "virtual constructor": it creates a similar
object regardless of its class. It cannot be a real constructor, because
constructors cannot be virtual. Note: it is extremely important that the
copy-method is correctly implemented for all derived PicObject classes!
If the method is left unimplemented, the base class version (if any) is used
instead with erroneus results. The macros (MAKE_)PICOBJECTCLASS take care of
implementing this method correctly.

:ret PicObject*: Pointer to a dynamically allocated copy of the receiver.
                 The caller must take care of deleting the object.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&) =0;

/*
The virtual assignment operator copies objects of same PicObject class.

:arg const PicObject&: A reference to an PicObject of the same or
                       compatible class than the receiver.

:ret PicObject&: Reference to the receiver after assignment.

:err typeMismatch: The argument was not of a compatible class. The
                   receiver was left intact. 
*/

// --------------------------------------------------------------------------
//      GLOBAL NULL OBJECT, CLASS IDs AND VALUE
//:fun ----------------------------------------------------------------------

    static PicObject& null ();

/*
The global null object is used as a return value when no valid PicObject can
be returned from some function returning a reference to an PicObject. It
must never be deleted.

:ret PicObject&: A reference to a statically allocated PicNullObject.
*/

// --------------------------------------------------------------------------
//      COMPARISONS
//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator == (const PicObject&) const =0;
            PICBOOL operator != (const PicObject&) const;

/*
Equality operator should be defined for every derived PicObject class.
Usually it performs a deep comparison, i.e. the objects are compared
component by component, recursively if necessary. Unequality operator is
defined as inline and needs not be redefined. Objects of different classes
(even if compatible) are always considered unequal.

:arg const PicObject&: Any PicObject.

:ret boolTrue:  If the argument and the receiver are of the same class and
                equal.
:ret boolFalse: If the argument and the receiver are of different classes
                or unequal.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator < (const PicObject&) const =0;

/*
Less than -operator is used mainly for sorting purposes. It is defined for
objects having a meaningful order. Otherwise it raises an error.

:arg const PicObject&: PicObject of the same class than the receiver.

:ret boolTrue:  If the argument and the receiver are comparable and the
                argument is greater than the receiver.
:ret boolFalse: If the argument and the receiver are uncomparable or the
                argument is smaller than the receiver.

:err uncomparable: The argument and the receiver cannot be compared for
                   magnitude.
*/

// --------------------------------------------------------------------------
//      NULL & DEFAULT
//:fun ----------------------------------------------------------------------

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

    virtual PICBOOL isDefault  () const =0;
    virtual void    setDefault ()       =0;

/*
Null and default values are defined in a class-by-class basis. They must
be defined for every derived PicObject class.
*/

// --------------------------------------------------------------------------
//      FOR COMPATIBILITY
// --------------------------------------------------------------------------

    PICSHORT       asInt     () const;
    PICBOOL        asBoolean () const;
    char        asChar    () const;
    const char* asChars   () const;

          PicObject& at (PICINDEX index);
    const PicObject& at (PICINDEX index) const;
          PicObject& at (const BaseString& name);
    const PicObject& at (const BaseString& name) const;

          PicObject& first ();
    const PicObject& first () const;
          PicObject& last  ();
    const PicObject& last  () const;

    PICBOOL putAt (PICINDEX index, const PicObject&);
    PICBOOL putAt (const BaseString& name, const PicObject&);
    PICBOOL add   (const PicObject&);

    PICBOOL find (PICINDEX index, PicIterator&) const;
    PICBOOL find (const BaseString& name, PicIterator&) const;
    PICBOOL find (const PicObject&, PicIterator&) const;

    PICBOOL contains (PICINDEX index) const;
    PICBOOL contains (const BaseString& name)  const;
    PICBOOL contains (const PicObject&) const;

    BaseString nameOf  (const PicObject&) const;
    PICINDEX   indexOf (const PicObject&) const;

    void deleteAt (PICINDEX index, PicPurgeMode = noPurge);
    void deleteAt (const BaseString& name, PicPurgeMode = noPurge);

    void deleteObject (const PicObject&, PicPurgeMode = noPurge);
    void deleteAll    ();

    PICSIZE    size         () const;
    PICBOOL    isEmpty      () const;

    BaseString firstName    () const;
    BaseString lastName     () const;

    PICINDEX   firstIndex   () const;
    PICINDEX   lastIndex    () const;

    PICBOOL    isPrimitive  () const;
    PICBOOL    isCollection () const;

    void       startIteration (PicIterator&) const;
    PICBOOL    next           (PicIterator&) const;
    PICBOOL    previous       (PicIterator&) const;
    PicObject& current        (const PicIterator&);
    const PicObject& current  (const PicIterator&) const;
    BaseString currentName    (const PicIterator&) const;
    PICINDEX   currentIndex   (const PicIterator&) const;

// --------------------------------------------------------------------------
//      TEXTUAL IMAGE
//:fun ----------------------------------------------------------------------

    virtual BaseString getImage () const =0;
    virtual BaseString getImage (const PicObject& format) const =0;

    virtual void setImage (const BaseString& image) =0;
    virtual void setImage (
        const PicObject&  format,
        const BaseString& image) =0;

/*
:list Method
:item getImage: Returns a textual representation of the receiver object.
:item setImage: Sets the value of the receiver object from the given
                textual representation.

:arg format: A class-specific formatting object. Without this argument
             a default format is used. Class-specific errors are raised
             if the format is invalid.

:ret BaseString: The value of the receiver in a textual, human-readable form.

:err undefinedOperation: These methods are defined for primitive PicObjects,
                         not for collection PicObjects.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const PicObject&);

    virtual PICOFFSET writeFmt (PicTextFmt&, PICOFFSET) const;

    PICBOOL pictureFmt (const PicObject& format, BaseString& image) const;

/*
Operator to write an PicObject to an PicFmtText.

The method writeFmt implements formatted output for an PicObject. The
default implementation of PicObject takes the default image for output.
This method should be redefined in classes supporting formatted output.

The method pictureFmt takes the image of the receiver object if the
given format object is a picture format string. It returns true if a format
was found and image stored in the second argument. This is used in
implementing the getImage method with a format argument.
*/

    PICBOOL typeGuard (const PicVirtClass& expected  DEBUG_FILE_LINE) const
    {
        return expected.typeGuard(*this  DEBUG_FILE_LINE_IMP);
    }

    protected:

    PicObject () {} // Otherwise compiler may generate a warning...

};


/*:class *** PRIMITIVE ******************************************************

PicPrimitive is a base class for primitive PicObjects. Primitive PicObjects
represent simple entities, like string, integer, real, etc. They do not
contain other objects as components.

PicPrimitive is an abstract class containing pure virtual methods, so
instances cannot be created. Pure virtual methods must be defined in the
derived classes.

****************************************************************************/

class PicPrimitive : public PicObject
{

    public:

    PICABSOBJECTCLASS(PicPrimitive)

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&) =0;

/*
Assignment operator must be defined in every derived class.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL operator == (const PicObject&) const =0;
    virtual PICBOOL operator <  (const PicObject&) const =0;

/*
Comparison operators must be defined in every derived class.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL isNull  () const =0;
    virtual void    setNull ()       =0;

    static PicPrimitive& null ();

/*
Null query and setting must be defined in every derived class.
*/

//:fun ----------------------------------------------------------------------

    virtual PICBOOL isDefault  () const =0;
    virtual void    setDefault ()       =0;

/*
Default query and setting must be defined in every derived class.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSHORT       asInt     () const =0;
    virtual PICBOOL        asBoolean () const =0;
    virtual char           asChar    () const =0;
    virtual const char*    asChars   () const =0;

/*
These methods provide some conversions to the simple types of C++. They are
defined for classes for which they are sensible.

:ret asInt:      The value of the receiver as a short integer.
:ret asBoolean:  The value of the receiver as a boolean value.
:ret asChar:     The value of the receiver as a character.
:ret asChars:    Pointer to a zero-terminated string contained in the
                 receiver, or NULLPTR if no such value is available.

:err undefinedOperation: A conversion is not meaningful for a derived class.
*/

//:fun ----------------------------------------------------------------------

    virtual BaseString getImage () const =0;
    virtual BaseString getImage (const PicObject& format) const =0;

    virtual void setImage (const BaseString& image) =0;
    virtual void setImage (
        const PicObject& format,
        const BaseString& image) =0;

/*
Image methods must be defined in every derived class.
*/

// --------------------------------------------------------------------------
//:fun  INCREMENT/DECREMENT

    virtual PICBOOL increment (PICSHORT);

/*
This method increments (or decrements) the value of the primitive
by the given amount. Return value is true if the object was able
to increment/decrement its value.
*/

    private:
// ==========================================================================
//:fun  PRIVATE METHODS

    virtual PICBOOL setTypedValue (const PicObject& constraints);

/*
This method must be defined for every class that is to support class-
specific type definition constraints. It determines whether the current
value of the receiver is correct for the given constraints, and possibly
adjusts the value. It returns boolTrue if the value is legal or otherwise
raises a suitable error. Default implementation only returns boolTrue.
This method is not called for enumerated attributes, because they cannot
have any further constraints.
*/

    friend class PicAttrType;
};


// --------------------------------------------------------------------------
//  INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline const PicObjectClass* PicObjectClass::derived (const char* className) const
       { return (PicObjectClass*)PicVirtClass::derived(className); }

inline PICBOOL PicObject::operator != (const PicObject& obj) const
       { return !(*this == obj); }

inline PICBOOL PicObject::isPrimitive () const
       { return id() >= PicPrimitive::classId(); }

#define SELF ((PicObject*)this)

inline const PicObject& PicObject::at (PICINDEX index) const
       { return SELF->at(index); }
inline const PicObject& PicObject::at (const BaseString& name) const
       { return SELF->at(name); }

inline const PicObject& PicObject::first () const
       { return SELF->first(); }
inline const PicObject& PicObject::last () const
       { return SELF->last(); }

inline PICBOOL PicObject::isEmpty () const
       { return size() == 0; }

inline const PicObject& PicObject::current (const PicIterator& i) const
       { return SELF->current(i); }

#undef SELF

inline PICBOOL PicPrimitive::increment (PICSHORT) { return boolFalse; }

#endif
