// --------------------------------------------------------------------------
//  STRING OBJECT
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PSTRING_HXX
#define PSTRING_HXX

#include <pobject.hxx>
#include <bstring.hxx>

/*:class ********************************************************************

PicString is a dynamic collection of ASCII characters. Size of a string is
limited only by the maximum allocation size. Character components are
referred to with positions within 1..size(). An empty string contains no
characters. It is not to be confused with null string, whose value is
considered unknown.

****************************************************************************/

class PicString : public PicPrimitive, public BaseString
{
    public:

    PICOBJECTCLASS(PicString)

    PICERRORCODES { ok,
        formatError      = ERR_STRING_FORMAT,
        charSetViolation = ERR_STRING_CHARSET_VIOLATION,
        notAlphabetic    = ERR_STRING_NOT_ALPHABETIC,
        notNumeric       = ERR_STRING_NOT_NUMERIC,
        notAlphanumeric  = ERR_STRING_NOT_ALPHANUMERIC,
        tooShort         = ERR_STRING_TOO_SHORT,
        emptyString      = ERR_STRING_EMPTY
    };

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

    PicString ();
    PicString (const char*);
    PicString (char, PICSIZE times);
    PicString (const PicString&);
    PicString (const BaseString&);

/*
Strings can be constructed in the following ways:
    
    1. PicString x1;
    2. PicString x2 ("cccc...cccc");
    3. PicString x4 ('c', 1);
    4. PicString x5 (anPicString);
    5. PicString x6 (aBaseString);

1. Default constructor creates a null PicString.
2. A zero-terminated string is copied into the newly created PicString.
   If the argument is NULLPTR a null PicString is created.
3. The given character is duplicated to fill the string. If count is 0 an
   empty PicString is created.
4. Copy constructor increments reference count. Character buffer is copied
   only when needed.
5. Also a kind of copy constructor: copy of a BaseString.

:err outOfMemory:  The character buffer is allocated dynamically which may
                   cause memory allocation failures.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);
            PicString& operator = (const PicString&);
            PicString& operator = (const char*);
            PicString& operator = (const BaseString&);

/*
When a zero-terminated string is assigned, the characters are always copied.
If the argument is NULLPTR the receiver is set null. Assigning an PicString
causes the reference count of the argument to be incremented.

:arg PicObject&: Reference to an PicString.
:ret PicString&, PicObject&: Reference to the receiver.
:err typeMismatch: PicObject&-argument is not an PicString.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&)  const;
            PICBOOL operator != (const PicObject&)  const;
    virtual PICBOOL operator <  (const PicObject&)  const;

            PICBOOL operator == (const PicString&)  const;
            PICBOOL operator != (const PicString&)  const;
            PICBOOL operator <  (const PicString&)  const;
            PICBOOL operator == (const BaseString&) const;
            PICBOOL operator != (const BaseString&) const;
            PICBOOL operator <  (const BaseString&) const;

/*
These methods compare the argument and the receiver. Operators are 
case-sensitive. Null strings are considered equal and smaller than 
any non-null string.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL is (const PicString&)  const;
    PICBOOL is (const BaseString&) const;
    PICBOOL is (const PicObject&)  const;

/*
Two strings are the same string if they occupy the same memory location.
Note that two strings referring to the same StringRep are not necessarily
the same string.
*/

// --------------------------------------------------------------------------
//:fun  NULL

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

/*
The method setNull makes the receiver a null PicString. It deletes the
character buffer if such exists.

:ret isNull: boolTrue if the receiver is a null PicString.
*/

// --------------------------------------------------------------------------
//:fun  DEFAULT

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();
    
/*
Default PicString is an empty string.
*/

// --------------------------------------------------------------------------
//:fun  CONVERSIONS

    virtual char        asChar  () const;
    virtual const char* asChars () const;

/*
:list Method
:item asChar:  Defined, if the string contains exactly one character.
:item asChars: Defined always. Behaves like the method BaseString::chars.
:end
*/

//:fun ----------------------------------------------------------------------

          BaseString& base ();
    const BaseString& base () const;

/*
This method is used to access the BaseString contained in the PicString
object. It may be needed as a workaround when the compiler cannot use
methods of BaseString directly.

Another use is to call BaseString methods that have counterparts with same
names in PicObject. Calling such methods directly causes compilation errors,
because compiler cannot decide which versions to elect. The problematic
methods are the following:

    char      at (const PICPOSITION)                const;
    PicString at (const PICPOSITION, const PICSIZE) const;

    char first () const;
    char last  () const;

    void deleteAt (const PICPOSITION, const PICSIZE);

    BaseString copy () const;
*/

// --------------------------------------------------------------------------
//:fun  ASKING ATTRIBUTES

    virtual PICSIZE size    () const;
    virtual PICBOOL isEmpty () const;

/*
These are defined for PicString as strings possess characteristics of
collections (a collection of characters).
*/

// --------------------------------------------------------------------------
//:fun  TEXTUAL IMAGE

    virtual BaseString getImage () const;
    virtual BaseString getImage (const PicObject& format) const;

/*
Image of a string is the string itself, plus possible formatting
(left/right/justify). The formatting object may contain the following
components:

    {
        SIZE       = <PICSIZE>; // Field width
        JUSTIFY    = <FORMAT_LEFT or FORMAT_CENTER or FORMAT_RIGHT>;
        FILLER     = <char>;
    }

Constants are defined in picbase.erh.
*/

//:fun ----------------------------------------------------------------------

    virtual void setImage (const BaseString& image);
    virtual void setImage (
        const PicObject&  format,
        const BaseString& image);

/*
These methods act like assignment operator. They are implemented for
symmetry. Formatting object is not used.
*/

//:fun ----------------------------------------------------------------------

    static BaseString& format (BaseString& image, const PicObject& format);

/*
This method is used in implementing the formatting of other primitives.
It formats the given string according to the given formatting map (see
PicString::getImage for contents). Return value is the image-argument.
*/

    friend PicFmtText& operator << (PicFmtText&, const PicString&);

// --------------------------------------------------------------------------
//:fun  UNDEFINED OPERATIONS

    virtual PICSHORT asInt     () const;
    virtual PICBOOL  asBoolean () const;

/*
These raise PicObject::undefinedOperation when called.
*/

    private:
// ==========================================================================
//:fun  PRIVATE METHODS

    virtual PICBOOL setTypedValue (const PicObject& constraints);

/*
Checks the current value of the receiver for PicString-specific type
definition constraints. The constraint object must be an indexed
dictionary; it may contain the following constraints:

    {
        MAXSIZE  = <PICSIZE>;
        // The maximum size of the string; if the string is longer, it is
        // cut to fit.
        MINSIZE  = <PICSIZE>;
        // The minimum size of the string; if the string is shorter, an
        // error is raised.
        CHARSET  = <String or ALPHABETIC or NUMERIC or ALPHANUMERIC>;
        // Restricts the set of characters allowed in the string. Its value
        // may be either a string (as a list of allowed characters) or one
        // of the above constants.
        CHARCASE = <UPPERCASE or LOWERCASE or PROPER>;
        // This setting determines, whether the string is adjusted to
        // uppercase, lowercase or properly for names.
        CHARLEADER = "c" or <ASCII code>;
        // This causes the string to be filled with the given character
        // up to MAXSIZE (with justification to right). Typical application
        // is automatic addition of leading zeros. Without MAXSIZE this
        // setting has no effect. MINSIZE is checked before the leader is
        // prepended.
    }

The constants are defined in picbase.erh.
*/

};

// --------------------------------------------------------------------------
//  INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicString& PicString::operator = (const char* chars)
       { BaseString::operator = (chars); return *this; }
inline PicString& PicString::operator = (const BaseString& string)
       { BaseString::operator = (string); return *this; }
inline PicString& PicString::operator = (const PicString& string)
       { BaseString::operator = (string); return *this; }

inline PICBOOL PicString::operator != (const PicObject& obj) const
       { return !(PicString::operator == (obj)); }
inline PICBOOL PicString::operator == (const PicString& string) const
       { return BaseString::operator == (string); }
inline PICBOOL PicString::operator != (const PicString& string) const
       { return BaseString::operator != (string); }
inline PICBOOL PicString::operator < (const PicString& string) const
       { return BaseString::operator < (string); }
inline PICBOOL PicString::operator == (const BaseString& string) const
       { return BaseString::operator == (string); }
inline PICBOOL PicString::operator != (const BaseString& string) const
       { return BaseString::operator != (string); }
inline PICBOOL PicString::operator < (const BaseString& string) const
       { return BaseString::operator < (string); }

inline PICBOOL PicString::is (const PicString& string) const
       { return this == &string; }
inline PICBOOL PicString::is (const BaseString& string) const
       { return this == &string; }
inline PICBOOL PicString::is (const PicObject& obj) const
       { return this == &obj; }

inline PICBOOL PicString::isNull () const { return BaseString::isNull(); }
inline void    PicString::setNull ()      { BaseString::setNull(); }

inline PICBOOL PicString::isDefault () const { return BaseString::isEmpty(); }
inline void    PicString::setDefault ()      { BaseString::setEmpty(); }

inline const char* PicString::asChars () const
       { return BaseString::chars(); }

inline PICSIZE PicString::size    () const { return BaseString::size(); }
inline PICBOOL PicString::isEmpty () const { return BaseString::isEmpty(); }

inline BaseString PicString::getImage () const
       { return BaseString::getImage(); }
inline void PicString::setImage (const BaseString& image)
       { BaseString::operator = (image); }

inline BaseString& PicString::base ()
       { return *(BaseString*)this; }
inline const BaseString& PicString::base () const
       { return *(const BaseString*)this; }

inline PicFmtText& operator << (PicFmtText& txt, const PicString& obj)
       { return operator << (txt, (const PicObject&)obj); }

inline PICBOOL PicString::asBoolean () const
       { undefined ("asBoolean"); return boolFalse; }
inline PICSHORT PicString::asInt () const
       { undefined ("asInt"); return 0; }

#endif
