// --------------------------------------------------------------------------
//  PicVirtObject
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PVIRTOBJ_HXX
#define PVIRTOBJ_HXX

#include <picbase.hxx>
#include <picerror.erh>
class PicVirtObject;
class PicVirtClass;
class BaseString;

/*:des ----------------------------------------------------------------------

          className& AS (className,       PicVirtObject& obj);
    const className& AS (className, const PicVirtObject& obj);
          className& AS (className,       PicVirtObject* obj);
    const className& AS (className, const PicVirtObject* obj);

This macro converts a reference or pointer to an PicVirtObject into a
reference to an object of the given derived class. It
terminates the program with FATAL_EXIT() if an error is detected. File
name and line number information for the error message is taken at the
point of conversion.

:arg className: Name of a class derived from PicVirtObject (without quotes).
:arg obj:       Reference or pointer to an object of the class className.

:ret className&:       Reference to the object obj.
:ret const className&: Reference to the constant object obj.

:err nullPointer:  Pointer argument obj was a null pointer.
:err typeMismatch: Argument obj didn't refer to an object of className.
*/

/*:des ----------------------------------------------------------------------

    PICBOOL PicVirtObject::IS_A (const PicVirtClass&);

This macro returns whether the object for which the operation is applied
is compatible with the argument class. In the practice this looks as
follows:

    if (object.IS_A(className)) ...

or with a pointer:

    if (objPtr->IS_A(className)) ...
*/

/*:des ----------------------------------------------------------------------

    PICBOOL TYPE_GUARD (const PicVirtClass&, const PicVirtObject&);

This macro tells whether an object is of the given class or some class
derived from it. It raises an error in case of type mismatch.

:arg PicVirtClass:  Identifier of a class derived from PicVirtObject.
:arg PicVirtObject: Reference to an object of the given class.

:ret boolTrue:  The object was of the expected or compatible class.
:ret boolFalse: The object was of uncompatible class.

:err typeMismatch: Raised when boolFalse is returned. 
*/

/*:des ----------------------------------------------------------------------

    void DEBUG_ASSERT (const PicVirtClass&, const PicVirtObject&);
    void DEBUG_ASSERT (const PicVirtClass&, const PicVirtObject*);

This macro is used by the AS-conversion. It checks the argument object
against null pointer and the expected class id. Failing check causes a
FATAL_EXIT(). 

:arg PicVirtClass:  Identifier of a class derived from PicVirtObject.
:arg PicVirtObject: Reference or pointer to an object of the given class.

:err nullPointer:   Pointer argument was a null pointer.
:err typeMismatch:  PicVirtObject argument didn't refer to an object of
                    the given class.
*/

/*:des ---------------------------------------------------------------------

    DEBUG_FILE_LINE
    DEBUG_FILE_LINE_IMP

The FILE_LINE-macros are used to carry file name and line number information
across call levels. DEBUG_FILE_LINE is used in the header of a function and
DEBUG_FILE_LINE_IMP in the implementation. They declare two
additional arguments:

    const PICUSHORT lineNumber
    const char*     fileName
*/

// --------------------------------------------------------------------------
//  Macros for run-time type checking

#define AS(className,object) \
    className::typecast((object), __LINE__, __FILE__)
#define TYPE_GUARD(expected,found) \
    (found).typeGuard((expected), __LINE__, __FILE__)
#define DEBUG_ASSERT(expected,found) \
    (expected).nassert((found), __LINE__, __FILE__)
#define DEBUG_ASSERT_FILE_LINE(expected,found) \
    (expected).nassert((found), lineNumber, fileName)
#define DEBUG_FILE_LINE \
    , const PICUSHORT lineNumber, const char* fileName
#define DEBUG_FILE_LINE_IMP \
    , lineNumber, fileName

#define IS_A(className) id().isKindOf(className::classId())

// --------------------------------------------------------------------------
//  Macros to initialize PicVirtClass or any class id:

#define PICANYCLASS(className, classIdName, baseName)                      \
    static const className& typecast (                                     \
        const baseName& obj  DEBUG_FILE_LINE)                              \
        { DEBUG_ASSERT_FILE_LINE(classId(), obj);                          \
          return (const className&)obj; }                                  \
    static const className& typecast (                                     \
        const baseName* obj  DEBUG_FILE_LINE)                              \
        { DEBUG_ASSERT_FILE_LINE(classId(), obj);                          \
          return *(const className*)obj; }                                 \
    static className& typecast (baseName& obj  DEBUG_FILE_LINE)            \
        { DEBUG_ASSERT_FILE_LINE(classId(), obj);                          \
          return (className&)obj; }                                        \
    static className& typecast (baseName* obj  DEBUG_FILE_LINE)            \
        { DEBUG_ASSERT_FILE_LINE(classId(), obj);                          \
          return *(className*)obj; }                                       \
    static const classIdName& classId();                                   \
    virtual const PicVirtClass& virtId() const;

#define MAKE_PICCLASSID(className)                                         \
    const PicVirtClass& className::virtId () const { return classId(); }   \
    class className##_Initer {                                             \
        public: className##_Initer () { className::classId(); }            \
    }; className##_Initer init_##className##;

#define MAKE_PICANYCLASS(className, classIdName, base)                     \
    const classIdName& className::classId() {                              \
      static PicVirtClassInfo info = { #className,                         \
            sizeof(className), (PICFVIRTCLASSID*)base::classId };          \
      static classIdName id (&info); return id; }                          \
    MAKE_PICCLASSID(className)

#define PICVIRTCLASS(className) \
    PICANYCLASS(className, PicVirtClass, PicVirtObject)

#define MAKE_PICVIRTCLASS(className, base) \
    MAKE_PICANYCLASS(className, PicVirtClass, base)

// --------------------------------------------------------------------------
//  A helper class to enable static initialization of class info:

typedef const PicVirtClass& PICFVIRTCLASSID();

struct PicVirtClassInfo
{
    const char*   pClassName;
    PICUSHORT        usInstSize;
    PICFVIRTCLASSID* pfBase;
};

/*:class ********************************************************************

PicVirtClass is a class identifier designed for classes in virtual
hierarchies. It works hand-in-hand with the abstract class PicVirtObject.
Every virtual hierarchy should begin from PicVirtObject.

PicVirtClass and PicVirtObject together provide the following run-time
type support:

    * Class name, instance size.
    * Class directory records information about the class hierarchy;
      the hierarchy can be searched in two directions: either for
      ancestor classes or for descendant classes.

Like the simple PicClass objects, PicVirtClasses are created as static
objects in the implementation of the corresponding PicVirtObject. Macros
are provided for proper initialization:

    // In the class header:
    PICVIRTCLASS(className)

    // In the class implementation:
    MAKE_PICVIRTCLASS(className, baseClassName)

These macros declare and implement two member functions:

    static  const PicVirtClass& classId ();
    virtual const PicVirtClass& virtId  () const;

The static function is used to access a class id directly and virtual
function is used to determine the class of an arbitrary PicVirtObject.

These macros are actually used seldom, because various class hierarchies
define their own class identifiers as derived from PicVirtClass, e.g:

    PicObject hierarchy: PicObjectClass, (MAKE_)PICOBJECTCLASS
    PicStream hierarchy: PicStreamClass, (MAKE_)PICSTREAMCLASS
    etc.

The derived hierarchies are advised to define an inline member function
along the following example:

    const PicObjectClass& id () { return (const PicObjectClass&)virtId(); }

Type cast is safe here, because every PicObject uses PicObjectClass for
its class identification.

****************************************************************************/

class PicVirtClass : public PicClass
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTOR & DESTRUCTOR
//:fun ----------------------------------------------------------------------

    PicVirtClass (PicVirtClassInfo*);
   ~PicVirtClass ();

/*
The constructor of PicVirtClass is usually not used directly, but only
through the macro MAKE_PICVIRTCLASS. It updates the class directory to
contain the newly created class. Destructor removes the class from the
directory.
*/

// --------------------------------------------------------------------------
//      CLASS DIRECTORY
//:fun ----------------------------------------------------------------------

    PICBOOL hasBase     (const PicVirtClass&) const;
    PICBOOL hasAncestor (const PicVirtClass&) const;
    PICBOOL isKindOf    (const PicVirtClass&) const;

    friend PICBOOL operator >  (const PicVirtClass&, const PicVirtClass&);
    friend PICBOOL operator >= (const PicVirtClass&, const PicVirtClass&);
    friend PICBOOL operator <  (const PicVirtClass&, const PicVirtClass&);
    friend PICBOOL operator <= (const PicVirtClass&, const PicVirtClass&);

/*
These methods use the class directory to examine the inheritance structure.
For comparison operators note that the usual rules don't hold:

    not(c1 >= c2)  is not equivalent to  (c1 < c2)
    not(c1 <= c2)  is not equivalent to  (c1 > c2)

This is because the order defined by the inheritance tree is partial.

:ret hasBase:     boolTrue if the argument is the direct base class of the
                  receiver.
:ret hasAncestor: boolTrue if the argument is a direct or indirect
                  base class of the receiver.
:ret isKindOf:    boolTrue if the receiver is the same class as
                  the argument, or a publicly derived class.
:ret operator >:  Argument 1 is derived from argument 2
                  (either directly or indirectly).
:ret operator >=: Argument 1 isA argument 2.
:ret operator <:  Argument 2 is derived from argument 1
                  (either directly or indirectly).
:ret operator <=: Argument 2 isA argument 1.
*/

//:fun ----------------------------------------------------------------------

    const PicVirtClass* base         () const;
    const PicVirtClass* firstDerived () const;
    const PicVirtClass* nextSibling  () const;

    const PicVirtClass* derived (const char* className) const;

/*
These methods examine the class directory:

:ret base:         Base class of the receiver, if any.
:ret firstDerived: First class derived from the receiver or NULLPTR if there
                   are no (known) derived classes.
:ret nextSibling:  Next class in the set of classes derived from a common
                   base. Use these two methods to iterate over the set of
                   derived classes. For example, the second derived class
                   is base.firstDerived()->nextSibling() (but remember
                   to check against NULLPTR return value indicating the
                   end of the list).
:ret derived:      Returns the class id of the given class derived from
                   the receiver or NULLPTR if no such class exists. With
                   this method you can access classes by name.
:ret stdClass:     Returns the class id of the given standard PIC class
                   or NULLPTR, if the class doesn't exist.
*/

//:fun ----------------------------------------------------------------------

    void nassert (const PicVirtObject&  DEBUG_FILE_LINE) const;
    void nassert (const PicVirtObject*  DEBUG_FILE_LINE) const;

    PICBOOL typeGuard (const PicVirtObject&  DEBUG_FILE_LINE) const;

/*
These methods are used in detecting and reporting type mismatch errors.
They are normally not used directly, but only through the macros TYPE_GUARD
and DEBUG_ASSERT, and as part of AS-conversions.
*/

    PICCLASS()

    private:
// ==========================================================================
//      DATA MEMBERS

    PicVirtClass* pBase;
                                // 2-way ring list of derived classes:
    PicVirtClass* pDerived;     // mark node ("first" derived class)
    PicVirtClass* pNextSibling; // forward link
    PicVirtClass* pPrevSibling; // backward link

    DENY_COPYING_FOR (PicVirtClass)

    PicVirtClass (); // To construct PicVirtObject

    friend class PicVirtObject;

};

/*:class ********************************************************************

PicVirtObject is a base class for virtual class hierarchies. It defines
the common elements of run-time type checking. PicVirtClass (or derived)
objects are always used to identify classes derived from PicVirtObject.
PicVirtObjects cannot be created, but pointers and references may be used.

****************************************************************************/

class PicVirtObject
{
    public:

//:fun ----------------------------------------------------------------------

    virtual ~PicVirtObject ();

/*
A virtual destructor is defined for PicVirtObject so that instances are
deleted correctly also through a pointer or reference to the base class.
*/

// --------------------------------------------------------------------------
//      CLASS ID
//:fun ----------------------------------------------------------------------

    static  const PicVirtClass& classId ();
    virtual const PicVirtClass& virtId  () const =0;
            const PicVirtClass& id      () const;

/*
:ret classId: Class identifier of PicVirtObject.
:ret virtId:  Class identifier of the receiver.
:ret id:      Same as virtId but may be redefined to return a derived
              class identifier.
*/

    PICERRORCODES { ok,
        typeMismatch       = ERR_OBJECT_TYPE_MISMATCH,
        nullPointer        = ERR_OBJECT_NULL_POINTER,
        undefinedOperation = ERR_OBJECT_UNDEFINED_OPERATION
    };

//:fun ----------------------------------------------------------------------

    void raiseTypeMismatch (
        const char* expectedClass
        DEBUG_FILE_LINE) const;

    void undefined (const char* method) const;

/*
The method raiseTypeMismatch is useful if you do type checking directly
and need to raise an error. 
Note that you must use the DEBUG_FILE_LINE and DEBUG_FILE_LINE_IMP
macros to give suitable arguments.

The undefined-method is called in a virtual method that is undefined in
some derived class. It raises an appropriate error.

:arg expectedClass: Name of the expected class.
:arg method:        Name of the undefined method (excluding class name).

:err typeMismatch:       Always raised in the method raiseTypeMismatch.
:err undefinedOperation: Always raised in the method undefined.
*/

// --------------------------------------------------------------------------
//      COMPARISON
//:fun ----------------------------------------------------------------------

    PICBOOL is (const PicVirtObject&) const;

/*
Two PicVirtObjects are the same object if they occupy the same location in
memory. The sameness comparison cannot be redefined.
*/

    PICBOOL typeGuard (const PicVirtClass& expected  DEBUG_FILE_LINE) const
    {
        return expected.typeGuard(*this  DEBUG_FILE_LINE_IMP);
    }

    protected:

    PicVirtObject () {} // Otherwise compiler may generate a warning.

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PICBOOL PicVirtObject::is (const PicVirtObject& obj) const
       { return this == &obj; }

inline PICBOOL PicVirtClass::isKindOf (const PicVirtClass& vc) const
       { return (this == &vc) || this->hasAncestor(vc); }

inline PICBOOL operator >  (const PicVirtClass& vc1, const PicVirtClass& vc2)
       { return vc1.hasAncestor(vc2); }
inline PICBOOL operator >= (const PicVirtClass& vc1, const PicVirtClass& vc2)
       { return (vc1 == vc2) || vc1.hasAncestor(vc2); }
inline PICBOOL operator <  (const PicVirtClass& vc1, const PicVirtClass& vc2)
       { return vc2.hasAncestor(vc1); }
inline PICBOOL operator <= (const PicVirtClass& vc1, const PicVirtClass& vc2)
       { return (vc1 == vc2) || vc2.hasAncestor(vc1); }

inline PICBOOL PicVirtClass::hasBase (const PicVirtClass& vc) const
       { return pBase == &vc; }
inline const PicVirtClass* PicVirtClass::base () const
       { return pBase; }

inline const PicVirtClass* PicVirtClass::firstDerived () const
       { return pDerived; }
inline const PicVirtClass* PicVirtClass::nextSibling () const
       { return (pNextSibling == pBase->pDerived) ? NULLPTR : pNextSibling; }

inline PicVirtClass::PicVirtClass () :
    pBase    (NULLPTR),
    pDerived (NULLPTR)
{
    pClassName = "PicVirtObject";
    usInstSize = sizeof(PicVirtObject);
}

inline const PicVirtClass& PicVirtObject::id () const { return virtId(); }

#endif
