// --------------------------------------------------------------------------
//  'GENERIC' ARRAY BASE CLASS implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <stdlib.h>
#include <string.h>
}

#include <bcollect.hxx>
#include <palloc.hxx>
#include <picerror.hxx>
#include <picbase.erh>

MAKE_PICCLASS(BaseCollection)

ERROR_MESSAGES(BaseCollection)
PARAMSTR(full, "Can not increase collection to size of @*")
PARAMSTR(illegalIndex, "Illegal index @*")
PARAMSTR(illegalLowerBound, "Illegal lower bound @*")
PARAMSTR(putNullObject, "Null object put")
PARAMSTR(unsortable, "Collection is unsortable")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      COPY
// --------------------------------------------------------------------------

void BaseCollection::copyFrom (const BaseCollection& array)
{
    DEBUG_REQUIRE(info->createItemCopy != NULLPTR);
    setNull();

    if (array.ub > 0)
    {
        allocate(array.ub + 1, boolFalse);
        DEBUG_REQUIRE(alloc <= array.alloc);

        if (content)
        {
            for (PICPOSITION ix = 1; ix < alloc; ix++)
                (*info->createItemCopy)(THIS[ix], array[ix]);
            ub = array.ub;
        }
    }
}

// --------------------------------------------------------------------------
//      COMPARISON
// --------------------------------------------------------------------------

PICBOOL BaseCollection::operator == (const BaseCollection& array) const
{
    DEBUG_REQUIRE(info->ascending != NULLPTR);

    if (lastPos() != array.lastPos())
        return boolFalse;

    for (PICPOSITION ix = 1; ix <= ub; ix++)
        if ((*info->ascending)(THIS[ix], array[ix]) != 0)
            return boolFalse;

    return boolTrue;
}

// --------------------------------------------------------------------------
//      ITERATORS
// --------------------------------------------------------------------------

void BaseCollection::forAll (PROC_SETITEM* setItem)
{
    PICPOSITION ix = 0;

    while (++ix <= ub)
        (*setItem)(THIS[ix]);
}

PICSIZE BaseCollection::count (PROC_TESTITEM* testItem, PICUSHORT result) const
{
    PICSIZE count = 0;

    for (PICPOSITION ix = 1; ix <= ub; ix++)
        if ((*testItem)(THIS[ix]) == result)
            count++;

    return count;
}

PICPOSITION BaseCollection::findNext (
    PICPOSITION    ix,
    PROC_TESTITEM* testItem,
    PICUSHORT      result) const
{
    while (++ix <= ub)
    {
        if ((*testItem)(THIS[ix]) == result)
            return ix;
    }
    return 0;
}

PICPOSITION BaseCollection::findPrev (
    PICPOSITION    ix,
    PROC_TESTITEM* testItem,
    PICUSHORT      result) const
{
    if (ix == 0)
        ix = ub + 1;

    while (--ix > 0)
    {
        if ((*testItem)(THIS[ix]) == result)
            return ix;
    }
    return 0;
}

// --------------------------------------------------------------------------
//      ACCESS
// --------------------------------------------------------------------------

void BaseCollection::positionError (PICPOSITION ix) const
{
    ERAISEFMT(BaseCollection, illegalIndex) << ix;
}

// --------------------------------------------------------------------------
//      PUT & INSERT
// --------------------------------------------------------------------------

PICBOOL BaseCollection::put (PICPOSITION ix)
{
    if (ix > 0)
    {
        if (ix <= ub)
            return boolTrue;
        else
            return addLoc(ix);
    }
    return boolFalse;
}

PICBOOL BaseCollection::addLoc (PICPOSITION ix)
{
    DEBUG_REQUIRE(ix > ub);

    if (ix >= alloc && !resize(ix + 1))
        return boolFalse;

    ub = ix;

    return boolTrue;
}

// --------------------------------------------------------------------------

PICPOSITION BaseCollection::insert (PICPOSITION ix)
{
    if (ix == 0)
        ix = lastPos() + 1;

    if (ix <= ub)
    {
        if ((*info->isItemNull)(THIS[ix]))
            return ix;
        else
            return insertLoc(ix) ? ix : 0;
    }
    return addLoc(ix) ? ix : 0;
}

PICBOOL BaseCollection::insertLoc (PICPOSITION locIx)
{
    DEBUG_REQUIRE(locIx > 0);

    PICPOSITION ix;

    for (ix = locIx + 1; ix <= ub; ix++)
        if ((*info->isItemNull)(THIS[ix]))
            goto MakeRoom;

    if (!addLoc(ix))
        return boolFalse;

  MakeRoom:
    memmove(THIS[locIx + 1], THIS[locIx], (ix - locIx) * info->itemSize);
    (*info->createItem)(THIS[locIx]);
    return boolTrue;
}

// --------------------------------------------------------------------------
//      DELETE, REMOVE & PURGE
// --------------------------------------------------------------------------

void BaseCollection::deleteAll ()
{
    if (content)
    {
        forAll(info->setItemNull);
        ub = 0;
    }
}

void BaseCollection::removeAll ()
{
    if (content)
    {
        picFree(content);
        content = NULLPTR;
        alloc   = 0;
        ub      = 0;
    }
}

void BaseCollection::setNull ()
{
    if (content)
    {
        forAll(info->setItemNull);
        picFree(content);
        content = NULLPTR;
        alloc   = 0;
        ub      = 0;
    }
}

// --------------------------------------------------------------------------

void BaseCollection::purge (PICPOSITION ix, PICBOOL inOrder)
{
    if (ix > 0 && ix <= lastPos())
    {
        (*info->setItemNull)(THIS[ix]);

        if (inOrder)
        {
            // Move all elements after emptied position
            // ---> order is preserved
            memmove(THIS[ix], THIS[ix + 1], (ub - ix) * info->itemSize);
        }
        else
        {
            // Move last to emptied position
            // ---> order is not preserved
            memmove(THIS[ix], THIS[ub], info->itemSize);
        }
        (*info->createItem)(THIS[ub]);
        lastPos(); // Reset upper bound
    }
}

void BaseCollection::purge (PICBOOL inOrder)
{
    lastPos();

    for (PICPOSITION ix = 1; ix < ub; ix++)
    {
        if ((*info->isItemNull)(THIS[ix]))
        {
            if (inOrder)
            {
                // Move all elements after emptied position
                // ---> order is preserved
                memmove(THIS[ix], THIS[ix + 1], (ub - ix) * info->itemSize);
            }
            else
            {
                // Move last to emptied position
                // ---> order is not preserved
                memmove(THIS[ix], THIS[ub], info->itemSize);
            }
            (*info->createItem)(THIS[ub]);
            lastPos();
        }
    }
}

// --------------------------------------------------------------------------
//      ATTRIBUTES
// --------------------------------------------------------------------------

PICPOSITION BaseCollection::firstFree () const
{
    PICPOSITION ix;

    if (ix = findNext(0, info->isItemNull, boolTrue))
        return ix;
    else if (ub < maxNewSize / info->itemSize && ub < ::maxSize)
        return ub + 1;
    else
        return 0;
}

// --------------------------------------------------------------------------
//      SORT & SEARCH
// --------------------------------------------------------------------------

PICPOSITION BaseCollection::findEqual (
    const void*        obj,
    PROC_COMPAREITEMS* compare) const
{
    for (PICPOSITION ix = 1; ix <= ub; ix++)
        if ((*compare)(obj, THIS[ix]) == 0)
            return ix;

    return 0;
}

PICPOSITION BaseCollection::findSorted (
    const void*        obj,
    PROC_COMPAREITEMS* compare) const
{
    if (lastPos() > 0)
    {
        PicErrorContext e;

        // Apply the bsearch-function from C run-time library:

        void* loc = bsearch (
            obj, content + info->itemSize, ub, info->itemSize, compare);

        if (PicError::isDetected()) // Array contains uncomparable elements.
            EHOLD(BaseCollection, unsortable);
        else
            return loc ? ((char*)loc - content) / info->itemSize : 0;
    }
    return 0;
}

PICPOSITION BaseCollection::findMin (PROC_COMPAREITEMS* compare) const
{
    if (lastPos() > 0)
    {
        PicErrorContext e;
        PICPOSITION ix = 1, atIx = 1;

        while (++ix <= ub)
            if ((*compare)(THIS[ix], THIS[atIx]) < 0)
                atIx = ix;

        if (PicError::isDetected()) // Array contains uncomparable elements.
            EHOLD(BaseCollection, unsortable);
        else
            return atIx;
    }
    return 0;
}

PICBOOL BaseCollection::sort (PROC_COMPAREITEMS* compare)
{
    purge(boolFalse); // Remove nulls to make sorting easier

    if (ub > 1)
    {
        PicErrorContext e;

        // Apply the qsort-function from C run-time library:

        qsort(content + info->itemSize, ub, info->itemSize, compare);

        if (PicError::isDetected()) // Array contains uncomparable elements.
        {
            EHOLD(BaseCollection, unsortable);
            return boolFalse;
        }
    }
    return boolTrue;
}

// --------------------------------------------------------------------------

PICBOOL BaseCollection::contains (const BaseCollection& array) const
{
    PICPOSITION ix = 0;

    while (ix = array.findNext(ix, info->isItemNull, boolFalse))
    {
        if (!findEqual(array[ix], info->ascending))
            return boolFalse;
    }
    return boolTrue;
}

// --------------------------------------------------------------------------
//      ALLOCATION
// --------------------------------------------------------------------------

PICBOOL BaseCollection::allocate (PICSIZE amount, PICBOOL clear)
{
    if (amount <= (maxNewSize / info->itemSize))
    {
        PICUSHORT bytes = picCalculateSize(amount * info->itemSize);
        if (content = (char*)picAllocate(bytes, clear))
        {
            alloc = bytes / info->itemSize;
            return boolTrue;
        }
        else
            MEMORY_ERROR("BaseCollection::allocate failed");
    }
    else
    {
        ERAISEFMT(BaseCollection, full) << amount;
    }

    content = NULLPTR;
    alloc   = 0;
    ub      = 0;
    return boolFalse;
}

PICBOOL BaseCollection::allocate (PICSIZE amount)
{
    if (allocate(amount, !info->initItem))
    {
        if (info->initItem)
        {
            for (PICOFFSET ix = 0; ix < alloc; ix++)
                (*info->initItem)(THIS[ix]);
        }
        return boolTrue;
    }
    return boolFalse;
}

PICBOOL BaseCollection::resize (PICSIZE amount, PICBOOL clear)
{
    if (amount <= (maxNewSize / info->itemSize))
    {
        PICUSHORT bytes = picCalculateSize(amount * info->itemSize);
        void*     buf   = picResize (
                             content, alloc * info->itemSize, bytes, clear);
        if (buf)
        {
            content = (char*)buf;
            alloc   = bytes / info->itemSize;
            return boolTrue;
        }
        else
            MEMORY_ERROR("BaseCollection::resize failed");
    }
    else
    {
        ERAISEFMT(BaseCollection, full) << amount;
    }

    return boolFalse;
}

PICBOOL BaseCollection::resize (PICSIZE amount)
{
    PICOFFSET ix = alloc;

    if (resize(amount, !info->initItem))
    {
        if (info->initItem)
        {
            while (ix < alloc)
                (*info->initItem)(THIS[ix++]);
        }
        return boolTrue;
    }
    return boolFalse;
}

