// --------------------------------------------------------------------------
//  BASE STRING implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
}

#include <bstring.hxx>
#include <picerror.hxx>
#include <palloc.hxx>
#include <picbase.erh>

#define PICDEBUGSTRING

static const char emptyAsciizString[] = "";
static const char thisFile[]          = __FILE__;

MAKE_PICCLASS(BaseString)

ERROR_MESSAGES(BaseString)
PARAMSTR(nullInstance, "Operation on Null instance.")
PARAMSTR(positionError, "Position @* is out of range")
PARAMSTR(sizeError, "Size @* is greater than maximum string size.")
END_PARAM_STRINGS

const BaseString& BaseString::null ()
{
    static BaseString nullString;
    return nullString;
}

const BaseString& BaseString::empty ()
{
    static BaseString emptyString (emptyAsciizString);
    return emptyString;
}

// Hides the access to string information

#define STRSIZE  info().size
#define STRALLOC info().alloc
#define STRCOUNT info().count

// --------------------------------------------------------------------------
//      STATIC (LOCAL) FUNCTIONS
// --------------------------------------------------------------------------
//
//      BaseString has maximum size maxNewSize - sizeof(StringRep).
//      isValidSize(PICSIZE) tests if given size is valid. The second one
//      is used when the size of the receiver is about to increase.
//      Count between returns how many characters are between two positions.
//      countBetween (pos, pos) == 1.

static void stringSizeError (PICSIZE size, PICSIZE addition, PICUSHORT line)
{
    PicStringText text;
    text("@* + @*") << size << addition;

    ERAISEFMT_LINE_FILE (
        BaseString::errors, BaseString::classId,
        BaseString::sizeError, line, thisFile) << text;
}

static void stringPosError (PICPOSITION pos, PICUSHORT line)
{
    ERAISEFMT_LINE_FILE (
        BaseString::errors, BaseString::classId,
        BaseString::positionError, line, thisFile) << pos;
}

inline PICSIZE bufferSize (PICSIZE stringSize)
{
    return stringSize + sizeof(PicStringInfo) + 1;
}

inline PICSIZE stringSize (PICSIZE bufferSize)
{
    return bufferSize - sizeof(PicStringInfo) - 1;
}

inline PICBOOL isValidSize (PICSIZE aSize)
{
    return aSize <= maxStringSize;
}

inline PICBOOL isValidSize (PICSIZE aSize, PICSIZE addition)
{
    return (maxStringSize - aSize) >= addition;
}

inline PICSIZE countBetween (PICPOSITION from, PICPOSITION to)
{
    return from <= to ? to - from + 1 : 0;
}

// --------------------------------------------------------------------------
//      CHARACTER/ASCIIZ COPYING
//
//      asciiz* functions copy characters and append a NUL characer.
//      char* functions copy characters but don't append NUL.
//      If you replace memmove with some other routine take care that
//      it handles overlapping regions correctly.

inline void asciizCopy (char* to, const char* from, PICSIZE sz)
{
    memmove(to, (void*)from, sz);
    *(to + sz) = '\0';
}

inline void charCopy (char* to, const char* from, PICSIZE sz)
{
    memmove(to, (void*)from, sz);
}

inline void asciizSet (char* to, char ch, PICSIZE times)
{
    memset(to, ch, times);
    *(to + times) = '\0';
}

inline void charSet (char* to, char ch, PICSIZE times)
{
    memset(to, ch, times);
}

// --------------------------------------------------------------------------
//      CREATION
//
//      Allocation must reserve memory both for StringRep and character
//      buffer. Because buffer starts directly after other instance
//      variables there can't be constructor for the buffer. After allocation
//      memory map is:
//      |alloc|size|count|content...|
//        2     2     2    alloc+1    bytes
//
//      Note that  0 <= allocSize <= maxStringSize
//      and        newSize <= allocSize <= maxStringSize.

static const char allocFail[] = "Allocation of BaseString failed.";

static char* createString (
    const char* azStr,
    const PICSIZE  newSize,
    const PICSIZE  allocSize)
{
    PICSIZE size = bufferSize(allocSize);
    PicStringInfo* info = (PicStringInfo*)picAllocate(size);

    if (info)
    {
        info->size  = newSize;
        info->alloc = stringSize(size);
        info->count = 1;
        if (azStr)
            asciizCopy(info->chars(), azStr, newSize);
        return info->chars();
    }

    PicError::raiseMemoryError(allocFail);
    return NULLPTR;
}

static char* createString (char ch, PICSIZE times)
{
    PICSIZE size = bufferSize(times);
    PicStringInfo* info = (PicStringInfo*)picAllocate(size);

    if (info)
    {
        info->size  = times;
        info->alloc = stringSize(size);
        info->count = 1;
        asciizSet(info->chars(), ch, times);
        return info->chars();
    }

    PicError::raiseMemoryError(allocFail);
    return NULLPTR;
}

inline void destroy (PicStringInfo* str)
{
    picFree(str);
}

// --------------------------------------------------------------------------
//      USE
//
//      Use allocates string buffer which can hold at least reqSize
//      characters. It must create new string buffer if count > 1 or there
//      is not enough space (either null string or alloc < reqSize).

PICBOOL BaseString::use (PICSIZE reqSize, PICBOOL saveOld)
{
    if (rep == NULLPTR)
        rep = createString(NULLPTR, 0, reqSize);
    else if (STRCOUNT > 1)
    {
        STRCOUNT--;
        if (saveOld)
            rep = createString(rep, STRSIZE,
                reqSize > STRSIZE ? reqSize : STRSIZE);
        else
            rep = createString(NULLPTR, 0, reqSize);
    }
    else if (STRALLOC < reqSize)
    {
        PicStringInfo* old = &info();
        if (saveOld)
            rep = createString(rep, old->size,
                reqSize > old->size ? reqSize : old->size);
        else
            rep = createString(NULLPTR, 0, reqSize);
        destroy(old);
    }
    else
        return boolTrue;

    return rep != NULLPTR;
}

void BaseString::deleteStringRep ()
{
    if (rep && --STRCOUNT == 0)
        destroy(&info());
    rep = NULLPTR;
}

// --------------------------------------------------------------------------
//      private constructor which is used to create substrings, so
//      aSize should always be valid String size.

BaseString::BaseString (const char* azStr, PICSIZE aSize)
{
    rep = createString(azStr, aSize, aSize);
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNEMENT

BaseString::BaseString (const char* azStr)
{
    if (azStr)
    {
        PICSIZE sz = strlen(azStr);

        if (isValidSize(sz))
        {
            rep = createString(azStr, sz, sz);
            return;
        }
        else
        {
            ERAISEFMT(BaseString, sizeError) << sz;
        }
    }
    rep = NULLPTR;
}

BaseString::BaseString (char ch, PICSIZE times)
{
    if (isValidSize(times))
        rep = createString(ch, times);
    else
    {
        rep = NULLPTR;
        ERAISEFMT(BaseString, sizeError) << times;
    }
}

BaseString::BaseString (const BaseString& str)
{
    rep = str.rep;
    if (rep)
        STRCOUNT++;
}

BaseString& BaseString::operator = (const char* azStr)
{
    if (azStr == NULLPTR)
        deleteStringRep();
    else
    {
        PICSIZE sz = strlen(azStr);

        if (!isValidSize(sz))
        {
            ERAISEFMT(BaseString, sizeError) << sz;
        }
        else if (rep)
        {
            if (STRCOUNT > 1)
            {
                STRCOUNT--;
                rep = createString(azStr, sz, sz);
            }
            else if (STRALLOC < sz)
            {
                PicStringInfo* old = &info();      // Note: azStr may be
                rep = createString(azStr, sz, sz); // inside this string!
                destroy(old);
            }
            else
            {
                asciizCopy(rep, azStr, sz);
                STRSIZE = sz;
            }
        }
        else
            rep = createString(azStr, sz, sz);
    }
    return *this;
}

BaseString& BaseString::operator = (const BaseString& str)
{
    if (rep != str.rep)
    {
        if (rep && --STRCOUNT == 0)
            destroy(&info());
        rep = str.rep;
        if (rep)
            STRCOUNT++;
    }
    return *this;
}

BaseString BaseString::copy () const
{
    if (rep)
        return BaseString(rep, STRSIZE);
    return BaseString();
}

// --------------------------------------------------------------------------
//      compare implements string comparison by using the given character
//      comparison function (PicChar::compare or PicChar::caseCompare).
//      Strings must not be NULLPTR, but sizes may be zero.

static PicCompareResult compare (
    const char* str1,
    PICSIZE        size1,
    const char* str2,
    PICSIZE        size2,
    PicCompareResult (*compChars)(char, char))
{
    const char *last1 = str1 + size1, *last2 = str2 + size2;

    while (str1 < last1 && str2 < last2)
    {
        PicCompareResult result = compChars(*str1++, *str2++);
        if (result != PicChar::isEqual)
            return result;
    }
    return PicCompareResult((size1 > size2) - (size1 < size2));
}

// --------------------------------------------------------------------------
//      SEARCHING
//
//      searchCS implements case-sensitive searching.
//      searchCI implements case-insensitive searching.
//       - pointers must not be NULL, but findSize > strSize is OK.
//       - findSize must not be zero
//       - case-sensitive searching is more efficient (uses memcmp).

static const char* searchCS (
    const char* string,
    PICSIZE        strSize,
    const char* find,
    PICSIZE        findSize)
{
    const char *ctry, *last = string + strSize - findSize;

    for (ctry = string; ctry <= last; ctry++)
    {
        if (memcmp(ctry, find, findSize) == 0)
            return ctry;
    }
    return NULLPTR;
}

static const char* searchCI (
    const char* string,
    PICSIZE        strSize,
    const char* find,
    PICSIZE        findSize)
{
    const char *ctry, *last = string + strSize - findSize;

    for (ctry = string; ctry <= last; ctry++)
    {
        if (::compare(ctry, findSize, find, findSize, PicChar::compare)
             == PicChar::isEqual)
            return ctry;
    }
    return NULLPTR;
}

// --------------------------------------------------------------------------

const char* BaseString::find (
    const char*     azStr,
    const char*     search,
    PicCaseSensitivity caseSense)
{
    if (!azStr || !search)
        return NULLPTR;
    PICSIZE sz1 = strlen(azStr), sz2 = strlen(search);
    if (sz1 && sz2)
    {
        return (caseSense == caseSensitive)
            ? searchCS(azStr, sz1, search, sz2)
            : searchCI(azStr, sz1, search, sz2);
    }
    return NULLPTR;
}

const char* BaseString::find (
    const char*     azStr,
    char            search,
    PicCaseSensitivity caseSense)
{
    if (!azStr)
        return NULLPTR;
    PICSIZE sz1 = strlen(azStr);
    if (sz1)
    {
        return (caseSense == caseSensitive)
            ? (const char*)memchr(azStr, search, sz1)
            : searchCI(azStr, sz1, &search, 1);
    }
    return NULLPTR;
}

// --------------------------------------------------------------------------

PICPOSITION BaseString::positionOf (
    const char*     azStr,
    PICPOSITION        startPos,
    PicCaseSensitivity aCase) const
{
    if (startPos) startPos--;
    if (azStr && rep && startPos < STRSIZE)
    {
        PICSIZE sz = strlen(azStr);

        if (sz > 0)
        {
            const char* start = rep + startPos;
            PICSIZE        rest  = STRSIZE - startPos;
            const char* pch   = (aCase == caseSensitive)
                ? searchCS(start, rest, azStr, sz)
                : searchCI(start, rest, azStr, sz);
            return pch ? pch - rep + 1 : outOfRange;
        }
    }
    return outOfRange;
}

PICPOSITION BaseString::positionOf (
    char            ch,
    PICPOSITION        startPos,
    PicCaseSensitivity aCase) const
{
    if (startPos) startPos--;
    if (rep && startPos < STRSIZE)
    {
        const char* start = rep + startPos;

        if (aCase == caseInsensitive)
        {
            for (const char* end = rep + STRSIZE; start < end; start++)
                if (PicChar::compare(*start, ch) == PicChar::isEqual)
                    return start - rep + 1;
            return outOfRange;
        }
        else
        {
            char* pch = (char*)memchr(start, ch, STRSIZE - startPos);
            return pch ? pch - rep + 1 : outOfRange;
        }
    }
    return outOfRange;
}

PICPOSITION BaseString::positionOf (
    const BaseString&  str,
    PICPOSITION        startPos,
    PicCaseSensitivity aCase) const
{
    if (startPos) startPos--;
    if (!str.isEmpty() && rep && startPos < STRSIZE)
    {
        PICSIZE sz = str.size();

        const char* start = rep + startPos;
        PICSIZE        rest  = STRSIZE - startPos;
        const char* pch   = (aCase == caseSensitive)
            ? searchCS(start, rest, str.rep, sz)
            : searchCI(start, rest, str.rep, sz);
        return pch ? pch - rep + 1 : outOfRange;
    }
    return outOfRange;
}

// --------------------------------------------------------------------------
//      COMPARISON
// --------------------------------------------------------------------------

PicCompareResult BaseString::compare (
    const char*     str1,
    const char*     str2,
    PicCaseSensitivity caseSense)
{
    if (str1 == str2)
        return PicChar::isEqual;
    else if (!str1 || !str2)
        return PicCompareResult(!str2 - !str1); // Either or both are null
    else
        return caseSense == caseSensitive ?
            ::compare(str1, strlen(str1), str2, strlen(str2), PicChar::caseCompare) :
            ::compare(str1, strlen(str1), str2, strlen(str2), PicChar::compare);
}

// --------------------------------------------------------------------------

PICBOOL BaseString::operator == (const BaseString& str) const
{
    if (rep == str.rep)
        return boolTrue;
    else if (isNull() || str.isNull() || STRSIZE != str.info().size)
        return boolFalse;
    else
        return ::compare(rep, STRSIZE, str.rep, str.info().size,
            PicChar::caseCompare) == PicChar::isEqual;
}

PICBOOL BaseString::operator >= (const BaseString& str) const
{
    if (rep == str.rep)
        return boolTrue;
    else if (isNull() || str.isNull())
        return str.isNull();
    else switch (::compare(rep, STRSIZE, str.rep, str.info().size,
                 PicChar::caseCompare))
         {
             case PicChar::isEqual:   return boolTrue;
             case PicChar::isGreater: return boolTrue;
             case PicChar::isLess:    return boolFalse;
         }
    return boolTrue;
}

PICBOOL BaseString::operator > (const BaseString& str) const
{
    if (rep == str.rep)
        return boolFalse;
    else if (isNull() || str.isNull())
        return str.isNull();
    else
        return ::compare(rep, STRSIZE, str.rep, str.info().size,
            PicChar::caseCompare) == PicChar::isGreater;
}

// --------------------------------------------------------------------------

PICBOOL BaseString::equal (
    const BaseString&  str,
    PicCaseSensitivity aCase) const
{
    if (rep == str.rep)
        return boolTrue;
    else if (isNull() || str.isNull() || STRSIZE != str.info().size)
        return boolFalse;
    else
        return PicChar::isEqual == (aCase == caseSensitive ?
        ::compare(rep, STRSIZE, str.rep, str.info().size, PicChar::caseCompare) :
        ::compare(rep, STRSIZE, str.rep, str.info().size, PicChar::compare));
}

PICBOOL BaseString::greaterEqualThan (
    const BaseString&  str,
    PicCaseSensitivity aCase) const
{
    if (rep == str.rep)
        return boolTrue;
    else if (isNull() || str.isNull())
        return str.isNull();
    else switch ((aCase == caseSensitive ?
        ::compare(rep, STRSIZE, str.rep, str.info().size, PicChar::caseCompare) :
        ::compare(rep, STRSIZE, str.rep, str.info().size, PicChar::compare)))
         {
             case PicChar::isEqual:   return boolTrue;
             case PicChar::isGreater: return boolTrue;
             case PicChar::isLess:    return boolFalse;
         }
    return boolTrue;
}

PICBOOL BaseString::greaterThan (
    const BaseString&  str,
    PicCaseSensitivity aCase) const
{
    if (rep == str.rep)
        return boolFalse;
    else if (isNull() || str.isNull())
        return str.isNull();
    else
        return PicChar::isGreater == (aCase == caseSensitive ?
        ::compare(rep, STRSIZE, str.rep, str.info().size, PicChar::caseCompare) :
        ::compare(rep, STRSIZE, str.rep, str.info().size, PicChar::compare));
}

// --------------------------------------------------------------------------
//      ACCESS

char BaseString::at (PICPOSITION pos) const
{
    if (inRange(pos))
        return rep[pos - 1];

    if (isNull())
    {
        ERAISE(BaseString, nullInstance);
    }
    else
        stringPosError(pos, __LINE__);
    return '\0';
}

char BaseString::last () const
{
    return isEmpty() ? '\0' : rep[STRSIZE - 1];
}

// --------------------------------------------------------------------------
//      LOW-LPICL ACCESS

const char* BaseString::chars () const
{
    return rep ? rep : emptyAsciizString;
}

void BaseString::setChars (const char* buf, PICSIZE count)
{
    if (!buf)
        setNull();
    else if (isValidSize(count))
    {
        if (use(count, boolFalse))
        {
            asciizCopy(rep, buf, count);
            STRSIZE = count;
        }
    }
    else
        stringSizeError(count, 0, __LINE__);
}

void BaseString::setChars (char ch, PICSIZE times)
{
    if (isValidSize(times))
    {
        if (use(times, boolFalse))
        {
            asciizSet(rep, ch, times);
            STRSIZE = times;
        }
    }
    else
        stringSizeError(times, 0, __LINE__);
}

char* BaseString::chars (PICSIZE setSize)
{
    if (isValidSize(setSize))
    {
        if (use(setSize, boolFalse))
        {
            STRSIZE = setSize;
            rep[setSize] = '\0';
            return rep;
        }
    }
    else
        stringSizeError(setSize, 0, __LINE__);

    return NULLPTR;
}

// --------------------------------------------------------------------------
//      SUBSTRING

BaseString BaseString::at (PICPOSITION pos, PICSIZE sz) const
{
    if (rep && STRSIZE > 0 && pos <= STRSIZE)
    {
        if (pos) pos--;
        if (maxStringSize - pos >= sz && pos + sz < STRSIZE)
            return BaseString(rep + pos, sz);
        else
            return BaseString(rep + pos, STRSIZE - pos);
    }
    else
        return empty();
}

BaseString BaseString::between (PICPOSITION from, PICPOSITION to) const
{
    PICPOSITION startPos = from ? from : 1;

    return at(startPos, countBetween(startPos, to));
}

// --------------------------------------------------------------------------
//      DELETE

void BaseString::deleteAt (PICPOSITION pos, PICSIZE sz)
{
    if (sz && !isEmpty())
    {
        if (pos <= STRSIZE)
        {
            if (use(STRALLOC))
            {
                if (pos) pos--;
                if (sz > STRSIZE - pos)
                    sz = STRSIZE - pos;

                STRSIZE -= sz;
                asciizCopy(rep + pos, rep + pos + sz, STRSIZE - pos);
            }
        }
        else
            stringPosError(pos, __LINE__);
    }
}

void BaseString::deleteBetween (PICPOSITION from, PICPOSITION to)
{
    PICPOSITION startPos = from ? from : 1;

    deleteAt(startPos, countBetween(startPos, to));
}

// --------------------------------------------------------------------------
//      INSERT

void BaseString::insert (PICPOSITION to, const char* from, PICSIZE sz)
{
    if (from && (sz || (sz = strlen(from))))
    {
        if (--to <= size())
        {
            if (isValidSize(size(), sz))
            {
                if (use(size() + sz))
                {
                    char* start = rep + to;
                    asciizCopy(start + sz, start, STRSIZE - to);
                    charCopy(start, from, sz);
                    STRSIZE += sz;
                }
            }
            else
                stringSizeError(size(), sz, __LINE__);
        }
        else
            stringPosError(to + 1, __LINE__);
    }
}

void BaseString::insert (PICPOSITION to, char ch, PICSIZE times)
{
    if (times == 0)
        return;
    if (--to <= size())
    {
        if (isValidSize(size(), times))
        {
            if (use(size() + times))
            {
                char* start = rep + to;
                asciizCopy(start + times, start, STRSIZE - to);
                charSet(start, ch, times);
                STRSIZE += times;
            }
        }
        else
            stringSizeError(size(), times, __LINE__);
    }
    else
        stringPosError(to + 1, __LINE__);
}

// --------------------------------------------------------------------------
//      REPLACE

void BaseString::replace (PICPOSITION at, const char* from, PICSIZE sz)
{
    if (from && (sz || (sz = strlen(from))))
    {
        if (--at <= size())
        {
            if (isValidSize(at, sz))
            {
                PICSIZE newSz = at + sz;
                if (newSz < size())
                    newSz = size();
                if (use(newSz))
                {
                    charCopy(rep + at, from, sz);
                    if (newSz > STRSIZE)
                    {
                        STRSIZE = newSz;
                        rep[newSz] = '\0';
                    }
                }
            }
            else
                stringSizeError(at, sz, __LINE__);
        }
        else
            stringPosError(at + 1, __LINE__);
    }
}

void BaseString::replace (PICPOSITION at, char ch, PICSIZE times)
{
    if (times == 0)
        return;
    if (--at <= size())
    {
        if (isValidSize(at, times))
        {
            PICSIZE newSz = at + times;
            if (newSz < size())
                newSz = size();
            if (use(newSz))
            {
                charSet(rep + at, ch, times);
                if (newSz > STRSIZE)
                {
                    STRSIZE = newSz;
                    rep[newSz] = '\0';
                }
            }
        }
        else
            stringSizeError(at, times, __LINE__);
    }
    else
        stringPosError(at + 1, __LINE__);
}

void BaseString::replace (
    const BaseString&  keyword,
    const BaseString&  replacement,
    PicCaseSensitivity aCase)
{
    PICPOSITION pos = 1;
    
    while ((pos = positionOf(keyword, pos, aCase)) != outOfRange)
    {
        deleteAt(pos, keyword.size());
        insert(pos, replacement);
        pos += replacement.size();
    }    
}

void BaseString::replaceKeychar (
    char            keyChar,
    char            replacement,
    PicCaseSensitivity aCase)
{
    PICPOSITION pos = positionOf(keyChar, 1, aCase);

    if (pos != outOfRange && use(STRALLOC))
    {
        do
            THIS[pos] = replacement;
        while ((pos = positionOf(keyChar, pos + 1, aCase)) != outOfRange);
    }
}

// --------------------------------------------------------------------------
//      CATENATION

BaseString& BaseString::append (const char* from, PICSIZE sz)
{
    if (from && (sz || (sz = strlen(from))))
    {
        PICOFFSET to = size();

        if (isValidSize(to, sz))
        {
            if (use(to + sz))
            {
                asciizCopy(rep + to, from, sz);
                STRSIZE += sz;
            }
        }
        else
            stringSizeError(to, sz, __LINE__);
    }
    return *this;
}

BaseString& BaseString::append (char ch, PICSIZE times)
{
    if (times)
    {
        PICOFFSET to = size();

        if (isValidSize(to, times))
        {
            if (use(to + times))
            {
                asciizSet(rep + to, ch, times);
                STRSIZE += times;
            }
        }
        else
            stringSizeError(to, times, __LINE__);
    }
    return *this;
}

// --------------------------------------------------------------------------
//      MISC

PICSIZE BaseString::size (const char* azStr)
{
    return azStr ? strlen(azStr) : 0;
}

PICSIZE BaseString::size () const
{
    return rep ? info().size : 0;
}

BaseString& BaseString::trim ()
{
    if (isEmpty())
        return *this;

    PICPOSITION pos;

    for (pos = STRSIZE; pos; pos--)
        if (!PicChar::isSpace(THIS[pos]))
            break;
    if (pos)
    {
        deleteAfter(pos);

        PICPOSITION last = STRSIZE;

        for (pos = 1; pos <= last; pos++)
            if (!PicChar::isSpace(THIS[pos]))
                break;
        deleteBefore(pos);
    }
    else
        setEmpty(); // It was all blank

    return *this;
}

BaseString& BaseString::reverse ()
{
  if (!isEmpty() && use(STRALLOC)){
    // strrev(rep);
    PICSIZE i = 0, middle= size()/2, last=size()-1;
    char t;
    while(i<middle){
      t = rep[i];
      rep[i] = rep[last];
      rep[last] = t;
      i++; last--;
    }
  }
    return *this;
}

BaseString& BaseString::format (
    PICSIZE          fieldWidth,
    PicJustification justify,
    char          filler)
{
    if (size() == fieldWidth)
        return *this;
    else if (size() > fieldWidth)
        deleteAfter(fieldWidth);
    else
    {
        PICUSHORT fillers = fieldWidth - size();

        if (justify == toRight)
            prepend(filler, fillers);
        else if (justify == toLeft)
            append(filler, fillers);
        else
        {
            PICSIZE appendAmount = fillers / 2;
            prepend(filler, fillers - appendAmount);
            append(filler, appendAmount);
        }
    }
    return *this;
}

// --------------------------------------------------------------------------
//      EMPTY & NULL
//
//      setEmpty must make sure that null String is changed to empty
//      String.

void BaseString::setEmpty ()
{
    if (use(1, boolFalse))
    {
        rep[0]  = '\0';
        STRSIZE = 0;
    }
}

// --------------------------------------------------------------------------
//      CONVERSIONS & CLASSIFICATION

BaseString& BaseString::makeUppercase ()
{
    if (!isEmpty() && use(STRALLOC))
    {
        char *pch = rep + STRSIZE;
        while (--pch >= rep)
            *pch = PicChar::uppercase(*pch);
    }
    return *this;
}

BaseString& BaseString::makeLowercase ()
{
    if (!isEmpty() && use(STRALLOC))
    {
        char *pch = rep + STRSIZE;
        while (--pch >= rep)
            *pch = PicChar::lowercase(*pch);
    }
    return *this;
}

BaseString& BaseString::makeProper ()
{
    if (!isEmpty() && use(STRALLOC))
    {
        PICBOOL convertUp = boolTrue;
        char *pch = rep, *end = rep + STRSIZE;
        for (; pch < end; pch++)
        {
            if (PicChar::isSpace(*pch) || *pch == '-')
                convertUp = boolTrue;
            else if (convertUp)
            {
                *pch = PicChar::uppercase(*pch);
                convertUp = boolFalse;
            }
            else
                *pch = PicChar::lowercase(*pch);
        }
    }
    return *this;
}

PICBOOL BaseString::isNumeric () const
{
    if (isEmpty())
        return boolTrue;

    char *pch = rep + STRSIZE;
    while (--pch >= rep)
        if (!PicChar::isDigit(*pch))
            return boolFalse;

    return boolTrue;
}

PICBOOL BaseString::isAlphabetic () const
{
    if (isEmpty())
        return boolTrue;

    char *pch = rep + STRSIZE;
    while (--pch >= rep)
        if (!PicChar::isAlphabetic(*pch))
            return boolFalse;

    return boolTrue;
}

PICBOOL BaseString::isAlphanumeric () const
{
    if (isEmpty())
        return boolTrue;

    char *pch = rep + STRSIZE;
    while (--pch >= rep)
        if (!PicChar::isAlphanumeric(*pch))
            return boolFalse;

    return boolTrue;
}

PICBOOL BaseString::isSpace () const
{
    if (isEmpty())
        return boolTrue;

    char *pch = rep + STRSIZE;
    while (--pch >= rep)
        if (!PicChar::isSpace(*pch))
            return boolFalse;

    return boolTrue;
}

PICBOOL BaseString::containsChars (const BaseString& charSet) const
{
    if (isEmpty())
        return boolTrue;

    char *pch = rep + STRSIZE;
    while (--pch >= rep)
        if (charSet.positionOf(*pch) == outOfRange)
            return boolFalse;

    return boolTrue;
}

// --------------------------------------------------------------------------
//      CONVERSIONS

BaseString asString (long value)
{
    char buffer [34];
    sprintf(buffer, "%ld", value);
    return buffer;
//      return ltoa(value, buffer, 10);
}

BaseString asString (unsigned long value)
{
    char buffer [34];
    sprintf(buffer, "%ld", value);
    return buffer;
//      return ultoa(value, buffer, 10);
}

//  BaseString asString (PICFLOAT value)  in BREAL.CXX

BaseString asString (short value)
{
    char buffer [18];
    sprintf(buffer, "%d", value);
    return buffer;
//    return itoa(value, buffer, 10);
}

BaseString asString (unsigned short value)
{
    char buffer [18];
    sprintf(buffer, "%d", value);
    return buffer;
//    return ultoa(value, buffer, 10);
}

BaseString asString (int value)
{
    char buffer [34];
    sprintf(buffer, "%d", value);
    return buffer;
//    return itoa(value, buffer, 10);
}

BaseString asString (unsigned int value)
{
    char buffer [34];
    sprintf(buffer, "%d", value);
    return buffer;
//    return ultoa(value, buffer, 10);
}

// --------------------------------------------------------------------------
//      FOR COMPATIBILITY

BaseString& operator << (BaseString& toStr, const BaseString& str)
{
    toStr.append(str);
    return toStr;
}

BaseString& operator << (BaseString& toStr, const char* azStr)
{
    toStr.append(azStr);
    return toStr;
}

BaseString& operator << (BaseString& toStr, char character)
{
    toStr.append(character);
    return toStr;
}

BaseString& operator << (BaseString& toStr, long value)
{
    toStr.append(asString(value));
    return toStr;
}

BaseString& operator << (BaseString& toStr, unsigned long value)
{
    toStr.append(asString(value));
    return toStr;
}

//  BaseString& operator << (BaseString& toStr, PICFLOAT value) in BREAL.CXX

BaseString& operator << (BaseString& toStr, short value)
{
    toStr.append(asString(value));
    return toStr;
}

BaseString& operator << (BaseString& toStr, unsigned short value)
{
    toStr.append(asString(value));
    return toStr;
}

BaseString& operator << (BaseString& toStr, int value)
{
    toStr.append(asString(value));
    return toStr;
}

BaseString& operator << (BaseString& toStr, unsigned int value)
{
    toStr.append(asString(value));
    return toStr;
}

