// --------------------------------------------------------------------------
//  BaseTime implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <stdlib.h>
#include <time.h>
}

#include <bdatetim.hxx>
#include <picerror.hxx>
#include <pdefault.hxx>

MAKE_PICCLASS(BaseTime)

ERROR_MESSAGES(BaseTime)
PARAMSTR(nullInstance, "Operation on Null instance.")
PARAMSTR(deformatError, "String '@*' is not of the given format.")
PARAMSTR(formatError, "String '@*' is not a valid time.")
PARAMSTR(timeIllegal, "Illegal time component.")
END_PARAM_STRINGS

static const PICSECONDCOUNT    secondsInDay    = 86400;
static const PICSECONDCOUNT    secondsInHour   = 3600;
static const PICSECONDCOUNT    secondsInMinute = 60;
static const PICMINUTECOUNT    minutesInHour   = 60;
static const PICHOURCOUNT      hoursInDay      = 24;
static const PICHOURCOUNT      hoursInHalfDay  = 12;

static       PicTimeResolution resolution      = defaultTimeResolution;
static       PicTimeStyle      style           = defaultTimeStyle;
static       char              delimiter       = defaultTimeDelimiter;
static       char              leader          = defaultTimeLeader;

PicTimeStyle BaseTime::defaultStyle ()
{
    return style;
}

PicTimeResolution BaseTime::defaultResolution ()
{
    return resolution;
}

char BaseTime::defaultDelimiter ()
{
    return delimiter;
}

char BaseTime::defaultLeader ()
{
    return leader;
}

// --------------------------------------------------------------------------
//      INITIALIZATION

PicTimeIniter::PicTimeIniter (
    PicTimeStyle      style,
    PicTimeResolution resolution,
    char           delimiter,
    char           leader)
    : tStyle(::style), tResol(::resolution), tDelim(::delimiter),
      tLeader(::leader)
{
    ::resolution = resolution;
    ::style      = style;
    ::delimiter  = delimiter;
    ::leader     = leader;
}

PicTimeIniter::~PicTimeIniter()
{
    resolution = tResol;
    style      = tStyle;
    delimiter  = tDelim;
    leader     = tLeader;
}


// --------------------------------------------------------------------------
//      INLINE FUNCTIONS
// --------------------------------------------------------------------------

inline PICBOOL hourIsLegal (PICHOUR anHour, PICHOUR maxH = maxHour)
{
    return (anHour >= minHour && anHour <= maxH);
}

inline PICBOOL minuteIsLegal (PICMINUTE aMinute)
{
    return (aMinute >= minMinute && aMinute <= maxMinute);
}

inline PICBOOL secondIsLegal (PICSECOND second)
{
    return (second >= minSecond && second <= maxSecond);
}

inline PICBOOL isLegal (
    PICSHORT anHour, 
    PICSHORT aMinute, 
    PICSHORT aSecond)
{
    return hourIsLegal((PICHOUR)anHour)    &&
           minuteIsLegal((PICMINUTE)aMinute) &&
           secondIsLegal((PICSECOND)aSecond);
}

// --------------------------------------------------------------------------
//      BASIC TIME OPERATIONS
// --------------------------------------------------------------------------

static BaseTime secondsToTime (PICSECONDCOUNT seconds)
{
    PICLONG anHour, aMinute, aSecond;

    anHour   = seconds / secondsInHour;
    aMinute  = seconds % secondsInHour;
    aSecond  = aMinute % secondsInMinute;
    aMinute /= secondsInMinute;

    BaseTime aTime;
    
    aTime.setTime(PICHOUR(anHour),PICMINUTE(aMinute),PICSECOND(aSecond));
    return aTime;
}

PICSECONDCOUNT BaseTime::asSeconds() const
{
    return theHour * secondsInHour + theMinute * secondsInMinute +
           theSecond;
}

static void getCurrentTime (
    PICHOUR&   anHour,
    PICMINUTE& aMinute,
    PICSECOND& second)
{
    time_t timer;
    tm*    today;

    time(&timer);
    today   = localtime(&timer);

    anHour  = PICHOUR(today->tm_hour);
    aMinute = PICMINUTE(today->tm_min);
    second  = PICSECOND(today->tm_sec);
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
// --------------------------------------------------------------------------

BaseTime::BaseTime() : theHour(nullHour) {}

BaseTime::BaseTime (PICHOURCOUNT hourInterval) : theHour(nullHour)
{
    setRelativeTime(hourInterval);
}

BaseTime::BaseTime (
    PICHOUR   anHour,
    PICMINUTE aMinute,
    PICSECOND aSecond)
    :
    theHour(nullHour)
{
    setTime(anHour,aMinute,aSecond);
}

BaseTime::BaseTime (const BaseString& aTime, char aDelimiter) :
    theHour(nullHour)
{
    setImage(aTime, aDelimiter);
}

BaseTime::BaseTime (const BaseString& aTime) :
    theHour(nullHour)
{
    setImage(aTime, defaultDelimiter());
}

BaseTime::BaseTime (const BaseTime &right)
{
    theHour   = right.theHour;
    theMinute = right.theMinute;
    theSecond = right.theSecond;
}

BaseTime& BaseTime::operator = (const BaseTime &right)
{
    theHour   = right.theHour;
    theMinute = right.theMinute;
    theSecond = right.theSecond;

    return *this;
}

// --------------------------------------------------------------------------
//      SET TIME
// --------------------------------------------------------------------------

void BaseTime::setTime (PICHOUR anHour, PICMINUTE aMinute, PICSECOND aSecond)
{
    if (isLegal(anHour, aMinute, aSecond))
    {
        theHour = anHour;
        theMinute = aMinute;
        theSecond = aSecond;
    }
    else
    {
        ERAISE(BaseTime, timeIllegal);
    }
}

void BaseTime::setRelativeTime (
    PICHOURCOUNT   hoursToAdd,
    PICMINUTECOUNT minutesToAdd,
    PICSECONDCOUNT secondsToAdd)
{
    getCurrentTime(theHour, theMinute, theSecond);

    if (hoursToAdd)
        addHours(hoursToAdd);

    if (minutesToAdd)
        addMinutes(minutesToAdd);

    if (secondsToAdd)
        addSeconds(secondsToAdd);
}

// --------------------------------------------------------------------------
//      INTERVAL
// --------------------------------------------------------------------------

BaseTime BaseTime::operator - (const BaseTime &right) const
{
    if (isNull() || right. isNull())
    {
        ERAISE(BaseTime, nullInstance);
        return BaseTime();
    }

    PICSECONDCOUNT secondsDiff = asSeconds() - right.asSeconds();

    if (secondsDiff < 0)
        secondsDiff += secondsInDay;

    return secondsToTime(secondsDiff);
}

// --------------------------------------------------------------------------

void BaseTime::addHours (PICHOURCOUNT hoursToAdd)
{
    if (isNull())
        ERAISE(BaseTime, nullInstance);
    else
    {
        PICHOURCOUNT anHour;

        anHour = (PICHOURCOUNT(theHour) + (hoursToAdd % hoursInDay))
                  % hoursInDay;
        
        if (anHour < minHour)
            anHour += hoursInDay;
    
        theHour = PICHOUR(anHour);
    }
}

void BaseTime::addMinutes (PICMINUTECOUNT minutesAdd)
{
    if (isNull())
    {
        ERAISE(BaseTime, nullInstance);
        return;
    }

    PICHOURCOUNT   hoursToAdd   = minutesAdd / minutesInHour;
    PICHOURCOUNT   anHour       = theHour;
    PICMINUTECOUNT minutesToAdd = minutesAdd % minutesInHour;

    if (theMinute + minutesToAdd < 0)
    {
        hoursToAdd--;
        minutesToAdd += minutesInHour;
    }
    else if (theMinute + minutesToAdd > maxMinute)
    {
        hoursToAdd++;
        minutesToAdd -= minutesInHour;
    }

    hoursToAdd %= hoursInDay;

    anHour += hoursToAdd;
    anHour %= hoursInDay;
    if (anHour < minHour)
        anHour += hoursInDay;
    
    theHour = PICHOUR(anHour);
    theMinute += PICMINUTE(minutesToAdd);
}

void BaseTime::addSeconds (PICSECONDCOUNT secondsAdd)
{
    if (isNull())
        ERAISE(BaseTime, nullInstance);
    else
    {
        PICSECONDCOUNT secondsToAdd = secondsAdd % secondsInDay;
        PICSECONDCOUNT theTime      = asSeconds();

        theTime += secondsToAdd;
        theTime %= secondsInDay;

        if (theTime < 0)
            theTime += secondsInDay;

        *this = secondsToTime(theTime);
    }
}

// --------------------------------------------------------------------------

void BaseDateTime::addHours (PICHOURCOUNT h)
{
    if (date.isNull() || time.isNull())
        ERAISE(BaseTime, nullInstance);
    else
    {
        PICHOURCOUNT anHour;
        PICDAYCOUNT  addDays;

        anHour  = PICHOURCOUNT(time.theHour) + h;
        addDays = anHour / hoursInDay;
        anHour %= hoursInDay;

        if (anHour < minHour)
        {
            anHour += hoursInDay;
            addDays--;
        }
    
        time.theHour = PICHOUR(anHour);
        date.addDays(addDays);
    }
}

void BaseDateTime::addMinutes (PICMINUTECOUNT m)
{
    if (date.isNull() || time.isNull())
    {
        ERAISE(BaseTime, nullInstance);
        return;
    }

    PICHOURCOUNT   hoursToAdd   = m / minutesInHour;
    PICMINUTECOUNT minutesToAdd = m % minutesInHour;

    if (time.theMinute + minutesToAdd < 0)
    {
        hoursToAdd--;
        minutesToAdd += minutesInHour;
    }
    else if (time.theMinute + minutesToAdd > maxMinute)
    {
        hoursToAdd++;
        minutesToAdd -= minutesInHour;
    }

    time.theMinute += PICMINUTE(minutesToAdd);
    if (hoursToAdd)
        addHours(hoursToAdd);
}

void BaseDateTime::addSeconds (PICSECONDCOUNT s)
{
    if (date.isNull() || time.isNull())
        ERAISE(BaseTime, nullInstance);
    else
    {
        PICDAYCOUNT    daysToAdd    = s / secondsInDay;
        PICSECONDCOUNT secondsToAdd = s % secondsInDay;
        PICSECONDCOUNT theTime      = time.asSeconds();

        theTime += secondsToAdd;

        if (theTime > secondsInDay)
        {
            daysToAdd++;
            theTime -= secondsInDay;
        }
        else if (theTime < 0)
        {
            daysToAdd--;
            theTime += secondsInDay;
        }

        time = secondsToTime(theTime);
        if (daysToAdd)
            date.addDays(daysToAdd);
    }
}

// --------------------------------------------------------------------------

PICHOURCOUNT BaseTime::hoursBetween (const BaseTime &right) const
{
    if (isNull() || right. isNull())
    {
        ERAISE(BaseTime, nullInstance);
        return 0;
    }
    return (asSeconds() - right.asSeconds()) / secondsInHour;
}

PICMINUTECOUNT BaseTime::minutesBetween (const BaseTime &right) const
{
    if (isNull() || right. isNull())
    {
        ERAISE(BaseTime, nullInstance);
        return 0;
    }
    return (asSeconds() - right.asSeconds()) / secondsInMinute;
}

PICSECONDCOUNT BaseTime::secondsBetween (const BaseTime &right) const
{
    if (isNull() || right. isNull())
    {
        ERAISE(BaseTime, nullInstance);
        return 0;
    }
    return asSeconds() - right.asSeconds();
}

// --------------------------------------------------------------------------
//      COMPARISONS
// --------------------------------------------------------------------------

PICBOOL BaseTime::operator == (const BaseTime& right) const
{
    if (isNull() || right.isNull())
        return isNull() && right.isNull();

    return theHour   == right.theHour   &&
           theMinute == right.theMinute &&
           theSecond == right.theSecond;
}

PICBOOL BaseTime::operator > (const BaseTime& right) const
{
    if (isNull())
        return boolFalse;
    else if (right.isNull())
        return boolTrue;
    else if (theHour > right.theHour)
        return boolTrue;
    else if (theHour < right.theHour)
        return boolFalse;
    else if (theMinute > right.theMinute)
        return boolTrue;
    else
        return theMinute == right.theMinute && theSecond > right.theSecond;
}

PICBOOL BaseDateTime::operator < (const BaseDateTime& bdt) const
{
    if (date < bdt.date)
        return boolTrue;
    else if (date > bdt.date)
        return boolFalse;
    else
        return time < bdt.time;
}

// --------------------------------------------------------------------------
//      FORMATTING AND DEFORMATTING
// --------------------------------------------------------------------------

static const char* imageFmts[2][3] =
{
//    hoursOnly   hoursToMinutes   hoursToSeconds
//    ---------   --------------   --------------
    { "@[+a AP]", "@[+a~nn AP]",   "@[+a~nn~ss AP]" }, // hours12
    { "@[+h]",    "@[+h~nn]",      "@[+h~nn~ss]"    }  // hours24
};

BaseString BaseTime::getImage () const
{
    DEBUG_REQUIRE(::style > 0 && ::style < 3);
    DEBUG_REQUIRE(::resolution > 0 && ::resolution < 4);

    PicStringText image;
    image(imageFmts[::style - 1][::resolution - 1]) << *this;
    return image.flush();
}

BaseString BaseTime::getImage (
    PicTimeResolution aResolution,
    PicTimeStyle      aStyle,
    char           aDelimiter,
    char           aLeader) const
{
    PicTimeResolution oldResolution = ::resolution;
    PicTimeStyle      oldStyle      = ::style;
    char           oldDelimiter  = ::delimiter;
    char           oldLeader     = ::leader;

    ::resolution = aResolution;
    ::style      = aStyle;
    ::delimiter  = aDelimiter;
    ::leader     = aLeader;

    BaseString image = getImage();

    ::resolution = oldResolution;
    ::style      = oldStyle;
    ::delimiter  = oldDelimiter;
    ::leader     = oldLeader;

    return image;
}

void BaseTime::setImage (const BaseString& timeImage)
{
    BaseString image = timeImage;
    image.trim();

    if (image.isEmpty())
        setNull();
    else if (image == "*")
        setRelativeTime(0);
    else
    {
        BaseDateTime dt;
        PICOFFSET i = dt.deformat(image.chars(), 0, BaseDate::yearFirst, 2);

        if (i == 0 || i < image.size())
        {
            ERAISEFMT(BaseTime, deformatError) << image;
        }
        else if (i == 0xffff)
        {
            ERAISEFMT(BaseTime, formatError) << image;
        }
        else
            *this = dt.time;
    }
}

void BaseTime::setImage (const BaseString& aTime, char aDelimiter)
{
    char oldDelimiter  = ::delimiter;

    ::delimiter = aDelimiter;

    setImage(aTime);

    ::delimiter = oldDelimiter;
}

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT
// --------------------------------------------------------------------------

static PICOFFSET wrTime (const BaseTime* obj, PicTextFmt& fmt, PICOFFSET i)
{
    static const char typeName[] = "BaseTime";

    BaseDateTime datetime = *obj;

    if (obj->isNull())
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return datetime.writeFmt(nullFmt, i, typeName);
    }
    return datetime.writeFmt(fmt, i, typeName);
}

static BaseTime* wrobjTime (BaseTime* obj, PICBOOL cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    return new BaseTime(*obj);
}

PicFmtText& operator << (PicFmtText& txt, const BaseTime& obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrTime, (PICWROBJFUNC*)wrobjTime);
}

// --------------------------------------------------------------------------
//      BaseDateTime
// --------------------------------------------------------------------------

BaseString BaseDateTime::getImage () const
{
    BaseString image;

    if (!date.isNull())
        image.append(date.getImage());
    if (!time.isNull())
        image.append(' ').append(time.getImage());

    return image;
}

void BaseDateTime::setImage (const BaseString& dtImage)
{
    BaseString image = dtImage;
    image.trim();

    if (image.isEmpty())
        setNull();
    else if (image == "*")
    {
        date.setRelativeDate(BaseDate::today, 0);
        time.setRelativeTime(0);
    }
    else
    {
        BaseDateTime dt;
        PICOFFSET i = dt.deformat (
            image.chars(), 0, BaseDate::defaultStyle(), 3);

        if (i == 0 || i < image.size())
        {
            ERAISEFMT(BaseTime, deformatError) << image;
        }
        else if (i == 0xffff)
        {
            ERAISEFMT(BaseTime, formatError) << image;
        }
        else
            *this = dt;
    }
}

// --------------------------------------------------------------------------

static PICOFFSET wrDateTime (
    const BaseDateTime* obj,
    PicTextFmt&         fmt,
    PICOFFSET           i)
{
    static const char typeName[] = "BaseDateTime";

    if (obj->isNull())
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return obj->writeFmt(nullFmt, i, typeName);
    }
    return obj->writeFmt(fmt, i, typeName);
}

static BaseDateTime* wrobjDateTime (BaseDateTime* obj, PICBOOL cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    return new BaseDateTime(*obj);
}

PicFmtText& operator << (PicFmtText& txt, const BaseDateTime& obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrDateTime, (PICWROBJFUNC*)wrobjDateTime);
}

