// --------------------------------------------------------------------------
//  INDEX-OBJECT ASSOCIATION ARRAY implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
}
// Edit this include:
// ------------------
#include <passarr.hxx>
#include <bstring.hxx>

// --------------------------------------------------------------------------
//      STATIC PARAMETER FUNCTIONS
//
//  Modify these functions to suit the component class. Default
//  implementations are suitable for most Pic classes. You may replace
//  ITEM_CLASS with the real component class name for the sake of
//  readability. For an example, see BSTRARR.CXX.
//
// --------------------------------------------------------------------------
//  Create a new component at the address pointed to by 'to'.
//  The component should be created as null-valued.

static void createItem (PicAssoc* to)
{
//    new(to) PicAssoc();

    PicAssoc tmp;
    *to = tmp;
}

// --------------------------------------------------------------------------
//  Create a new component at the address pointed to by 'to'.
//  Copy the value of the argument 'from'.

static void createItemCopy (PicAssoc* to, const PicAssoc* from)
{
//    new(to) PicAssoc(*from);

    *to = *from;
}

// --------------------------------------------------------------------------
//  Return boolTrue if the given component is in null state.
//  Such components are regarded as non-existing, so locations
//  occupied by them are empty.

static PICBOOL isItemNull (const PicAssoc* item)
{
    return item->isNull();
}

// --------------------------------------------------------------------------
//  Set the given component in null state. After that it must not
//  allocate dynamic memory.

static void setItemNull (PicAssoc* item)
{
    item->setNull();
}

// --------------------------------------------------------------------------
//  Compare two components for ascending order. Return
//   0  if they are equal
//   1  if 'i1' sorts after 'i2'
//  -1  if 'i1' sorts before 'i2'
//  Make sure that null components are sorted after all others.

static int compareAsc (PicAssoc* i1, PicAssoc* i2)
{
    if (i1->isNull() || i2->isNull())
        return (i1->isNull() - i2->isNull()); // isNull is PICBOOL
    if (i1->index() == i2->index())
        return 0;
    return i1->index() < i2->index() ? -1 : 1;
}

// --------------------------------------------------------------------------
//  Compare two components for descending order. Return
//   0  if they are equal
//  -1  if 'i1' sorts after 'i2'
//   1  if 'i1' sorts before 'i2'
//  Make sure that null components are sorted after all others.

static int compareDesc (PicAssoc* i1, PicAssoc* i2)
{
    if (i1->isNull() || i2->isNull())
        return (i1->isNull() - i2->isNull()); // isNull is PICBOOL
    if (i1->index() == i2->index())
        return 0;
    return i1->index() < i2->index() ? 1 : -1;
}

// --------------------------------------------------------------------------
//      TYPE INFORMATION
// --------------------------------------------------------------------------

static PicGenArrayInfo typeInfo =
{
    sizeof(PicAssoc),

    (PROC_SETITEM*)      createItem,
    (PROC_SETITEM*)      NULLPTR, // or NULLPTR to init as zeros
    (PROC_COPYITEM*)     createItemCopy,
    (PROC_SETITEM*)      setItemNull,
    (PROC_TESTITEM*)     isItemNull,
    (PROC_COMPAREITEMS*) compareAsc,
    (PROC_COMPAREITEMS*) compareDesc
};

#include <genarray.cgn>

PICPOSITION PicAssocArray::locationFor (PICINDEX index)
{
    PICPOSITION firstFree = 0;

    for (PICPOSITION ind = ub; ind >= 1; ind--)
    {
        if (THIS[ind].object())
        {
            if (THIS[ind].index() == index)
                return ind;
        }
        else
            firstFree = ind;
    }

    if (firstFree || (firstFree = insert()))
    {
        THIS[firstFree].setIndex(index);
        return firstFree;
    }

    return 0;
}

PICPOSITION PicAssocArray::locationOf (PICINDEX index) const
{
    for (PICPOSITION ind = ub; ind >= 1; ind--)
    {
        if (THIS[ind].object() && THIS[ind].index() == index)
            return ind;
    }
    return 0;
}

PICPOSITION PicAssocArray::locationOf (const PicObject& obj) const
{
    for (PICPOSITION ind = 1; ind <= ub; ind++)
    {
        if (THIS[ind].object() && *THIS[ind].object() == obj)
            return ind;
    }
    return 0;
}

PICBOOL PicAssocArray::equals (const PicAssocArray& anArray) const
{
    PICPOSITION thisInd  = 1;
    PICPOSITION arrayInd = 1;

    do {

        while (thisInd <= ub && THIS[thisInd].isNull())
            thisInd++;

        while (arrayInd <= anArray.ub && anArray[arrayInd].isNull())
            arrayInd++;

        if (thisInd > ub)
            return arrayInd > anArray.ub;
        if (arrayInd > anArray.ub)
            return thisInd > ub;

    } while (THIS[thisInd++] == anArray[arrayInd++]);

    return boolFalse;
}
