// --------------------------------------------------------------------------
//  COLLECTION OBJECT base class implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
#include <stdlib.h>
}

#include <pcollect.hxx>
#include <picerror.hxx>
#include <bcollect.hxx>
#include <piterimp.hxx>
#include <picbase.erh>

MAKE_PICABSOBJECTCLASS(PicCollection, PicObject)
MAKE_PICABSOBJECTCLASS(PicIxCollect, PicCollection)
MAKE_PICABSOBJECTCLASS(PicKeyIxCollect, PicIxCollect)
MAKE_PICABSOBJECTCLASS(PicSeqIxCollect, PicIxCollect)
MAKE_PICABSOBJECTCLASS(PicNmCollect, PicCollection)

// --------------------------------------------------------------------------
//      PicObject
// --------------------------------------------------------------------------

#define CALL_GUARD(className, func) \
    if (TYPE_GUARD(className::classId(), *this)) \
        ((className*)this)->func

#define RET_CALL_GUARD(className, func, defval) \
    if (TYPE_GUARD(className::classId(), *this)) \
        return ((className*)this)->func; \
    return defval

PICBOOL PicObject::isCollection () const
{
    return id() >= PicCollection::classId();
}

PicObject& PicObject::at (PICINDEX index)
{
    RET_CALL_GUARD(PicIxCollect, at(index), PicObject::null());
}

PicObject& PicObject::at (const BaseString& name)
{
    RET_CALL_GUARD(PicNmCollect, at(name), PicObject::null());
}

PicObject& PicObject::first ()
{
    RET_CALL_GUARD(PicCollection, first(), PicObject::null());
}

PicObject& PicObject::last ()
{
    RET_CALL_GUARD(PicCollection, last(), PicObject::null());
}

PICBOOL PicObject::putAt (PICINDEX index, const PicObject& obj)
{
    RET_CALL_GUARD(PicIxCollect, putAt(index, obj), boolFalse);
}

PICBOOL PicObject::putAt (const BaseString& name, const PicObject& obj)
{
    RET_CALL_GUARD(PicNmCollect, putAt(name, obj), boolFalse);
}

PICBOOL PicObject::add (const PicObject& obj)
{
    RET_CALL_GUARD(PicSeqIxCollect, add(obj), boolFalse);
}

PICBOOL PicObject::find (PICINDEX index, PicIterator& i) const
{
    RET_CALL_GUARD(PicIxCollect, find(index, i), boolFalse);
}

PICBOOL PicObject::find (const BaseString& name, PicIterator& i) const
{
    RET_CALL_GUARD(PicNmCollect, find(name, i), boolFalse);
}

PICBOOL PicObject::find (const PicObject& obj, PicIterator& i) const
{
    RET_CALL_GUARD(PicCollection, find(obj, i), boolFalse);
}

BaseString PicObject::nameOf (const PicObject& obj) const
{
    RET_CALL_GUARD(PicNmCollect, nameOf(obj), BaseString());
}

PICINDEX PicObject::indexOf (const PicObject& obj) const
{
    RET_CALL_GUARD(PicIxCollect, indexOf(obj), outOfIndex);
}

void PicObject::deleteAt (PICINDEX index, PicPurgeMode purge)
{
    CALL_GUARD(PicIxCollect, deleteAt(index, purge));
}

void PicObject::deleteAt (const BaseString& name, PicPurgeMode purge)
{
    CALL_GUARD(PicNmCollect, deleteAt(name, purge));
}

void PicObject::deleteObject (const PicObject& obj, PicPurgeMode purge)
{
    CALL_GUARD(PicCollection, deleteObject(obj, purge));
}

void PicObject::deleteAll ()
{
    CALL_GUARD(PicCollection, deleteAll());
}

// PicObject::size is located in earray.cxx

BaseString PicObject::firstName () const
{
    RET_CALL_GUARD(PicNmCollect, firstName(), BaseString());
}

BaseString PicObject::lastName () const
{
    RET_CALL_GUARD(PicNmCollect, lastName(), BaseString());
}

PICINDEX PicObject::firstIndex () const
{
    RET_CALL_GUARD(PicIxCollect, firstIndex(), outOfIndex);
}

PICINDEX PicObject::lastIndex () const
{
    RET_CALL_GUARD(PicIxCollect, lastIndex(), outOfIndex);
}

void PicObject::startIteration (PicIterator& i) const
{
    CALL_GUARD(PicCollection, startIteration(i));
}

PICBOOL PicObject::next (PicIterator& i) const
{
    RET_CALL_GUARD(PicCollection, next(i), boolFalse);
}

PICBOOL PicObject::previous (PicIterator& i) const
{
    RET_CALL_GUARD(PicCollection, previous(i), boolFalse);
}

PicObject& PicObject::current (const PicIterator& i)
{
    RET_CALL_GUARD(PicCollection, access(i), PicObject::null());
}

BaseString PicObject::currentName (const PicIterator& i) const
{
    RET_CALL_GUARD(PicNmCollect, nameAt(i), BaseString());
}

PICINDEX PicObject::currentIndex (const PicIterator& i) const
{
    RET_CALL_GUARD(PicIxCollect, indexAt(i), outOfIndex);
}

// --------------------------------------------------------------------------

PICBOOL PicObject::contains (PICINDEX index) const
{
    PicIterator here;
    RET_CALL_GUARD(PicIxCollect, find(index, here), boolFalse);
}

PICBOOL PicObject::contains (const BaseString& name) const
{
    PicIterator here;
    RET_CALL_GUARD(PicNmCollect, find(name, here), boolFalse);
}

PICBOOL PicObject::contains (const PicObject& obj) const
{
    PicIterator here;
    RET_CALL_GUARD(PicCollection, find(obj, here), boolFalse);
}

// --------------------------------------------------------------------------
//      PicCollection
// --------------------------------------------------------------------------

class PicNullCollection : public PicCollection
{
    public:

    virtual PicObject* copy () const 
        { return NULLPTR; }
    virtual PicObject& operator = (const PicObject&)
        { return (*this); }
    virtual PICBOOL operator == (const PicObject& obj) const 
        { return (&obj == this); }
    virtual PICBOOL isNull  () const { return boolTrue; }
    virtual void    setNull () {}

    virtual PicObject& access (const PicIterator&) 
        { return PicObject::null(); }
    virtual PicObject* replace (PicIterator&, PicObject*) 
        { return NULLPTR; }
    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue)
        { return boolFalse; }
    virtual PICBOOL allocate (PICSIZE forComponents)
        { return boolFalse; }
};

PicCollection& PicCollection::null()
{
    static PicNullCollection nullCollect;
    return nullCollect;
}

PICBOOL PicCollection::isDefault () const
{
    PicIterator i;
    ITERATE(*this, i)
        if (!at(i).isDefault())
            return boolFalse;
    return boolTrue;
}

void PicCollection::setDefault ()
{
    PicIterator i;
    ITERATE(*this, i)
        at(i).setDefault();
}

PicObject& PicCollection::first ()
{
    PicIterator here;
    return findFirst(here) ? access(here) : PicObject::null();
}

PicObject& PicCollection::last ()
{
    PicIterator here;
    return findLast(here) ? access(here) : PicObject::null();
}

PICBOOL PicCollection::putAt (PicIterator& here, const PicObject& obj)
{
    PicObject* pobj = replace(here, obj.copy());
    delete pobj;
    return boolTrue;
}

// --------------------------------------------------------------------------

PICBOOL PicCollection::findObject (
    const PicObject& obj,
    PicIterator&     here) const
{
    ITERATE(*this, here)
        if (at(here) == obj)
            return boolTrue;
    return boolFalse;
}

PICBOOL PicCollection::findMinMax (PicIterator& here, PICBOOL smallest) const
{
    startIteration(here);
    if (next(here))
    {
        PicIterator      i   = here;
        const PicObject* obj = &at(here);

        while (next(i))
            if (smallest ? (at(i) < *obj) : (*obj < at(i)))
            {
                obj  = &at(i);
                here = i;
            }

        return boolTrue;
    }
    return boolFalse;
}

PICBOOL PicCollection::findPosition (PICPOSITION n, PicIterator& here) const
{
    ITERATE(*this, here)
        if (--n == 0)
            return boolTrue;
    return boolFalse;
}

// --------------------------------------------------------------------------

PICBOOL PicCollection::deleteAt (PicIterator& here, PicPurgeMode purgeMd)
{
    PicObject* pobj = replace(here, NULLPTR);
    delete pobj;
    return (purgeMd == noPurge) ?
        previous(here) :
        purge(here, (purgeMd == withPurgeOnOrder));
}

void PicCollection::deleteObject (const PicObject& obj, PicPurgeMode purgeMd)
{
    PicIterator here;

    if (find(obj, here))
    {
        PicObject* pobj = replace(here, NULLPTR);
        if (pobj && !pobj->is(obj))
            delete pobj;
        if (purgeMd != noPurge)
            purge(here, (purgeMd == withPurgeOnOrder));
    }
}

void PicCollection::deleteAll ()
{
    PicIterator i;
    PicObject*  pobj;

    ITERATE(*this, i)
    {
        pobj = replace(i, NULLPTR);
        delete pobj;
    }
}

void PicCollection::removeAll ()
{
    PicIterator i;

    ITERATE(*this, i)
        replace(i, NULLPTR);
}

// --------------------------------------------------------------------------

void PicCollection::startIteration (PicIterator& i) const
{
    SET_ITERATE_INDEX(i, 0);
}

PICBOOL PicCollection::next (PicIterator&) const
{
    return boolFalse;
}

PICBOOL PicCollection::previous (PicIterator&) const
{
    return boolFalse;
}

// --------------------------------------------------------------------------

PICSIZE PicCollection::size () const
{
    PicIterator i;
    PICSIZE     size = 0;

    ITERATE(*this, i)
        size++;

    return size;
}

// --------------------------------------------------------------------------

PicCollection& PicCollection::operator *= (const PicCollection& coll)
{
    PicIterator i, here;

    ITERATE_BACKWARDS(*this, i)
        if (!coll.find(at(i), here))
        {
            PicObject* pobj = replace(i, NULLPTR);
            delete pobj;
        }

    return *this;
}

PicCollection& PicCollection::operator -= (const PicCollection& coll)
{
    PicIterator i, here;

    ITERATE(coll, i)
        if (find(coll.at(i), here))
        {
            PicObject* pobj = replace(here, NULLPTR);
            delete pobj;
        }

    return *this;
}

// --------------------------------------------------------------------------

PICBOOL PicCollection::operator < (const PicObject&) const
 { undefined("operator <"); return boolFalse; }

BaseString PicCollection::getImage () const
{
    undefined("getImage");
    return BaseString();
}

BaseString PicCollection::getImage (const PicObject&) const
 { undefined("getImage(format)"); return BaseString(); }

void PicCollection::setImage (const BaseString&)
 { undefined("setImage"); }

void PicCollection::setImage (const PicObject&, const BaseString&)
 { undefined("setImage(format)"); }

// --------------------------------------------------------------------------
//      PicIxCollect
// --------------------------------------------------------------------------

class PicNullIxCollect : public PicIxCollect
{
    public:

    virtual PicObject* copy () const 
        { return NULLPTR; }
    virtual PicObject& operator = (const PicObject&)
        { return (*this); }
    virtual PICBOOL operator == (const PicObject& obj) const 
        { return (&obj == this); }
    virtual PICBOOL isNull  () const { return boolTrue; }
    virtual void    setNull () {}

    virtual PicObject& access (const PicIterator&) 
        { return PicObject::null(); }
    virtual PicObject* replace (PicIterator&, PicObject*) 
        { return NULLPTR; }
    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue)
        { return boolFalse; }
    virtual PICBOOL allocate (PICSIZE forComponents)
        { return boolFalse; }

    virtual PICINDEX indexAt (const PicIterator&) const
        { return outOfIndex; }
};

PicIxCollect& PicIxCollect::null()
{
    static PicNullIxCollect nullIxCollect;
    return nullIxCollect;
}

PicObject& PicKeyIxCollect::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(PicIxCollect::classId(), obj))
    {
        setNull();
        *this += (const PicIxCollect&)obj;
    }
    return *this;
}

// --------------------------------------------------------------------------

PicObject& PicIxCollect::at (PICINDEX index)
{
    PicIterator here;
    if (find(index, here))
        return access(here);

    ERAISEFMT(BaseCollection, illegalIndex) << index;
    return PicObject::null();
}

// --------------------------------------------------------------------------

PICBOOL PicIxCollect::putAtIndex (
    PICINDEX         index,
    const PicObject& obj,
    PicIterator&     here)
{
    if (find(index, here))
    {
        PicObject* pobj = replace(here, obj.copy());
        delete pobj;
        return boolTrue;
    }
    return boolFalse;
}

PICBOOL PicKeyIxCollect::putAtIndex (
    PICINDEX         index,
    const PicObject& obj,
    PicIterator&     here)
{
    if (insert(index, here))
    {
        PicObject* pobj = replace(here, obj.copy());
        delete pobj;
        return boolTrue;
    }
    return boolFalse;
}

PicObject* PicIxCollect::remove (PICINDEX index, PicPurgeMode purgeMd)
{
    PicIterator here;
    PicObject* pobj = NULLPTR;

    if (find(index, here))
    {
        pobj = replace(here, NULLPTR);
        if (purgeMd != noPurge)
            purge(here, (purgeMd == withPurgeOnOrder));
    }

    return pobj;
}

// --------------------------------------------------------------------------

PICINDEX PicIxCollect::indexOf (const PicObject& obj) const
{
    PicIterator here;
    return find(obj, here) ? indexAt(here) : outOfIndex;
}

void PicIxCollect::deleteAt (PICINDEX index, PicPurgeMode purge)
{
    PicIterator here;
    if (find(index, here))
        deleteAt(here, purge);
}

PICINDEX PicIxCollect::firstIndex () const
{
    PicIterator here;
    return findFirst(here) ? indexAt(here) : outOfIndex;
}

PICINDEX PicIxCollect::lastIndex () const
{
    PicIterator here;
    return findLast(here) ? indexAt(here) : outOfIndex;
}

PICBOOL PicIxCollect::findIndex (PICINDEX index, PicIterator& here) const
{
    ITERATE(*this, here)
        if (indexAt(here) == index)
            return boolTrue;
    return boolFalse;
}

// --------------------------------------------------------------------------

PICSHORT PicIxCollect::ascending (PICINDEX ix1, PICINDEX ix2)
{
    if (ix1 > ix2)
        return 1;

    if (ix1 == ix2)
        return 0;

    return -1;
}

PICSHORT PicIxCollect::descending (PICINDEX ix1, PICINDEX ix2)
{
    if (ix1 > ix2)
        return -1;

    if (ix1 == ix2)
        return 0;

    return 1;
}

// --------------------------------------------------------------------------

PicKeyIxCollect& PicKeyIxCollect::operator += (const PicIxCollect& ixColl)
{
    PicIterator i;

    ITERATE(ixColl, i)
        if (!putAt(ixColl.indexAt(i), ixColl.at(i)))
            break;

    return *this;
}

PicKeyIxCollect& PicKeyIxCollect::operator |= (const PicIxCollect& ixColl)
{
    PicIterator i;

    ITERATE(ixColl, i)
        if (!contains(ixColl.indexAt(i)))
            if (!putAt(ixColl.indexAt(i), ixColl.at(i)))
                break;

    return *this;
}

PicIxCollect& PicIxCollect::operator &= (const PicIxCollect& ixColl)
{
    PicIterator i, here;

    ITERATE(*this, i)
        if (ixColl.find(indexAt(i), here))
            if (!putAt(i, ixColl.at(here)))
                break;

    return *this;
}

// --------------------------------------------------------------------------
//      PicSeqIxCollect
// --------------------------------------------------------------------------

PicObject& PicSeqIxCollect::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(PicCollection::classId(), obj))
    {
        setNull();
        *this += (const PicCollection&)obj;
    }
    return *this;
}

// --------------------------------------------------------------------------

PICBOOL PicSeqIxCollect::add (const PicObject& obj, PicIterator& here)
{
    if (insert(outOfIndex, here))
    {
        replace(here, obj.copy());
        return boolTrue;
    }
    return boolFalse;
}

PICBOOL PicSeqIxCollect::addAt (
    PICINDEX         index,
    const PicObject& obj,
    PicIterator&     here)
{
    if (insert(index, here))
    {
        replace(here, obj.copy());
        return boolTrue;
    }
    return boolFalse;
}

PICBOOL PicSeqIxCollect::putAtIndex (
    PICINDEX         index,
    const PicObject& obj,
    PicIterator&     here)
{
    if (put(index, here))
    {
        PicObject* pobj = replace(here, obj.copy());
        delete pobj;
        return boolTrue;
    }
    return boolFalse;
}

// --------------------------------------------------------------------------

PicSeqIxCollect& PicSeqIxCollect::operator += (const PicCollection& coll)
{
    PicIterator i, here;

    ITERATE(coll, i)
        if (!add(coll.at(i), here))
            break;

    return *this;
}

// --------------------------------------------------------------------------

PICSHORT PicSeqIxCollect::ascending (const PicObject& o1, const PicObject& o2)
{
    if (o2 < o1)
        return 1;

    if (o1 == o2)
        return 0;

    return -1;
}

PICSHORT PicSeqIxCollect::descending (const PicObject& o1, const PicObject& o2)
{
    if (o2 < o1)
        return -1;

    if (o1 == o2)
        return 0;

    return 1;
}

// --------------------------------------------------------------------------
//      PicNmCollect
// --------------------------------------------------------------------------

class PicNullNmCollect : public PicNmCollect
{
    public:

    virtual PicObject* copy () const 
        { return NULLPTR; }
    virtual PicObject& operator = (const PicObject&)
        { return (*this); }
    virtual PICBOOL operator == (const PicObject& obj) const 
        { return (&obj == this); }
    virtual PICBOOL isNull  () const { return boolTrue; }
    virtual void    setNull () {}

    virtual PicObject& access (const PicIterator&) 
        { return PicObject::null(); }
    virtual PicObject* replace (PicIterator&, PicObject*) 
        { return NULLPTR; }
    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue)
        { return boolFalse; }
    virtual PICBOOL allocate (PICSIZE forComponents)
        { return boolFalse; }

    virtual const char* nameAt (const PicIterator&) const
        { return NULLPTR; }
    virtual PICBOOL insert (const char* name, PicIterator&)
        { return boolFalse; }
};

PicNmCollect& PicNmCollect::null()
{
    static PicNullNmCollect nullNmCollect;
    return nullNmCollect;
}

PicObject& PicNmCollect::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
    {
        setNull();
        *this += (PicNmCollect&)obj;
    }
    return *this;
}

// --------------------------------------------------------------------------

PicObject& PicNmCollect::at (const char* name)
{
    PicIterator here;
    if (find(name, here))
        return access(here);

    ERAISEFMT(BaseCollection, illegalIndex) << name;
    return PicObject::null();
}

PICBOOL PicNmCollect::putAt (
    const char*   name,
    const PicObject& obj,
    PicIterator&     here)
{
    if (insert(name, here))
    {
        PicObject* pobj = replace(here, obj.copy());
        delete pobj;
        return boolTrue;
    }
    return boolFalse;
}

// --------------------------------------------------------------------------

BaseString PicNmCollect::nameOf (const PicObject& obj) const
{
    PicIterator here;
    if (find(obj, here))
        return BaseString(nameAt(here));
    return BaseString();
}

void PicNmCollect::deleteAt (const char* name, PicPurgeMode purge)
{
    PicIterator here;
    if (find(name, here))
        deleteAt(here, purge);
}

PicObject* PicNmCollect::remove (const char* name, PicPurgeMode purgeMd)
{
    PicIterator here;
    PicObject* pobj = NULLPTR;

    if (find(name, here))
    {
        pobj = replace(here, NULLPTR);
        if (purgeMd != noPurge)
            purge(here, (purgeMd == withPurgeOnOrder));
    }

    return pobj;
}

BaseString PicNmCollect::firstName () const
{
    PicIterator here;
    if (findFirst(here))
        return BaseString(nameAt(here));
    return BaseString();
}

BaseString PicNmCollect::lastName () const
{
    PicIterator here;
    if (findLast(here))
        return BaseString(nameAt(here));
    return BaseString();
}

PICBOOL PicNmCollect::findName (
    const char* name,
    PicIterator&   here) const
{
    ITERATE(*this, here)
        if (strcmp(nameAt(here), name) == 0)
            return boolTrue;
    return boolFalse;
}

// --------------------------------------------------------------------------

PicNmCollect& PicNmCollect::operator += (const PicNmCollect& nmColl)
{
    PicIterator i;

    ITERATE(nmColl, i)
        if (!putAt(nmColl.nameAt(i), nmColl.at(i)))
            break;

    return *this;
}

PicNmCollect& PicNmCollect::operator &= (const PicNmCollect& nmColl)
{
    PicIterator i, here;

    ITERATE(*this, i)
        if (nmColl.find(nameAt(i), here))
            if (!putAt(i, nmColl.at(here)))
                break;

    return *this;
}

PicNmCollect& PicNmCollect::operator |= (const PicNmCollect& nmColl)
{
    PicIterator i;

    ITERATE(nmColl, i)
        if (!contains(nmColl.nameAt(i)))
            if (!putAt(nmColl.nameAt(i), nmColl.at(i)))
                break;

    return *this;
}

// --------------------------------------------------------------------------

PICSHORT PicNmCollect::ascending (const BaseString& n1, const BaseString& n2)
{
    if (n1 > n2)
        return 1;

    if (n1 == n2)
        return 0;

    return -1;
}

PICSHORT PicNmCollect::descending (const BaseString& n1, const BaseString& n2)
{
    if (n1 > n2)
        return -1;

    if (n1 == n2)
        return 0;

    return 1;
}

