// --------------------------------------------------------------------------
//  FILE implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
  #include <unistd.h>
  #include <sys/stat.h>
  #include <sys/types.h>
  #include <fcntl.h>
  #include <errno.h>
  #include <sys/file.h>
  #include <dirent.h>
  #include <stdio.h>
}

extern "C" {
#include <string.h>
#include <time.h>
#include <stdlib.h>
}

#include <pfile.hxx>
#include <penviron.hxx>
#include <palloc.hxx>
#include <bstrarr.hxx>
#include <picerror.hxx>
#include <picbase.erh>

MAKE_PICCLASS(PicFile)

ERROR_MESSAGES(PicFile)
PARAMSTR(error, "File error: @*")
PARAMSTR(fileNotOpened, "File not opened: @*")
PARAMSTR(fileStillOpen, "File still open: @*")
PARAMSTR(fileAlreadyOpen, "File already open: @*")
PARAMSTR(illegalDrive, "Illegal drive: @*")
PARAMSTR(openModeError, "Open mode error: @*")
PARAMSTR(fileNotFound, "File not found: @*")
PARAMSTR(nameError, "File name error: @*")
PARAMSTR(sharingError, "File sharing error: @*")
PARAMSTR(accessError, "File access error: @*")
PARAMSTR(driveNotReady, "Drive not ready: @*")
PARAMSTR(pathError, "File path error: @*")
PARAMSTR(eIsDir, "File is a directory: @*")
PARAMSTR(eBadFileNbr, "File error: Bad file number")
PARAMSTR(eInvalidArg, "File error: Invalid argument")
END_PARAM_STRINGS

#define FILE_RAISE(status)  ERAISEFMT_LINE_FILE (PicFile::errors, \
    PicFile::classId, status, __LINE__, __FILE__)

const PICSIZE maxPathAndFile = MAXNAMLEN -1 ;
const PICSIZE maxRootSize    = 8;       // used by PicFile::makeUniqueName

const char pathDelimiter  = '/';
const char pathSeparator  = ':';
const char extDelimiter   = '.';

static void errorInfo(
          PICUINT     err,
    const BaseString& name,
          PICUSHORT&  status,
          BaseString& info)
{
    switch(err)
    {
        case EACCES:
        case EROFS:
        case ETXTBSY:
            status = PicFile::accessError;
            info   = name;
            break;
        case ENOENT:
            status = PicFile::fileNotFound;
            info   = name;
            break;
        case EBADF:
	    status = PicFile::eBadFileNbr; // fall through
        case EISDIR:
	    status = PicFile::eIsDir; // fall through
        case EINVAL:
	    status = PicFile::eInvalidArg; // fall through
        case ENAMETOOLONG:
	    status = PicFile::nameError;
            info   = name;
            break;
        case ENOTDIR:
            status = PicFile::pathError;
            info   = name;
            break;
        default:
            status = PicFile::error;
            info = "Status: ";
            info.append(asString(err));
            break;
    }
}

PICPOSITION PicFile::skipDrive() const
{
    return firstPos;
}

PICPOSITION PicFile::skipPath() const
{
    PICPOSITION pathStart = skipDrive();
    const char* start     = fName.chars();
    const char* pathEnd   = strrchr(start+(pathStart-1),pathDelimiter);

    if (pathEnd)
        return (pathEnd-start)+2;
    else
        return pathStart;
}

PICPOSITION PicFile::skipRoot() const
{
    PICPOSITION nameStart = skipPath();
    const char* start     = fName.chars();
    const char* rootEnd   = strrchr(start+(nameStart-1),extDelimiter);

    if (rootEnd)
        return rootEnd-start+2;
    else
        return outOfRange;
}

// --------------------------------------------------------------------------
//      file find
//
//      Used search algorithm is :
//          match pattern in start directory
//          if (recurse)
//              recursively match at each subdirectory

PICBOOL PicFile::fileFind(
          PicFile&    match,
    const BaseString& pattern,
    const BaseString& startDir,
          PICBOOL     recurse,
          PICUINT     attr)
{
    PicFile file(startDir);

    if (startDir.last() != pathDelimiter)
        file.fName.append(pathDelimiter,1);
    file.fName.append(pattern);

    PicFileIterator i(file);
    while (i.next())
    {
        match = i.file;
        return boolTrue;
    }

    if (recurse)
    {
        file.fName = startDir;
        if (startDir.last() != pathDelimiter)
            file.fName.append(pathDelimiter,1);
        file.fName.append('*');

        PicFileIterator d(file, PicFile::subdirectories);
        while (d.next())
            if (fileFind(match, pattern, d.file.fName, boolTrue, attr))
                return boolTrue;
    }
    return boolFalse;
}

void PicFile::fileFind(
    BaseStringArray&  store,
    const BaseString& pattern,
    const BaseString& startDir,
          PICBOOL     recurse,
          PICUINT     attr)
{
    PicFile file(startDir);

    if (startDir.last() != pathDelimiter)
        file.fName.append(pathDelimiter,1);
    file.fName.append(pattern);

    PicFileIterator i(file);
    while (i.next())
    {
        if (!store.add(i.file.fName))
            break;
    }

    if (recurse)
    {
        file.fName = startDir;
        if (startDir.last() != pathDelimiter)
            file.fName.append(pathDelimiter,1);
        file.fName.append('*');

        PicFileIterator d(file, PicFile::subdirectories);
        while (d.next()){
            fileFind(store, pattern, d.file.fName, boolTrue, attr);
	}
    }
}

// --------------------------------------------------------------------------
//      METHODS

BaseString PicFile::path() const
{
    if (isNull())
        return BaseString();

    PICPOSITION pathStart = skipDrive();
    PICPOSITION nameStart = skipPath();

    if (pathStart != nameStart){
        nameStart--;
        if(pathStart == nameStart)
            nameStart++;
    }
    return fName.before(nameStart);
}

BaseString PicFile::name() const
{
    if (isNull())
        return BaseString();

    return fName.after(skipPath() - 1);
}

BaseString PicFile::root() const
{
    if (isNull())
        return BaseString();

    PICPOSITION nameStart = skipPath(), extStart = skipRoot();

    if (extStart == outOfRange)
        return fName.after(nameStart - 1);
    else
        return fName.between(nameStart, extStart - 2);
}

BaseString PicFile::suffix() const
{
    if (isNull())
        return BaseString();

    PICPOSITION extStart = skipRoot();

    if (extStart == outOfRange)
        return BaseString("");
    else
        return fName.after(extStart - 1);
}

void PicFile::setFullName(const BaseString& newName)
{
    fName = newName;
}

void PicFile::setPath(const BaseString& path)
{
    fName.deleteBefore(skipPath());

    if (path.isEmpty())
        return;

    if (path.last() != pathDelimiter)
        fName.prepend(pathDelimiter);
    
	fName.prepend(path);
}

void PicFile::setName(const BaseString& name)
{
    fName.deleteAfter(skipPath()-1);
    fName.append(name);
}

void PicFile::setRoot(const BaseString& root)
{
    PICPOSITION extStart  = skipRoot();
    PICPOSITION rootStart = skipPath();

    if (extStart == outOfRange)
        fName.deleteAfter(rootStart-1);
    else if (extStart > 2)
        fName.deleteBetween(rootStart,extStart-2);

    fName.insert(rootStart,root);
}

void PicFile::setSuffix(const BaseString& suffix)
{
    PICPOSITION extStart = skipRoot();

    if (extStart != outOfRange)
        fName.deleteAfter(extStart-2);

    if (!suffix.isEmpty())
    {
        fName.append(extDelimiter);
        fName.append(suffix);
    }
}

PICBOOL PicFile::exists() const
{
    if (fName.isSpace())
        return boolFalse;
    PICUINT osErr;

    if(access(fullName().chars(), F_OK)==0)
        return boolTrue;
    osErr = errno;
    if (osErr != ENOENT)
    {
        PICUSHORT  status;
        BaseString info;
        errorInfo(osErr, fName, status, info);
        FILE_RAISE(status) << info;
    }
    return boolFalse;
}

PICBOOL PicFile::findPath() const
{
    while(fName.positionOf("./")==1)
        ((PicFile*)this)->fName.deleteBefore(3);
    return boolTrue;
}

PicFile PicFile::find(
    const BaseString& pattern,
    const BaseString& dirs,
          PICBOOL     recurse,
          PICUINT     attributes)
{
    BaseString pathSpec(dirs);
    
    pathSpec.trim();
    if (pathSpec.isEmpty())
        return PicFile();

    if (pathSpec.last() != pathSeparator)
        pathSpec.append(pathSeparator, 1);

    PicFile     file;
    BaseString  startDir;
    PICPOSITION start = firstPos;
    PICPOSITION end   = pathSpec.positionOf(pathSeparator);

    while (end != outOfRange)
    {
        startDir = pathSpec.between(start,end-1);

        if (fileFind(file, pattern, startDir, recurse, attributes))
            break;

        start = end+1;
        end = pathSpec.positionOf(pathSeparator, start);
    }
    return file;
}

PICBOOL PicFile::find(
    BaseStringArray&  files,
    const BaseString& pattern,
    const BaseString& dirs,
          PICBOOL     recurse,
          PICUINT     attributes)
{
    BaseString pathSpec(dirs);

    files.deleteAll();
    pathSpec.trim();
    if (pathSpec.isEmpty())
        pathSpec = currentDirectory();

    if (pathSpec.last() != pathSeparator)
        pathSpec.append(pathSeparator, 1);
    
    BaseString  startDir;
    PICPOSITION start = firstPos;
    PICPOSITION end   = pathSpec.positionOf(pathSeparator);
    
    while (end != outOfRange)
    {
        startDir = pathSpec.between(start, end-1);

        fileFind(files, pattern, startDir, recurse, attributes);
        start = end+1;
        end = pathSpec.positionOf(pathSeparator, start);
    }
    return !files.isEmpty();
}

static BaseString unique(const BaseString& base)
{
    static PICBOOL isInitialized = boolFalse;

    if (!isInitialized)
    {
        clock_t clk;
        clk = clock();

        srand((unsigned)clk);
        isInitialized = boolTrue;
    }

    BaseString uniq = base;

    if (uniq.size() < maxRootSize)
        uniq.append('0', maxRootSize-uniq.size());
    else if (uniq.size() > maxRootSize)
        uniq.deleteAfter(maxRootSize);

    uniq.replace(4, asString(rand() % 99998));   // 1..99999

    return uniq;
}

PicFile PicFile::makeUniqueName(const BaseString& baseName)
{
    PicFile    result(baseName);
    BaseString root = result.root();

    if (root.isEmpty())
        root = "tmp";

    for (PICINDEX ind = 0; ind < 10; ind++)
    {
        result.setRoot(unique(root));

        if (!(result.exists()))
            return result;
    }
    return PicFile();
}

PICBOOL PicFile::copyTo(const PicFile& newFile) const
{
    return boolFalse;
}

PICBOOL PicFile::rename(const BaseString& newName)
{
    PICUINT err;
    if(::rename(fName.chars(), newName.chars())==0)
        return boolTrue;
    err = errno;

    PICUSHORT  status;
    BaseString info;
    errorInfo(err, fName, status, info);
    FILE_RAISE(status) << info;
    return boolFalse;
}

PICBOOL PicFile::deleteFile()
{
    PICUINT err;
    if(unlink(fName.chars())==0)
        return boolTrue;
    err = errno;

    PICUSHORT  status;
    BaseString info;
    errorInfo(err, fName, status, info);
    FILE_RAISE(status) << info;
    return boolFalse;
}

PICBOOL PicFile::changeDirectory(const BaseString& toDir)
{
    PICUINT err;
    if(chdir(toDir.chars())==0)
        return boolTrue;
    err = errno;

    PICUSHORT  status;
    BaseString info;
    errorInfo(err, toDir, status, info);
    FILE_RAISE(status) << info;
    return boolFalse;
}

PICBOOL PicFile::makeDirectory(const BaseString& dir)
{
    PICUINT err;
    if(mkdir(dir.chars(), 0660)==0)
        return boolTrue;
    err = errno;

    PICUSHORT  status;
    BaseString info;
    errorInfo(err, dir, status, info);
    FILE_RAISE(status) << info;
    return boolFalse;
}

PICBOOL PicFile::removeDirectory(const BaseString& dir)
{
    PICUINT err;
    if(rmdir(dir.chars())==0)
        return boolTrue;
    err = errno;

    PICUSHORT  status;
    BaseString info;
    errorInfo(err, dir, status, info);
    FILE_RAISE(status) << info;
    return boolFalse;
}

BaseString PicFile::currentDirectory()
{
    char curDir[maxPathAndFile];
    if (getcwd(curDir, maxPathAndFile))
    {
        return curDir;
    }
    else
    {
        PICUSHORT  status;
        BaseString info;
        errorInfo(errno, BaseString(curDir), status, info);
        FILE_RAISE(status) << info;
    }

    return BaseString::null();
}

PICSIZE PicFile::itemCount(PICSIZE itemSize) const
{
    PICSIZE   itemSz = itemSize == 0 ? 1 : itemSize;
    PICULONG  fileSize;
    PICUSHORT osErr = 0;
    struct stat info;
    osErr = (stat(fName.chars(), &info) != 0);
    fileSize = info.st_size;
    return osErr ? 0 : (PICSIZE)(fileSize / itemSz);
}

PICUINT PicFile::attributes() const
{
    struct stat info;
    int osErr = (stat(fName.chars(), &info) != 0);
    return osErr ? 0 : info.st_mode;
}

// --------------------------------------------------------------------------
//      PicFileIterator
//
//      Special attribute subdirectories is used to ignore '.', '..' and
//      non-directory entries. 

inline PICBOOL isParentOrSelf(const char* name)
{
    return *name == '.' && 
        (*(name+1)=='\0' || (*(name+1)=='.' && *(name+2)=='\0'));
}

PicFileIterator::~PicFileIterator()
{
    if(handle) closedir(handle);
}

PICBOOL fileMatches(BaseString str, BaseString patt)
{
    PICPOSITION i, j, k;
    char pattCh;

    patt.replace("**", "*");
    if(str.isEmpty()) 
        return patt.containsChars("*");
    if(!patt.positionOf('*'))
        return patt == str;

    j = 1;
    for(i = 1; i <= patt.size(), j <= str.size(); i++)
      {
        pattCh = patt.at(i);
        if(pattCh != '*'){
            if(pattCh != str.at(j))
                return boolFalse;
            j++;
        }else{
            if(i == patt.size())
                return boolTrue;
            BaseString patt2(patt.after(i));
            for(k = j; k <= str.size(); k++){
                if(fileMatches(str.after(k-1),patt2))
                    return boolTrue;
            }
            break;
        }
      }
    return boolFalse;
}

PICBOOL PicFileIterator::next()
{
    BaseString path = file.path();
    if (!handle)  // FindFirst
    {
        file.findPath();

        if(path.isEmpty())
            path = ".";

        if ((handle = opendir(path.chars())) == NULLPTR)
            return boolFalse;
        pattern = file;
    }
    struct dirent *diren;
    while(1){
      if (((diren = readdir(handle)) == NULLPTR) || (!(diren->d_name)))
        return boolFalse;
      if(fileMatches(diren->d_name, pattern.name()))
	break;
    }

    struct stat info;

    PicFile trueFile(diren->d_name); 
    trueFile.setPath(path);
    if(stat(trueFile.fullName().chars(), &info) != 0){
        return boolFalse;
    }

    while ((attr & 020000) &&
           (!(PicFile::directory & info.st_mode) ||
            isParentOrSelf(diren->d_name)) &&
           fileMatches(diren->d_name, pattern.name()))
    {
        if ((diren = readdir(handle)) == NULLPTR)
            return boolFalse;
        trueFile.setName(diren->d_name);
        if(stat(trueFile.fullName().chars(), &info) != 0)
            return boolFalse;
    }
    file.fName.deleteAfter(file.skipPath() - 1);
    file.fName.append(diren->d_name);

    return boolTrue;
}



