// --------------------------------------------------------------------------
//  PUBLIC INTERFACE OF ERROR MANAGER - implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <stdlib.h>
}

#include <picerror.hxx>
#include <perrmgr.hxx>
#include <picbase.erh>

MAKE_PICCLASS(PicError)

ERROR_MESSAGES(PicError)
PARAMSTR(functionNotImplemented, "Function not implemented.")
PARAMSTR(outOfMemory, "Out of memory.")
PARAMSTR(memoryAllocation, "Memory allocation failure.")
PARAMSTR(nullInstance, "Operation on Null instance.")
PARAMSTR(assertCondition, "Assertion failed: @*")
PARAMSTR(assertNotNull, "Not Null assertion failed: @*")
PARAMSTR(assertNotNullPtr, "Assertion failed: @* is null pointer")
PARAMSTR(error, "Error")
PARAMSTR(fatalError, "Fatal error")
END_PARAM_STRINGS

static PicErrMgrParam* param = &PicErrorManager::param();

// --------------------------------------------------------------------------
//      IMPLEMENTATION FOR PicErrorText
// --------------------------------------------------------------------------

PicErrorText::PicErrorText (
    const PicClass& errClass,
    PICSTATUS       errStatus,
    PICUSHORT       errLine,
    const char*  errFile)
    :
    errorText   (),
    errorClass  (&errClass),
    errorStatus (errStatus),
    errorLine   (errLine),
    errorFile   (errFile)
{}

PicErrorText::~PicErrorText () {}

PicFmtErrorText::~PicFmtErrorText () {}

// --------------------------------------------------------------------------
//      IMPLEMENTATION FOR PicFmtErrorText
// --------------------------------------------------------------------------

PicFmtErrorText::PicFmtErrorText (
    PIC_PARAMSTR_FUNC* messages,
    const PicClass&    errorClass(),
    PICSTATUS          errorStatus,
    PICUSHORT          errorLine,
    const char*        errorFile)
    :
    PicErrorText (errorClass(), errorStatus, errorLine, errorFile),
    strStream    (msg()),
    fmtText      (strStream)
{
    if (messages)
        fmtText.format(messages(errorStatus));
}

void PicRaiseFmtError::raise (
    PIC_PARAMSTR_FUNC* messages,
    const PicClass&    errClass(),
    PICSTATUS          errStatus,
    PICUSHORT          errLine,
    const char*        errFile,
    PICUSHORT          errKind)  // 0=raise 1=hold 2=fatal
{
    PicErrorText err (errClass(), errStatus, errLine, errFile);
    if (messages)
        err.msg(messages(errStatus));
    switch (errKind)
    {
        case 0: PicError::raise(err); break;
        case 1: PicError::raise(err, boolTrue); break;
        case 2: PicError::raise(err, boolTrue);
                PicError::exitOnFatalError(errLine, errFile);
    }
}

PicRaiseFmtError::~PicRaiseFmtError ()
{
    text() ENDFMT;
    PicError::raise(THIS);
}

PicHoldFmtError::~PicHoldFmtError ()
{
    text() ENDFMT;
    PicError::raise(THIS, boolTrue);
}

PicRaiseFatalError::~PicRaiseFatalError ()
{
    text() ENDFMT;
    PicError::raise(THIS, boolTrue);
    PicError::exitOnFatalError(line(), file());
}

// --------------------------------------------------------------------------
//      Handling of memory allocation failures: The PicMemErrHandler
//      constructor sets the 'new' error handling function, using
//      set_new_handler.

static const char memoryFailMessage [] =
    "memoryFailHandler: Memory allocation with 'new' failed!";

static void memoryFailHandler ()
{
    PicError::raiseMemoryError(memoryFailMessage);
}

struct PicMemErrHandler
{
    PicMemErrHandler ()
    {
        set_new_handler(&memoryFailHandler);
    }
};

static PicMemErrHandler initMemoryErrorHandler;

// --------------------------------------------------------------------------
//      ERROR INFORMATION OF THE GLOBAL CONTEXT
//
//      The global variable picGlobalError holds the error information
//      when the error manager is in single-context state. If it is
//      NULLPTR, the global context is clear.

PicError* picGlobalError = NULLPTR;
PICBOOL   picGlobalMemoryError = boolFalse;

static void raiseGlobal (const PicErrorText& message, PICBOOL hold)
{
    PicError* error = new PicError(message);

    if (picGlobalError)
    {
        if (hold)
            error->appendError(picGlobalError);
        else
            delete picGlobalError;
    }

    picGlobalError = error;
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS & DESTRUCTOR

static PICUSHORT errorCounter = 0;

PicError::PicError (const PicErrorText& message, PicErrorManager* mgr) :
    errorText   (message.msg()),
    errorClass  (message.id()),
    errorStatus (message.status()),
    errorLine   (message.line()),
    errorFile   (message.file()),
    errorNum    (++errorCounter),
    manager     (mgr),
    next        (NULLPTR),
    last        (this)
{}

PicError::PicError (const PicError& copy) :
    errorText   (copy.errorText),
    errorClass  (copy.errorClass),
    errorStatus (copy.errorStatus),
    errorLine   (copy.errorLine),
    errorFile   (copy.errorFile),
    errorNum    (copy.errorNum),
    manager     (copy.manager),
    next        (NULLPTR),
    last        (NULLPTR)
{
    if (copy.next)
    {
        // Copy linked PicErrors recursively:

        next = new PicError (*copy.next);
        last = next->last;
    }
    else
        last = this;
}

PicError::~PicError ()
{
    if (next)
        delete next;
}

// --------------------------------------------------------------------------
//      CLEARING ERRORS

void PicError::clearCurrent ()
{
    PicErrorManager* mgr = PicErrorManager::get();

    if (mgr)
        mgr->clearCurrent();
    else
    {
        if (picGlobalError)
            delete picGlobalError;
        picGlobalError = NULLPTR;
        picGlobalMemoryError = boolFalse;
    }
}

void PicError::resetErrorManager ()
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->reset();
    else
        PicError::clearCurrent();
}

// --------------------------------------------------------------------------
//      GETTING ERROR DESCRIPTION

void PicError::description (PicFmtText& descr, const char* templ) const
{
    descr(templ)
        >>"TEXT"    << text()
        >>"CLASS"   << errorClass.className()
        >>"STATUS"  << errorStatus
        >>"LINE"    << errorLine
        >>"FILE"    << errorFile
        >>"NUMBER"  << errorNum
        ENDFMT;
}

BaseString PicError::description (
    const char* raiseTemplate,
    const char* holdTemplate) const
{
    PicStringText   wholeDescr;
    const PicError* err = this;

    if (!holdTemplate)
        holdTemplate = raiseTemplate ? raiseTemplate : param->emgrHoldTemplate;
    if (!raiseTemplate)
        raiseTemplate = param->emgrRaiseTemplate;

    while (err)
    {
        err->description(wholeDescr.text(), raiseTemplate);
        raiseTemplate = holdTemplate;
        err = err->next;
    }

    return wholeDescr;
}

// --------------------------------------------------------------------------

PICBOOL PicError::isDetected ()
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        return (mgr->isDetected());
    else
        return picGlobalError || picGlobalMemoryError;
}

PICBOOL PicError::isDetected (const PicClass& errorClass)
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        return (
            mgr->currentError() &&
            mgr->currentError()->errorClass == errorClass);
    else
        return (
            picGlobalError &&
            picGlobalError->errorClass == errorClass);
}

// --------------------------------------------------------------------------
//      NULL ERROR & CURRENT ERROR

const PicError& PicError::current ()
{
    PicError*        error = NULLPTR;
    PicErrorManager* mgr   = PicErrorManager::get();

    if (mgr)
        error = mgr->currentError();
    else
        error = picGlobalError;

    if (error)
        return *error;
    else
        return nullError();
}

const PicError& PicError::nullError ()
{
    static PicErrorText nulltxt (
        PicClass::nullClass(), 0, 0, NULLPTR);
    static PicError nullPicError (nulltxt);

    return nullPicError;
}

// --------------------------------------------------------------------------
//       RAISE & HOLD

void PicError::raise (const PicErrorText& message, PICBOOL hold)
{
    PicErrorManager* mgr = PicErrorManager::get();

    if (mgr)
    {
        mgr->raise(message, hold);
    }
    else
        raiseGlobal(message, hold);
}

// --------------------------------------------------------------------------
//      CUMULATE

void PicError::cumulate (const BaseString& cumulationText)
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->cumulate(cumulationText);
    else if (picGlobalError)
        picGlobalError->appendText(cumulationText);
}

// --------------------------------------------------------------------------
//      MEMORY ERRORS

void PicError::raiseMemoryError (const char* explanation)
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->raiseMemoryError(explanation);
    else
        picGlobalMemoryError = boolTrue;
}

PICBOOL PicError::checkMemoryError (
    PICBOOL     wantToContinue,
    const char* cumulationText,
    PICUSHORT   lineNumber,
    const char* fileName)
{
    PicErrorManager* mgr = PicErrorManager::get();

    if (mgr)
        return mgr->checkMemoryError (
            wantToContinue, cumulationText, lineNumber, fileName);
    else if (picGlobalMemoryError)
    {
        if (wantToContinue)
            return boolTrue;
        else
            PicError::exitOnMemoryError (
                cumulationText, lineNumber, fileName);
    }
    return boolFalse;
}

// --------------------------------------------------------------------------
//      FATAL EXITING

void PicError::exitOnFatalError (PICUSHORT line, const char* file)
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
    {
        mgr->notifier().fatalExit(line, file);
    }

    exit(1);
}

void PicError::exitOnMemoryError (
    const char* message,
    PICUSHORT   lineNumber,
    const char* fileName)
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->notifier().memoryExit(message, lineNumber, fileName);

    exit(1);
}

// --------------------------------------------------------------------------
//      RPICRSE & REMOVE ERROR INFO

void PicError::reverseCurrent ()
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->reverseError();
    else if (picGlobalError)
        picGlobalError = picGlobalError->reverse();
}

PicError* PicError::reverse ()
{
    PicError* nextError = this->next;
    PicError* prevError = this;
    PicError* error     = this;

    last = this;
    next = NULLPTR;

    while (nextError)
    {
        error = nextError;
        nextError = error->next;
        error->next = prevError;
        error->last = this;
        prevError = error;
    }
    return error;
}

PicError* PicError::remove ()
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        return mgr->removeError();
    else
    {
        if (picGlobalError)
        {
            PicError* error = picGlobalError;
            picGlobalError = NULLPTR;
            return error;
        }
        else
            return NULLPTR;
    }
}

// --------------------------------------------------------------------------
//      APPEND ERROR INFORMATION

void PicError::appendError (PicError* nextError)
{
    PicError* err = this;

    last->next = nextError;

    while (err && err != nextError)
    {
        err->last = nextError->last;
        err = err->next;
    }
}

void PicError::appendText (const BaseString& cumulationText)
{
    if (errorText.isNull())
    {
        errorText = cumulationText;
        if (errorText.first() != '\n')
            errorText.prepend("\n");
    }
    else
    {
        if (errorText.last() != '\n')
            errorText.append("\n");
        errorText.append(cumulationText);
    }
}
