// --------------------------------------------------------------------------
//  DYNAMIC INTEGER ARRAY implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
#include <stdlib.h>
}

#include <pintarr.hxx>
#include <palloc.hxx>
#include <bstring.hxx>
#include <picerror.hxx>
#include <piterimp.hxx>
#include <picbase.erh>
#include <pcollect.hxx>
#include <bcollect.hxx>

MAKE_PICCLASS(PicIntArray)

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

PicIntArray::PicIntArray (PICINDEX lowerBound, PICSIZE initialSize) :
    lb (lowerBound), ub (lowerBound - 1), content (NULLPTR)
{
    if (lb == outOfIndex)
    {
        ERAISEFMT(BaseCollection, illegalLowerBound) << lb;
        setNull();
    }
    else
        allocate (initialSize);
}

PicIntArray::PicIntArray (const PicObject& theObject) :
    content (NULLPTR), lb (1), ub (0)
{
    PicErrorContext e1;

    if (theObject.isNull())
    {
        setNull();
        return;
    }

    allocate (AS(PicCollection, theObject).size());

    if (content)
    {
        PicIterator theIterator;
        ITERATE (theObject, theIterator)
            add (theObject.current (theIterator).asInt ());
    }
}

PicIntArray::PicIntArray (
    PICINDEX lowerBound,
    const PicObject& theObject) : lb (lowerBound), ub (lowerBound - 1)
{
    if (!theObject.isCollection() || AS(PicCollection, theObject).isEmpty()
        || (lowerBound == outOfIndex))
    {
        setNull();
        return;
    }

    allocate(AS(PicCollection, theObject).size());

    if (content)
    {
        PicIterator theIterator;
        ITERATE (theObject, theIterator)
            add (theObject.current (theIterator).asInt ());
    }
}

PicIntArray::PicIntArray (const PicIntArray& anArray) :
    lb(anArray.lb), ub(anArray.ub), content(NULLPTR)
{
    if (anArray.isNull())
        setNull();
    else
    {
        allocate (anArray.size());
        if (content)
            copy (anArray);
    }
}

PicIntArray::~PicIntArray ()
{
    picFree (content);
}

PicIntArray& PicIntArray::operator = (const PicIntArray& anArray)
{
    PICSIZE arraySize = anArray.ub - anArray.lb + 1;

    if (anArray.isNull())
    {
        setNull();
        return *this;
    }

    // If the present allocation is too small or far too big, reallocate:
    if (alloc < arraySize || alloc > (arraySize << 1))
    {
        picFree (content);
        allocate (arraySize);
    }

    if (content)
    {
        lb = anArray.lb;
        ub = anArray.ub;
        copy (anArray);
    }

    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISON

PICBOOL PicIntArray::operator == (const PicIntArray& anArray) const
{
    if (lb != anArray.lb || ub != anArray.ub) 
        return boolFalse;

    if (isEmpty ())
         return boolTrue;

    if (memcmp(content, anArray.content, size() * sizeof(PICINT)))
        return boolFalse;

    return boolTrue;
}

// --------------------------------------------------------------------------
//      ACCESS

PICINT PicIntArray::at (PICINDEX anIndex) const
{
    if (inRange (anIndex))
        return (*this)[anIndex];

    ERAISEFMT(BaseCollection, illegalIndex) << anIndex;

    return outOfIndex;
}

// --------------------------------------------------------------------------
//      NULL ARRAY

void PicIntArray::setNull ()
{
    if (content)
        picFree(content);

    content = NULLPTR;
    alloc   = 0;
    ub      = lb - 1;
}

// --------------------------------------------------------------------------
//      ADD & REPLACE

void PicIntArray::add (PICINT value)
{
    if (isNull())
    {
        allocate (defaultSize);
        if (isNull())
            return;
    }

    if (ub == ::maxIndex)
    {
        ERAISEFMT(BaseCollection, illegalIndex) << (asString(ub) & " + 1");
        return;
    }

    // long-casts are used to ensure correct answers even when dealing
    // with numbers near the limits of short integer.

    if (((PICLONG)ub - (PICLONG)lb + 1) >= (PICLONG)alloc)
    {
        if (!resize (alloc + 1))
            return;
    }

    ub++;
    (*this)[ub] = value;
}

void PicIntArray::putAt (PICINDEX anIndex, PICINT value)
{
    if (anIndex < lb)
    {
        ERAISEFMT(BaseCollection, illegalIndex) << anIndex;
        return;
    }

    if (isNull())
    {
        allocate (anIndex - lb + 1);
        if (isNull())
            return;
    }

    if (anIndex >= (lb + (PICLONG)alloc))
    {
        if (!resize (anIndex - lb + 1))
            return;
    }

    (*this)[anIndex] = value;

    if (anIndex > ub)
        ub = anIndex;
}

void PicIntArray::fillBetween (PICINDEX from, PICINDEX to, PICINT value)
{
    if (to < from || to < lb)
        return;
    
    if (to >= lb + (PICLONG)alloc)
    {
        if (!resize (to - lb + 1))
            return;
    }

    for (PICINDEX ind = (from < lb) ? lb : from; ind <= to; ind++)
        (*this)[ind] = value;

    if (to > ub)
        ub = to;
}

// --------------------------------------------------------------------------
//      SET EMPTY

void PicIntArray::setEmpty ()
{
    memset(content, 0, alloc * sizeof(PICINT));
    ub = lb - 1;
}

// --------------------------------------------------------------------------
//      SEARCH

PICINDEX PicIntArray::indexOf (PICINT value) const
{
    for (PICINDEX ind = lb; ind <= ub; ind++)
        if ((*this)[ind] == value)
            return ind;

    return outOfIndex;
}

// --------------------------------------------------------------------------
//      MAXIMUM INDEX

PICINDEX PicIntArray::maxIndex () const
{
    PICLONG maxInd = lb + (PICLONG)alloc - 1;
    return (maxInd > ::maxIndex) ? ::maxIndex : (PICINDEX)maxInd;
}

// --------------------------------------------------------------------------
//      STANDARD ITERATOR

void PicIntArray::startIteration (PicIterator& theIterator) const
{
    SET_ITERATE_INDEX (theIterator, minLongInt);
}

PICBOOL PicIntArray::next (PicIterator& theIterator) const
{
    PICLONG currentInd = GET_ITERATE_INDEX (theIterator);

    if (currentInd == minLongInt)
        currentInd = lb;
    else
        currentInd++;  // Cannot overflow (ub is smaller integer)
    
    SET_ITERATE_INDEX (theIterator, currentInd);

    return PICBOOL   (currentInd <= (PICLONG)ub);
}

PICBOOL PicIntArray::previous (PicIterator& theIterator) const
{
    PICLONG currentInd = GET_ITERATE_INDEX (theIterator);

    if (currentInd == minLongInt)
        currentInd = ub;
    else
        currentInd--;  // Cannot underflow (lb is smaller integer)
    
    SET_ITERATE_INDEX (theIterator, currentInd);

    return PICBOOL    (currentInd >= (PICLONG)lb);
}

PICINT PicIntArray::current (PicIterator& theIterator) const
{
    return (*this)[(PICINDEX) GET_ITERATE_INDEX (theIterator)];
}

PICINDEX PicIntArray::currentIndex (PicIterator& theIterator) const
{
    return (PICINDEX) GET_ITERATE_INDEX (theIterator);
}

// --------------------------------------------------------------------------
//      SORT
//
//      Compare functions for qsort
//      ---------------------------
//      These functions are provided for the C run-time library
//      function qsort that is used to implement sorting in
//      either ascending or descending order. See documentation
//      of the qsort function for further details.

extern "C" {

static int compareAsc (const void* elem1, const void* elem2)
{
    if (*(PICINT*)elem1 < *(PICINT*)elem2)
        return -1;

    if (*(PICINT*)elem1 == *(PICINT*)elem2)
        return 0;

    return 1;
}

static int compareDesc (const void* elem1, const void* elem2)
{
    if (*(PICINT*)elem1 < *(PICINT*)elem2)
        return 1;

    if (*(PICINT*)elem1 == *(PICINT*)elem2)
        return 0;

    return -1;
}

} // extern "C"

void PicIntArray::sort (PicSortOrder order)
{
    if (size() > 1)
    {
        // Apply the qsort-function from C run-time library:

        if (order == ascending)
            qsort(content, ub - lb + 1, sizeof(PICINT), compareAsc);
        else
            qsort(content, ub - lb + 1, sizeof(PICINT), compareDesc);
    }
}

// --------------------------------------------------------------------------
//      PRIVATE METHODS

void PicIntArray::copy (const PicIntArray& anArray)
{
    memcpy(content, anArray.content, alloc * sizeof(PICINT));
}

void PicIntArray::allocate (PICSIZE amount)
{
    if (amount > maxNewWordSize)
    {
        ERAISEFMT(BaseCollection, full) << amount;
        setNull();
        return;
    }

    PICSIZE bytes = picCalculateSize(amount * sizeof (PICINT));

    content = (PICINT*)picAllocate(bytes);

    if (content)
        alloc = bytes / sizeof (PICINT);
    else
    {
        setNull();
        MEMORY_ERROR("PicIntArray::allocate: Memory allocation failed");
    }
}

PICBOOL PicIntArray::resize (PICSIZE toSize)
{
    if (toSize == alloc)
        return boolTrue;

    if (toSize > maxNewWordSize)
    {
        ERAISEFMT(BaseCollection, full) << toSize;
        return boolFalse;
    }

    PICSIZE bytes = picCalculateSize (toSize * sizeof(PICINT));
    PICINT* buf = (PICINT*)picResize(content, alloc * sizeof(PICINT), bytes);

    if (buf)
    {
        content = buf;
        alloc = bytes / sizeof(PICINT);
    }
    else
    {
        MEMORY_ERROR("PicIntArray::resize: Memory allocation failed");
        return boolFalse;
    }
    return boolTrue;
}
