// --------------------------------------------------------------------------
//  INTEGER SET implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
#include <stdarg.h>
}

#include <pintset.hxx>
#include <palloc.hxx>
#include <piterimp.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pcollect.hxx>
#include <bcollect.hxx>

MAKE_PICCLASS(PicIntSet)

// --------------------------------------------------------------------------
//  PicIntSetBlock
// --------------------------------------------------------------------------

struct PicIntSetBlock1
{
    PICUSHORT count;  // total number of allocated extra blocks
    PICULONG  bits;   // bits for numbers 0..31

    void set   (PICUSHORT value) { bits |= 1UL << value; }
    void clear (PICUSHORT value) { bits &= ~(1UL << value); }
    PICBOOL inRange (PICUSHORT value)
        { return value < 32; }
    PICBOOL contains (PICUSHORT value)
        { return (bits & (1UL << value)) != 0; }
    PICBOOL next (PICUSHORT& value);
    PICBOOL previous (PICUSHORT& value);

};

PICBOOL PicIntSetBlock1::next (PICUSHORT& value)
{
    if (!inRange(value))
        value = 0xffff;
    while (++value < 32)
        if (contains(value))
            return boolTrue;
    return boolFalse;
}

PICBOOL PicIntSetBlock1::previous (PICUSHORT& value)
{
    if (!inRange(value))
        value = 32;
    while (value-- > 0)
        if (contains(value))
            return boolTrue;
    return boolFalse;
}

struct PicIntSetBlock
{
    PICUSHORT start;  // number represented by the first bit (zero if no bits)
    PICULONG  bits;   // 32 bits for subsequent numbers start..start+31

    void set   (PICUSHORT value) { bits |= 1UL << (value - start); }
    void clear (PICUSHORT value) { bits &= ~(1UL << (value - start)); }
    void setRange (PICUSHORT value) 
        { start = value & 0xffe0; set(value); }
    void clearRange (PICUSHORT value) 
        { clear(value); if (!bits) start = 0; }
    PICBOOL inRange (PICUSHORT value) const
        { return value >= start && value < start + 32; }
    PICBOOL contains (PICUSHORT value)
        { return (bits & (1UL << (value - start))) != 0; }
    PICBOOL next (PICUSHORT& value);
    PICBOOL previous (PICUSHORT& value);

};

PICBOOL PicIntSetBlock::next (PICUSHORT& value)
{
    if (!inRange(value))
        value = start - 1;
    while (++value < start + 32)
        if (contains(value))
            return boolTrue;
    return boolFalse;
}

PICBOOL PicIntSetBlock::previous (PICUSHORT& value)
{
    if (!inRange(value))
        value = start + 32;
    while (--value >= start)
        if (contains(value))
            return boolTrue;
    return boolFalse;
}

// --------------------------------------------------------------------------
//  PicIntSetData
// --------------------------------------------------------------------------

struct PicIntSetData
{
    PicIntSetBlock1 first;
    PicIntSetBlock  second;
    // continues...

    static PicIntSetData* create  ();
    static void           destroy (PicIntSetData*);
           PicIntSetData* copy    ();
    static PICBOOL        grow    (PicIntSetData*&);

    const PicIntSetBlock& at (PICUSHORT i) const { return (&second)[i]; }
          PicIntSetBlock& at (PICUSHORT i)       { return (&second)[i]; }

    PICSIZE alloc () const;

    static void add      (PICUSHORT value, PicIntSetData*&);
           void remove   (PICUSHORT value);
           void setEmpty ();
        PICBOOL isEmpty  () const;

    static void unite     (const PicIntSetData&, PicIntSetData*& me);
           void intersect (const PicIntSetData&);
           void subtract  (const PicIntSetData&);

};

inline PICSIZE PicIntSetData::alloc () const
{ 
    return sizeof(PicIntSetData) + first.count * sizeof(PicIntSetBlock);
}

PicIntSetData* PicIntSetData::create ()
{
    PicIntSetData* data = (PicIntSetData*)picAllocate(sizeof(PicIntSetData));
    if (!data)
        MEMORY_ERROR("PicIntSetData::create");
    return data;
}

PICBOOL PicIntSetData::grow (PicIntSetData*& me)
{
    PicIntSetData* buf = (PicIntSetData*)picResize (
        me, me->alloc(), me->alloc() + 4 * sizeof(PicIntSetBlock));

    if (buf)
    {
        me = buf;
        me->first.count += 4;
        return boolTrue;
    }
    else
        MEMORY_ERROR("PicIntSetData::grow");

    return boolFalse;
}

inline void PicIntSetData::destroy (PicIntSetData* data)
{
    picFree(data);
}

PicIntSetData* PicIntSetData::copy ()
{
    PicIntSetData* theCopy = (PicIntSetData*)picAllocate(alloc());
    memcpy(theCopy, this, alloc());
    return theCopy;
}

inline void PicIntSetData::setEmpty ()
{
    memset(&first.bits, 0, alloc() - sizeof(PICUSHORT));
}

PICBOOL PicIntSetData::isEmpty () const
{
    if (first.bits != 0)
        return boolFalse;
    for (PICUSHORT i = 0; i <= first.count; i++)
        if (at(i).start)
            return boolFalse;
    return boolTrue;
}

// --------------------------------------------------------------------------

void PicIntSetData::add (PICUSHORT value, PicIntSetData*& data)
{
    if (data->first.inRange(value))
        data->first.set(value);
    else
    {
        PICUSHORT i;
        for (i = 0; i <= data->first.count; i++)
            if (data->at(i).inRange(value))
            {
                data->at(i).set(value);
                return;
            }
        for (i = 0; i <= data->first.count; i++)
            if (!data->at(i).start)
            {
                data->at(i).setRange(value);
                return;
            }
        if (grow(data))
            data->at(i).setRange(value);
    }
}

void PicIntSetData::remove (PICUSHORT value)
{
    if (first.inRange(value))
        first.clear(value);
    else
    {
        for (PICUSHORT i = 0; i <= first.count; i++)
            if (at(i).inRange(value))
            {
                at(i).clearRange(value);
                return;
            }
    }
}

// --------------------------------------------------------------------------

void PicIntSetData::unite (const PicIntSetData& data, PicIntSetData*& me)
{
    me->first.bits |= data.first.bits;

    for (PICUSHORT i = 0; i <= data.first.count; i++)
    {
        PICUSHORT start = data.at(i).start;
        if (start == 0)
            continue;
        PICULONG bits = data.at(i).bits;
        PICUSHORT free = endOfList, j;
        for (j = 0; j <= me->first.count; j++)
        {
            if (free == endOfList && me->at(j).start == 0)
                free = j;
            else if (me->at(j).start == start)
            {
                me->at(j).bits |= bits;
                goto Next;
            }
        }
        if (free != endOfList)
        {
            me->at(free).start = start;
            me->at(free).bits  = bits;
        }
        else if (grow(me))
        {
            me->at(j).start = start;
            me->at(j).bits  = bits;
        }
        Next: ;
    }
}

void PicIntSetData::intersect (const PicIntSetData& data)
{
    first.bits &= data.first.bits;

    for (PICUSHORT i = 0; i <= first.count; i++)
    {
        PicIntSetBlock& block = at(i);
        if (block.start == 0)
            continue;
        for (PICUSHORT j = 0; j <= data.first.count; j++)
        {
            if (data.at(j).start == block.start)
            {
                block.bits &= data.at(j).bits;
                if (block.bits == 0)
                    block.start = 0;
                goto Next;
            }
        }
        block.bits = 0;
        block.start = 0;
        Next: ;
    }
}

void PicIntSetData::subtract (const PicIntSetData& data)
{
    first.bits &= ~data.first.bits;

    for (PICUSHORT i = 0; i <= data.first.count; i++)
    {
        PICUSHORT start = data.at(i).start;
        if (start == 0)
            continue;
        PICULONG bits = data.at(i).bits;
        for (PICUSHORT j = 0; j <= first.count; j++)
        {
            if (at(j).start == start)
            {
                at(j).bits &= ~bits;
                if (at(j).bits == 0)
                    at(j).start = 0;
            }
        }
    }
}

// --------------------------------------------------------------------------
//      PicIntSet
// --------------------------------------------------------------------------

PicIntSet::PicIntSet (PICUSHORT int1, PICUSHORT int2, ...) :
    content(PicIntSetData::create())
{
    if (!content)
        return;

    add(int1);

    va_list theMembers;
    va_start(theMembers, int2);

    PICUSHORT currentArgument = int2;

    while (currentArgument != endOfList)
    {
        add(currentArgument);
        currentArgument = va_arg(theMembers, PICUSHORT);
    }
}

PicIntSet::PicIntSet (const PicObject& theObject) : content(NULLPTR)
{
    if (!theObject.isCollection() || AS(PicCollection, theObject).isEmpty())
        return;

    content = PicIntSetData::create();

    if (content)
    {
        PicIterator theIterator;
        ITERATE(theObject, theIterator)
            add(theObject.current(theIterator).asInt());
    }
}

PicIntSet::PicIntSet (const PicIntSet& aSet) : content(NULLPTR)
{
    if (!aSet.isNull())
        content = aSet.content->copy();
}

PicIntSet::~PicIntSet ()
{
    if (content)
        PicIntSetData::destroy(content);
}

PicIntSet& PicIntSet::operator = (const PicIntSet& aSet)
{
    if (&aSet == this)
        return *this;

    if (aSet.isNull())
    {
        setNull();
        return *this;
    }

    if (!content || content->alloc() != aSet.content->alloc())
    {
        PicIntSetData::destroy(content);
        content = aSet.content->copy();
    }
    else
        memcpy(content, aSet.content, content->alloc());

    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISON

PICBOOL PicIntSet::operator == (const PicIntSet& aSet) const
{
    if (isNull() || aSet.isNull())
        return PICBOOL(isNull() && aSet.isNull());

    PicIntSetData* diff = content->copy();
    diff->subtract(*aSet.content);
    PICBOOL ok = diff->isEmpty();
    PicIntSetData::destroy(diff);
    if (ok)
    {
        diff = aSet.content->copy();
        diff->subtract(*content);
        ok = diff->isEmpty();
        PicIntSetData::destroy(diff);
    }
    return ok;
}

// --------------------------------------------------------------------------
//      NULL SET

void PicIntSet::setNull ()
{
    if (content)
    {
        PicIntSetData::destroy(content);
        content = NULLPTR;
    }
}

// --------------------------------------------------------------------------
//      ADD & REPLACE

void PicIntSet::add (PICUSHORT int1)
{
    if (isNull())
    {
        if (!(content = PicIntSetData::create()))
            return;
    }

    PicIntSetData::add(int1, content);
}

void PicIntSet::add (PICUSHORT int1, PICUSHORT int2, ...)
{
    add(int1);

    va_list theMembers;
    va_start(theMembers, int2);

    PICUSHORT currentArgument = int2;

    while (currentArgument != endOfList)
    {
        PicIntSetData::add(currentArgument, content);
        currentArgument = va_arg(theMembers, PICUSHORT);
    }
}

void PicIntSet::addBetween (PICUSHORT from, PICUSHORT to)
{
    if (to < from)
        return;

    if (isNull())
    {
        if (!(content = PicIntSetData::create()))
            return;
    }

    for ( ; from <= to; from++)
        PicIntSetData::add(from, content);
}

void PicIntSet::replace (PICUSHORT oldInt, PICUSHORT newInt)
{
    if (contains(oldInt))
    {
        remove(oldInt);
        add(newInt);
    }
}

// --------------------------------------------------------------------------
//      REMOVE

void PicIntSet::remove (PICUSHORT anInteger)
{
    if (content)
        content->remove(anInteger);
}

void PicIntSet::removeBetween (PICUSHORT from, PICUSHORT to)
{
    if (to < from)
        return;

    for ( ; from <= to; from++)
        content->remove(from);
}

void PicIntSet::setEmpty ()
{
    if (content)
        content->setEmpty();
}

// --------------------------------------------------------------------------
//      SIMPLE ITERATOR

PICBOOL PicIntSet::next (PICUSHORT& current) const
{
    if (!content)
        return boolFalse;

    PICUSHORT theNext = current;

    if (current == endOfList || content->first.inRange(current))
    {
        if (content->first.next(theNext))
        {
            current = theNext;
            return boolTrue;
        }
    }
    theNext = endOfList;
    for (PICUSHORT i = 0; i <= content->first.count; i++)
    {
        PicIntSetBlock& block = content->at(i);
        if (block.start)
        {
            PICUSHORT n = block.inRange(current) ? current : endOfList;
            if (block.next(n) && n < theNext &&
                (n > current || current == endOfList))
                theNext = n;
        }
    }
    current = theNext;
    return theNext != endOfList;
}

PICBOOL PicIntSet::previous (PICUSHORT& current) const
{
    if (!content)
        return boolFalse;

    PICUSHORT thePrev = 0;
    for (PICUSHORT i = 0; i <= content->first.count; i++)
    {
        PicIntSetBlock& block = content->at(i);
        if (block.start)
        {
            PICUSHORT n = block.inRange(current) ? current : endOfList;
            if (block.previous(n) && n > thePrev && n < current)
                thePrev = n;
        }
    }
    if (thePrev == 0)
        return content->first.previous(current);
    current = thePrev;
    return boolTrue;
}

// --------------------------------------------------------------------------
//      STANDARD ITERATOR

void PicIntSet::startIteration (PicIterator& theIterator) const
{
    SET_ITERATE_INDEX(theIterator, (PICULONG)endOfList);
}

PICBOOL PicIntSet::next (PicIterator& theIterator) const
{
    PICUSHORT currentIndex = (PICUSHORT)GET_ITERATE_INDEX(theIterator);
    PICBOOL ok = next(currentIndex);
    SET_ITERATE_INDEX(theIterator, (PICULONG)currentIndex);
    return ok;
}

PICBOOL PicIntSet::previous (PicIterator& theIterator) const
{
    PICUSHORT currentIndex = (PICUSHORT)GET_ITERATE_INDEX(theIterator);
    PICBOOL ok = previous(currentIndex);
    SET_ITERATE_INDEX(theIterator, (PICULONG)currentIndex);
    return ok;
}

PICUSHORT PicIntSet::current (PicIterator& theIterator) const
{
    return (PICUSHORT)GET_ITERATE_INDEX(theIterator);
}

// --------------------------------------------------------------------------
//      SET OPERATORS

PicIntSet& PicIntSet::operator += (const PicIntSet& aSet)
{
    if (aSet.isEmpty())
        ;
    else if (isEmpty())
        *this = aSet;
    else
        PicIntSetData::unite(*aSet.content, content);

    return *this;
}

PicIntSet& PicIntSet::operator *= (const PicIntSet& aSet)
{
    if (aSet.isEmpty())
        setEmpty();
    else if (isEmpty())
        ;
    else
        content->intersect(*aSet.content);

    return *this;
}

PicIntSet& PicIntSet::operator -= (const PicIntSet& aSet)
{
    if (!aSet.isEmpty() && !isEmpty())
        content->subtract(*aSet.content);

    return *this;
}

// --------------------------------------------------------------------------
//      ATTRIBUTES

PICBOOL PicIntSet::contains (PICUSHORT anInteger) const
{
    if (content)
    {
        if (content->first.inRange(anInteger))
            return content->first.contains(anInteger);
        for (PICUSHORT i = 0; i <= content->first.count; i++)
        {
            if (content->at(i).inRange(anInteger))
                return content->at(i).contains(anInteger);
        }
    }
    return boolFalse;
}

PICSIZE PicIntSet::size () const
{
    if (!content)
        return 0;

    PICSIZE aCount = 0;
    PICUSHORT n;

    for (n = 0; n < 32; n++)
        aCount += content->first.contains(n);

    for (PICUSHORT i = 0; i <= content->first.count; i++)
    {
        PicIntSetBlock& block = content->at(i);
        PICUSHORT start = block.start;
        if (start)
            for (n = 0; n < 32; n++)
                aCount += block.contains(n + start);
    }

    return aCount;
}

PICBOOL PicIntSet::isEmpty () const
{
    return content ? content->isEmpty() : boolTrue;
}

PICUSHORT PicIntSet::smallest () const
{
    PICUSHORT smallest = endOfList;

    if (content)
    {
        if (content->first.bits)
        {
            for (PICUSHORT n = 0; n < 32; n++)
                if (content->first.contains(n))
                    return n;
        }
        for (PICUSHORT i = 0; i <= content->first.count; i++)
        {
            PicIntSetBlock& block = content->at(i);
            PICUSHORT start = block.start;
            if (start && start < smallest)
                for (PICUSHORT n = 0; n < 32; n++)
                    if (block.contains(n + start))
                    {
                        smallest = n + start;
                        break;
                    }
        }
    }
    return smallest;
}

PICUSHORT PicIntSet::greatest () const
{
    if (content)
    {
        PICUSHORT greatest = 0;

        if (content->first.bits)
        {
            for (PICSHORT n = 31; n >= 0; n--)
                if (content->first.contains(n))
                {
                    greatest = n;
                    break;
                }
        }
        for (PICUSHORT i = 0; i <= content->first.count; i++)
        {
            PicIntSetBlock& block = content->at(i);
            PICUSHORT start = block.start;
            if (start && start > greatest)
                for (PICSHORT n = 31; n >= 0; n--)
                    if (block.contains(n + start))
                    {
                        greatest = n + start;
                        break;
                    }
        }
        return greatest ?
            greatest : (content->first.contains(0) ? 0 : endOfList);
    }
    return endOfList;
}

PICSIZE PicIntSet::maxSize () const
{
    return content ? (content->first.count + 2) * 32 : 0;
}

