// --------------------------------------------------------------------------
//  INDEXED MAP OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <pixmap.hxx>
#include <psemaph.hxx>
#include <piterimp.hxx>
#include <picerror.hxx>
#include <picbase.erh>

#define DEBUG_CHECK_INDEX(index) \
    if (index == outOfIndex) { \
        ERAISEFMT(BaseCollection, illegalIndex) << index; \
        return boolFalse; }

MAKE_PICABSOBJECTCLASS(PicIxMapObject, PicKeyIxCollect)
MAKE_PICOBJECTCLASS(PicIxMap, PicIxMapObject)
MAKE_PICOBJECTCLASS(PicSortedIxMap, PicIxMapObject)

// --------------------------------------------------------------------------
//      PicIxMapObject
// --------------------------------------------------------------------------
//      ASSIGNMENT

PicObject& PicIxMapObject::operator = (const PicObject& obj)
{
    if (obj.id() >= id())
        return *this = (const PicIxMapObject&)obj;
    else
        return PicKeyIxCollect::operator=(obj);
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicIxMapObject::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return equals((PicIxMapObject&)obj);

    return boolFalse;
}

// --------------------------------------------------------------------------
//      NULL

PICBOOL PicIxMapObject::isNull  () const { return assocs.isNull(); }
void    PicIxMapObject::setNull ()       { assocs.setNull(); }

// --------------------------------------------------------------------------
//      ACCESSING COMPONENTS

PicObject& PicIxMapObject::access (const PicIterator& here)
{
    PICPOSITION accessPos = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocs.contains(accessPos));
    return *assocs[accessPos].object();
}

PicObject* PicIxMapObject::replace (PicIterator& here, PicObject* pobj)
{
    PICPOSITION replaceLoc = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocs.inRange(replaceLoc));
    return assocs.replace(replaceLoc, pobj);
}

PICBOOL PicIxMapObject::findIndex (PICINDEX index, PicIterator& here) const
{
    PICPOSITION ind = location(index);
    SET_ITERATE_INDEX(here, ind);
    return (ind != 0);
}

PICINDEX PicIxMapObject::indexAt (const PicIterator& here) const
{
    PICPOSITION indexPos = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocs.contains(indexPos));
    return assocs[indexPos].index();
}

// --------------------------------------------------------------------------
//      STANDARD ITERATOR

PICBOOL PicIxMapObject::next (PicIterator& i) const
{
    PICPOSITION nextPos = (PICPOSITION)GET_ITERATE_INDEX(i);
    nextPos = assocs.next(nextPos);
    SET_ITERATE_INDEX(i, nextPos);
    return nextPos != 0;
}

PICBOOL PicIxMapObject::previous (PicIterator& i) const
{
    PICPOSITION prevPos = (PICPOSITION)GET_ITERATE_INDEX(i);
    prevPos = assocs.previous(prevPos);
    SET_ITERATE_INDEX(i, prevPos);
    return prevPos != 0;
}

// --------------------------------------------------------------------------
//      DELETING & REMOVING COMPONENTS

void PicIxMapObject::deleteAll ()
{
    assocs.deleteAll();
}

// --------------------------------------------------------------------------
//      ASKING ATTRIBUTES

PICSIZE PicIxMapObject::size () const { return assocs.size(); }

PICINDEX PicIxMapObject::minIndex () const
{
    PICPOSITION pos = assocs.smallest();
    return pos ? assocs[pos].index() : outOfIndex;
}

PICINDEX PicIxMapObject::maxIndex () const
{
    PICPOSITION pos = assocs.greatest();
    return pos ? assocs[pos].index() : outOfIndex;
}


// --------------------------------------------------------------------------
//      PicIxMap
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicIxMap::PicIxMap () : PicIxMapObject() {}

PicIxMap::PicIxMap (PICSIZE initSize) : PicIxMapObject(initSize) {}

PicIxMap::PicIxMap (const PicIxMap& map) : PicIxMapObject(map) {}

PicIxMap::PicIxMap (const PicAssocArray& assocarr) :
    PicIxMapObject(assocarr) {}

PicIxMap::PicIxMap (const PicIxCollect& fromIxCollect) : PicIxMapObject()
{
    *this = fromIxCollect;
}

PicIxMap::~PicIxMap () {}

// --------------------------------------------------------------------------
//      MODIFYING COMPONENTS

PICBOOL PicIxMap::purge (PicIterator& here, PICBOOL inOrder)
{
    PICPOSITION purgeLoc = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocArray().inRange(purgeLoc));
    assocArray().purge(purgeLoc, inOrder);
    return previous(here);
}

PICBOOL PicIxMap::insert (PICINDEX index, PicIterator& here)
{
    DEBUG_CHECK_INDEX(index);

    PICPOSITION loc = assocArray().locationFor(index);
    SET_ITERATE_INDEX(here, loc);
    return (loc != 0);
}

PICBOOL PicIxMap::putPtrAt (PICINDEX index, PicObject* objPtr)
{
    DEBUG_CHECK_INDEX(index);

    PICPOSITION loc = assocArray().locationFor(index);
    if (loc)
        assocArray()[loc].setObject(objPtr);
    return loc != 0;
}

// --------------------------------------------------------------------------
//      PicSortedIxMap
// --------------------------------------------------------------------------
//      CONSTRUCTORS AND DESTRUCTOR

PicSortedIxMap::PicSortedIxMap () :
    PicIxMapObject(), order(PicIxCollect::ascending) {}

PicSortedIxMap::PicSortedIxMap (PICSIZE initSize) :
    PicIxMapObject(initSize), order(PicIxCollect::ascending) {}

PicSortedIxMap::PicSortedIxMap (const PicSortedIxMap& map) :
    PicIxMapObject(map), order(map.order) {}

PicSortedIxMap::PicSortedIxMap (const PicAssocArray& assocarr) :
    PicIxMapObject(assocarr), order(PicIxCollect::ascending) {}

PicSortedIxMap::PicSortedIxMap (const PicIxCollect& fromIxCollect) :
    PicIxMapObject(), order(PicIxCollect::ascending)
{
    *this = fromIxCollect;
}

PicSortedIxMap::~PicSortedIxMap () {}

// --------------------------------------------------------------------------
//      FIND SORTED LOCATION

static PICPOSITION findLoc (
    PicAssocArray& assocs,
    PICINDEX       index,
    PICIXSORTER*   order)
{
    PICSHORT result = 1;

    PICPOSITION loc;
    for (loc = 1; loc <= assocs.lastPos(); loc++)
    {
        PicAssoc& assoc = assocs[loc];
        if (assoc.object())
        {
            result = (*order)(assoc.index(), index);
            if (result >= 0)
                break;
        }
    }

    if (result == 0)
        return loc;
    else if (loc > 1 && !assocs.contains(loc - 1))
    {
        assocs[loc - 1].setIndex(index);
        return loc - 1;
    }
    else if (assocs.insert(loc))
    {
        assocs[loc].setIndex(index);
        return loc;
    }
    else
        return 0;
}

// --------------------------------------------------------------------------
//      MODIFYING COMPONENTS

PICBOOL PicSortedIxMap::purge (PicIterator& here, PICBOOL)
{
    PICPOSITION purgeLoc = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocArray().inRange(purgeLoc));
    assocArray().purge(purgeLoc, boolTrue); // Must retain order
    return previous(here);
}

PICBOOL PicSortedIxMap::insert (PICINDEX index, PicIterator& here)
{
    DEBUG_CHECK_INDEX(index);

    PICPOSITION loc = findLoc(assocArray(), index, order);
    SET_ITERATE_INDEX(here, loc);
    return (loc != 0);
}

PICBOOL PicSortedIxMap::putPtrAt (PICINDEX index, PicObject* objPtr)
{
    PicIterator here;
    if (insert(index, here))
    {
        objPtr = replace(here, objPtr);
        delete objPtr;
        return boolTrue;
    }
    return boolFalse;
}

// --------------------------------------------------------------------------
//      SORTING

static PICIXSORTER* pfSorter;

static int sortCompare (const void* elem1, const void* elem2)
{
//      if (((PicAssoc*)elem1)->isNull() || ((PicAssoc*)elem2)->isNull())
//          return (((PicAssoc*)elem1)->isNull() - ((PicAssoc*)elem2)->isNull());

    PICINDEX i1 = ((PicAssoc*)elem1)->index();
    PICINDEX i2 = ((PicAssoc*)elem2)->index();

    return (*pfSorter)(i1, i2);
}

static PicSemaphore semIxMap;

void PicIxMap::sort (PICIXSORTER* pfCompare)
{
    PicCritSec sec (semIxMap);

    ::pfSorter = pfCompare;
    assocArray().sort(::sortCompare);
}

void PicSortedIxMap::sort (PICIXSORTER* pfCompare)
{
    PicCritSec sec (semIxMap);

    ::pfSorter = pfCompare;
    assocArray().sort(::sortCompare);
    order = pfCompare;
}

PICBOOL PicSortedIxMap::findIndex (PICINDEX index, PicIterator& here) const
{
    PICPOSITION pos = 0;

    if (!isEmpty())
    {
        PicAssoc assoc (index, NULLPTR);

        PicCritSec sec (semIxMap);

        ::pfSorter = order;
        pos = assocArray().findSorted(&assoc, ::sortCompare);

        SET_ITERATE_INDEX(here, pos);
    }
    return pos != 0 && assocArray()[pos].object();
}
