// --------------------------------------------------------------------------
//  LOGIC OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <plogic.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pstring.hxx>
#include <breal.hxx>

MAKE_PICOBJECTCLASS(PicLogic,PicPrimitive)

ERROR_MESSAGES(PicLogic)
PARAMSTR(nullInstance, "Operation on Null instance")
PARAMSTR(deformatError, "String '@*' is not of the given format")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      STATIC VARIABLES & CONSTANTS

const PICSHORT falseValue     = 0;
const PICSHORT trueValue      = 1;
const PICSHORT unknownValue   = 2;
const PICSHORT nullValue      = 3;

static const char* trueIm    = "true";
static const char* falseIm   = "false";
static const char* unknownIm = "unknown";

// --------------------------------------------------------------------------
//     INITIALIZATION

PicLogicIniter::PicLogicIniter (
    const char* trueKeyword,
    const char* falseKeyword,
    const char* unknownKeyword)
    :
    prevTrueKeyword    (trueIm),
    prevFalseKeyword   (falseIm),
    prevUnknownKeyword (unknownIm)
{
    if (trueKeyword)
        trueIm = trueKeyword;
    if (falseKeyword)
        falseIm = falseKeyword;
    if (unknownKeyword)
        unknownIm = unknownKeyword;
}

PicLogicIniter::~PicLogicIniter ()
{
    trueIm    = prevTrueKeyword;
    falseIm   = prevFalseKeyword;
    unknownIm = prevUnknownKeyword;
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS, COPY & ASSIGNMENT

PicLogic::PicLogic () : PicPrimitive(), val(nullValue) {}

PicLogic::PicLogic (const PicLogic& logic) :
    PicPrimitive(), val(logic.val) {}

PicLogic::PicLogic (PICBOOL b) :
    PicPrimitive(), val(b ? trueValue : falseValue) {}

PicLogic::PicLogic (const BaseString& image) :
    PicPrimitive(), val(nullValue)
{
    setImage(image);
}

PicObject& PicLogic::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
        val = ((const PicLogic&)obj).val;
    return *this;
}

PicLogic& PicLogic::operator = (PICBOOL b)
{
    val = b ? trueValue : falseValue;
    return *this;
}

// --------------------------------------------------------------------------
//      LOGIC CONSTANTS

const PicLogic& PicLogic::False ()
{
    static PicLogic aFalse (boolFalse);
    return aFalse;
}

const PicLogic& PicLogic::True ()
{
    static PicLogic aTrue (boolTrue);
    return aTrue;
}

const PicLogic& PicLogic::Unknown ()
{
    static PicLogic anUnknown;
    anUnknown.val = unknownValue;
    return anUnknown;
}

const PicLogic& PicLogic::Null ()
{
    static PicLogic aNull;
    return aNull;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicLogic::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return operator == ((const PicLogic&)obj);
    return boolFalse;
}

PICBOOL PicLogic::operator < (const PicObject& obj) const
{
    if (id() == obj.id())
        return val < ((const PicLogic&)obj).val;

    ERAISEFMT(PicObject, uncomparable)
        >>"OBJ1" << *this
        >>"OBJ2" << obj;
    return boolFalse;
}

// --------------------------------------------------------------------------
//      NULL

PICBOOL PicLogic::isNull () const
{
    return val == nullValue;
}

void PicLogic::setNull ()
{
    val = nullValue;
}

// --------------------------------------------------------------------------
//      DEFAULT 

void PicLogic::setDefault ()
{
    val = unknownValue;
}

// --------------------------------------------------------------------------
//      CONVERSIONS

const char* PicLogic::asChars () const
{
    switch (val)
    {
        case nullValue:  return NULLPTR;
        case falseValue: return falseIm;
        case trueValue:  return trueIm;
        default:         return unknownIm;
    }
}

// --------------------------------------------------------------------------
//      LOGIC OPERATIONS

const PicLogic& PicLogic::operator ! () const
{
    switch (val)
    {
        case nullValue:  ERAISE(PicLogic, nullInstance);
                         return Null   ();
        case falseValue: return True   ();
        case trueValue:  return False  ();
        default:         return Unknown();
    }
}

const PicLogic& PicLogic::operator || (const PicLogic& logic) const
{
    if (val == nullValue || logic.val == nullValue)
    {
        ERAISE(PicLogic, nullInstance);
        return Null();
    }
    else if ((val == falseValue) && (logic.val == falseValue))
        return False();
    else if ((val == trueValue) || (logic.val == trueValue))
        return True();
    else
        return Unknown();
}

const PicLogic& PicLogic::operator && (const PicLogic& logic) const
{
    if (val == nullValue || logic.val == nullValue)
    {
        ERAISE(PicLogic, nullInstance);
        return Null();
    }
    else if ((val == falseValue) || (logic.val == falseValue))
        return False();
    else if ((val == trueValue) && (logic.val == trueValue))
        return True();
    else
        return Unknown();
}

// --------------------------------------------------------------------------
//      SURELY, SURELY NOT, UNCERTAIN

PICBOOL surely (const PicLogic& logic)
{
    return logic.val == trueValue;
}

PICBOOL surelyNot (const PicLogic& logic)
{
    return logic.val == falseValue;
}

PICBOOL uncertain (const PicLogic& logic)
{
    return logic.val == unknownValue;
}

// --------------------------------------------------------------------------
//      TEXTUAL IMAGE

BaseString PicLogic::getImage () const
{
    return BaseString(this->PicLogic::asChars());
}

BaseString PicLogic::getImage (const PicObject& format) const
{
    BaseString image;

    if (!pictureFmt(format, image))
        PicString::format(image = getImage(), format);
    return image;
}

void PicLogic::setImage (const BaseString& imageStr)
{
    BaseString image = imageStr;
    BaseString comp  = trueIm;

    image.trim();

    if (image.isEmpty())
        val = nullValue;
    else if (image.equal(comp))
        val = trueValue;
    else if (image.equal(comp = falseIm))
        val = falseValue;
    else if (image.equal(comp = unknownIm))
        val = unknownValue;
    else
        ERAISEFMT(PicLogic, deformatError) << image;
}

// --------------------------------------------------------------------------

PICBOOL PicLogic::increment (PICSHORT inc)
{
    if (isNull() || val == unknownValue)
        val = falseValue;
    else
        (val += inc) %= 2;
    return boolTrue;
}

