// --------------------------------------------------------------------------
//  NAME-OBJECT ASSOCIATION ARRAY implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
}

#include <pnassarr.hxx>
#include <palloc.hxx>
#include <piterimp.hxx>
#include <psemaph.hxx>
#include <picerror.hxx>

MAKE_PICCLASS(PicNamAssocArray)

// --------------------------------------------------------------------------
//      NAME ENTRY & BUFFER CLASSES
//
//      Name buffer is a data structure consisting of name entries. One
//      entry stores one name. Space is allocated in 4-byte increments
//      and size is stored in 4-byte units, so the maximum size of a name
//      is 1024 bytes (including zero terminator). The actual name string
//      may or may not fill the allocated space - reused name entry may
//      contain shorter name it was originally allocated for. The purging
//      process reclaims the wasted space.
//
//      Name buffer contains some housekeeping information:
//          - Total allocated size for the buffer.
//          - Space actually required by all the name entries. This is
//            the size of the buffer after purging (minus the size of
//            the NameBuffer object itself).
//          - Offset to the first free entry in the end of the buffer.
//            That entry is guaranteed to be null.
//
//      Inline functions:
//
//      nameEntrySize           Calculates the size of the name entry from
//                              the length of the name string (as returned
//                              by strlen-function).
//      initialNameBufferSize   Calculates the initial size for name buffer
//                              when given the amount of space allocated for
//                              objects (as object count).

inline NAMEENTRYSIZE nameEntrySize (PICSIZE nameLength)
{
    return (nameLength + 1 + 3) & ~3;
}

inline PICUSHORT initialNameBufferSize (PICSIZE alloc)
{
    return (alloc > 1024) ? 8192 : (alloc ? alloc << 3 : 32);
}


struct NameEntry
{

    PICFLAG size;      // Reserved size in 4-byte units
    char name [1];  // Name area


// --------------------------------------------------------------------------
//      METHODS
//
//      operator =      Copies the name entry exactly as it is. Does not
//                      check whether there is enough space. The other
//                      variant copies the string assuming it is short
//                      enough to fit.
//      setName         Sets the name entry to contain the given string of
//                      the given size. Assumes there is enough space.
//                      Size shouldn't exceed 1024 bytes.
//      adjustSize      Adjusts the size of the entry to fit exactly for
//                      the current name string.
//      getSize         Returns the size of the entry (not the length of
//                      the name string).
//      required        Returns the total number of bytes required for
//                      this name entry.
//
//      next            Returns a pointer to the next entry assuming the
//                      entries are located contiguously.
//
//      setNull         Entry is null when its size is zero. Thus it
//      isNull          contains only the size field.
//      isEmpty         Entry is empty if it contains an empty string.

    NameEntry& operator = (const NameEntry& copy)
    {
        size = copy.size;
        strcpy(name, copy.name);
        return *this;
    }

    NameEntry& operator = (const char* newName)
    {
        strcpy(name, newName);
        return *this;
    }

    void setName (NAMEENTRYSIZE entrySize, const char* newName)
    {
        size = entrySize >> 2;
        strcpy(name, newName);
    }

    void          adjustSize () { size = nameEntrySize(strlen(name)) >> 2; }
    NAMEENTRYSIZE getSize    () const { return size << 2; }
    PICUSHORT     required   () const { return getSize() + sizeof(PICFLAG); }

    NameEntry* next () { return (NameEntry*)((char*)this + required()); }

    void      setNull  () { size = 0; }
    PICBOOL   isNull   () const { return !size; }
    PICBOOL   isEmpty  () const { return name[0] == '\0'; }

};


struct NameBuffer
{

    PICSIZE   allocated;   // Size allocated for the buffer
    PICSIZE   required;    // Size required for all names
    PICOFFSET freeSpace;   // Offset to the free buffer area
    NameEntry firstName;   // The first name entry


// --------------------------------------------------------------------------
//      METHODS
//
//      entryAt     Returns a reference to the name entry at the given
//                  offset. Doesn't check the validity of the offset.
//      at          Returns the name string at the given offset.
//      offsetOf    Returns the offset of the given name entry.
//      add         Adds the given string of the given entry size to
//                  the end of the buffer. Assumes that there is enough
//                  space in the buffer. Returns the offset of the created
//                  name entry.
//      create      Allocates a new name buffer. The argument is the size
//                  required for name entries (requiredAlloc). Returns
//                  NULLPTR if not successful.
//      free        Frees the given entry. Decrements the required amount.
//      reuse       Reuses an existing name entry. Assumes that the entry
//                  is large enough the contain the given name string.
//      setEmpty    Sets the buffer empty. After this there are no entries.
//      directCopyTo  Copies the name buffer directly, ie. without purging.
//                    New buffer is assumed to be large enough.
//
//      fragmented  Returns whether the name buffer has been considerably
//                  fragmented and should be purged.
//      requiredAlloc Returns the required allocation size for this name
//                    buffer.

    NameEntry& entryAt (PICOFFSET offset)
    {
        return *(NameEntry*)((char*)this + offset);
    }

    const NameEntry& entryAt (PICOFFSET off) const
    {
        return *(const NameEntry*)((char*)this + off);
    }

    const char* at (PICOFFSET offset) const
    {
        return entryAt(offset).name;
    }

    PICOFFSET offsetOf (const NameEntry& entry) const
    {
        return (char*)&entry - (char*)this;
    }

    PICOFFSET add (NAMEENTRYSIZE entrySize, const char* name);

    static NameBuffer* create (PICSIZE requiredSize);

    void free (PICOFFSET offset)
    {
        required -= (nameEntrySize(strlen(at(offset))) + 1);
    }

    void reuse (
        PICOFFSET      offset,
        NAMEENTRYSIZE  entrySize,
        const char* name)
    {
        required += (entrySize + 1);
        entryAt(offset) = name;
    }

    void setEmpty ();

    void directCopyTo (NameBuffer* newBuffer) const
    {
        // Don't copy allocation amount
        memcpy (
            (char*)newBuffer + sizeof(PICSIZE),
            (char*)this + sizeof(PICSIZE),
            allocated - sizeof(PICSIZE));
    }

    PICBOOL fragmented () const
    {
        return (freeSpace - 6) / 1.1 - 32 > required;
    }

    PICSIZE requiredAlloc () const
    {
        return required + sizeof(NameBuffer);
    }

};

PICOFFSET NameBuffer::add (NAMEENTRYSIZE entrySize, const char* name)
{
    PICOFFSET next = freeSpace;
    entryAt(freeSpace).setName(entrySize, name);
    required += (entrySize + 1);
    freeSpace += (entrySize + 1);
    entryAt(freeSpace).setNull();
    return next;
}

void NameBuffer::setEmpty ()
{
    required = 0;
    freeSpace = offsetOf(firstName);
    firstName.setNull();
}

NameBuffer* NameBuffer::create (PICSIZE requiredSize)
{
    PICUSHORT size = picCalculateSize (requiredSize);
    NameBuffer* buffer = (NameBuffer*)picAllocate(size);
    if (MEMORY_ERROR("PicNamAssocArray::NameBuffer::create"))
        return NULLPTR;
    buffer->allocated = size;
    buffer->setEmpty();
    return buffer;
}

// --------------------------------------------------------------------------
//      PicNamAssocArray
// --------------------------------------------------------------------------
//      PRIVATE METHODS

PICBOOL PicNamAssocArray::allocateNameBuffer (PICSIZE newSize)
{
    NameBuffer* newBuffer;

    if (!names || names->allocated < newSize)
    {
        picFree(names);
        if (newBuffer = NameBuffer::create(newSize))
            names = newBuffer;
        else
            return boolFalse;
    }
    return boolTrue;
}

void PicNamAssocArray::copyNamesFrom (const PicNamAssocArray& copy)
{
    if (copy.names->fragmented())
    {
        if (allocateNameBuffer(copy.names->requiredAlloc()))
            copy.copyNamesTo(*this, names);
    }
    else if (allocateNameBuffer(copy.names->allocated))
        copy.names->directCopyTo(names);
}

void PicNamAssocArray::copyNamesTo (
    PicNamAssocArray& array,
    NameBuffer*       dest) const
{
    NameEntry* currentEntry = &(dest->firstName);
    dest->required = 0;

    for (PICPOSITION ind = 1; ind <= ub; ind++)
    {
        if (THIS[ind].isNull())
            array[ind].setIndex(0);
        else
        {
            *currentEntry = names->entryAt(THIS[ind].index());
            currentEntry->adjustSize();
            dest->required += currentEntry->required();
            array[ind].setIndex(array.names->offsetOf(*currentEntry));
            currentEntry = currentEntry->next();
        }
    }

    currentEntry->setNull();
    dest->freeSpace = dest->offsetOf(*currentEntry);
}

PICBOOL PicNamAssocArray::resizeNameBuffer (PICSIZE increment)
{
    if (::maxNewSize - names->requiredAlloc() < increment)
    {
        ERAISEFMT(BaseCollection, full) << "(name buffer overflow)";
        return boolFalse;
    }

    NameBuffer* newBuffer = NULLPTR;

    if (names->fragmented() ||
        ::maxNewSize - names->allocated < increment)
    {
        newBuffer = NameBuffer::create(names->requiredAlloc() + increment);
        if (newBuffer)
            copyNamesTo(*this, newBuffer);
        else
            return boolFalse;
    }
    else
    {
        newBuffer = NameBuffer::create(names->allocated + increment);
        if (newBuffer)
            names->directCopyTo(newBuffer);
        else
            return boolFalse;
    }
    picFree(names);
    names = newBuffer;
    return boolTrue;
}

PICOFFSET PicNamAssocArray::addName (
    NAMEENTRYSIZE  nameSize,
    const char* name)
{
    if (names->allocated - names->freeSpace < nameSize + 4)
    {
        if (!resizeNameBuffer(nameSize + 4))
            return 0;
    }
    return names->add(nameSize, name);
}

// --------------------------------------------------------------------------
//      NAME BUFFER

void PicNamAssocArray::purgeNameBuffer () const
{
    #define SELF ((PicNamAssocArray*)this)

    if (names && names->fragmented())
    {
        NameBuffer* newBuffer = NameBuffer::create(names->requiredAlloc());

        if (newBuffer)
        {
            copyNamesTo(*SELF, newBuffer);
            picFree(SELF->names);
            SELF->names = newBuffer;
        }
    }

    #undef SELF
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS & DESTRUCTOR

PicNamAssocArray::PicNamAssocArray (const PicNamAssocArray& copy) :
    PicAssocArray(copy), names(NULLPTR)
{
    if (copy.names)
        copyNamesFrom(copy);
}

PicNamAssocArray::~PicNamAssocArray ()
{
    if (names)
        picFree(names);
}

PicNamAssocArray& PicNamAssocArray::operator = (const PicNamAssocArray& copy)
{
    PicAssocArray::operator = (copy);

    if (copy.names)
        copyNamesFrom(copy);
    else
    {
        picFree(names);
        names = NULLPTR;
    }

    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISON

inline PICBOOL assocsEqual (
    const PicAssoc& assoc1, const PicAssoc& assoc2,
    const NameBuffer* names1, const NameBuffer* names2)
{
    return (assoc1.isNull() && assoc2.isNull()) ||
           (!strcmp(names1->at(assoc1.index()), names2->at(assoc2.index()))
            && assoc1.object() && assoc2.object() &&
            (*assoc1.object() == *assoc2.object()));
}

PICBOOL PicNamAssocArray::equals (const PicNamAssocArray& anArray) const
{
    PICPOSITION thisInd  = 1;
    PICPOSITION arrayInd = 1;

    do {

        while (thisInd <= ub && THIS[thisInd].isNull())
            thisInd++;

        while (arrayInd <= anArray.ub && anArray[arrayInd].isNull())
            arrayInd++;

        if (thisInd > ub)
            return arrayInd > anArray.ub;
        if (arrayInd > anArray.ub)
            return thisInd > ub;

    } while (assocsEqual (
        THIS[thisInd++], anArray[arrayInd++], names, anArray.names));

    return boolFalse;
}

// --------------------------------------------------------------------------
//      ADD, REPLACE & REMOVE

PICPOSITION PicNamAssocArray::putName (const char* name)
{
    PICPOSITION   firstFree  = 0;
    PicAssoc*     current    = NULLPTR;
    PICPOSITION   bestFitInd = 0;
    PICSHORT      bestFit    = SHRT_MAX;
    PICSHORT      stry       = 0;
    NAMEENTRYSIZE nameSize   = nameEntrySize(strlen(name));

    if (!names && !(names = NameBuffer::create(initialNameBufferSize(alloc))))
        return 0;

    for (PICPOSITION ind = ub; ind >= 1; ind--)
    {
        current = &THIS[ind];
        if (current->isNull())
        {
            firstFree = ind;
            if (current->index())
            {
                stry = names->entryAt(current->index()).getSize() - nameSize;
                if (stry >= 0 && stry < bestFit)
                {
                    bestFitInd = ind;
                    bestFit    = stry;
                }
            }
        }
        else if (strcmp(names->at(current->index()), name) == 0)
        {
            return ind;
        }
    }

    if (bestFitInd)
    {
        names->reuse(THIS[bestFitInd].index(), nameSize, name);
        return bestFitInd;
    }
    else if (firstFree || (firstFree = insert()))
    {
        PICOFFSET offset = addName(nameSize, name);
        if (offset)
        {
            THIS[firstFree].setIndex(offset);
            return firstFree;
        }
    }
    return 0;
}

PICPOSITION PicNamAssocArray::putName (const char* name, PICNMSORTER* order)
{
    PICPOSITION   loc;
    NAMEENTRYSIZE nameSize = nameEntrySize(strlen(name));

    if (!names && !(names = NameBuffer::create(initialNameBufferSize(alloc))))
        return 0;

    if (searchName(name, loc, order))
        return loc;
    else
    {
        if (loc <= ub)
        {
            if (THIS[loc].isNull())
                ;
            else if (loc > 1 && THIS[loc - 1].isNull())
                loc--;
            else if (!insert(loc))
                return 0;
        }
        else if (!put(loc))
            return 0;

        PICOFFSET offset = THIS[loc].index();

        if (offset)
        {
            names->free(offset);

            if (names->entryAt(offset).getSize() >= nameSize)
            {
                names->reuse(offset, nameSize, name);
                return loc;
            }
        }

        if (offset = addName(nameSize, name))
        {
            THIS[loc].setIndex(offset);
            return loc;
        }
    }
    return 0;
}

void PicNamAssocArray::deleteAll ()
{
    PicAssocArray::deleteAll();

    if (names)
    {
        picFree(names);
        names = NULLPTR;
    }
}

void PicNamAssocArray::removeAll ()
{
    PicAssocArray::removeAll();

    if (names)
    {
        picFree(names);
        names = NULLPTR;
    }
}

void PicNamAssocArray::purge (PICPOSITION loc, PICBOOL inOrder)
{
    PICOFFSET offset = THIS[loc].index();
    if (offset)
    {
        names->free(offset);
        THIS[loc].setIndex(0);
    }
    PicAssocArray::purge(loc, inOrder);
}

void PicNamAssocArray::purge (PICBOOL inOrder)
{
    PICPOSITION loc = 0;
    while (loc = findNext(loc, info->isItemNull, boolTrue))
    {
        PICOFFSET offset = THIS[loc].index();
        if (offset)
        {
            names->free(offset);
            THIS[loc].setIndex(0);
        }
    }
    PicAssocArray::purge(inOrder);
}

// --------------------------------------------------------------------------
//      SEARCH

const char* PicNamAssocArray::nameOf (const PicObject& obj) const
{
    PICPOSITION index = locationOf(obj);
    if (index && names)
        return names->at(THIS[index].index());
    return NULLPTR;
}

PICBOOL PicNamAssocArray::contains (const PicNamAssocArray& anArray) const
{
    for (PICPOSITION ind = 1; ind <= anArray.ub; ind++)
    {
        if (!anArray[ind].isNull())
        {
            PICPOSITION index = locationOfName(anArray.nameAt(ind));
            if (index == 0)
                return boolFalse;
            if (!(*THIS[index].object() == *anArray[ind].object()))
                return boolFalse;
        }
    }
    return boolTrue;
}

// --------------------------------------------------------------------------
//      SORT

// A global variable is needed, because the qsort-function of C run-time
// library cannot pass contextual information. Here the needed information
// is the name buffer. 

static PICNMSORTER*      pfSorter;
static const NameBuffer* sortedArrayNames;

static int sortCompare (const void* elem1, const void* elem2)
{
    static BaseString n1, n2;

//      if (((PicAssoc*)elem1)->isNull() || ((PicAssoc*)elem2)->isNull())
//          return (((PicAssoc*)elem1)->isNull() - ((PicAssoc*)elem2)->isNull());

    n1 = sortedArrayNames->at(((PicAssoc*)elem1)->index());
    n2 = sortedArrayNames->at(((PicAssoc*)elem2)->index());

    return (*pfSorter)(n1, n2);
}

static PicSemaphore semPicNamAssocArray;

void PicNamAssocArray::sort (PICNMSORTER* pfCompare)
{
    PicCritSec sec(semPicNamAssocArray);

    ::pfSorter = pfCompare;
    ::sortedArrayNames = this->names;
    PicAssocArray::sort(::sortCompare);
}

// --------------------------------------------------------------------------
//      INTERNAL LOCATIONS

const char* PicNamAssocArray::nameAt (PICPOSITION location) const
{
    if (!isEmpty())
    {
        PICOFFSET offset = THIS[location].index();
        if (offset)
            return names->at(offset);
    }
    return NULLPTR;
}

PICPOSITION PicNamAssocArray::locationOfName (const char* name) const
{
    if (!names)
        return 0;

    const PicAssoc* current;

    for (PICPOSITION ind = 1; ind <= ub; ind++)
    {
        current = &THIS[ind];
        if (!current->isNull() && !strcmp(names->at(current->index()), name))
            return ind;
    }

    return 0;
}

PICBOOL PicNamAssocArray::searchName (
    const char* name,
    PICPOSITION&   loc,
    PICNMSORTER*   order) const
{
    PICOFFSET offset;
    PICSHORT  result;
    BaseString n1, n2 = name;

    for (loc = 1; loc <= ub; loc++)
    {
        if (offset = THIS[loc].index())
        {
            result = (*order)(n1 = names->at(offset), n2);
            if (result >= 0)
                return result == 0;
        }
        else
            break;
    }
    return boolFalse;
}

