// --------------------------------------------------------------------------
//  NAMED MAP OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
}

#include <pnmmap.hxx>
#include <picerror.hxx>
#include <bstring.hxx>
#include <piterimp.hxx>
#include <picbase.erh>

#define DEBUG_CHECK_NAME(pname) \
    if (!(pname)) { \
        ERAISEFMT(BaseCollection, illegalIndex) \
            << ("\"" & BaseString(pname) & "\""); \
        return boolFalse; }

MAKE_PICABSOBJECTCLASS(PicNmMapObject, PicNmCollect)
MAKE_PICOBJECTCLASS(PicNmMap, PicNmMapObject)
MAKE_PICOBJECTCLASS(PicSortedNmMap, PicNmMapObject)

// --------------------------------------------------------------------------
//      PicNmMapObject
// --------------------------------------------------------------------------
//      ASSIGNMENT

PicObject& PicNmMapObject::operator = (const PicObject& obj)
{
    if (obj.id() >= id())
        return *this = (const PicNmMapObject&)obj;
    else
        return PicNmCollect::operator=(obj);
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicNmMapObject::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return equals((PicNmMapObject&)obj);

    return boolFalse;
}

// --------------------------------------------------------------------------
//      NULL

PICBOOL PicNmMapObject::isNull  () const { return assocs.isNull(); }
void    PicNmMapObject::setNull ()       { assocs.setNull(); }

// --------------------------------------------------------------------------
//      ACCESSING COMPONENTS

PicObject& PicNmMapObject::access (const PicIterator& here)
{
    PICPOSITION accessPos = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocs.contains(accessPos));
    return *assocs[accessPos].object();
}

PicObject* PicNmMapObject::replace (PicIterator& here, PicObject* pobj)
{
    PICPOSITION replaceLoc = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocs.inRange(replaceLoc));
    return assocs.replace(replaceLoc, pobj);
}

PICBOOL PicNmMapObject::findName (const char* name, PicIterator& here) const
{
    PICPOSITION ind = location(name);
    SET_ITERATE_INDEX(here, ind);
    return (ind != 0);
}

const char* PicNmMapObject::nameAt (const PicIterator& here) const
{
    PICPOSITION namePos = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocs.contains(namePos));
    return assocs.nameAt(namePos);
}

// --------------------------------------------------------------------------
//      STANDARD ITERATOR

PICBOOL PicNmMapObject::next (PicIterator& i) const
{
    PICPOSITION nextPos = (PICPOSITION)GET_ITERATE_INDEX(i);
    nextPos = assocs.next(nextPos);
    SET_ITERATE_INDEX(i, nextPos);
    return nextPos != 0;
}

PICBOOL PicNmMapObject::previous (PicIterator& i) const
{
    PICPOSITION prevPos = (PICPOSITION)GET_ITERATE_INDEX(i);
    prevPos = assocs.previous(prevPos);
    SET_ITERATE_INDEX(i, prevPos);
    return prevPos != 0;
}

// --------------------------------------------------------------------------
//      DELETING & REMOVING COMPONENTS

void PicNmMapObject::deleteAll ()
{
    assocs.deleteAll();
}

// --------------------------------------------------------------------------
//      ASKING ATTRIBUTES

PICSIZE PicNmMapObject::size () const { return assocs.size(); }

const char* PicNmMapObject::minName () const
{
    PICPOSITION pos = assocs.smallest();
    return pos ? assocs.nameAt(pos) : NULLPTR;
}

const char* PicNmMapObject::maxName () const
{
    PICPOSITION pos = assocs.greatest();
    return pos ? assocs.nameAt(pos) : NULLPTR;
}


// --------------------------------------------------------------------------
//      PicNmMap
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicNmMap::PicNmMap () : PicNmMapObject() {}

PicNmMap::PicNmMap (PICSIZE initSize) : PicNmMapObject(initSize) {}

PicNmMap::PicNmMap (const PicNmMap& map) : PicNmMapObject(map) {}

PicNmMap::PicNmMap (const PicNamAssocArray& assocarr) :
    PicNmMapObject(assocarr) {}

PicNmMap::PicNmMap (const PicNmCollect& fromNmCollect) : PicNmMapObject()
{
    *this = fromNmCollect;
}

PicNmMap::~PicNmMap () {}

// --------------------------------------------------------------------------
//      MODIFYING COMPONENTS

PICBOOL PicNmMap::purge (PicIterator& here, PICBOOL inOrder)
{
    PICPOSITION purgeLoc = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocArray().inRange(purgeLoc));
    assocArray().purge(purgeLoc, inOrder);
    return previous(here);
}

PICBOOL PicNmMap::insert (const char* name, PicIterator& here)
{
    DEBUG_CHECK_NAME(name);

    PICPOSITION loc = assocArray().putName(name);
    SET_ITERATE_INDEX(here, loc);
    return (loc != 0);
}

PICBOOL PicNmMap::putPtrAt (const char* name, PicObject* objPtr)
{
    DEBUG_CHECK_NAME(name);

    PICPOSITION loc = assocArray().putName(name);
    if (loc)
        assocArray()[loc].setObject(objPtr);
    return loc != 0;
}

// --------------------------------------------------------------------------
//      PicSortedNmMap
// --------------------------------------------------------------------------
//      CONSTRUCTORS AND DESTRUCTOR

PicSortedNmMap::PicSortedNmMap () :
    PicNmMapObject(), order(PicNmCollect::ascending) {}

PicSortedNmMap::PicSortedNmMap (PICSIZE initSize) :
    PicNmMapObject(initSize), order(PicNmCollect::ascending) {}

PicSortedNmMap::PicSortedNmMap (const PicSortedNmMap& map) :
    PicNmMapObject(map), order(map.order) {}

PicSortedNmMap::PicSortedNmMap (const PicNamAssocArray& assocarr) :
    PicNmMapObject(assocarr), order(PicNmCollect::ascending) {}

PicSortedNmMap::PicSortedNmMap (const PicNmCollect& fromNmCollect) :
    PicNmMapObject(), order(PicNmCollect::ascending)
{
    *this = fromNmCollect;
}

PicSortedNmMap::~PicSortedNmMap () {}

// --------------------------------------------------------------------------
//      ACCESSING COMPONENTS

PICBOOL PicSortedNmMap::findName (
    const char* name,
    PicIterator&   here) const
{
    PICPOSITION loc;

    if (assocArray().searchName(name, loc, order))
    {
        SET_ITERATE_INDEX(here, loc);
        return !assocArray()[loc].isNull();
    }
    return boolFalse;
}

// --------------------------------------------------------------------------
//      MODIFYING COMPONENTS

PICBOOL PicSortedNmMap::purge (PicIterator& here, PICBOOL)
{
    PICPOSITION purgeLoc = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(assocArray().inRange(purgeLoc));
    assocArray().purge(purgeLoc, boolTrue); // Must retain order
    return previous(here);
}

PICBOOL PicSortedNmMap::insert (const char* name, PicIterator& here)
{
    DEBUG_CHECK_NAME(name);

    PICPOSITION loc = assocArray().putName(name, order);
    SET_ITERATE_INDEX(here, loc);
    return (loc != 0);
}

PICBOOL PicSortedNmMap::putPtrAt (const char* name, PicObject* objPtr)
{
    DEBUG_CHECK_NAME(name);

    PICPOSITION loc = assocArray().putName(name, order);
    if (loc)
    {
        objPtr = assocArray().replace(loc, objPtr);
        delete objPtr;
        return boolTrue;
    }
    return boolFalse;
}
