// --------------------------------------------------------------------------
//  STRING OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <pstring.hxx>
#include <pcollect.hxx>
#include <pinteger.hxx>
#include <picerror.hxx>
#include <piterate.hxx>
#include <picbase.erh>

MAKE_PICOBJECTCLASS(PicString,PicPrimitive)

ERROR_MESSAGES(PicString)
PARAMSTR(formatError, "Undefined format")
PARAMSTR(charSetViolation, "@* not restricted to defined character set")
PARAMSTR(notAlphabetic, "@* is not alphabetic")
PARAMSTR(notNumeric, "@* is not numeric")
PARAMSTR(notAlphanumeric, "@* is not alphanumeric")
PARAMSTR(tooShort, "String is shorter than the minimum of @*")
PARAMSTR(emptyString, "String is empty")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      CONSTRUCTORS, COPY & ASSIGNMENT

PicString::PicString () : PicPrimitive(), BaseString() {}

PicString::PicString(const char* chars) :
    PicPrimitive(), BaseString(chars) {}

PicString::PicString (char chr, PICSIZE times) :
    PicPrimitive(), BaseString(chr, times) {}

PicString::PicString (const BaseString& string) :
    PicPrimitive(), BaseString(string) {}

PicString::PicString (const PicString& string) :
    PicPrimitive(), BaseString(string) {}

PicObject& PicString::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
        BaseString::operator = ((const PicString&)obj);
    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicString::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return BaseString::operator == ((const PicString&)obj);

    return boolFalse;
}

PICBOOL PicString::operator < (const PicObject& obj) const
{
    if (classId() == obj.id())
        return BaseString::operator < ((const PicString&)obj);

    ERAISEFMT(PicObject, uncomparable)
        >>"OBJ1" << *this
        >>"OBJ2" << obj;
    return boolFalse;
}

// --------------------------------------------------------------------------
//      CONVERSIONS

char PicString::asChar () const
{
    if (BaseString::size() == 1)
        return BaseString::first ();

    ERAISEFMT(BaseString, sizeError) << BaseString::size();
    PicError::cumulate (
        "\nNot one-character string");
    return '\0';
}

// --------------------------------------------------------------------------
//      TEXTUAL IMAGE

BaseString PicString::getImage (const PicObject& format) const
{
    BaseString image;

    if (!pictureFmt(format, image))
        image = PicString::format(image = BaseString::getImage(), format);
    return image;
}

void PicString::setImage (const PicObject& fmt, const BaseString& image)
{
    BaseString::operator = (image);
    if (!fmt.IS_A(PicString))
        format(*this, fmt);
}

BaseString& PicString::format (BaseString& image, const PicObject& format)
{
    PICSIZE          size;
    char          filler;
    PicJustification justify;

    if (!format.contains(SIZE))
        return image;

    size = format.at(SIZE).asInt();

    if (format.contains(FILLER))
        filler = format.at(FILLER).asChar();
    else
        filler = ' ';

    if (format.contains(JUSTIFY))
        switch (format.at(JUSTIFY).asInt())
        {
            case FORMAT_LEFT:
                justify = BaseString::toLeft;
                break;
            case FORMAT_CENTER:
                justify = BaseString::center;
                break;
            case FORMAT_RIGHT:
                justify = BaseString::toRight;
                break;
            default:
                ERAISE(PicString, formatError);
                justify = BaseString::toRight;
                break;
        }
    else
        justify = toRight;

    return image.format(size,justify,filler);
}

// --------------------------------------------------------------------------
//      SET TYPED VALUE

PICBOOL PicString::setTypedValue (const PicObject& constraints)
{
    if (isNull())
        return boolTrue;

    PicIterator here;
    const PicIxCollect& cntr = AS(PicIxCollect, constraints);

    if (cntr.find(MINSIZE, here))
    {
        PICSIZE minSize = cntr.at(here).asInt();
        if (size() < minSize)
        {
            if (isEmpty())
                ERAISE(PicString, emptyString);
            else
                ERAISEFMT(PicString, tooShort) << minSize;
            return boolFalse;
        }
    }

    if (cntr.find(CHARSET, here))
    {
        if (cntr.at(here).virtId() == PicString::classId())
        {
            const PicString& charSet = AS(PicString, cntr.at(here));
            if (!containsChars(charSet))
            {
                ERAISEFMT(PicString, charSetViolation) << *(BaseString*)this;
                return boolFalse;
            }
        }
        else switch (cntr.at(here).asInt())
        {
            case ALPHABETIC:
                if (!isAlphabetic())
                {
                    ERAISEFMT(PicString, notAlphabetic) << *(BaseString*)this;
                    return boolFalse;
                }
                break;

            case NUMERIC:
                if (!isNumeric())
                {
                    ERAISEFMT(PicString, notNumeric) << *(BaseString*)this;
                    return boolFalse;
                }
                break;

            case ALPHANUMERIC:
                if (!isAlphanumeric())
                {
                    ERAISEFMT(PicString, notAlphanumeric) << *(BaseString*)this;
                    return boolFalse;
                }
                break;
        }
    }

    if (cntr.find(MAXSIZE, here))
    {
        PICSIZE add, maxSize = cntr.at(here).asInt();
        if (size() > maxSize)
            deleteAfter(maxSize);
        else if (cntr.find(CHARLEADER, here) && (add = maxSize - size()) > 0)
        {
            prepend(cntr.at(here).asChar(), add);
        }
    }

    if (cntr.find(CHARCASE, here))
    {
        switch (cntr.at(here).asInt())
        {
            case UPPERCASE:
                makeUppercase(); break;
            case LOWERCASE:
                makeLowercase(); break;
            case PROPER:
                makeProper();    break;
        }
    }

    return boolTrue;
}
