// --------------------------------------------------------------------------
//  STRING STREAM implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
}

#include <pstrstm.hxx>
#include <picerror.hxx>

MAKE_PICSTREAMCLASS(PicInStringStream,PicStream)
MAKE_PICSTREAMCLASS(PicStringStream,PicInStringStream)
MAKE_PICSTREAMCLASS(PicInCharStream,PicStream)
MAKE_PICSTREAMCLASS(PicCharStream,PicInCharStream)

// --------------------------------------------------------------------------
//      PicInStringStream
// --------------------------------------------------------------------------

PicInStringStream::PicInStringStream (const BaseString& str) :
    PicStream(), stream((BaseString*)&str), current(1) {}

PicInStringStream::~PicInStringStream () {}

PICSIZE PicInStringStream::read (void* to, PICSIZE length)
{
    if (current == 0 || current > stream->size())
    {
        current = 0;
        return 0;
    }

    PICSIZE sz = stream->size() - (current - 1);

    if (length < sz)
        sz = length;       // Something but not so much

    char* start = (char*)stream->chars() + (current - 1);

    memcpy(to, start, sz);
    current += sz;

    return sz;
}

char* PicInStringStream::readAhead (PICSIZE& amount)
{
    if (current == 0 || current > stream->size())
    {
        current = 0;
        return NULLPTR;
    }
    amount = stream->size() - (current - 1);
    return (char*)stream->chars() + (current - 1);
}

void PicInStringStream::moveAhead (PICSIZE amount)
{
    if (amount > stream->size() - current)
        current = stream->size() + 1;
    else
        current += amount;
}

PICSTATUS PicInStringStream::status () const
{
    return readable | positionable | (current == 0 ? ended : 0);
}

BaseString PicInStringStream::name () const
{
    return BaseString(id().className()) & "(" & stream->before(10) & "...)";
}

void PicInStringStream::move (PICLONG offset, PicStreamPosition startAt)
{
    current = (PICPOSITION)movePointer (
        current - 1, stream->size(), offset, startAt) + 1;
}

// --------------------------------------------------------------------------
//      PicStringStream
// --------------------------------------------------------------------------

PicStringStream::PicStringStream (BaseString& str) :
    PicInStringStream(str) {}

PICSIZE PicStringStream::write (const void *from, PICSIZE length)
{
    PICSIZE origSize = stream->size();

    if (current > origSize)
        stream->append(' ', length);
    else
        stream->insert(current, ' ', length);

    PICSIZE size = stream->size() - origSize;

    if (size)
    {
        char *pCh = (char*)stream->chars() + (current - 1);

        current += size;

        memcpy(pCh, from, size);
    }

    return size;
}

PICSTATUS PicStringStream::status () const
{
    return writeable | readable | positionable | (current == 0 ? ended : 0);
}

// --------------------------------------------------------------------------
//      PicInCharStream
// --------------------------------------------------------------------------

PicInCharStream::PicInCharStream (const char* buf, PICSIZE len) :
    PicStream(), stream((char*)buf), stmSize(len), current(0) {}

PICSIZE PicInCharStream::read (void* to, PICSIZE length)
{
    if (current >= stmSize)
    {
        current = 0xffff;
        return 0;
    }

    PICSIZE sz = stmSize - current;

    if (length < sz)
        sz = length;       // Something but not so much

    memcpy(to, stream + current, sz);
    current += sz;

    return sz;
}

char* PicInCharStream::readAhead (PICSIZE& amount)
{
    if (current >= stmSize)
    {
        current = 0xffff;
        return NULLPTR;
    }
    amount = stmSize - current;
    return stream + current;
}

void PicInCharStream::moveAhead (PICSIZE amount)
{
    if (amount >= stmSize - current)
        current = stmSize;
    else
        current += amount;
}

PICSTATUS PicInCharStream::status () const
{
    PICSTATUS result = positionable;
    if (stream)
        result |= readable;
    if (current == 0xffff || !stream)
        result |= ended;
    return result;
}

BaseString PicInCharStream::name () const
{
    PicStringText text;

    text("@*") << id().className();

    return text.flush();
}

void PicInCharStream::move (PICLONG offset, PicStreamPosition startAt)
{
    current = (PICOFFSET)movePointer(current, stmSize, offset, startAt);
}

// --------------------------------------------------------------------------
//      PicCharStream
// --------------------------------------------------------------------------

PicCharStream::PicCharStream (char* buf, PICSIZE len) :
    PicInCharStream(buf, len) {}

PICSIZE PicCharStream::write (const void *from, PICSIZE length)
{
    PICSIZE sz = stmSize - current;
    if (sz > length)
        sz = length;

    if (sz)
    {
        memcpy(stream + current, from, sz);
        current += sz;
    }

    return sz;
}

PICSTATUS PicCharStream::status () const
{
    return PicInCharStream::status() | (stream ? writeable : 0);
}
