// --------------------------------------------------------------------------
//  PicVirtObject implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
}

#include <picbase.erh>
#include <pvirtobj.hxx>
#include <picerror.hxx>

MAKE_PICCLASS(PicClass)
MAKE_PICCLASS(PicVirtClass)

ERROR_MESSAGES(PicVirtObject)
PARAMSTR(typeMismatch, "Expected CLASS@*, found IMAGE@*")
PARAMSTR(nullPointer, "Using NULL-pointer (class @*)")
PARAMSTR(undefinedOperation, "Undefined operation METHOD@* for class CLASS@*")
END_PARAM_STRINGS

const PicClass& PicClass::nullClass ()
{
    static const PicClass null = { "NULL", 0 };
    return null;
}

// --------------------------------------------------------------------------
//      CONSTRUCTOR & DESTRUCTOR

PicVirtClass::PicVirtClass (PicVirtClassInfo* info) :
    pBase          (NULLPTR),
    pDerived       (NULLPTR)
{
    pClassName = info->pClassName;
    usInstSize = info->usInstSize;

    pBase = (PicVirtClass*)&info->pfBase();

    if (!pBase->pDerived)
    {   // This is the first derived class:
        pBase->pDerived = pNextSibling = pPrevSibling = this;
    }
    else
    {   // This is a later derived class:
        pPrevSibling = pBase->pDerived;
        pNextSibling = pPrevSibling->pNextSibling;
        pPrevSibling->pNextSibling = this;
        pNextSibling->pPrevSibling = this;
    }
}

PicVirtClass::~PicVirtClass ()
{
    if (pBase)
    {
        if (pBase->pDerived == this)
        {   // We are the mark node:
            pBase->pDerived = (pNextSibling == this) ? NULLPTR : pNextSibling;
        }
        pPrevSibling->pNextSibling = pNextSibling;
        pNextSibling->pPrevSibling = pPrevSibling;
    }
}

// --------------------------------------------------------------------------
//      CLASS DIRECTORY

PICBOOL PicVirtClass::hasAncestor (const PicVirtClass& vc) const
{
    // Search parent tree recursively:

    if (!pBase)
        return boolFalse;
    else if (pBase == &vc)
        return boolTrue;
    else
        return pBase->hasAncestor(vc);
}

const PicVirtClass* PicVirtClass::derived (const char* className) const
{
    const PicVirtClass *pder, *pder2;

    for (pder = firstDerived(); pder; pder = pder->nextSibling())
    {
        if (strcmp(pder->className(), className) == 0)
            return pder;
        else if (pder2 = pder->derived(className))
            return pder2;
    }
    return NULLPTR;
}

// --------------------------------------------------------------------------
//      TYPE CHECKING

void PicVirtClass::nassert (const PicVirtObject& found  DEBUG_FILE_LINE) const
{
    if (found.virtId().isKindOf(*this))
        return;

    found.raiseTypeMismatch(className()  DEBUG_FILE_LINE_IMP);
    FATAL_EXIT();
}

void PicVirtClass::nassert (const PicVirtObject* found  DEBUG_FILE_LINE) const
{
    if (found)
    {
        nassert(*found  DEBUG_FILE_LINE_IMP);
        return;
    }
    ERAISEFMT_LINE_FILE (
        PicVirtObject::errors,
        PicVirtObject::classId,
        PicVirtObject::nullPointer,
        lineNumber, fileName) << className();
    FATAL_EXIT();
}

PICBOOL PicVirtClass::typeGuard (
    const PicVirtObject& found
    DEBUG_FILE_LINE) const
{
    if (found.virtId().isKindOf(*this))
        return boolTrue;

    found.raiseTypeMismatch(className()  DEBUG_FILE_LINE_IMP);
    return boolFalse;
}

// --------------------------------------------------------------------------
//      DESTRUCTOR

PicVirtObject::~PicVirtObject () {}

// --------------------------------------------------------------------------
//      CLASS ID

const PicVirtClass& PicVirtObject::classId ()
{
    static PicVirtClass id;
    return id;
}

// --------------------------------------------------------------------------
//      RAISING ERRORS

void PicVirtObject::raiseTypeMismatch (
    const char* expectedClass
    DEBUG_FILE_LINE) const
{
    ERAISEFMT_LINE_FILE (
        PicVirtObject::errors,
        PicVirtObject::classId,
        typeMismatch, lineNumber, fileName)
        >>"CLASS" << expectedClass
        >>"IMAGE" << virtId().className();
}

void PicVirtObject::undefined (const char* method) const
{
    ERAISEFMT(PicVirtObject, undefinedOperation)
        >>"CLASS"  << virtId().className()
        >>"METHOD" << method;
}

// --------------------------------------------------------------------------

const char* picParamString (const PicNumStringPair* strs, PICUSHORT id)
{
    if(!strs) 
        return NULLPTR;
    
    while(!(strs->str==NULLPTR && strs->num==0)){
        if(strs->num == id)
            return strs->str;
        strs++;
    }
    return NULLPTR;
}

