// **********************************************************************
//  PicGUI class definition
//  Part of : PicGUI
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICGUI_HXX
#define PICGUI_HXX

#include "bstring.hxx"
#include "bstrarr.hxx"
#include "picbase.hxx"
#include "parray.hxx"
#include "pstring.hxx"
#include "pinteger.hxx"
#include "tk.h"

static Tcl_Interp *interpreter;

/**
  * PicGUI interface class.
  * PicGUI contains various functions called by 
  * PICClient system and Tcl/TK to provide communication between them.
  *
  * @version 0.1 
  * author Timo Virtanen
  **/

class PicGUI
{
public:

//  *  *  *  *  * NetPicSearch added functionality *  *  *  *  * :

  /**
    * This function prints the info text if the server failed to complete the given request
    *
    **/
  static PICBOOL FailedToCompleteRequest();
 
  /**
    * This function prints the info text if there is no more queryresults
    *
    **/
  static PICBOOL NoMoreQueryResults();

  /**
    * This function prints the info text if there is no query matches
    *
    **/
  static PICBOOL NoQueryMatches();

  /**
    * This function prints the info text if the server succeeded to complete the given request
    *
    **/
  static PICBOOL OperationComplete();

  /**
    * This function prints the info text if the password was incorrect
    *
    **/
  static PICBOOL PasswordIncorrect();


	//
	// These methods are artifacts from the v2 era. They are needed
	// for the time being. Sod.
	//
	static PICBOOL setSegmentationAlgorithms(BaseStringArray &names);


private:

  /**
    * Get the segment algorithm params
    * 
    * @param algname  algorithm name
    **/
  static void tr_segalgParams_get(char *algname);

  /** 
    * User has changed the segment algorithm params
    * 
    * @param algname    algorithm name
    * @param algparams  algorithm params in one string
    **/
  static void tr_segalgParams_set(char *algname, char *algparams);




/*** NEW IN V3 OR KEPT FROM V2 ***/

public:

	//
	// Dummy constructor.
	//
	PicGUI();

	//
	// Dummy destructor.
	//
	~PicGUI();

	//
	// Tcl/Tk GUI initializer.
	//
	friend int Tcl_AppInit(Tcl_Interp *interp);

	//
	// A dummy function that calls the Tcl/Tk GUI initializer
	//
	PICBOOL start(int argc, char *argv[]);

	//
	// A method used to print text in the info window of the GUI.
	//
	static PICBOOL infoPrint(BaseString message);

	//
	// This method sets the results of a query (given by the parameter
	// 'pics'.) to the main window. Called by PicMain after a query has
	// completed or when the user scrolls the results.
	//
	static PICBOOL setQueryResults(const PicIxCollect &pics);

	//
	// Notify the GUI about the arrival of server options.
	//
	static void setServerOptions(BaseStringArray options);

	//
	// This method sets the query progress (the number of images
	// scanned) to the GUI.
	//
	static PICBOOL QueryProgressReport(PICULONG completed, PICULONG total = 0);

	//
	// Tihs method sets the available search types to the GUI.
	//
	static PICBOOL setSearchTypes(BaseStringArray &searchTypes);

private:

	//
	// This function receives commands from the GUI and sends them to the
	// main program. Called by the GUI every time it wants to communicate
	// with other parts of the system.
	//
	static PICINT transmitCmd(ClientData clientdata, Tcl_Interp *interp, 
		PICINT argc, char *argv[]);

	//
	// Auxiliary method to set the 'actions' Tcl variable
	// (see procedures.tk: proc actionloop).
	//
	static void setActions(char *action, const BaseStringArray &args);

	inline static void setActions(char *action, const BaseString &args) {
		BaseStringArray argsArray; argsArray.add(args);
		setActions(action, argsArray);
	}

 	inline static void setActions(char *action) {
		BaseStringArray argsArray; argsArray.setNull();
		setActions(action, argsArray);
	}

	//
	// Auxiliary method to set transmit's return value from a BaseStringArray.
	//
	inline static void setResult(const BaseStringArray &results) {
		Tcl_SetResult(interpreter, BaseStringArrayToList(results), TCL_VOLATILE);
	}

	inline static void setResult(const BaseString &result) {
		BaseStringArray resultArray; resultArray.add(result);
		setResult(resultArray);
	}

	//
	// Auxiliary method to set the data associated with a new query image.
	//
	static void newQueryImage(PicPicture &newPicture, char *destImageName,
		char *listIndex);

	//
	// Convert a PicArray to a BaseStringArray.
	//
	static BaseStringArray arrayToBaseStringArray(PicArray &array);
	
	//
	// Convert a BaseStringArray to a Tcl list.
	//
	static char *BaseStringArrayToList(const BaseStringArray &array);



	//
	// Methods called by transmitCmd
	//
	static void acceptRights (char *argv[]);		// Confirm changes in user rights.
	static void addUser(char *argv[]);			// Add a new user.
	static void addAlgorithm (char *argv[]);		// Add a new algorithm.
	static void batchInsert(char *argv[]);			// Insert a set of images.
	static void browseResults(char *argv[]);		// Scroll query results.
	static void cancelQuery(char *argv[]);			// Stop an ongoing query.
	static void changePassword(char *argv[]);		// Change a user's password.
	static void changeSearchType(char *argv[]);		// Change search type.
	static void changeUserDbAuthorization(char *argv[]);	// Change a user's database access rights.
	static void changeUserServerAuthorization(char *argv[]);// Change a user's server access rights.
	static void closeQueryImage(char *argv[]);		// Close an open query image.
	static void connect(char *argv[]);			// Attempt to establish connection.
	static void createDatabase(char *argv[]);		// Create a new database.
	static void deleteImage(char *argv[]);			// Delete the given image from the database.
	static void disconnect(char *argv[]);			// Disconnect.
	static void getAlgorithmInfo(char *argv[]);		// Get an algorithm's description.
	static void getAlgorithmParams(char *argv[]);		// Get the current parameters of an algorithm.
	static void getAllDbNames(char *argv[]);		// Get the names of all the databases residing at the server.
	static void getAllUsers(char *argv[]);			// Get the names of all the users known by the current server.
	static void getClientOptions(char *argv[]);		// Get general client options.
	static void getDatabaseAlgNames(char *argv[]);		// Get names of the algorithms in the specified database.
	static void getDatabaseNames(char *argv[]);		// Get the names of the databases the user has access to.
	static void getImage(char *argv[]);			// Make the selected image a new query image.
	static void getServerAddress(char *argv[]);		// Get server's address.
	static void getServerAdvert(char *argv[]);		// Get server's advertisement.
	static void getServerOptions(char *argv[]);		// Get server specific options.
	static void getSignature(char *argv[]);			// Get a textual representation of an image's signature.
	static void getThumbnail(char *argv[]);			// Make the selected thumbnail a new query image.
	static void getUserDbAuthorization(char *argv[]);	// Get user's rights to a specific database.
	static void getUserServerAuthorization(char *argv[]);	// Get user's server authorization.
	static void guiStarted(char *argv[]);			// Notify PicMain that the GUI has started.
	static void insertImage(char *argv[]);			// Insert a given image to the selected database.
	static void openQueryImage(char *argv[]);		// Load new query image.
	static void query(char *argv[]);			// Initiate a query.
	static void quit(char *argv[]);				// End program execution.
	static void removeAlgorithm(char *argv[]);		// Remove an algorithm.
	static void removeDatabase(char *argv[]);		// Remove a database.
	static void removeUser(char *argv[]);			// Remove a user.
	static void replaceAlgorithm(char *argv[]);		// Replace an existing algorithm.
	static void saveQueryImage(char *argv[]);		// Save the active query image.
	static void scaleImage(char *argv[]);			// Scale the given image if it's larger than the specified canvas.
	static void setAlgorithmParams(char *argv[]);		// Set an algorithm's parameters.
	static void setClientOptions(char *argv[]);		// Store changed client options.
	static void setDatabaseAlgorithms (char *argv[]);	// Set/unset algorithms used by a database	
	static void setImageAttributes(char *argv[]);		// Set changes in active query image's attributes.
	static void setNumOfDbFrames(char *argv[]);		// Sets the number of thumbnails received in one go.
	static void setServerOptions(char *argv[]);		// Send changed options to the server.

	struct funcPointer {
		char *name;				// Method's name.
		char argc;				// How many arguments it takes.
		void (*fptr)(char *[]);	// Function pointer to the method.
	};

	static const funcPointer transmitFuncs[];	// Declare transmitFuncs.
};

#endif
