// **********************************************************************
//
//      PICHost.cpp
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************    

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdio.h>
#include <PICDebug.h>
#include <PICHost.h>
#include <PICServer.h>
#include <sys/stat.h>           // get 'mkdir'
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>

#include <message.h>

PICHost *host_;

// ===========================================================

PICHost::PICHost(void) : params("database/params.cfg"), 
  users("database/users.cfg"), DBs("database/DBs.cfg") {

  //saving these gives the cfgFiles a cleaner look
  params.save();
  users.save();
  DBs.save();

  maxClient = 1;
  numClient = 0;
  superIn = false;
  if (!net.HostSocketInit(PORTNUMBER))
    exit(1);

  // ensure that 'the database' directory exists 
  mkdir("database",0700);

  signal(SIGCHLD,SignalHandler);
  signal(SIGINT,SignalHandler);
  signal(SIGUSR1,SignalHandler);
  signal(SIGPIPE,SignalHandler);

  if(!InitializeServerOptionsTable()) 
    ERROR_MSG("[HST] options init failed\n"); 
  InitializeUserTable();
}

// ===========================================================

void PICHost::SignalHandler(int a) {
  if( a == SIGCHLD ) {
    pid_t pid=wait(NULL);
    LOG_MSG("[HST] Child died (waitpid="<<pid<<")"<<endl);
    host_->RemoveServer(pid);
    LOG_MSG("[HST] Server removed."<<endl);
    signal(SIGCHLD,SignalHandler);
  } else if( a == SIGINT ) {
    LOG_MSG("[HST] Shutting down...\n");

    signal(SIGCHLD,SIG_IGN);
    signal(SIGINT,SIG_IGN);
    signal(SIGUSR1,SIG_IGN);
    signal(SIGPIPE,SIG_IGN);

    // kill servers
    DEBUG_MSG( "[HST] clients: " << host_->numClient << '\n' );
    for( int i = 0; i < host_->numClient; i++ ) {
      LOG_MSG("[HST] sending SIGUSR2 to " << host_->pidArray[i] << '\n');
      kill( host_->pidArray[i], SIGUSR2 );
    }
    if (host_->superIn) {
      kill(host_->pidSuper,SIGUSR2);
      host_->superIn=false;
    }

    exit(0);
  } else if( a == SIGUSR1 ) {
    ERROR_MSG("[HST] Error: network read error\n");
  } else if( a == SIGPIPE ) {
    ERROR_MSG("[HST] Error: network write error\n");
  }
}

// ===========================================================

//void PICHost::Main(int maxClient)
void PICHost::Main( )
{
  // Loop forever.
  LOG_MSG("[HST] Host ready\n");
  for(;;) {
    DEBUG_MSG("[HST] Waiting for client..."<<endl);   
    // Read the request from the socket.
    if (!ReadRequest()) {
      LOG_MSG("[HST] ReadRequest failed..."<<endl);   
      continue;
    }
    //reload all cfgFiles here to make sure you've got the latest versions.
    params.reload();
    users.reload();
    DBs.reload();

    // Check the authorization.
    // If the authorization level can not be read, the username is not
    // valid in the system, and the access is denied.
    if (!GetAuthorization()) {
      LOG_MSG("[HST] NAK1..."<<endl);
      SendAcknowledgement(ack_unauthorised);
      continue;
    }
    // Else check if the user is a superuser or an ordinary 'picture hunter'
    if (IsSuperuser()) {
      // If the user is the superuser then check if some other
      // superuser is already in. If the user is the first superuser
      // send ACK (AUTHORIZED) to the client (user has a superuser 
      // rights!) and fork a dedicated PICServer.
      if (IsFirstSuperuser()) {
	LOG_MSG("[HST] ACK..."<<endl);
	SendAcknowledgement(ack_OK);
	superIn = true;
	CreateServer();
	pidSuper = process;
	continue;
      }
      // else change the users superuser right to normal rights and
      // check is there any free place for a new client. If there is
      // then send ACK (AUTHORIZED) to the client (user has a normal 
      // user rights!) and fork a dedicated PICServer.
      if(!IsFreeSpace()) {
	LOG_MSG("[HST] NAK2..."<<endl);
	SendAcknowledgement(ack_busy);
	continue;
      }
      DropRights();
      LOG_MSG("[HST] ACK..."<<endl);
      SendAcknowledgement(ack_OK);
      CreateServer();
      AddServer();  
      continue;
    }
    // Else check if there any free place for a new client. If there is
    // then send ACK (AUTHORIZED) to the client (user has a normal user 
    // rights!) and fork a dedicated PICServer.
    if(!IsFreeSpace()) {
      LOG_MSG("[HST] NAK2..."<<endl);
      SendAcknowledgement(ack_busy);
    }
    LOG_MSG("[HST] ACK..."<<endl);
    SendAcknowledgement(ack_OK);
    CreateServer();
    AddServer();
  }
  net.CloseHostSocket();
} 

// ===========================================================

PICHost::~PICHost()
{
}

// ===========================================================

bool PICHost::ReadRequest()
{
  if (!net.AcceptClientRequest()){
    ERROR_MSG("[HST] Error in AcceptClientRequest"<<endl);
    return false;
  }

  xdrH=new xdrHandler(net.serverSocketID());

  struct addUser_msg usr;
  xdrH->receive(&usr);
  userName=strdup(usr.name);
  passWord=strdup(usr.passW);
  xdrH->free(&usr);

  return true;
}

void PICHost::SendAcknowledgement(unsigned int replyValue)
{
  struct ack_msg ack;

  ack.value=(ackValue)replyValue;
  if(!xdrH->send(&ack)) {
    cout << "send ack FAILED! fuck!" << endl;
  }
}

// ===========================================================

bool PICHost::InitializeServerOptionsTable()
{
  LOG_MSG("[HST] Checking server options...\n");
  parameterSection hostOpt=params["host"];

  if(hostOpt.maxclients<=0) 
    hostOpt.maxclients=1;
  /*if(hostOpt.timeout ) //can't say what would be an invalid value.
    hostOpt.timeout=0;*/
  if((string)hostOpt.http=="")
    hostOpt.http="http://www.cs.helsinki.fi/research/picsearch";
  if((string)hostOpt.info=="")
    hostOpt.info="Powered by PICSearch 3.1";

  maxClient = hostOpt.maxclients;

  DEBUG_MSG("[HST] maximum number of non-superuser clients is " << 
	    maxClient << endl);
  DEBUG_MSG("[HST] Server options initialized\n");
  return true;
}

// ===========================================================

bool PICHost::GetAuthorization()
{
  const string name=userName;
  const string pass=passWord;
  const string anonName="anonymous";

  // if the password is not correct and the user is not 'anonymous', then
  // return

  //FIXME: does this problem still exist?
  //.compare because old g++ claims != is ambiguous for strings.
  string PW=users[name].password;
  if(PW.compare(pass)!=0 && name.compare(anonName)!=0)
    return false;

    //compiler trouble.
    //if((string)users[name].password!=pass && name!=anonName)

  authLevel = users[name].authLevel;
  changePassword = users[name].changePass;

  return true;
}

// ===========================================================

bool PICHost::InitializeUserTable()
{
  LOG_MSG("[HST] Checking user table\n"); 

  if(!users.exists("SuperUser"))
    AddUser("SuperUser","ResuRepus",SuperUser,1);
  if(!users.exists("anonymous"))
    AddUser("anonymous","neppari",UserAccess,0);

  return true;
}

// ===========================================================

bool PICHost::AddUser(string name,string pass,ServerAuth auth,bool changePW) {
  LOG_MSG("[HST] adding user '"<< name << "' with rigths '" << 
	  (int)auth << "'\n");

  // first, insert a new user to the user table
  users[name].password=pass;
  users[name].authLevel=auth;
  users[name].changePass=changePW;

  // next, create anon access level to all databases for the new user
  for(cfgIterator i=DBs.begin();i!=DBs.end();++i)
    users[name].dbAuth[i.name()]=DB_ReadAccess;
  return true;
}

// ===========================================================

void PICHost::CreateServer()
{
  DEBUG_MSG("[HST] In CreateServer method..."<<endl);

  // Fork the Server process.
  int myPid = fork();

  if (myPid == -1)
  {
    ERROR_MSG("[HST] Error: Unable to fork a child." << endl);
    exit(1);
  }

  else if (myPid == 0)    // Child 
  {
    net.CloseHostSocket();
    PICServer *serv=new PICServer(net, xdrH);
    DEBUG_MSG("[HST] Executing server 'Main' with authlevel " << (int)authLevel
	      << "," << (int)changePassword << "\n" << 
	      "and name " << userName);
    serv->Main(userName,authLevel,changePassword);
    delete serv;
    exit(0);
  }
  else                    // Mother
  {
    process = myPid;
    DEBUG_MSG("[HST] Server process id = "<<process<<endl);
  }
}

// ===========================================================

// Adds server to the pidArray list.
void PICHost::AddServer()
{
  LOG_MSG("[HST] Server added (pid = " << process << ").\n");
  pidArray[numClient++] = process;
}

// ===========================================================

// Removes server from the pidArray list by compressing
// and reduce the counter.
void PICHost::RemoveServer(int childPid)
{
  if (childPid == pidSuper)
  {
    pidSuper = 0;
    superIn = false;
  }
  else
  {
    int i = 0;
    LOG_MSG("[HST] Removing server..." << endl;  );
    while ((pidArray[i] != process) && (i < numClient))
      i++;
    if (i<numClient) {
      for (int j = i; j < (numClient-1); j++)
	pidArray[j] = pidArray[j+1];
      numClient--;
    }
  }
}

// ===========================================================

bool PICHost::IsFreeSpace()
{
  if (numClient < maxClient)
    return true;
  return false;
}

// ===========================================================

bool PICHost::IsSuperuser()
{
  if (authLevel == SuperUser)
    return true;
  return false;  
}

// ===========================================================

bool PICHost::IsAccess()
{
  if (authLevel == UserAccess || authLevel == SuperUser)
    return true;
  return false;  
}

// ===========================================================

bool PICHost::IsFirstSuperuser()
{
  if (superIn == false )
    return true;
  return false;
}

// ===========================================================

void PICHost::DropRights()
{
  if (authLevel == SuperUser)
    authLevel = UserAccess;
}

// ===========================================================

int main(int argc, char *argv[])
{
  RESET_LOG;

  if (argc > 2)
    cout << "Usage: PICServer [path]." << endl;
  else
  {
    /*if (argc < 2)
      host_=new PICHost( NULL );
    else
    host_=new PICHost( argv[1] );*/

    host_=new PICHost();
    host_->Main();

    delete host_;
    return 0;
  }
}




