// **********************************************************************
//  AverageColor search algorithm library 
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1999 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <string.h>
#include <math.h>
#include <picerror.hxx>
#include <pstring.hxx>
#include <PicAverageColor.hxx>
#include <mmain.hxx>

#include <iostream.h>

PicLibAlgorithm::PicLibAlgorithm()
{
  name="AverageColor";
  info="AverageColor algorithm\n";
}

PICBOOL PicLibAlgorithm::setParams(PicIxCollect &params)
{
  if(params.contains(1)){
    PicObject &curr = params.at(1);
    if(TYPE_GUARD(PicString::classId(), curr)){
      PicString &parVal = AS(PicString, curr);
      if(parVal == BaseString("1"))
        useRed = boolTrue;
      else
        useRed = boolFalse;
    }
  }
  if(params.contains(2)){
    PicObject &curr = params.at(2);
    if(TYPE_GUARD(PicString::classId(), curr)){
      PicString &parVal = AS(PicString, curr);
      if(parVal == BaseString("1"))
        useGreen = boolTrue;
      else
        useGreen = boolFalse;
    }
  }
  if(params.contains(3)){
    PicObject &curr = params.at(3);
    if(TYPE_GUARD(PicString::classId(), curr)){
      PicString &parVal = AS(PicString, curr);
      if(parVal == BaseString("1"))
        useBlue = boolTrue;
      else
        useBlue = boolFalse;
    }
  }
  if(!useRed && !useGreen && !useBlue){
    useRed = useGreen = useBlue = boolTrue;
  }
  
  return boolTrue;
}
  

PICUSHORT PicLibAlgorithm::compare(PicSignature &firstSig, 
                                          PicSignature &secSig)
{
  static double scaleFactor3 = 2.2641187; // Assume retVal range 0-1000
  static double scaleFactor2 = 2.7729678;
  static double scaleFactor1 = 3.9215686;

  PICUSHORT retVal = 0;
  PICBOOL   compsUsed = 0;

#ifdef DEBUG
	cout << "AVE: 1stsize: " << firstSig.signature_size << "\n";
	cout << "AVE: 2ndsize: " << firstSig.signature_size << "\n";
#endif

  PicAverageColorSignature baseSig(&firstSig);
  PicAverageColorSignature actSig(&secSig);
  if(baseSig.sigFailure || actSig.sigFailure)
  	return(0);
	
  PICLONG  sum = 0;
  PICLONG diff;
  if(useRed){
    diff = (baseSig.avgRed - actSig.avgRed);
    diff *= diff;
    sum += diff;
    compsUsed++;
  }
  if(useGreen){
    diff = (baseSig.avgGreen - actSig.avgGreen);
    diff *= diff;
    sum += diff;
    compsUsed++;
  }
  if(useBlue){
    diff = (baseSig.avgBlue - actSig.avgBlue);
    diff *= diff;
    sum += diff;
    compsUsed++;
  }
  if(sum<0) sum=0;      // Little paranoia

  double colorDistance = sqrt(sum);
  switch(compsUsed)
  {
    case 1:
      retVal = (PICUSHORT)(colorDistance * scaleFactor1);
      break;
    case 2:
      retVal = (PICUSHORT)(colorDistance * scaleFactor2);
      break;
    case 3:
      retVal = (PICUSHORT)(colorDistance * scaleFactor3);
      break;
    default:
      retVal = 0;
      break;
  }
  
  return (1000 - retVal);
}

PicSignature *PicLibAlgorithm::computeSignature(PicPicture &pic)
{
  PicSignature *retVal;

#ifdef DEBUG
	cout << "AVE: Entered computeSig()\n";
#endif

  if(!pic.raw || pic.raw->isNull()){
#ifdef DEBUG
	cout << "AVE: No raw data to calc from!\n";
#endif
	return(NULL);
  }else{
    PICULONG redSum=0, blueSum=0, greenSum=0;
    PICULONG ix, len, size;
    unsigned char *data = pic.raw->getDataPtr();

    size = pic.raw->getWidth() * pic.raw->getHeight();
    len = 3*size;
    for(ix=0; ix<len; ix += 3){
      redSum += data[ix];
      greenSum += data[ix+1];
      blueSum += data[ix+2];
    }
    if(size<=0) size=1;

#ifdef DEBUG
	cout << "AVE: Entering newsig\n";
#endif
	
	PicAverageColorSignature *tmp=new PicAverageColorSignature(redSum/size, greenSum/size, blueSum/size);
  	if(tmp->sigFailure)
		return(0);
	
	retVal=tmp;
	
#ifdef DEBUG 
  	cout << "AVE: Newsig passed\n";
#endif
  }

#ifdef DEBUG
  cout << "AVE: passed\n";
#endif

  return retVal;
}

/*
BaseString PicAverageColorAlgorithm::getImage() const
{
  return name & " [" & asString(id) & "]";
}
*/

// **********************************************************************

PicAverageColorSignature::PicAverageColorSignature()
  : avgRed(0), avgGreen(0), avgBlue(0)
{
	sigFailure=0;
}

PicAverageColorSignature::PicAverageColorSignature(PICSHORT red, 
                                               PICSHORT green, 
                                               PICSHORT blue)
  : avgRed(red), avgGreen(green), avgBlue(blue)
{
	sigFailure=0;
	char *data = new char[3*sizeof(char)];
	if(!data)
	{
#ifdef DEBUG
		cout << "AVE: Memory allocation error\n";
#endif
		sigFailure=1;
		return;
	}
		
	data[0] = (unsigned char)avgRed;
	data[1] = (unsigned char)avgGreen;
	data[2] = (unsigned char)avgBlue;
	
	this->signature_size=3*sizeof(char);
	this->signature=data;
}

PicAverageColorSignature::PicAverageColorSignature(PicSignature *from)
{
	sigFailure=0;
	this->signature_size=0;
	this->signature=NULL;
	
#ifdef DEBUG
	cout << "AVE: fromBin() coming...\n";
#endif
	
	PICULONG size=from->signature_size;
	char *data=from->signature;

  if(size!=(3*sizeof(char)) || !data){
#ifdef DEBUG
	cout << "AverageCol: Wrong type of signature! (size: " << size << ")\n";
#endif
	avgRed = avgGreen = avgBlue = 0;
    sigFailure=1;
	return;
  }
  avgRed = (unsigned char)(data[0]);
  avgGreen = (unsigned char)(data[1]);
  avgBlue = (unsigned char)(data[2]);

#ifdef DEBUG
	cout << "AVE: fromBin() passed...\n";
#endif
}

/*
BaseString PicAverageColorSignature::getImage() const
{
  return "Avg.Red = " & asString(avgRed) &
         "\nAvg.Green = " & asString(avgGreen) &
         "\nAvg.Blue = " & asString(avgBlue);
}
*/

