// **********************************************************************
//  This file contains PicWaveletAlgorithm implementation
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>

#include <PicWavelet.hxx>

// **********************************************************************

PicLibAlgorithm::PicLibAlgorithm()
{
  name="Wavelet";
  info="Wavelet algorithm\n";
}

PICBOOL PicLibAlgorithm::setParams(PicIxCollect &params)
{
  return boolTrue;
}

inline double dabs(double num)
{
  if (num<0.0)
    return -num;
  else
    return num;
}

PicWaveletSignature::PicWaveletSignature(PicSignature *from)
{
	sigFailure=0;
	this->signature=NULL;
	this->signature_size=0;
	
	PICULONG size=from->signature_size;
	char *data=from->signature;

/* debugging
	FILE *fp;
	char *buffer;
	buffer=tmpnam(NULL);
	fp=fopen(buffer, "w");
	fwrite(data, 1, size, fp);
	fclose(fp);
*/

  PICULONG i, j;
  if(!data ||
     (size != 3*sizeof(double) + 2*3*sizeof(PICUSHORT) + 
              3*NUM_OF_WAVELET_COEFS*2*sizeof(PICUSHORT)))
  {
#ifdef DEBUG
	cout << "Wavelet: Sig size " << size << " doesn't match expected!\n";
#endif
	sigFailure=1;
	return;
  }
  const char *cursor = data;

  for(i=0; i<3; i++){
    memcpy(&avgVals[i], cursor, sizeof(double));
    cursor+=sizeof(double);
    memcpy(&posCoefs[i], cursor, sizeof(PICUSHORT));
    cursor+=sizeof(PICUSHORT);
    memcpy(&negCoefs[i], cursor, sizeof(PICUSHORT));
    cursor+=sizeof(PICUSHORT);

    if(posCoefs[i] + negCoefs[i] != NUM_OF_WAVELET_COEFS)
	{
#ifdef DEBUG
		cout << "Wavelet: UUURGH, Wavelet Coefs num mismatch!\n";
#endif
		sigFailure=1;
		return;
	}
		
    for(j=0; j<NUM_OF_WAVELET_COEFS; j++){
      memcpy(&(bigCoefs[i][j].x), cursor, sizeof(PICUSHORT));
      cursor+=sizeof(PICUSHORT);
      memcpy(&(bigCoefs[i][j].y), cursor, sizeof(PICUSHORT));
      cursor+=sizeof(PICUSHORT);
    }
  }
}

PICUSHORT PicLibAlgorithm::compare(PicSignature &firstSig,
                                       PicSignature &secSig)
{
//    static double w0[3] = { 4.5, 17.0, 26.0 };
  static double w0[3] = { 5.0, 5.0, 5.0 };
//    static double w1[3] = { 0.5, 0.4, 0.3 };
  static double w1[3] = { 0.5, 0.5, 0.5 };

  static double maxVal = w0[0] + w0[1] + w0[2];
  static double minVal = -1.0 * NUM_OF_WAVELET_COEFS*(w1[0] + w1[1] + w1[2]);
  static double interval = maxVal - minVal;

  double score = minVal;
  PICUSHORT i, j, colorInd;

#ifdef DEBUG
	cout << "WAV: 1size: " << firstSig.signature_size << "\n";
	cout << "WAV: 2size: " << secSig.signature_size << "\n";
	if(firstSig.signature)
		cout << "WAV: Sig1 present\n";
	if(secSig.signature)
		cout << "WAV: Sig2 present\n";
#endif

	PicWaveletSignature q(&firstSig);
	if(q.sigFailure)
		return(0);
	
	PicWaveletSignature t(&secSig);
	if(t.sigFailure)
		return(0);

#ifdef DEBUG
	cout << "WAV: Conversion passed.\n";
#endif

    if(q.posCoefs[0]==0 && q.negCoefs[0]==0)
      return 0;
    score = 0.0;
    for (colorInd=0; colorInd<3; colorInd++) {
      score += w0[colorInd] * (dabs(q.avgVals[colorInd]-t.avgVals[colorInd]));
    
      for(i=0; i<q.posCoefs[colorInd]; i++){
        for(j=0; j<t.posCoefs[colorInd]; j++){
          if(q.bigCoefs[colorInd][i].x == t.bigCoefs[colorInd][j].x &&
             q.bigCoefs[colorInd][i].y == t.bigCoefs[colorInd][j].y)
          {
            score -= w1[colorInd];
          }
        }
      }
      for(i=q.posCoefs[colorInd]; i<q.posCoefs[colorInd]+q.negCoefs[colorInd]; i++){
        for(j=t.posCoefs[colorInd]; j<t.posCoefs[colorInd]+t.negCoefs[colorInd]; j++){
          if(q.bigCoefs[colorInd][i].x == t.bigCoefs[colorInd][j].x &&
             q.bigCoefs[colorInd][i].y == t.bigCoefs[colorInd][j].y)
          {
            score -= w1[colorInd];
          }
        }
      }
    }

#ifdef DEBUG
	cout << "WAV: Score1: " << score << "\n";
#endif
  
  score -= minVal;            // score is now in [0, interval]
  score = interval - score;   // small is good
  score /= interval;          // now in [0, 1]

#ifdef DEBUG
	cout << "WAV: Score2: " << score << "\n";
#endif

  return (PICUSHORT)(score*1000.0);
}

int valCompare(const void *a, const void *b)
{
  double v1 = ((const struct coords *)a)->val;
  double v2 = ((const struct coords *)b)->val;

  if(v1==v2)
    return 0;
  if(dabs(v1) < dabs(v2))
    return 1;   // big values to the head, small to the tail

  return -1;
}

int coordCompare(const void *a, const void *b)
{
  const struct xy *sp1 = (const struct xy *)a;
  const struct xy *sp2 = (const struct xy *)b;

  if (sp1->y < sp2->y)
    return -1;
  if (sp1->y > sp2->y)
    return 1;
  if (sp1->x < sp2->x)
    return -1;

  return 1;
}

void decomposeArray(struct coords *array, PICULONG size)
{
  if(size==0)
    return;

  PICULONG i, i2, h=size;
  const double sqrtH = sqrt(h), sqrt2 = sqrt(2);
  double *tempArray = (double *)malloc(sizeof(double) * size);

  for(i=0; i<h; i++){
    array[i].val /= sqrtH;
  }
  while(h>1){
    h /= 2;
    for(i=0, i2=0; i<h; i++, i2+=2){
      tempArray[i] = (array[i2].val + array[i2+1].val) / sqrt2;
      tempArray[h+i] = (array[i2].val - array[i2+1].val) / sqrt2;
    }
    for(i=0; i<size; i++){    // or (i=0; i<2*h; i++) ??
      array[i].val = tempArray[i];
    }
  }
  free(tempArray);
}

void transpose(struct coords *array, PICULONG width, PICULONG height)
{
  PICULONG i, j, ind, ind2;
  double    temp;

  for(i=0; i<height; i++){
    for(j=i+1; j<width; j++) {
      ind = i*width + j;
      ind2 = j*width + i;
      temp = array[ind].val;
      array[ind].val = array[ind2].val;
      array[ind2].val = temp;
    }
  }
}

PicSignature *PicLibAlgorithm::computeSignature(PicPicture &pic)
{
#ifdef DEBUG
	cout << "Wav entered compute\n";
#endif

  if(!pic.raw || pic.raw->isNull()){
#ifdef DEBUG
	cout << "WAV: Argh, no raw data received to calculate on.\n";
#endif
	return(NULL);
  }
  PicRawPicture *scaled = PicTools::scaleImage(*(pic.raw), 128, 128, boolFalse);
  // NOTE: image must be 'square' and width a power of 2 

  PICULONG  i, j, ind, colorInd;
  PICULONG  width = scaled->getWidth();
  PICULONG  height = scaled->getHeight();
  PICULONG  size = width*height;

  struct coords *coefs = 
    (struct coords *)malloc(sizeof(struct coords) * size);

  double      avgVals[3]; 
  PICUSHORT   posCoefs[3], negCoefs[3];
  struct xy   bigCoefs[3][NUM_OF_WAVELET_COEFS];

  for (colorInd=0; colorInd<3; colorInd++) {
    unsigned char *cursor = scaled->getDataPtr() + colorInd;
    for(i=0; i<height; i++){
      for(j=0; j<width; j++) {
        ind = i*width + j;
        coefs[ind].x = i;
        coefs[ind].y = j;
        coefs[ind].val = ((double)*cursor) / 255.0;  // scale to [0 - 1]
        cursor++;
      }
    }
    for(i=0; i<height; i++){ // decompose rows
      decomposeArray(&coefs[width*i], width);
    }
    transpose(coefs, width, height);
    for(i=0; i<height; i++){ // in effect, decompose columns
      decomposeArray(&coefs[width*i], width);
    }
    transpose(coefs, width, height);  // transpose back

    // exclude (0,0) from sort:
    qsort(&coefs[1], size-1, sizeof(struct coords), valCompare); 

    avgVals[colorInd] = coefs[0].val;
    posCoefs[colorInd] = 0;

    ind = 0;
    for (i=1; i<=NUM_OF_WAVELET_COEFS; i++){    // exclude (0,0) 
      if(coefs[i].val > 0.0){   // Found a positive coeff
        bigCoefs[colorInd][ind].x = coefs[i].x;
        bigCoefs[colorInd][ind].y = coefs[i].y;
        ind++;
      }
    }
    posCoefs[colorInd] = ind;
    qsort(&bigCoefs[colorInd][0], posCoefs[colorInd], 
          sizeof(struct xy), coordCompare);   // sort by coordinates

    for (i=1; i<=NUM_OF_WAVELET_COEFS; i++){    // exclude (0,0) 
      if(coefs[i].val <= 0.0){ // Found a negative coeff
        bigCoefs[colorInd][ind].x = coefs[i].x;
        bigCoefs[colorInd][ind].y = coefs[i].y;
        ind++;
      }
    }
    negCoefs[colorInd] = NUM_OF_WAVELET_COEFS - posCoefs[colorInd];
    qsort(&bigCoefs[colorInd][posCoefs[colorInd]], negCoefs[colorInd], 
          sizeof(struct xy), coordCompare);   // sort by coordinates
  }
  free(coefs);
  delete scaled;

#ifdef DEBUG
	cout << "WAV: Entering sig creat\n";
#endif

  PicWaveletSignature *tmp=new PicWaveletSignature(avgVals, posCoefs,
  							negCoefs, bigCoefs);
  if(tmp->sigFailure)
  	return(NULL);
	
  PicSignature *signature=tmp;

#ifdef DEBUG
	cout << "WAV: Exiting sig creat\n";
#endif

  return(signature);
}

// **********************************************************************

PicWaveletSignature::PicWaveletSignature()
{
	sigFailure=0;
  
  PICUSHORT i;
  for(i=0; i<3; i++){
    avgVals[i] = 0.0;
    posCoefs[i] = negCoefs[i] = 0;
  }
  // bigCoefs are left uninitialized
}

PicWaveletSignature::PicWaveletSignature(
    double newAvgVals[3],
    PICUSHORT newPosCoefs[3], 
    PICUSHORT newNegCoefs[3],
    struct xy  newBigCoefs[3][NUM_OF_WAVELET_COEFS])
{
	sigFailure=0;
  
  PICULONG i, j;
  for(i=0; i<3; i++){
    avgVals[i] = newAvgVals[i];
    posCoefs[i] = newPosCoefs[i];
    negCoefs[i] = newNegCoefs[i];
    for(j=0; j<NUM_OF_WAVELET_COEFS; j++){
      bigCoefs[i][j] = newBigCoefs[i][j];
    }
  }

  int size = 3*sizeof(double) + 2*3*sizeof(PICUSHORT) + 
			3*NUM_OF_WAVELET_COEFS*2*sizeof(PICUSHORT);

  char *cursor, *binary = new char[size];

  if(!binary){
    size = 0;
	sigFailure=1;
#ifdef DEBUG
	cout << "WAV: Memory allocation error!\n";
#endif
	return;
  }
  cursor=binary;
  for(i=0; i<3; i++){
    memcpy(cursor, &avgVals[i], sizeof(double));
    cursor+=sizeof(double);
    memcpy(cursor, &posCoefs[i], sizeof(PICUSHORT));
    cursor+=sizeof(PICUSHORT);
    memcpy(cursor, &negCoefs[i], sizeof(PICUSHORT));
    cursor+=sizeof(PICUSHORT);
    for(j=0; j<posCoefs[i]; j++){
      memcpy(cursor, &(bigCoefs[i][j].x), sizeof(PICUSHORT));
      cursor+=sizeof(PICUSHORT);
      memcpy(cursor, &(bigCoefs[i][j].y), sizeof(PICUSHORT));
      cursor+=sizeof(PICUSHORT);
    }
    for(j=posCoefs[i]; j<negCoefs[i]+posCoefs[i]; j++){
      memcpy(cursor, &(bigCoefs[i][j].x), sizeof(PICUSHORT));
      cursor+=sizeof(PICUSHORT);
      memcpy(cursor, &(bigCoefs[i][j].y), sizeof(PICUSHORT));
      cursor+=sizeof(PICUSHORT);
    }
  }
  this->signature=binary;
  this->signature_size=size;
}

PicWaveletSignature::~PicWaveletSignature()
{

}

/*
BaseString PicWaveletSignature::getImage() const
{
  PICSHORT i, j;
  BaseString text;
     
  for(i=0; i<3; i++){
    switch(i){
      case 0:
        text.append("Red:\n");
        break;
      case 1:
        text.append("Green:\n");
        break;
      case 2:
        text.append("Blue:\n");
        break;
    }
    text.append("Average value = " & asString(avgVals[i]) & "\n");
    text.append("Number of positive coeffs: " & asString(posCoefs[i]) & "\n");
    for(j=0; j<posCoefs[i]; j++){
      text.append("(" & asString(bigCoefs[i][j].x) &
                  "," & asString(bigCoefs[i][j].y) & ") " );
    }
    text.append("\nNumber of negative coeffs: " & asString(negCoefs[i]) & "\n");
    for(j=posCoefs[i]; j<negCoefs[i]+posCoefs[i]; j++){
      text.append("(" & asString(bigCoefs[i][j].x) &
                  "," & asString(bigCoefs[i][j].y) & ") " );
    }
    text.append("\n------\n");
  }
  return text;
}
*/

