// **********************************************************************
//
//      Maint.cpp
//
//  This file contains the functions related to performing the
//  algorithm maintenance operations on the server side.
//
// **********************************************************************
// Copyright (C) 1999 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <PICDebug.h>
#include <PICNet.h>
#include <PICServer.h>
#include <pstrstm.hxx>
#include <newdb.h> //for DB stuff
#include <string>  //for strings.

bool checkAlg(const char *libPath, algorithmID_t algID);
bool crapCopy(const char *dst,const char *src);

bool PICServer::addAlgorithm(addAlg_msg &msg)
{
  char *libPath=msg.algLibPath;
  char *paramWinPath=msg.paramWinPath;
  bool haveParamWin=false;
  int algClass=msg.algClass;
  algorithmID_t newID=0;	

  newID=DB.newAlgID();			  // Get new unique ID

  const string newLocation="database/alg" + itos(newID) + ".so";
  const string tkLocation="database/alg" + itos(newID) + ".tk";

  if(!crapCopy(newLocation.c_str(), libPath)) {
    cout << "[ADDA]: Unable to copy " << libPath << endl;
    return(false);
  }

  // Let's check if the algorithm seems ok. This also adds it to the algList.
  if(!checkAlg(newLocation.c_str(), newID))
  {
    cout << "[MAINT]: Unsuitable libfile " << libPath << endl;
    unlink(newLocation.c_str());
    return(false);
  }			

  // Copy the parameter window code  
  if(paramWinPath && paramWinPath[0])
  {
    if(!crapCopy(tkLocation.c_str(), paramWinPath))
    {
      cout << "[ADDA]: Unable to copy " << paramWinPath << endl;
      return(false);
    }
    haveParamWin=true;
  } 
	
  // Insert the new algorithm into the algorithm database.
  DB.algs[newID].paramWin=haveParamWin;
  DB.algs[newID].algClass=algClass;
  DB.algs[newID].active=true;

  // send "ack_OK" acknowledgement.
  SendComplete();

  return(true);
}

bool PICServer::modifyAlgorithms(modifyAlg_msg &msg)
{
	char *libPath=msg.algLibPath;
	char *paramWinPath=msg.paramWinPath;
	algorithmID_t oldID=msg.ID, newID=0;
	int algClass=msg.algClass;
	bool recalcSigs=msg.recalcSigs;
	bool haveParamWin=false;

	// check that the old algorithm really exists.
	if(!DB.algs.exists(oldID)) {
	  cout << "[MODA]: algoritm doesn't exist. " << endl;
	  return false;
	}

	newID=DB.newAlgID();	      // Get new unique ID

	const string newLocation="database/alg" + itos(newID) + ".so";
	const string tkLocation="database/alg" + itos(newID) + ".tk";

	if(!crapCopy(newLocation.c_str(), libPath)) {
	  cout << "[MODA]: Could not copy " << libPath << endl;
	  return(false);
	}
	
	// Let's check if the algorithm seems ok. This also adds it to the 
	// algList.
	if(!checkAlg(newLocation.c_str(), newID))
	{
		cout << "[MODA]: Unsuitable libfile " << libPath << endl;
		unlink(newLocation.c_str());
		return(false);	
	}			
	
	// Copy the parameter window code
	if(paramWinPath && paramWinPath[0]) //parameter window given
	{
	  if(!crapCopy(tkLocation.c_str(), paramWinPath))
	    {
	      cout << "[MODA]: Unable to copy " << paramWinPath << endl;
	      return(false);
	    }
	  haveParamWin=true;
	} 
	else //no parameter window given, use the old one.
	{
	  if(DB.algs[oldID].paramWin) { //did the old alg have a paramWin?
	    const string oldTkLocation="database/alg" + itos(oldID) + ".tk";
			
	    if(!crapCopy(tkLocation.c_str(), oldTkLocation.c_str())) {
	      cout << "[MODA]: Unable to copy " << oldTkLocation << endl;
	      return(false);
	    }
	    haveParamWin=true;
	  }
	}

	// Insert the new algorithm into the algorithm database.
	DB.algs[newID].paramWin=haveParamWin;
	DB.algs[newID].algClass=algClass;
	DB.algs[newID].active=true;

	if(recalcSigs) //must recalculate signatures
	{
	  PicSearchAlgorithm *thisalg;
	  PicSignature *sigtmp;
	  PicPicture *picture;

	  thisalg=PicMain::globData.getSearchAlg(newID);
	  if(!thisalg)
  	    return(false);
		
	  // DOTHIS: Loop all picture databases where this algorithm 
	  // is active.
	  // For each old sig table, create a new signature table
	  // associated with the new algorithm ID. 
	  // 
	  // Loop all the pictures in the db.
	  // half-DONE: the syntax is not clean or uniform, but it 
	  // does the job. There's a spot to fill in.
	  
	  //loop through all databases
	  for(cfgIterator i=DB.DBs.begin();i!=DB.DBs.end();++i) {
	    
	    //boolean indicates if algo is active
	    if(DB.DBs[i.name()].algs[(int)newID]==false) 
	      continue; //skip the ones where it's not.

	    databaseID_t DBID=DB.DBs[i.name()].ID; //get DB ID

	    imageDB *imgDB=DB.getImgDB(DBID); //get image database.
	    if(imgDB==NULL) {
	      cout << "[MODA]: Couldn't access DB " << DBID << endl;
	      continue; //no reason to stop just because one failed..
	    }

	    imgDB->addSigDB(newID); //add sig DB for new alg.
	    
	    //next go through all pictures.
	    struct picture *pic;
	    pictureID_t picID=imgDB->firstPicKey(); //get first key
	    while((pic=imgDB->fetchPic(picID))!=NULL) { //iterate all
	      
	      //this interface really sucks. I'll think of something..
	      
	      string imgFileName=pic->fileName; //just get the filename.
	      free(pic); //free mem mallocated by GDBM.
	      
	      //this is how images are loaded..
	      //image img=imgDB->loadImage(imgFileName);
	      //you probably want to initialize a picImage straight from
	      //the file, though.. so you need the full fileName. This
	      //is generated as shown below.
	      //string imgPathName=imgDB->imagePathName(imgFileName);

	      //OKAY, each pic in the DB is now iterated.
	      //saving a sig is done as follows:
	      
	      /* 
		 the signature looks like this, pardon the ugly hack..
		 
		 struct signature {
		   unsigned int sigSize;
		   char sig[0];
		 };
		 
		 //proto save signature..
			 
		 signature *sig=(signature*)new char[sizeof(signature)+
		                                     sigSize];
		 sig->sigSize=sigSize;
		 memcpy(sig->sig,sigData,sigSize);

		 imgDB->storeSig(newID,picID,sig);

		 delete sig;

	      */

	      picID=imgDB->nextPicKey(picID);
	    }		  
	  }

	  // FIXME: I didn't finish the job, this part should really be
	  // in the innermost loop.		
	  // sigtmp=thisalg->computeSignature(picture);
	  
	  // DOTHIS: Write the new signature to the new table.
	  // half-DONE: there's a proto above showing what to do.
	  // delete sigtmp;

	  // End loop
	}
	else  //!recalcSigs: Ok, let's just copy the old tables
	{
	  //again, loop through all databases
	  for(cfgIterator i=DB.DBs.begin();i!=DB.DBs.end();++i) {

	    //boolean indicates if algo is active
	    if(!DB.DBs[i.name()].algs[newID])
	      continue; //skip the ones where it's not.

	    databaseID_t DBID=DB.DBs[i.name()].ID; //get DB ID

	    imageDB *imgDB=DB.getImgDB(DBID); //get image database.
	    if(imgDB==NULL) {
	      cout << "[MODA]: Couldn't access DB " << DBID << endl;
	      continue; //no reason to stop just because one failed..
	    }

	    //get the file names for the old & new signature DBs.
	    string oldSigDBPathName=imgDB->sigDBPathName(oldID);
	    string newSigDBPathName=imgDB->sigDBPathName(newID);
	    
	    //copy the old signature DB to the new one.
	    //FIXME: add some reasonable error handling.
	    crapCopy(newSigDBPathName.c_str(),
		     oldSigDBPathName.c_str());

	    //and add a signature DB. Can't do this earlier, it would create
	    //a new file (GDBM would, that is).
	    imgDB->addSigDB(newID);
	  }		  	  
	}
	
	// Activate the new algorithm

	//iterate through all DBs
	for(cfgIterator i=DB.DBs.begin();i!=DB.DBs.end();++i) {  
	  //boolean value set iff the algorithm is active.
	  if(DB.DBs[i.name()].algs[oldID])     //if active
	    DB.DBs[i.name()].algs[newID]=true; //set new alg active.
	}
	
	// Disable the old algorithm
	
	DB.algs[oldID].active=false; 
	DB.algs[oldID].replacedWithID=newID;
	
	// FIXME: again the same as in addAlgorithm--if the client does a 
	// synchronous wait SendFail should be called for all errors.

	SendComplete();

	return(true);
}

bool PICServer::removeAlgorithm(removeAlg_msg &msg)
{
	int idToRemove=msg.ID;

	if(!DB.algs.exists(idToRemove)) {
	  cout << "[REMA]: algoritm doesn't exist. " << endl;
	  return false;
	}
	
	DB.algs[idToRemove].active=false;
	DB.algs[idToRemove].replacedWithID=0;

	// DOTHIS: All files related to the removed algorithm
	// must be deleted on next time the host is started.
	// This must be done in some file of host/ probably.
	// FIXME: this not done currently.
	
	SendComplete();
	return(true);
}

/*
 * This is a crappy filecopy routine.
 *
 */

#define COPYBUFSIZE 65536

bool crapCopy(const char *dst,const char *src)
{
	char buffer[COPYBUFSIZE];
	FILE *fp, *dp;
	int bytesRead=0;

	fp=fopen(src, "r");
	dp=fopen(dst, "w");

	if(!dp || !fp)
		return(false); 

	while((bytesRead=fread(buffer, 1, COPYBUFSIZE, fp)))
	{
		fwrite(buffer, 1, bytesRead, dp);
	}

	fclose(fp);
	fclose(dp);

	return(true);
}

/*
 * This performs a check if the algorithm is ok. The checking is
 * actually done in the constructor.
 *
 * Notice that this also adds the algorithm to the servers list 
 * of algorithms.
 *
 */

bool checkAlg(const char *libPath, algorithmID_t algID)
{
	PicSearchAlgorithm *temp;

	temp = new PicSearchAlgorithm((char *)libPath, (int)algID);

	if(temp && temp->algID()) {                 // Alg seems OK
		PicMain::globData.searchAlgs.listAdd((void *)temp);
	}
	else
	{
		if(temp)
			delete temp;
		
		return(false);
	}			

	return(true);
}

