// **********************************************************************
//
//      Query.cpp
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <PICDebug.h>
#include <PICNet.h>
#include <PICServer.h>
#include <mmain.hxx> // for the V1 stuff...
#include <pstring.hxx> // for the V1 stuff... [PicString class]

bool PICServer::isInPICCache(long databaseId, long pictureId) {
  int i;
  int j;
  for(i = 0; i < cacheElements; i += 2) {
    if(cacheElementsArray[i + 1] > pictureId)
      return false;
    if(cacheElementsArray[i + 1] < pictureId)
      continue;
    break;
  }
  // Match found, search the premesis for a matching databaseId
  j = i;
  // Try our luck
  if(cacheElementsArray[j] == databaseId)
    return true;
  while(cacheElementsArray[(j - 2) + 1] == pictureId)
    j -= 2;
  while(j < cacheElements && cacheElementsArray[j + 1] == pictureId) {
    if(cacheElementsArray[j] == databaseId && cacheElementsArray[j + 1] == pictureId)
      return true;
    j += 2;
  }
  // Bummer...
  return false;
}

void PICServer::Query(query_msg &msg) {
  DEBUG_MSG("[SRV] in Query\n");

  // first, invalidate last query
  query_ok=0;
  current_query.deleteAll();

  // read in the query parameters from the message
  databaseID_t query_db_id=msg.DBID;
  algorithmID_t search_alg_id=msg.algID;
  unsigned int param_count=msg.params.params_len;
 
  // get a hold of the search algorithm object we're going to use for the query
  PicSearchAlgorithm *alg;

  PicMain::globData.searchAlgs.listRewind();

  while((alg=(PicSearchAlgorithm *)
	 PicMain::globData.searchAlgs.listNext())) {
    if(alg->algID()==(int)search_alg_id)
      break;
  }
  
  if (!alg) {
    ERROR_MSG("[SRV] '"<<user_name<<"' tried to query with an unknown "<< \
	      "algorithm "<<search_alg_id<<endl);
    SendFail();
    return;
  }

  // set up search algorithm parameters if any
  if (param_count) {
    DEBUG_MSG("[SRV] setting "<<param_count<<" parameters for '"<< \
	      alg->algName <<"'\n");
    PicArray array;
    array.deleteAll();

    for (unsigned int i=0; i<param_count; i++) {
      char *param_str=msg.params.params_val[i].str;
      array.putAt(i+1,PicString(param_str));
    }
    alg->setParams(array);
    DEBUG_MSG("[SRV] Parameters set for '"<<alg->algName()<<"'\n");
  }

  unsigned long total_results=msg.thumbMax;
  long oneshot_results=msg.thumbNum;
  unsigned long min_match=msg.minMatch*10;

  DEBUG_MSG("[SRV] Query: got thumbMax " << msg.thumbMax << ", thumbNum " <<
	    msg.thumbNum << " & minMatch " << msg.minMatch << endl);

  if (!total_results) {
    ERROR_MSG("[SRV] '"<<user_name<<"' tried to query with 0 total results\n");
    SendFail();
    return;
  }
  
  DEBUG_MSG("[SRV] query intialized [dbid:"<<query_db_id<<"/algid:"<< \
	    search_alg_id<<"/total:"<< \
	    total_results<<"/one:"<<oneshot_results<<"/minm:"<< \
	    min_match<<"]\n");

  PicSignature *ex_sign=NULL;

  current_query.startTime.setRelativeTime(0);
  
  // determine which type of query we're going to make and 'do the right thing'
  switch (msg.type) {
  case queryByPicture: // picture data included in the query message
    DEBUG_MSG("[SRV] fetching query picture from message data\n");
    {
      picData_msg qPic;

      if(!xdrH->receive(&qPic)) {
	DEBUG_MSG("[SRV] Query: failed to receive picture!\n");
	SendFail();
	return;
      }

      unsigned int image_size=qPic.picData.picData_len;
      DEBUG_MSG("[SRV] Query: got query image with size "<< image_size<< endl);

      if (!image_size) {
	ERROR_MSG("[SRV] '"<<user_name<<"' tried to query with an image "<< \
		  "with no image data\n");
	SendFail();
	return;
      }

      char *raw_image=new char[image_size];
      memcpy(raw_image,qPic.picData.picData_val,image_size);
      unsigned int image_width=qPic.picWidth;
      unsigned int image_height=qPic.picHeight;
      unsigned int format_id=qPic.picFormat;

      DEBUG_MSG("[SRV] Query: picture width "<< image_width << ", height " <<
		image_height << ", format " << format_id << endl);

      //FIXME: enable cache.
      /*// Section changed to include PICCache
      cacheElements = msg.ReadInt32();
      cacheIsDisabled = (cacheElements == 0) ? true : false;
      if(cacheElementsArray != NULL)
	free(cacheElementsArray);
      if(cacheIsDisabled == false) {
	cacheElementsArray = (long *) malloc(sizeof(long) * cacheElements);
	for(int i = 0; i < cacheElements; i += 2) {
	  cacheElementsArray[i] = msg.ReadInt32();
	  cacheElementsArray[i + 1] = msg.ReadInt32();
	}
      }*/

      DEBUG_MSG("[SRV] initializing picture object\n");
      PicPicture ex_pic;
      ex_pic.setNull();
      ex_pic.format=PicMain::globData.getImageFormat(format_id);
      DEBUG_MSG("[SRV] picture object format set to "<<format_id<<endl);
      PicRawPicture *new_raw;
      if (!ex_pic.format) {  // it's raw
	DEBUG_MSG("[SRV] loading RAW picture\n");
	ex_pic.setRaw(new PicRGBPicture(image_width,image_height,
					(unsigned char *)raw_image));
      } else {
	DEBUG_MSG("[SRV] loading picture from stream\n");
	PicInCharStream stream(raw_image,image_size);
	if ((new_raw=ex_pic.format->readPicture(stream,image_size)))
	  ex_pic.setRaw(new_raw);
      }
      delete raw_image;

      DEBUG_MSG("[SRV] calculating query image signature\n");
      ex_sign=alg->computeSignature(ex_pic);
      ex_sign->creator_id=alg->algID();
    }
    break;
  case queryByID: // picture ID + database ID where to get the signature
    DEBUG_MSG("[SRV] fetching query image signature from database\n");
    {
      ex_sign=new PicSignature;
      
      databaseID_t qpic_db_id=msg.qPicDB;
      pictureID_t qpic_id=msg.qPicID;

      //FIXME: enable cache.
      /*// Section changed to include PICCache
      cacheElements = msg.ReadInt32();
      cacheIsDisabled = (cacheElements == 0) ? true : false;
      if(cacheElementsArray != NULL)
	free(cacheElementsArray);
      if(cacheIsDisabled == false) {
	cacheElementsArray = (long *) malloc(sizeof(long) * cacheElements);
	for(int i = 0; i < cacheElements; i += 2) {
	  cacheElementsArray[i] = msg.ReadInt32();
	  cacheElementsArray[i + 1] = msg.ReadInt32();
	}
      }*/

      DEBUG_MSG("[SRV] db_id = "<<qpic_db_id<<", pic_id = "<<qpic_id<<endl);

      imageDB *imgDB;
      if((imgDB=DB.getImgDB(qpic_db_id))==NULL) {
	DEBUG_MSG("[SRV] Query: database nonexistent.\n");
	SendFail();
	return;
      }

      signature *theSig=imgDB->fetchSig(search_alg_id,qpic_id);

      if(theSig==NULL) {
	DEBUG_MSG("[SRV] Query: signature not found.\n");
	SendFail();
	return;
      }

      DEBUG_MSG("[SRV] initializing signature object (case 1)\n");

      ex_sign->signature_size=theSig->sigSize;
      ex_sign->signature=new char[ex_sign->signature_size];
      memcpy(ex_sign->signature,theSig->sig,ex_sign->signature_size);
      
      //the signature comes from GDBM and was malloced.
      free(theSig);

      ex_sign->creator_id=alg->algID();
    }
    break;
  }
  
  // the main part of the query is the same of all the query types.

  if (!ex_sign) {
    ERROR_MSG("[SRV] signature creation error in algorithm '"<< \
	      alg->algName()<<"'\n");
    SendFail();
    return;
  }
  
  SendQueryProgress(0);
  current_query.scannedPics=0;

  DEBUG_MSG("[SRV] starting query loop in db "<<query_db_id<<endl);

  imageDB *imgDB;
  if((imgDB=DB.getImgDB(query_db_id))==NULL) {
    DEBUG_MSG("[SRV] Query: database not found.\n");
    SendFail();
    return;
  }

  pictureID_t currentSig=imgDB->firstSigKey(search_alg_id);

  for(;;) { //infinite loop, break from inside.
    PicSignature test_sign;

    if (net.CheckSocket(0)) {
      DEBUG_MSG("[SRV] detected incoming message while executing query\n");

      if(xdrH->currentMessage()==msg_stopQuery) {
	DEBUG_MSG("[SRV] received stop query message while executing query\n");
	// since this query hasn't been finished yet, we can only abort.
	xdrH->proceed(); //no real need to read it.
	delete ex_sign;
	return;
      }
      // ignore any other messages that are received during the query process.
      xdrH->proceed(); //ignore.
      DEBUG_MSG("[SRV] received unexpected message in query - ignored\n");
    }
    DEBUG_MSG("[SRV] in Query loop, iteration " << current_query.scannedPics <<
	      endl);
    
    signature *theSig=imgDB->fetchSig(search_alg_id,currentSig);

    if(theSig==NULL) { //assume no more pictures
      break;
    }

    current_query.scannedPics++;

    DEBUG_MSG("[SRV] comparing signatures\n");

    // do the actual comparing of signatures	
    test_sign.signature=theSig->sig;
    test_sign.signature_size=theSig->sigSize;

    test_sign.creator_id=alg->algID();

    DEBUG_MSG("sigSize: " << test_sign.signature_size << endl);

    PICULONG match_rate=alg->compare(*ex_sign,test_sign);
    DEBUG_MSG("Match rate: " << match_rate << endl);
    
    if(match_rate>min_match)
      current_query.add(PicQueryObj(currentSig,match_rate));

    //theSig was malloced by GDBM.
    free(theSig);

    test_sign.signature=NULL;
    test_sign.signature_size=0;

    DEBUG_MSG("Query.add passed\n");

    if (current_query.scannedPics%10==0)
      SendQueryProgress(current_query.scannedPics);
    currentSig=imgDB->nextSigKey(search_alg_id,currentSig);
  }
  
  SendQueryProgress(current_query.scannedPics);

  current_query.updateTotalTime();
  current_query.crop(total_results);
  DEBUG_MSG("[SRV] "<<current_query.scannedPics<<" images scanned\n");
  
  delete ex_sign;

  current_query_db_id=query_db_id;
  query_ok=1;

  if (!current_query.contains(1)) {
    queryNoResults_msg reply;
    xdrH->send(&reply);
    LOG_MSG("[SRV] no query matches message sent\n");
    return;
  }

  DEBUG_MSG("[SRV] Query passed, sending results." << endl);
  SendNails(1,oneshot_results);
}

void PICServer::SendQueryProgress(long num) const {
  progReport_msg msg;

  msg.status=(int)num;
  xdrH->send(&msg);
  DEBUG_MSG("[SRV] query progress report sent ("<<num<<")\n");
}

void PICServer::StopQuery(void) {
  // this is an unexpected message since StopQuery-messages are expected
  // to arrive only while executing a query. anyway, this message _can_
  // arrive if the user's stop query message is very much delayed and
  // the query has already been completed
  DEBUG_MSG("[SRV] stop query message ignored - query already done");
}

void PICServer::SendNails(unsigned int start,unsigned int count) {
  DEBUG_MSG("[SRV] in SendNails\n");
  
  if(!query_ok || count<=0) {
    DEBUG_MSG("[SRV] SendNails: can't send any." << endl);
    return;
  }

  imageDB *imgDB;
  if((imgDB=DB.getImgDB(current_query_db_id))==NULL) {
    DEBUG_MSG("[SRV] SendNails: database not found.\n");
    SendFail();
    return;
  }

  DEBUG_MSG("[SRV] preparing " << count << " thumbnails for reply message\n");

  queryResults_msg reply;

  reply.nails.nails_len=count;
  reply.nails.nails_val=new thumbNail[reply.nails.nails_len];

  unsigned int nail_pos=0;
  bool success=true;
  for(unsigned int i=0;i<count;i++) {
    thumbNail &nail=reply.nails.nails_val[i];

    int match_ranking=i+start;
    if (current_query.contains(match_ranking)) {
      PicQueryObj &hit=AS(PicQueryObj,current_query.at(match_ranking));
      // first we get the thumbnail of the result and store to our
      // temporary NailStruct array

      picture *currentPic=imgDB->fetchPic(hit.imageId);

      if(currentPic==NULL) {
	DEBUG_MSG("[SRV] SendNails: picture " << hit.imageId << "not found\n");
	success=false;
	break;
      }

      //FIXME: if the pointer currentPic is not freed here or saved somewhere
      //this will continue to leak all memory given to it.
      nail.data.data_len=currentPic->thumbSize;
      nail.data.data_val=currentPic->thumbNail;
      nail.picID=(pictureID_t)hit.imageId;
      nail.match=hit.matchRate;
      nail.rank=match_ranking;
      nail.paramStr1=currentPic->paramStr1;
      nail.paramStr2=currentPic->paramStr2;
      nail_pos++;
    }
  }
  if (success) {
    DEBUG_MSG("[SRV] constructing thumbnail message ("<<nail_pos<< \
	      " nails found)\n");
    //set thumbnum to actual amount of thumbnails found.
    reply.nails.nails_len=nail_pos;
    xdrH->send(&reply);
  }

  //FIXME: I have no way to free the mem GDBM gave me. see FIXME above.
  DEBUG_MSG("[SRV] freeing mem\n");
  delete reply.nails.nails_val;
  DEBUG_MSG("[SRV] frees done.\n");
}

void PICServer::SendMoreNails(moreThumbs_msg &msg) {
  DEBUG_MSG("[SRV] in SendMoreNails\n");

  unsigned int start=msg.firstThumb;
  unsigned int count=msg.thumbNum;

  //FIXME: enable cache.
  /*// Section changed to include PICCache
  cacheElements = msg.ReadInt32();
  cacheIsDisabled = (cacheElements == 0) ? true : false;
  if(cacheElementsArray != NULL)
    free(cacheElementsArray);
  if(cacheIsDisabled == false) {
    cacheElementsArray = (long *) malloc(sizeof(long) * cacheElements);
    // Ok, fill the cache information; it will be used in SendNails()
    for(int i = 0; i < cacheElements; i += 2) {
      cacheElementsArray[i] = msg.ReadInt32();
      cacheElementsArray[i + 1] = msg.ReadInt32();
    }
  }*/
  DEBUG_MSG("[SRV] Sending "<<count<<" nails starting from #"<<start<<endl);
  SendNails(start,count);
}

