// **********************************************************************
//  Various generally helpful functions
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef MTOOLS_HXX
#define MTOOLS_HXX

#include <parray.hxx>
#include <pfile.hxx>

class PicPicture;
class PicRawPicture;
class PicEnhancedThread;
class PicQueryArray;

/*:class ********************************************************************

PicTools is a placeholder class for various generally useful
miscellaneous functions.

****************************************************************************/

class PicTools
{
 public:

//:fun
  static PicRawPicture* scaleImage(const PicRawPicture &raw,
                                   PICUSHORT newWidth,
                                   PICUSHORT newHeight,
                                   PICBOOL   keepAspectRatio = boolTrue);
/*
Creates a new scaled PicRawPicture from the argument image data.
NewWidth and newHeight are the (maximum) dimensions of the
resulting image. If keepAspectRatio is true, the proportions
of the resulting image are equal to that of the source. In this
case the resulting image dimensions may differ from newWidth and 
newHeight.
*/

//:fun
  static PICBOOL listToCollect(BaseString from,
                               PicIxCollect &to,
                               char delimiter = ' ');

  static PICBOOL collectToList(BaseString &to,
                               PicIxCollect &from,
                               char delimiter = ' ');
/*
These functions convert a list of parameters into an indexed
collection of PicStrings and vice versa. Empty strings and 
those having space-characters are surrounded by braces in
the BaseString:s (feature required by Tcl/Tk).
*/

//:fun
  static PICBOOL windowGrab(PicPicture &pic);
/*
Used in Win32 for grabbing the image of editor window.
*/

//:fun
  static PICBOOL sendMasterPrint(BaseString mesg);
  static PICBOOL sendMasterQueryResults(const PicQueryArray &pics);
  static PICBOOL sendThreadCommand(PicEnhancedThread &t);
  static PICBOOL listenToMaster(PicEnhancedThread &t);
/*
Used in Unix systems to send messages to between processes.
*/

};

/*:class ********************************************************************

PicBin is a class for representing binary data.

****************************************************************************/

class PicBin : public PicObject
{
 public:
  PICOBJECTCLASS(PicBin)

//:fun
  PicBin ();
  PicBin (const PicBin &other);
  ~PicBin ();
/*
Default constructor creates a null object.
Copy constructor copies data from argument.
Destructor frees the data allocated.
*/

//:fun
  virtual PicObject& operator = (const PicObject&);
  virtual PicObject& operator = (BaseString other);
  virtual PicObject& set(void *from, PICULONG size);
/*
These functions free the current data and copy new
from the argument.
*/
  virtual PICBOOL readFromFile(const PicFile &file);
  virtual PICBOOL writeToFile(const PicFile &file) const;

//:fun
  virtual const void *get() const;
  virtual PICULONG    size() const;
/*
Get returns a pointer to the internal buffer, don't store it
anywhere since it is not any more valid if the object is 
modified/destructed. Return value may be NULLPTR.
Size returns the size of the data (in bytes).
*/

//:fun
  virtual PICBOOL operator == (const PicObject&) const;
/*
Returns true if both objects are of equal size and the data
in the buffers match.
*/

  virtual PICBOOL operator <  (const PicObject&) const;

  virtual PICBOOL isNull() const;
//:fun
  virtual void    setNull();
/*
Deletes the data.
*/

  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;

//:fun
  virtual void setImage(const BaseString& image);
/*
Functions like  operator = (BaseString other);
*/
  virtual void setImage(const PicObject& format, const BaseString& image);

 protected:
  
  void      *data;
  PICULONG  dataSize;
};

// **********************************************************************

inline PICBOOL PicBin::isDefault() const 
  { return isNull(); }
inline void PicBin::setDefault()
  { setNull(); }
inline BaseString PicBin::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicBin::setImage(const PicObject& format, 
                             const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }

#endif

