// **********************************************************************
//  This file contains the initialization routines of PicMain
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <mmain.hxx>
#include <picerror.hxx>
#include <pinteger.hxx>
#include <pstring.hxx>
#include <bstrarr.hxx>
#include <iostream.h>

#include <unistd.h>

MAKE_PICCLASS(PicMain)

ERROR_MESSAGES(PicMain)
PARAMSTR(initFailed, "Initialization failed")
PARAMSTR(unKnownImageFormat, "Unknown image format")
PARAMSTR(unKnownSearchType, "Unknown search type \"@*\"")
PARAMSTR(unKnownSearchAlg, "Unknown search algorithm \"@*\"")
PARAMSTR(unKnownSegmAlg, "Unknown segmentation algorithm \"@*\"")
PARAMSTR(unKnownExampleImage, "Unknown query image (index @*)")
PARAMSTR(threadSuspendFailure, "Thread did not suspend within @* seconds")
PARAMSTR(threadAbortFailure, "Thread did not abort within @* seconds")
PARAMSTR(invalidParams, "Invalid parameters")
PARAMSTR(imageInsertFailed, "Image insertion failed")
PARAMSTR(imageReadFailed, "Image reading failed")
PARAMSTR(deleteImageFailed, "Image deletion failed")
PARAMSTR(exImageCreateFailed, "Could not create new query image")
PARAMSTR(exImageCloseFailed, "Could not close query image")
PARAMSTR(exImageClearFailed, "Could not clear query image")
END_PARAM_STRINGS

/*
static char *copyRightMesg = 
  "PicSearch version 3.1, Copyright (C) 1999 PicSearch Team\n" \
  "PicSearch comes with ABSOLUTELY NO WARRANTY; \n" \
  "This is free software, and you are welcome to redistribute it \n" \
  "under certain conditions; see General Public Licence for details.\n";
*/

// **********************************************************************
//  This function creates an object identified by class id and
//  adds it into given collection. Also descendants are inserted
//  recursively. Of course, all abstract classes are skipped.

static void addToArray(const PicObjectClass *id, PicSeqIxCollect *coll)
{
  if(!id || !coll)
    return;

  PicObject *obj = id->createNew();
  if(obj){
    // It's a non-abstract class, so add to collection
    coll->add(*obj);
    delete obj;
  }
  // Add children also:
  const PicVirtClass *curr = id->firstDerived();
  while(curr){
    DEBUG_REQUIRE(curr->hasAncestor(PicObject::classId()));
    addToArray((PicObjectClass *)curr, coll);
    curr = curr->nextSibling(); // carefully, now
  }
}

// **********************************************************************
// Construct globData first so others may use it

PicMainSharedData   PicMain::globData;

PicMain::PicMain() 
{
  PicErrorContext ecxt;
  PicIterator     iter;
  // First create some search types

  // Then iterate all known image formats and algorithms
  // to create one of each into suitable collection

  const PicVirtClass *classId = &PicImageFormat::classId();
  DEBUG_REQUIRE(classId && classId->hasAncestor(PicObject::classId()));
  classId = classId->firstDerived();
  while(classId){
    DEBUG_REQUIRE(classId->hasAncestor(PicObject::classId()));
    addToArray((PicObjectClass *)classId, &globData.imageFormats);
    classId = classId->nextSibling();
  }
  globData.params.formatForScaledImages = globData.getImageFormat(PIC_JPEG_IMAGEFORMAT_ID);

  classId = &PicSegmentAlgorithm::classId();
  DEBUG_REQUIRE(classId && classId->hasAncestor(PicObject::classId()));
  classId = classId->firstDerived();
  while(classId){
    DEBUG_REQUIRE(classId->hasAncestor(PicObject::classId()));
    addToArray((PicObjectClass *)classId, &globData.segmentationAlgs);
    classId = classId->nextSibling();
  }

  if(PicError::isDetected()){
    EHOLD(PicMain, initFailed);
    return;
  }
}

// **********************************************************************

PicMain::~PicMain()
{

}

// **********************************************************************

// PICBOOL      picSearchIsExiting; // set in PicMain::exitProgram

#include <picerror.hxx>
#include <sys/time.h>

// **********************************************************************

PicParams::PicParams()
  : 
  showAbortedQuery(boolFalse),
  askChangeConfirm(boolTrue),
  dbInsPhysically(boolFalse),
  picGUIDebug(boolFalse),
  picCoreDebug(boolFalse),
  picAlgDebug(boolFalse),
  picDbDebug(boolFalse),
  threadAbortTimeout(15),
  dbImageFramesOnScreen(12),
  stampWidth(128),
  stampHeight(128),
  formatForScaledImages(NULLPTR),
  defWidth(300),
  defHeight(300)
{

}

// **********************************************************************

PICBOOL PicParams::fromColl(PicIxCollect &coll)
{
  PicErrorContext ecxt;
  static PicString trueVal("1");

  if(coll.contains(PICPARAMS_SHOW_ABORT_QUERY)){
    showAbortedQuery = 
      (AS(PicString, coll.at(PICPARAMS_SHOW_ABORT_QUERY)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_ASK_CHANGE_CONF)){
    askChangeConfirm = 
      (AS(PicString, coll.at(PICPARAMS_ASK_CHANGE_CONF)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_DB_INS_PHYS)){
    dbInsPhysically = 
      (AS(PicString, coll.at(PICPARAMS_DB_INS_PHYS)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_GUI_DEBUG)){
    picGUIDebug = 
      (AS(PicString, coll.at(PICPARAMS_GUI_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_CORE_DEBUG)){
    picCoreDebug = 
      (AS(PicString, coll.at(PICPARAMS_CORE_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_ALG_DEBUG)){
    picAlgDebug = 
      (AS(PicString, coll.at(PICPARAMS_ALG_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_DB_DEBUG)){
    picDbDebug = 
      (AS(PicString, coll.at(PICPARAMS_DB_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_DEF_WIDTH)){
    PicString &ref = AS(PicString, coll.at(PICPARAMS_DEF_WIDTH));
    PicInteger newWidth(ref);
    if(!PicError::isDetected()){
      PICUSHORT trueWidth = setDefWidth((PICUSHORT)(newWidth.asInt()));
      ref = asString(trueWidth);
    }
  }
  if(coll.contains(PICPARAMS_DEF_HEIGHT)){
    PicString &ref = AS(PicString, coll.at(PICPARAMS_DEF_HEIGHT));
    PicInteger newHeight(ref);
    if(!PicError::isDetected()){
      PICUSHORT trueHeight = setDefHeight((PICUSHORT)(newHeight.asInt()));
      ref = asString(trueHeight);
    }
  }
  return boolTrue;
}

// **********************************************************************

PICBOOL PicParams::toColl(PicIxCollect &coll)
{
  PicErrorContext ecxt;
  static PicString falseVal("0");
  static PicString trueVal("1");

  coll.deleteAll();

  coll.putAt(PICPARAMS_SHOW_ABORT_QUERY, 
              showAbortedQuery ? trueVal : falseVal);
  coll.putAt(PICPARAMS_ASK_CHANGE_CONF, 
              askChangeConfirm ? trueVal : falseVal);
  coll.putAt(PICPARAMS_DB_INS_PHYS, 
              dbInsPhysically ? trueVal : falseVal);
  coll.putAt(PICPARAMS_GUI_DEBUG, 
              picGUIDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_CORE_DEBUG, 
              picCoreDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_ALG_DEBUG, 
              picAlgDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_DB_DEBUG, 
              picDbDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_DEF_WIDTH, PicString(asString(getDefWidth())));
  coll.putAt(PICPARAMS_DEF_HEIGHT, PicString(asString(getDefHeight())));

  return !PicError::isDetected();
}

// **********************************************************************

PICUSHORT PicParams::getDefWidth()
{
  return defWidth;
}

PICUSHORT PicParams::setDefWidth(PICUSHORT newWidth)
{
  if(newWidth > 20 && newWidth < 1025){
    defWidth = newWidth;
  }
  return defWidth;
}

PICUSHORT PicParams::getDefHeight()
{
  return defHeight;
}

PICUSHORT PicParams::setDefHeight(PICUSHORT newHeight)
{
  if(newHeight > 20 && newHeight < 769){
    defHeight = newHeight;
  }
  return defHeight;
}

